// ***************************************************************************
// bcdemux.exe - Copyright (C) 2002 - Ben Cooley
//
// bcdemux demultiplexes an ATSC transport stream and converts it to an MPEG2
// DVD compatible video file.  It can select any program in the TS file, and 
// can automatically read a sequence of TS files and convert them to any number
// of MPG files.
//
// Changelog:
//
// 1.0 - 8-1-2002 -		First version
//
// 1.01 - 8-1-2002 -	Fixed a problem with some TS streams (ABC) not having the RANDOM bit in
//						the adaption header set.  Now just looks for the 'start' bit.  I think
//						this will get A-B frames at the beginning of a video, but it seems to 
//						work.
//
// 1.02 - 13-1-2002 -   Fixed problem again with some TS streams not having an adaption header on
//                      TS packets marked 'START'.  This was causing some of the streams not to 
//                      convert correctly.
//
// 1.03 - 16-1-2002 -	Fixed problem with AC3 audio stream from HBO streams.  AC3 packets from HBO
//						aren't aligned to the MPEG 0xBD packets.  Bit of code realigns them.
//
// 1.04 - 19-1-2002 -	FINALLY fixed problem with AC3 audio stream from HBO streams.  AC3 packets from HBO
//						aren't aligned to the MPEG 0xBD packets, and they don't start in TS start packets 
//						occasionally.  Solution is to just skip packet starts where there is no AC3 packet
//						sync.  Seems to work pretty well.
//
//						Fixed pack PTS timestamps so that the time displays in DVD players are accurate.
//						One minute is now really one minute.  There's still a problem with some DVD players
//						picking up the PTS in the video frames, but we'll save that for later.
//
//						Integrated this file with both the HDTVtoMPEG2 projects and the bcdmux projects. 
//						That way changes to this file automatically are reflected in both projects.
//
// 1.05 - 2-18-2002 -	Added channel detection, support for outputting files greater than 2GB, and better
//						error correction.  Also changed the program to start a new file after a selectable
//						size has been reached.
//
// 1.06 - 2-19-2002 -	Forced all files to be cut on I-Frame boundries.  Now files start on I-Frame's with
//						GOP (group of picture) packets.
//
// 1.07 - 2-23-2002 - GR: Recognise AccessDTV's strange sync byte mangling (0x47, 0x72, 0x29).
//						Packet alignment per file support (.adtv files not ts packet aligned).
//						Support for progress bar of total conversion.
//						Improved I-frame detection, detect if picture packets are I-frames in case GOPs are missing.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#include <memory.h>

#include "bcdmux.h"

#ifdef HDTVtoMPEG2
#include "mpeg2stuff.h"
#endif

#define PACKETSIZE		(2048)						// MPEG Program stream block size
#define FILE_TIME		(SYSTEMCLOCKRATE * (__int64)60)	// Time for each file
#define DEMUX			(((int)STREAMRATE * 8) / 50)// Demux value for HD content STREAMRATE / 50

#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))
#endif

// ------------------------------------------------------------------------------------

#ifdef HDTVtoMPEG2

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

static inline void error(char *, char * = NULL)
{
}

static inline void warning(char *, char * = NULL)
{
}

#else

static void error(char* errorstr, char* extra = NULL)
{
	printf("ERROR: ");
	printf(errorstr, extra);
}

static void warning(char* errorstr, char* extra = NULL)
{
	printf("WARNING: ");
	printf(errorstr, extra);
}

#endif

// TSFileInfo

TSFileInfo::~TSFileInfo()
{
	Clear();
}

TSRegionInfo *TSFileInfo::Find(__int64 offset)
{
	TSRegionInfo *i = m_head;

	if (i && (offset < i->m_offset))
		return NULL;

	while (i && i->m_next && (offset >= i->m_next->m_offset))
		i = i->m_next;

	return i;
}

TSRegionInfo *TSFileInfo::Add(RegionType type, __int64 offset)
{
	// sorted add
	TSRegionInfo *info = new TSRegionInfo;
	if (info)
	{
		info->m_type = type;
		info->m_offset = offset;
		info->m_parent = this;

		TSRegionInfo *i = m_head;
		int r = 0;
		while (i && (offset > i->m_offset))
			i = i->m_next;
		if (i)
		{
			if (offset == i->m_offset)
			{
				// already have an entry, use that instead
				delete info;
				info = i;
				info->m_type = type;
			}
			else
			{
				// add before i
				info->m_prev = i->m_prev;
				info->m_next = i;
				if (i->m_prev)
					i->m_prev->m_next = info;
				i->m_prev = info;
				if (m_head == i)
					m_head = info;
				m_count++;
			}
		}
		else
		{
			// add to tail
			info->m_prev = m_tail;
			info->m_next = NULL;
			if (!m_head)
				m_head = info;
			if (m_tail)
				m_tail->m_next = info;
			m_tail = info;
			m_count++;
		}
	}
	return info;
}

void TSFileInfo::Delete(TSRegionInfo *info)
{
	if (info->m_next)
		info->m_next->m_prev = info->m_prev;
	if (info->m_prev)
		info->m_prev->m_next = info->m_next;
	if (m_head == info)
		m_head = info->m_next;
	if (m_tail == info)
		m_tail = info->m_prev;

	delete info;

	m_count--;
}

void TSFileInfo::Clear(void)
{
	while (m_head)
	{
		m_tail = m_head->m_next;
		delete m_head;
		m_head = m_tail;
	}

	m_head = NULL;
	m_tail = NULL;
	m_count = 0;
}

// TSFileList

TSFileList::~TSFileList()
{
	Clear();
	DeleteCriticalSection(&m_cSec);
}

TSFileInfo *TSFileList::Find(const char *name)
{
	EnterCriticalSection(&m_cSec);

	TSFileInfo *i = m_head;

//	while (i && stricmp(name, i->m_fileName))
	while( i && _stricmp(name, i->m_fileName ))
		i = i->m_next;

	LeaveCriticalSection(&m_cSec);
	return i;
}

TSFileInfo *TSFileList::Find(__int64 bytes)
{
	EnterCriticalSection(&m_cSec);

	TSFileInfo *i = m_head;

	while (i && (bytes > i->m_totalBytesPrior + i->m_fileSize))
		i = i->m_next;

	LeaveCriticalSection(&m_cSec);
	return i;
}

TSFileInfo *TSFileList::Add(const char *name, __int64 size, int offset)
{
	EnterCriticalSection(&m_cSec);

	// sorted add
	TSFileInfo *info = new TSFileInfo;
	if (info)
	{
		strcpy(info->m_fileName, name);
		info->m_fileSize = size;
		info->m_firstPacketOffset = offset;
		info->m_f = INVALID_HANDLE_VALUE;

		TSFileInfo *i = m_head;
		int r = 0;
//		while (i && ((r = stricmp(name, i->m_fileName)) > 0))
		while( i && ((r = _stricmp(name, i->m_fileName )) > 0 ))
			i = i->m_next;
		if (i)
		{
			if (!r)
			{
				// already have an entry, use that instead
				delete info;
				info = i;
				if (info->m_f != INVALID_HANDLE_VALUE)
					CloseHandle(info->m_f);

				info->m_fileSize = size;
				info->m_firstPacketOffset = offset;
				info->m_f = INVALID_HANDLE_VALUE;
			}
			else
			{
				// add before i
				info->m_prev = i->m_prev;
				info->m_next = i;
				if (i->m_prev)
					i->m_prev->m_next = info;
				i->m_prev = info;
				if (m_head == i)
					m_head = info;
				m_count++;
			}
		}
		else
		{
			// add to tail
			info->m_prev = m_tail;
			info->m_next = NULL;
			if (!m_head)
				m_head = info;
			if (m_tail)
				m_tail->m_next = info;
			m_tail = info;
			m_count++;
		}
		RecomputeBytes();
		RecomputeTotalIncludedBytes();
	}

	LeaveCriticalSection(&m_cSec);
	return info;
}

void TSFileList::Delete(const char *name)
{
	EnterCriticalSection(&m_cSec);

	Delete(Find(name));

	LeaveCriticalSection(&m_cSec);
}

void TSFileList::Delete(TSFileInfo *info)
{
	EnterCriticalSection(&m_cSec);

	if (info->m_next)
		info->m_next->m_prev = info->m_prev;
	if (info->m_prev)
		info->m_prev->m_next = info->m_next;
	if (m_head == info)
		m_head = info->m_next;
	if (m_tail == info)
		m_tail = info->m_prev;

	if (info->m_f != INVALID_HANDLE_VALUE)
		CloseHandle(info->m_f);

	delete info;

	m_count--;
	RecomputeBytes();
	RecomputeTotalIncludedBytes();

	LeaveCriticalSection(&m_cSec);
}

void TSFileList::Clear(void)
{
	EnterCriticalSection(&m_cSec);

	while (m_head)
	{
		if (m_head->m_f != INVALID_HANDLE_VALUE)
			CloseHandle(m_head->m_f);

		m_tail = m_head->m_next;
		delete m_head;
		m_head = m_tail;
	}

	m_head = NULL;
	m_tail = NULL;
	m_count = 0;
	m_totalBytes = 0;
	m_totalIncludedBytes = 0;

	LeaveCriticalSection(&m_cSec);
}

void TSFileList::RecomputeBytes(void)
{
	EnterCriticalSection(&m_cSec);

	m_totalBytes = 0;
	TSFileInfo *i = m_head;
	while (i)
	{
		i->m_totalBytesPrior = m_totalBytes;
		m_totalBytes += i->m_fileSize;
		i = i->m_next;
	}

	LeaveCriticalSection(&m_cSec);
}

void TSFileList::RecomputeTotalIncludedBytes(void)
{
	EnterCriticalSection(&m_cSec);

	m_totalIncludedBytes = 0;
	TSFileInfo *i = m_head;
	RegionType rt = rt_include;
	while( i )
	{
		__int64 offset = 0;
		TSRegionInfo *ri = i->m_head;

		while( ri )
		{
			if( rt == rt_include || rt == rt_cut )
				m_totalIncludedBytes += ri->m_offset - offset;

			offset = ri->m_offset;
			rt = ri->m_type;
			ri = ri->m_next;
		}

		if( rt == rt_include || rt == rt_cut )
			m_totalIncludedBytes += i->m_fileSize - offset;

		i = i->m_next;
	}

	LeaveCriticalSection(&m_cSec);
}

void TSFileList::CloseFilesOutsideRange(__int64 rmin, __int64 rmax)
{
	EnterCriticalSection(&m_cSec);

	TSFileInfo *i = m_head;
	while (i)
	{
		if (i->m_f != INVALID_HANDLE_VALUE)
		{
			if ((rmin >= i->m_totalBytesPrior + i->m_fileSize) ||
				(rmax < i->m_totalBytesPrior))
			{
				CloseHandle(i->m_f);
				i->m_f = INVALID_HANDLE_VALUE;
			}
		}
		i = i->m_next;
	}

	LeaveCriticalSection(&m_cSec);
}

TSRegionInfo *TSFileList::FindRegion(__int64 bytes)
{
	EnterCriticalSection(&m_cSec);

	TSFileInfo *i = m_head;
	TSRegionInfo *r = NULL;
	TSRegionInfo *r2 = NULL;

	while (i)
	{
		r2 = i->Find(bytes - i->m_totalBytesPrior);
		if (r2)
			r = r2;

		i = i->m_next;
	}

	LeaveCriticalSection(&m_cSec);
	return r;
}

TSRegionInfo *TSFileList::PrevRegion(TSRegionInfo *ri)
{
	if (!ri)
		return NULL;

	if (ri->m_prev)
		return ri->m_prev;

	EnterCriticalSection(&m_cSec);

	TSFileInfo *i = ri->m_parent->m_prev;
	while (i)
	{
		if (i->m_tail)
		{
			LeaveCriticalSection(&m_cSec);
			return i->m_tail;
		}
		i = i->m_prev;
	}

	LeaveCriticalSection(&m_cSec);
	return NULL;
}

TSRegionInfo *TSFileList::NextRegion(TSRegionInfo *ri)
{
	if (ri && ri->m_next)
		return ri->m_next;

	EnterCriticalSection(&m_cSec);

	TSFileInfo *i;
	if (ri)
		i = ri->m_parent->m_next;
	else
		i = m_head;
	while (i)
	{
		if (i->m_head)
		{
			LeaveCriticalSection(&m_cSec);
			return i->m_head;
		}
		i = i->m_next;
	}

	LeaveCriticalSection(&m_cSec);
	return NULL;
}

TSRegionInfo *TSFileList::AddRegion(RegionType type, __int64 bytes)
{
	EnterCriticalSection(&m_cSec);

	TSRegionInfo *r = NULL;
	TSFileInfo *ti = Find(bytes);

	if (ti)
	{
		r = ti->Add(type, bytes - ti->m_totalBytesPrior);
		RecomputeTotalIncludedBytes();
	}

	LeaveCriticalSection(&m_cSec);
	return r;
}

void TSFileList::DeleteRegion(TSRegionInfo *info)
{
	EnterCriticalSection(&m_cSec);

	if (info->m_parent)
		info->m_parent->Delete(info);
	RecomputeTotalIncludedBytes();

	LeaveCriticalSection(&m_cSec);
}

void TSFileList::ClearRegions(void)
{
	EnterCriticalSection(&m_cSec);

	TSFileInfo *i = m_head;
	while (i)
	{
		i->Clear();
		i = i->m_next;
	}
	RecomputeTotalIncludedBytes();

	LeaveCriticalSection(&m_cSec);
}

void TSFileList::ClearPrevIFrameOffsets(void)
{
	EnterCriticalSection( &m_cSec );

	TSFileInfo *i = m_head;
	while( i )
	{
		i->m_prevIFrameOffset_Info = NULL;
		i = i->m_next;
	}

	LeaveCriticalSection( &m_cSec );
}

void TSFileList::AdjustList( int movingIndex, int moveToIndex )
{
	EnterCriticalSection( &m_cSec );
	
	TSFileInfo* ts_MovingIndex = NULL;
	TSFileInfo* ts_MoveToIndex = NULL;
	TSFileInfo* ts_Current = NULL;
/*
// debug
ts_Current = m_head;
_Error("head:%x\n", m_head);
for( int i = 0; i < Count(); i++ )
{
	_Error("%s, %x, %x, %x\n", ts_Current->m_fileName, ts_Current->m_prev, ts_Current, ts_Current->m_next );
	ts_Current = ts_Current->m_next;
}
_Error("tail:%x\n", m_tail);
*/
	ts_Current = m_head;

	// find our objects
	for( int i = 0; i < Count(); i++ )
	{
		if( i == movingIndex )
			ts_MovingIndex = ts_Current;
		else if( i == moveToIndex )
			ts_MoveToIndex = ts_Current;

		if( ts_MovingIndex != NULL && ts_MoveToIndex != NULL )
			break; 
		else
		{
			if( i == ( Count() - 1 ))
				ts_MoveToIndex = ts_Current;

			ts_Current = ts_Current->m_next;
		}
	}

	// move node to end of list
	if( moveToIndex == -1 )
	{
		// if it's the head node
		if( ts_MovingIndex == m_head )
		{
			m_head = ts_MovingIndex->m_next;
			ts_MovingIndex->m_next->m_prev = NULL;
		}
		else
		{
			ts_MovingIndex->m_prev->m_next = ts_MovingIndex->m_next;
			ts_MovingIndex->m_next->m_prev = ts_MovingIndex->m_prev;
		}

		// insert the node to the end of the list
		ts_MoveToIndex->m_next = ts_MovingIndex;
		ts_MovingIndex->m_prev = ts_MoveToIndex;
		ts_MovingIndex->m_next = NULL;
		m_tail = ts_MovingIndex;
	}
	// move node to the beginning of list
	else if( moveToIndex == 0 )
	{
		// are we moving the tail?
		if( ts_MovingIndex == m_tail )
		{
			ts_MovingIndex->m_prev->m_next = NULL;
			m_tail = ts_MovingIndex->m_prev;
		}
		else
		{
			ts_MovingIndex->m_prev->m_next = ts_MovingIndex->m_next;
			ts_MovingIndex->m_next->m_prev = ts_MovingIndex->m_prev;
		}

		// insert the node to the beginning of the list
		ts_MovingIndex->m_next = ts_MoveToIndex;
		ts_MovingIndex->m_prev = NULL;
		ts_MoveToIndex->m_prev = ts_MovingIndex;
		m_head = ts_MovingIndex;
	}
	// move down
	else if( movingIndex < moveToIndex )
	{
		// are we moving the head to the middle?
		if( ts_MovingIndex == m_head )
		{
			m_head = ts_MovingIndex->m_next;
			m_head->m_prev = NULL;
		}
		else
		{
			ts_MovingIndex->m_prev->m_next = ts_MovingIndex->m_next;
			ts_MovingIndex->m_next->m_prev = ts_MovingIndex->m_prev;
		}

		// insert the node into the middle of the list
		ts_MoveToIndex->m_prev->m_next = ts_MovingIndex;
		ts_MovingIndex->m_next = ts_MoveToIndex;
		ts_MovingIndex->m_prev = ts_MoveToIndex->m_prev;
		ts_MoveToIndex->m_prev = ts_MovingIndex;
		
	}
	// move up
	else // if( movingIndex > moveToIndex )
	{
		// are we moving the tail to the middle?
		if( ts_MovingIndex == m_tail )
		{
			ts_MovingIndex->m_prev->m_next = NULL;
			m_tail = ts_MovingIndex->m_prev;
		}
		else
		{
			ts_MovingIndex->m_prev->m_next = ts_MovingIndex->m_next;
			ts_MovingIndex->m_next->m_prev = ts_MovingIndex->m_prev;
		}

		// insert the node into the middle of the list
		ts_MoveToIndex->m_prev->m_next = ts_MovingIndex;
		ts_MovingIndex->m_next = ts_MoveToIndex;
		ts_MovingIndex->m_prev = ts_MoveToIndex->m_prev;
		ts_MoveToIndex->m_prev = ts_MovingIndex;
	}

	RecomputeBytes();

/*
// debug
ts_Current = m_head;
_Error("\n");
_Error("head:%x\n", m_head);
for( int i = 0; i < Count(); i++ )
{
	_Error("%s, %x, %x, %x\n", ts_Current->m_fileName, ts_Current->m_prev, ts_Current, ts_Current->m_next );
	ts_Current = ts_Current->m_next;
}
_Error("tail:%x\n", m_tail);
*/
	LeaveCriticalSection( &m_cSec );
}

void TSFileList::Lock(void)
{
	EnterCriticalSection(&m_cSec);
}

void TSFileList::Unlock(void)
{
	LeaveCriticalSection(&m_cSec);
}


// ------------------------------------------------------------------------------------

int f64_error = 0;

int ferror64(void)
{
	return f64_error;
}

FILE64* fopen64(const char* name, const char* mode, int buffersize, bool overwrite)
{
	f64_error = 0;

	FILE64* f = new FILE64;

	// Just check first char..
	if (mode[0] == 'r')
	{
		f->handle = _open(name, _O_RDONLY | _O_SEQUENTIAL | _O_BINARY);
		if (f->handle < 0)
		{
			f64_error = errno;
			delete f;
			return NULL;
		}
		f->read = true;
		f->size = buffersize;
		f->buffer = new unsigned char[f->size];

		if (!_fstati64(f->handle, &f->info))
		{
			f->filelen = _lseeki64(f->handle, 0, SEEK_END);
			_lseeki64(f->handle, 0, SEEK_SET);
		}
		else
			f->filelen = f->info.st_size;
	}
	else
	{

		int oflag = _O_WRONLY | _O_CREAT | _O_TRUNC | _O_BINARY | _O_SEQUENTIAL;
		if (!overwrite)
			oflag |= _O_EXCL;
		f->handle = _open(name, oflag, _S_IREAD | _S_IWRITE);
		if (f->handle < 0)
		{
			f64_error = errno;
			delete f;
			return NULL;
		}
		f->read = false;
		f->size = buffersize;
		f->buffer = new unsigned char[f->size];
		f->filelen = 0;
	}

	f->filepos = 0;
	f->pos = f->len = 0;

	return f;
}

static void fillbuf(FILE64* f)
{
	f->filepos += f->len;
	f->len = _read(f->handle, f->buffer, f->size);
	f->pos = 0;
}

static bool flushbuf(FILE64* f)
{
	if (_write(f->handle, f->buffer, f->len) != f->len)
		return false;
	f->filepos += f->len;
	f->filelen += f->len;
	f->pos = f->len = 0;
	return true;
}

__int64 ftell64(FILE64* f)
{
	f64_error = 0;
	return f->filepos + (__int64)f->pos;
}

__int64 fseek64(FILE64* f, __int64 pos, int set)
{
	f64_error = 0;
	if ((pos >= f->filepos) && (pos < (f->filepos + (__int64)f->len)))
	{
		f->pos = (int)(pos - f->filepos);
	}
	else
	{
		_lseeki64(f->handle, pos, SEEK_SET);
		f->filepos = pos;
		f->pos = f->len = 0;
	}
	return pos;
}

__int64 filelength64(FILE64* f)
{
	f64_error = 0;
	return f->filelen;
}

int fread64(void* buf, int elsize, int elnum, FILE64* f)
{
	f64_error = 0;
	int size = elsize;
	if (elnum > 1)
		size *= elnum;
	
	int read = 0;

	if (size < f->len - f->pos)
	{
		memcpy(buf, f->buffer + f->pos, size);
		f->pos += size;
		read = size;
	}
	else
	{
		memcpy(buf, f->buffer + f->pos, f->len - f->pos);
		read += f->len - f->pos;

		fillbuf(f);

		if (f->len > size - read)
		{
			memcpy((unsigned char*)buf + read, f->buffer, size - read);
			f->pos += size - read;
			read = size;
		}
		else
		{
			memcpy((unsigned char*)buf + read, f->buffer, f->len);
			f->pos += f->len;
			read += f->len;
		}
	}

	if (elnum == 1 && read == size)
		return 1;
	else if (elsize == 1)
		return read;
	else
		return read / elsize;
}

int fwrite64(void* buf, int elsize, int elnum, FILE64* f)
{
	f64_error = 0;
	int size = elsize;
	if (elnum > 1)
		size *= elnum;
	
	int written = 0;

	if (size <= f->size - f->pos)
	{
		memcpy(f->buffer + f->pos, buf, size);
		f->pos += size;
		f->len = f->pos;
		written = size;
	}
	else
	{
		memcpy(f->buffer + f->pos, buf, f->size - f->pos);
		written += f->size - f->pos;
		f->pos += f->size - f->pos;
		f->len = f->pos;

		if (flushbuf(f))
		{
			memcpy(f->buffer, (unsigned char*)buf + written, size - written);
			f->pos += size - written;
			f->len = f->pos;
			written += size - written;
		}
	}


	if (elnum == 1 && written == size)
		return 1;
	else
		return written / elsize;
}

void fclose64(FILE64* f)
{
	f64_error = 0;
	if (!f->read)
		flushbuf(f);
	_close(f->handle);
	delete []f->buffer;
	delete f;
}



// ------------------------------------------------------------------------------------

int bitpos = 0;
unsigned int bitval = 0;
unsigned char* bitbuf = NULL;
unsigned int bitmask[] = {
	0x0,0x1,0x3,0x7,0xf,0x1f,0x3f,0x7f,0xff,
	0x1ff,0x3ff,0x7ff,0xfff,0x1fff,0x3fff,0x7fff,0xffff,
	0x1ffff,0x3ffff,0x7ffff,0xfffff,0x1fffff,0x3fffff,0x7fffff,0xffffff,
	0x1ffffff,0x3ffffff,0x7ffffff,0xfffffff,0x1fffffff,0x3fffffff,0x7fffffff,0xffffffff};

// ------------------------------------------------------------------------------------

int align_to_next_packet(uchar *buf, int size)
{
	int pos = 0;
	bool found = false;
	while (!found && (pos < 188))
	{
		found = true;
		for (int i = 0; pos+i < size; i += 188)
		{
			unsigned char c = buf[pos+i];
			// Check sync byte
			if ((c != 0x47) && (c != 0x72) && (c != 0x29))
			{
				// this offset failed, try next
				found = false;
				pos++;
				break;
			}
		}
	}
	return pos;
}

__int64 align_to_next_packet(FILE64* f)
{
	uchar buf[188*20];

	__int64 start = ftell64(f);
	__int64 pos = 0;

	if (fread64(buf, 188*20, 1, f) == 1)
		pos = (__int64)align_to_next_packet(buf, 188*20);

	if (pos == 188)
		pos = 0;		// failed to find anything!!!!!?

	fseek64(f, start+pos, SEEK_SET);

	return pos;
}

// ------------------------------------------------------------------------------------

static void set_pts(unsigned char* buf, __int64 PTS)
{
	set_buf(buf, 5);

	set_bits(2, 4);								// '0010'							4
	set_bits((unsigned int)(PTS >> 30), 3);		// PTS [32..30]						3
	set_bits(1, 1);								// marker bit						1
	set_bits((unsigned int)(PTS >> 15), 15);	// PTS [29..15]						15
	set_bits(1, 1);								// marker bit						1
	set_bits((unsigned int)PTS, 15);			// PTS [14..0]						15
	set_bits(1, 1);								// marker bit						1
}

static bool write_pack(FILE64* fout, __int64 time)
{
	unsigned char buf[64];
	set_buf(buf, 64);

	__int64 ext_time = time % 300;
	time = time / 300;

	set_bits(0x000001ba, 32);					// pack id								32
	set_bits(1, 2);								// 0x01									2
	set_bits((unsigned int)(time >> 30), 3);	// system_clock_reference_base			3
	set_bits(1, 1);								// marker_bit							1
	set_bits((unsigned int)(time >> 15), 15);	// system_clock_reference_base			15
	set_bits(1, 1);								// marker_bit							1
	set_bits((unsigned int)time, 15);			// system_clock_reference_base1			15
	set_bits(1, 1);								// marker_bit							1
	set_bits((unsigned int)ext_time, 9);		// system_clock_reference_extension		9
	set_bits(1, 1);								// marker_bit							1
	set_bits(DEMUX, 22);						// program_mux_rate						22
	set_bits(1, 1);								// marker_bit							1
	set_bits(1, 1);								// marker_bit							1
	set_bits(31, 5);							// reserved								5
	set_bits(0, 3);								// pack_stuffing_length					3

	return fwrite64(buf, buf_size(), 1, fout) == 1;
}

static bool write_system(FILE64* fout, 
	int videostream, int vidbscl, int vidbsize,
	int audiostream, int audbscl, int audbsize)
{
	unsigned char buf[64];
	set_buf(buf, 64);

	int len = 6;
	if (videostream != 0)
		len += 3;
	if (audiostream != 0)
		len += 3;

	set_bits(0x000001bb, 32);					// system id							32
	set_bits(len, 16);							// header_length						16
	set_bits(1, 1);								// marker_bit							1
	set_bits(DEMUX, 22);						// rate_bound							22
	set_bits(1, 1);								// marker_bit							1
	set_bits(1, 6);								// audio_bound							6
	set_bits(0, 1);								// fixed_flag							1
	set_bits(0, 1);								// CSPS_flag							1
	set_bits(1, 1);								// system_audio_lock_flag				1
	set_bits(1, 1);								// system_video_lock_flag				1
	set_bits(1, 1);								// marker_bit							1
	set_bits(1, 5);								// video_bound							5
	set_bits(0, 1);								// packet_rate_restriction_flag			1
	set_bits(127, 7);							// reserved_byte						7

	if (videostream != 0)
	{
		set_bits(videostream, 8);
		set_bits(0x3, 2);
		set_bits(vidbscl, 1);
		set_bits(vidbsize, 13);
	}

	if (audiostream != 0)
	{
		set_bits(audiostream, 8);
		set_bits(0x3, 2);
		set_bits(audbscl, 1);
		set_bits(audbsize, 13);
	}

	return fwrite64(buf, buf_size(), 1, fout) == 1;
}

static bool pad_buffer(FILE64* fout, int pad)
{
	pad -= 6;

	char buf[6];
	buf[0] = '\x0'; buf[1] = '\x0'; buf[2] = '\x1'; buf[3] = '\xbe';
	buf[4] = pad >> 8; buf[5] = pad & 0xff;

	if (fwrite64(buf, 6, 1, fout) != 1)
		return false;

	unsigned char padbyte = 0xff;
	for (int i = 0; i < pad; i++)
	{
		if (fwrite64(&padbyte, 1, 1, fout) != 1)
			return false;
	}

	return true;
}

int make_pes_header(unsigned char* buf, int streamid, int len, __int64 PTS, __int64 DTS)
{
	int hdrlen = 0;
	int PTS_DTS_flags = 0;
	if (PTS != -1)
	{
		if (DTS != -1)
		{
			PTS_DTS_flags = 3;
			hdrlen += 10;
		}
		else
		{
			PTS_DTS_flags = 2;
			hdrlen += 5;
		}
	}

	set_buf(buf, 9 + hdrlen);

	set_bits(0x000001, 24);						// packet_start_code_prefix				24
	set_bits((unsigned int)streamid, 8);		// directory_stream_id					8
	set_bits(len, 16);							// PES_packet_length					16
	set_bits(0x2, 2);							// '10'									2
	set_bits(0, 2);								// PES_scrambling_control				2
	set_bits(1, 1);								// PES_priority							1
	set_bits(0, 1);								// data_alignment_indicator				1
	set_bits(0, 1);								// copyright							1
	set_bits(0, 1);								// original_or_copy						1
	set_bits(PTS_DTS_flags, 2);					// PTS_DTS_flags						2
	set_bits(0, 1);								// ESCR_flag							1
	set_bits(0, 1);								// ES_rate_flag							1
	set_bits(0, 1);								// DSM_trick_mode_flag					1
	set_bits(0, 1);								// additional_copy_info_flag			1
	set_bits(0, 1);								// PES_CRC_flag							1
	set_bits(0, 1);								// PES_extension_flag					1
	set_bits(hdrlen, 8);						// PES_header_data_length				8
	
	if (PTS_DTS_flags == 2)
	{
		set_bits(2, 4);								// '0010'							4
		set_bits((unsigned int)(PTS >> 30), 3);		// PTS [32..30]						3
		set_bits(1, 1);								// marker bit						1
		set_bits((unsigned int)(PTS >> 15), 15);	// PTS [29..15]						15
		set_bits(1, 1);								// marker bit						1
		set_bits((unsigned int)PTS, 15);			// PTS [14..0]						15
		set_bits(1, 1);								// marker bit						1
	}
	else if (PTS_DTS_flags == 3)
	{
		set_bits(3, 4);								// '0011'							4
		set_bits((unsigned int)(PTS >> 30), 3);		// PTS [32..30]						3
		set_bits(1, 1);								// marker bit						1
		set_bits((unsigned int)(PTS >> 15), 15);	// PTS [29..15]						15
		set_bits(1, 1);								// marker bit						1
		set_bits((unsigned int)PTS, 15);			// PTS [14..0]						15
		set_bits(1, 1);								// marker bit						1
		set_bits(1, 4);								// '0001'							4
		set_bits((unsigned int)(DTS >> 30), 3);		// DTS [32..30]						3
		set_bits(1, 1);								// marker bit						1
		set_bits((unsigned int)(DTS >> 15), 15);	// DTS [29..15]						15
		set_bits(1, 1);								// marker bit						1
		set_bits((unsigned int)DTS, 15);			// DTS [14..0]						15
		set_bits(1, 1);								// marker bit						1
	}

	return buf_size();
}

static bool write_file_header(FILE64* fout, int vidstreamid, int audstreamid, int vidpid, int audpid)
{ 
	// Write the first pack header
	if (!write_pack(fout, 0))
	{
		error("Couldn't write system pack header!\n");
		return false;
	}

	// Write the system block
	if (audpid != -1)
	{
		if (!write_system(fout, vidstreamid, 1, 4096, audstreamid, 0, 58))	// 4096k for video packets 58 for audio??
		{
			error("Couldn't write system header!\n");
			return false;
		}
	}
	else
	{
		if (!write_system(fout, vidstreamid, 1, 4096, 0, 0, 0))		// 4096k for video packets
		{
			error("Couldn't write system header!\n");
			return false;
		}
	}

	// Pad out to even multiple of PACKETSIZE
	int beg_len = (int)ftell64(fout);
	if (!pad_buffer(fout, PACKETSIZE - beg_len))
	{
		error("Couldn't write pad buffer!\n");
		return false;
	}

	return true;
}

static int get_hex(char* str)
{
	if (strncmp(str, "0x", 2) != 0)
		return atol(str);

	int v = 0;	
	char *p = str + 2;
	while (*p && *p != ' ')
	{
		if (*p >= '0' && *p <= '9')
			v = (v << 4) | (*p - '0');
		else if (*p >= 'a' && *p <= 'f')
			v = (v << 4) | ((*p - 'a') + 10);
		else if (*p >= 'A' && *p <= 'F')
			v = (v << 4) | ((*p - 'A') + 10);
		p++;
	}

	return v;
}

bool inc_file_counter(char* filename)
{
	bool found_digits = false;
	char *p = filename + strlen(filename) - 1;
	while (p >= filename)
	{
		if ((*p >= '0') && (*p <= '9'))
		{
			found_digits = true;
			if (*p == '9')
				*p = '0';
			else
			{
				*p += 1;
				break;
			}
		}
		else if (found_digits)
			break;
		p--;
	}

	return found_digits;
}

// ------------------------------------------------------------------------------------
#if 1
extern payload_buff Stream_Header; // saved copy of stream_header
int convert_files_to_mpg(TSFileList *inputList, char *outputFilePattern,
		int vidPid, int audPid, int maxFileSize, FileReadBuffer *readBuffer,
		convert_callback callback, void *data, HWND parentWnd, CPtrArray *pPidInfoList )
{
	int rVal = 0;
	char errMsgBuf[256];
	int errMsgVal;
	long timeDiscrepancy = 0;

	__int64 filePos = 0;
	__int64 totalRead = 0;
	__int64 totalOut = 0;
	__int64 maxFileSize64 = (__int64)maxFileSize * (__int64)(1024 * 1024);
//	FILE64 *fIn = NULL;
	bool needToSeek = true;
	int counter = 0;

	TSFileInfo *fileInfo = inputList->GetHead();
	if (!fileInfo)
		return 1;

	TSRegionInfo *regionInfo = fileInfo->m_head;
	RegionType regionType = rt_include;

	// Open the destination file
	char outputFileName[256];
	strcpy(outputFileName, outputFilePattern);

#ifndef HDTVtoMPEG2
	bool overwrite = true;
#else
	bool overwrite = false;
#endif
	FILE64 *fOut = fopen64(outputFileName, "wb", WRBUFSIZE, overwrite);
	while (!fOut)
	{
#ifndef HDTVtoMPEG2
		printf("ERROR: Couldn't open output file.\n%s", outputFileName);
#else
		char *s = "Couldn't open output file\n%s";
		if (ferror64() == EEXIST)
			s = "File already exists!\n%s\nRetry and overwrite anyway?";
		sprintf(errMsgBuf, s, outputFileName);
		long tmp1, tmp2;
		time((time_t*)&tmp1);
		errMsgVal = ::MessageBox(parentWnd, errMsgBuf, "Warning!", MB_RETRYCANCEL | MB_ICONWARNING);
		time((time_t*)&tmp2);
		timeDiscrepancy += tmp2 - tmp1;
		if (errMsgVal == IDRETRY)
		{
			overwrite = (ferror64() == EEXIST);
			fOut = fopen64(outputFileName, "wb", WRBUFSIZE, overwrite);
		}
		else
#endif
			return 1;
	}

	// stream info
	bool vidFoundFirst = false;
	bool audFoundFirst = false;
	int vidPackets = 0;
	int audPackets = 0;
	int vidStreamId = 0xE0;	// MPEG2 video stream id
	int audStreamId; // AC3, MPEG-2 stream id. AC-3 is on 0xBD, MPEG-2 is on 0xC0

	// get the pid type for the audio pid
	pidType PIDType = PID_UNKNOWN;
	Pid_Info *pPidInfo;
	for( int i = 0; i < pPidInfoList->GetCount(); ++i )
	{
		pPidInfo = (Pid_Info*)pPidInfoList->GetAt( i );
		if( pPidInfo->original_pid == audPid )
		{
			PIDType = pPidInfo->type;
			break;
		}
	}

	ASSERT( PIDType != PID_UNKNOWN );

	PIDinfoGroup PIDs = { 2,
					{	{ vidPid, 0, STREAM_PES, PID_MPEG2_VIDEO, false, NULL, 0, 0, -1, false, false },
						{ audPid, 0, STREAM_PES, PIDType, false, NULL, 0, 0, -1, false, false }	}};

	PIDs.pids[0].buffer = new uchar[VIDEO_BUFFER_SIZE];
	PIDs.pids[0].bufferSize = VIDEO_BUFFER_SIZE;
	PIDs.pids[1].buffer = new uchar[AUDIO_BUFFER_SIZE];
	PIDs.pids[1].bufferSize = AUDIO_BUFFER_SIZE;

	// set the audio stream id
	if( PIDType == PID_AC3_AUDIO )
		audStreamId = 0xBD; // AC-3 carried on private_stream_1 code
	if( PIDType == PID_MPEG2_AUDIO || PIDType == PID_MPEG1_AUDIO ) // don't know if this will work for MPEG-1 audio
		audStreamId = 0xC0; // carried on ISO 13818-3 or ISO 11172-3 audio code

	ASSERT( PIDType == PID_AC3_AUDIO ||
		    PIDType == PID_MPEG2_AUDIO ||
			PIDType == PID_MPEG1_AUDIO );

	if (!PIDs.pids[0].buffer || !PIDs.pids[1].buffer)
	{
		if (PIDs.pids[0].buffer)
			delete PIDs.pids[0].buffer;
		if (PIDs.pids[1].buffer)
			delete PIDs.pids[1].buffer;
		if (fOut)
			fclose64(fOut);

		::MessageBox(parentWnd, "Out of memory", NULL, MB_OK | MB_ICONERROR);
		return 1;
	}

	// Set the starting time value
	__int64 packetTime = 0;
	__int64 fileTime = 0;
	__int64 PCR = -1;
	__int64 firstVideoPCR;
	bool doingIFrame;
	int frame = 0;

	// Write the first pack header
	if (!write_file_header(fOut, vidStreamId, audStreamId, vidPid, audPid))
	{
		if (PIDs.pids[0].buffer)
			delete PIDs.pids[0].buffer;
		if (PIDs.pids[1].buffer)
			delete PIDs.pids[1].buffer;
		if (fOut)
			fclose64(fOut);

#ifndef HDTVtoMPEG2
		printf("ERROR: Couldn't write file header to output file\n%s", outputFileName);
#else
		sprintf(errMsgBuf, "Couldn't write file header to output file\n%s", outputFileName);
		errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
		return 1;
	}

#ifndef HDTVtoMPEG2
	// Write out 'processing'..
	printf("Converting file %s to file %s\n"
		"Video stream 0x%x Audio stream 0x%x\n"
		"Please wait", ifilename, outputFileName, vidPid, audPid);
#endif

	// Here's the deal.  We simply read each packet in the transport file, decide if it is one of the PID's
	// we're interested in, detect the 'start' bit and 'random access' bit in the header, then once we find
	// one, simply add the payloads of all subsequent packets with that id into a buffer.  Once we get the
	// next packet for a PID with a 'start bit' for either stream, we dump the previous packet, to the 
	// MPEG2 program file with a pack header, 2048 bytes at a time, then pad the final pack to 2048 bytes.

	// We have to get the PCR timestamps from the transport stream and use them to set the timestamps for
	// the pack headers in the MPEG stream.  We just do a one to one copy when we find one, and use the 
	// number of bytes read times the demux rate to do any pack header that doesn't have a timestamp.

	// As for AC3, we have to add that little 4 byte substreamid, AC3 id value before the AC3 payload for
	// each MPEG2 AC3 2048 byte packet.  For MPEG-2 audio, no substream id (a descriptor actually) is needed
	// MPEG-2 audio is carried on stream 0xC0.

	// For demux rate, we just use the ATSC 19.2Mbps rate, since that's pretty close to what the file is.
	// For the decode buffer sizes, we err on the side of caution.  Perhaps they're too big, but I'm not
	// sure.

	// Finally, once we get to the end of the file, we add the MPEG stop packet id, and we're done.
	// Whahla.. we have a nice DVD compatible super high definition standard MPEG video file, all set for
	// compression into DivX or what-have-you.

	unsigned char buf[188];
	int bytesReadInPacket = 0;
	for (;;)
	{
		counter++;

		// Call callback
		if (((counter & 0x3f) == 0) && callback && fileInfo)
		{
			bool cont = callback(fileInfo->m_fileName, outputFileName,
					fileInfo->m_totalBytesPrior + filePos,
					totalRead, inputList->TotalIncludedBytes(),
					totalOut, timeDiscrepancy, data);
			timeDiscrepancy = 0;

			// Cancelled?
			if (!cont)
				break;
		}

		// check regions, skip data, open new files
		if (regionInfo && (filePos >= regionInfo->m_offset))
		{
			regionType = regionInfo->m_type;
			regionInfo = regionInfo->m_next;
		}
		if (regionType == rt_exclude)
		{
			if (regionInfo)
			{
				filePos = regionInfo->m_offset;
			}
			else
			{
				if (readBuffer->IsFileOpen())
//				if (fIn)
				{
					readBuffer->Close();
//					fclose64(fIn);
//					fIn = NULL;
				}
				fileInfo = fileInfo->m_next;
				if (!fileInfo)
					break;	// we're complete!
				regionInfo = fileInfo->m_head;
				filePos = 0;
			}
			PIDs.pids[0].reading = false;		// discard any packets in progress!
			PIDs.pids[1].reading = false;
			needToSeek = true;
			continue;
		}
		else
		{
			if (!readBuffer->IsFileOpen())
//			if (!fIn)
			{
				readBuffer->Open(fileInfo->m_fileName);
//				fIn = fopen64(fileInfo->m_fileName, "rb", RDBUFSIZE);
				while (!readBuffer->IsFileOpen())
//				while (!fIn)
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't open input file\n%s", fileInfo->m_fileName);
#else
					sprintf(errMsgBuf, "Couldn't open input file\n%s", fileInfo->m_fileName);
					long tmp1, tmp2;
					time((time_t*)&tmp1);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_RETRYCANCEL | MB_ICONWARNING);
					time((time_t*)&tmp2);
					timeDiscrepancy += tmp2 - tmp1;
					if (errMsgVal == IDRETRY)
					{
						readBuffer->Open(fileInfo->m_fileName);
//						fIn = fopen64(fileInfo->m_fileName, "rb", RDBUFSIZE);
					}
					else
#endif
					{
						rVal = 1;
						break;	// where is this file??
					}
				}
				if (rVal)
					break;
				needToSeek = true;
			}
			if (needToSeek)
			{
				// check packet alignment of new file
				__int64 offset = (188 - ((filePos + 188 - fileInfo->m_firstPacketOffset) % 188)) % 188;
				if (offset != ((188 - bytesReadInPacket) % 188))
				{
					// alignment does not match, discard partial packet
					bytesReadInPacket = 0;
					filePos += offset;
				}
				readBuffer->Seek(filePos);
//				fseek64(fIn, filePos, SEEK_SET);
				needToSeek = false;
			}
		}

		// Try to read packet..
		int bytesRead;
		if ((bytesRead = readBuffer->Read(buf + bytesReadInPacket, 188 - bytesReadInPacket)) != 188 - bytesReadInPacket)
//		if ((bytesRead = fread64(buf+bytesReadInPacket, 1, 188 - bytesReadInPacket, fIn)) != 188 - bytesReadInPacket)
		{
			if (bytesRead < 0)
				bytesRead = 0;
			bytesReadInPacket += bytesRead;
			totalRead += bytesRead;

			// Close this file
			readBuffer->Close();
//			fclose64(fIn);
//			fIn = NULL;
			fileInfo = fileInfo->m_next;
			if (!fileInfo)
				break;	// we're complete!
			regionInfo = fileInfo->m_head;
			filePos = 0;

			continue;
		}
		else
		{
			filePos += bytesRead;
			bytesReadInPacket = 0;
			totalRead += bytesRead;
		}

		// Check output for max filesize (Only cut files on I-Frames -- i.e. video packets containing Group of Picture id)
		if (ftell64(fOut) >= maxFileSize64 && doingIFrame)
		{
			// Mark last valid stream pos
			unsigned char end[4] = { 0, 0, 1, 0xb9 };
			fwrite64(&end, 4, 1, fOut);
			totalOut += 4;

			// Close files
			fclose64(fOut);
			fOut = NULL;

			// Try next file?
			if (inc_file_counter(outputFileName))
			{
				// Open the destination file
#ifndef HDTVtoMPEG2
				overwrite = true;
#else
				overwrite = false;
#endif
				fOut = fopen64(outputFileName, "wb", WRBUFSIZE, overwrite);
				while (!fOut)
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't open output file\n%s", outputFileName);
#else
					char *s = "Couldn't open output file\n%s";
					if (ferror64() == EEXIST)
						s = "File already exists\n%s\nRetry and overwrite anyway?";
					sprintf(errMsgBuf, s, outputFileName);
					long tmp1, tmp2;
					time((time_t*)&tmp1);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_RETRYCANCEL | MB_ICONWARNING);
					time((time_t*)&tmp2);
					timeDiscrepancy += tmp2 - tmp1;
					if (errMsgVal == IDRETRY)
					{
						overwrite = (ferror64() == EEXIST);
						fOut = fopen64(outputFileName, "wb", WRBUFSIZE, overwrite);
					}
					else
#endif
					{
						rVal = 1;
						break;
					}
				}
				if (rVal)
					break;

				// Set the starting time value
				PCR = -1;
				frame = 0;
				fileTime = totalRead * SYSTEMCLOCKRATE / STREAMRATE;

				// Write the first pack header
				if (!write_file_header(fOut, vidStreamId, audStreamId, vidPid, audPid))
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't write file header to output file\n%s", outputFileName);
#else
					sprintf(errMsgBuf, "Couldn't write file header to output file\n%s", outputFileName);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
					rVal = 1;
					break;
				}

#ifndef HDTVtoMPEG2
				printf("\nWriting next output file %s\n", outputFileName);
				printf("Please Wait");
#endif
			}
			else
				break;
		}

		// Check sync byte
		if ((buf[0] != 0x47) && (buf[0] != 0x72) && (buf[0] != 0x29))
		{
			warning("Bad transport packet (no sync byte 0x47)!\n");
			PIDs.pids[0].reading = false;		// discard any packets in progress!
			PIDs.pids[1].reading = false;
			continue;
		}

		// Get pid
		int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;

		PIDinfo *p = find_PID_no_new(&PIDs, pid);
		if (!p)
			continue;

		p->count++;

		// Get error
		bool errorBit = (buf[1] & 0x80) != 0;
		if (errorBit)
		{
			warning("Error bit set in packet\n");
			// invalidate accumulated content of this stream
			p->reading = false;
			continue;
		}

		// Get adaptation header info
		int adaptation = (buf[3] & 0x30) >> 4;
		int adaptLength = 0;
		if (adaptation == 0)
		{
			warning("Bad adaption code (code was 0)!\n");
			adaptLength = -1;	// bad!
		}
		else if (adaptation == 0x2)
			adaptLength = 184;
		else if (adaptation == 0x3)
			adaptLength = buf[4] + 1;
		if (adaptLength > 184)
		{
			warning("Invalid adapt len (was > 183)!\n");
			adaptLength = -1;	// too long!
		}

		if (adaptLength < 0)
		{
			PIDs.pids[0].reading = false;		// discard any packets in progress!
			PIDs.pids[1].reading = false;
			continue;
		}

		// Get start code
		bool start = (buf[1] & 0x40) != 0;

		if (start)
		{
			// Get PCR
			// if adaptation flag and PCR flag are set
			if ((adaptation & 0x2) && (buf[5] & 0x10))
			{
				__int64 PCRBase = ((__int64)buf[6] << 25) | ((__int64)buf[7] << 17) | 
					  ((__int64)buf[8] << 9) | ((__int64)buf[9] << 1) | ((__int64)buf[10] >> 7);
				__int64 PCRExt = ((__int64)(buf[10] & 0x1) << 8) | ((__int64)buf[11]);
				PCR = PCRBase * 300 + PCRExt;
			}

			if (p->reading && (p->bufferPos > 0))
			{
				unsigned char ac3SubstreamId[4];
				int ac3Len = 0;

				switch( p->pType )
				{
					case PID_MPEG2_VIDEO:
	#ifdef HDTVtoMPEG2
						// decode I frames
						if( doingIFrame )
							process_PES_video_packet( p->buffer, p->bufferPos, true );

						break;
	#endif
					case PID_AC3_AUDIO:
					{
						// Make a four byte ac3 streamid
						ac3SubstreamId[0] = 0x80;	// Four byte AC3 CODE??
						ac3SubstreamId[1] = 0x01;
						ac3SubstreamId[2] = 0x00;	// WHY???  OH WHY??
						ac3SubstreamId[3] = 0x02;
						ac3Len = 4;
					}
					break;

					//case PID_MPEG2_AUDIO:
					//case PID_MPEG1_AUDIO:
					// Note: for mpeg audio, we do nothing.  there is no descriptor(streamid) required because 
					//        the audio is not carried on a private_stream like AC-3 audio is
				}// end switch( p->pType )

				if (p->reading)
				{
					// Write a 2048 byte program stream packet..
					int written = 0;	// Bytes we've written to output file
					int pos = 0;		// Position in PES packet buffer
					for (;;)
					{
						__int64 fPos = ftell64(fOut);
						if ((fPos % (__int64)PACKETSIZE) != 0)
						{
#ifndef HDTVtoMPEG2
							printf("ERROR: Packets not falling on packetsize boundries";
#else
							errMsgVal = ::MessageBox(parentWnd, "Packets not falling on packetsize boundries", NULL, MB_OK | MB_ICONERROR);
#endif
							rVal = 1;
							break;
						}

						// Get total length of this pack
						int len = min(14 + ac3Len + p->bufferPos - pos, PACKETSIZE);

						// Figure out stuffing (if we have less than 16 bytes left)
						int stuffing = 0;
						if ((len < PACKETSIZE) && (PACKETSIZE - len < 16))
						{
							stuffing = PACKETSIZE - len;
							len += stuffing;
						}

						// Write out pack header
						packetTime = (totalRead * SYSTEMCLOCKRATE / STREAMRATE) - fileTime;
						if (!write_pack(fOut, packetTime))
						{
#ifndef HDTVtoMPEG2
							printf("ERROR: Couldn't write pack header to output file\n%s", outputFileName);
#else
							sprintf(errMsgBuf, "Couldn't write pack header to output file\n%s", outputFileName);
							errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
							rVal = 1;
							break;
						}

						switch (p->pType)
						{
							case PID_MPEG2_VIDEO:
								p->buffer[pos + 3] = vidStreamId;
								break;

							case PID_AC3_AUDIO:
							case PID_MPEG2_AUDIO:
							case PID_MPEG1_AUDIO: // not sure this will work for mpeg-1 audio
								p->buffer[pos + 3] = audStreamId;
								break;
						}

						// Packet length..
						// Subtract pack size (14) and pes id and len (6) from lenth
						p->buffer[pos + 4] = (len - 6 - 14) >> 8;
						p->buffer[pos + 5] = (len - 6 - 14) & 0xFF;

						// Add any stuffing bytes to header extra len
						int hdrSize = 9 + p->buffer[pos + 8];
						p->buffer[pos + 8] += stuffing;					// Add stuffing to header bytes

						// Write out id, sreamid, len
						if (fwrite64(p->buffer + pos, hdrSize, 1, fOut) != 1)	// Write pes id, streamid, and len
						{
#ifndef HDTVtoMPEG2
							printf("ERROR: Couldn't write to output file\n%s", outputFileName);
#else
							sprintf(errMsgBuf, "Couldn't write to output file\n%s", outputFileName);
							errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
							rVal = 1;
							break;
						}
						totalOut += hdrSize;

						// Write stuffing
						for (int i = 0; i < stuffing; i++)				// Write any stuffing bytes
						{
							unsigned char stuff = 0xff;
							if (fwrite64(&stuff, 1, 1, fOut) != 1)
							{
#ifndef HDTVtoMPEG2
								printf("ERROR: Couldn't write to output file\n%s", outputFileName);
#else
								sprintf(errMsgBuf, "Couldn't write to output file\n%s", outputFileName);
								errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
								rVal = 1;
								break;
							}
						}
						if (rVal)
							break;
						totalOut += stuffing;

						// Write ac3 streamid
						if (ac3Len != 0)
						{
							if (fwrite64(ac3SubstreamId, ac3Len, 1, fOut) != 1)
							{
#ifndef HDTVtoMPEG2
								printf("ERROR: Couldn't write to output file\n%s", outputFileName);
#else
								sprintf(errMsgBuf, "Couldn't write to output file\n%s", outputFileName);
								errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
								rVal = 1;
								break;
							}
							totalOut += ac3Len;
						}

						// Write rest of data len minus headersize (9) stuffing, and pack size (14)
						if (fwrite64(p->buffer + pos + hdrSize, len - hdrSize - 14 - stuffing - ac3Len, 1, fOut) != 1)	// Write data bytes
						{
#ifndef HDTVtoMPEG2
							printf("ERROR: Couldn't write to output file\n%s", outputFileName);
#else
							sprintf(errMsgBuf, "Couldn't write to output file\n%s", outputFileName);
							errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
							rVal = 1;
							break;
						}
						written += len;
						totalOut += len;

						// Add len minus stuff we added like the pack (14) and the stuffing.
						pos += len - 14 - stuffing - ac3Len;
						if (pos == p->bufferPos)
							break;

						// Add pes header for next packet
						pos -= 9;
						make_pes_header(p->buffer + pos, (p->pType == PID_MPEG2_VIDEO ? vidStreamId : audStreamId), 0, -1, -1);
					}
					if (rVal > 0)
						break;

					// Write padding
					if ((written % PACKETSIZE) != 0)
					{
						int left = PACKETSIZE - (written % PACKETSIZE);

						// Pad out to PACKETSIZE bytes
						if (!pad_buffer(fOut, left))
						{
#ifndef HDTVtoMPEG2
							printf("ERROR: Couldn't write to output file\n%s", outputFileName);
#else
							sprintf(errMsgBuf, "Couldn't write to output file\n%s", outputFileName);
							errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
							rVal = 1;
							break;
						}
						totalOut += left;
					}

					if (p->pType == PID_MPEG2_VIDEO)
					{
						vidPackets++;
#ifndef HDTVtoMPEG2
						// Print out a "." every 30 frames
						if (vidPackets % 30 == 0)
								printf(".");
#endif
					}
					else
						audPackets++;
				}// end if( p->reading )
			}// end if( p->reading && ( p->bufferPos > 0 ))
			// start a new PES packet
			if (p->buffer && !p->processed)
			{
				p->reading = true;
				p->bufferPos = 0;
				p->confirmedRead = false;
			}
		}// end if(start)
		if (p->reading)
		{
			// Add the payload for this packet to the current buffer
			int payloadLength = 184 - adaptLength;
			if (payloadLength > 0)
			{
				if ((p->bufferPos + payloadLength) <= p->bufferSize)
				{
					memcpy(p->buffer + p->bufferPos, buf + 4 + adaptLength, payloadLength);
					p->bufferPos += payloadLength;
				}
				else
				{
					// buffer overflow!
//					_Error("Error: buffer overflow! PID=0x%02X, PES packet, buffer size = %d\n", pid, p->bufferSize);
					p->reading = false;
				}
			}

			if (!p->confirmedRead && (p->bufferPos > 512))
			{
				switch (p->pType)
				{
					case PID_MPEG2_VIDEO:
					{
						// only want to continue reading this PES packet if it's and i-frame
						// check for a sequence_header_code, if we have it were good to go
						// check for a group_start_code, if we have it were good to go
						// check for a picture_start_code, if it's an i-frame, add a
						//  stream_header to the beginning of the buffer
						doingIFrame = false;
						uint start_code = 0;
						for( int i = 0; i < p->bufferPos; i++ )
						{
							start_code = ( start_code << 8 ) | p->buffer[i];
							// sequence_header_code || group_start_code
							if( start_code == 0x000001b3 || start_code == 0x000001b8 ) 
							{  
								// this is all we need to continue
								doingIFrame = true;
								break;
							}

							// picture_start_code
							if( start_code == 0x00000100 ) 
							{
								// check if picture_coding_type == 1
								if(( p->buffer[i + 2] & ( 0x7 << 3 )) == ( 1 << 3 ))
								{
									// add a sequence_header to the beginning of the buffer
									if( Stream_Header.buffer )
									{
										unsigned char *pTemp = new unsigned char[ p->bufferPos ];
										memcpy( pTemp, p->buffer, p->bufferPos );
										memcpy( p->buffer, Stream_Header.buffer, Stream_Header.bytes );
										memcpy( p->buffer + Stream_Header.bytes, pTemp, p->bufferPos );
										p->bufferPos += Stream_Header.bytes;

										delete pTemp;

										// i-frame found
										doingIFrame = true;
										break;
									}
									else // can't use this i-frame, no sequence_header found yet
										break;
								}// end if(( p->buffer[i + 2] & ( 0x7 << 3 )) == ( 1 << 3 ))
							}// end if( start_code == 0x00000100 )
						}// end for( int i = 0; i < p->bufferPos; i++ )

						if( doingIFrame )
						{
							if( !vidFoundFirst )
							{
								vidFoundFirst = true;
								firstVideoPCR = PCR;
							}
						}
						if( vidFoundFirst || doingIFrame )
							p->confirmedRead = true;
						if( !p->confirmedRead )
							p->reading = false;
					}
					break;

					case PID_AC3_AUDIO:
					case PID_MPEG2_AUDIO:
					{
						// set audio found first ONLY after first video frame found...
						if( !audFoundFirst && vidFoundFirst )
							audFoundFirst = true;
						if( !audFoundFirst )
							p->confirmedRead = false;
					}

				}// end switch (p->pType)
			}// end if (!p->confirmedRead && (p->bufferPos > 512))
		}// end if (p->reading)
	}// end for (;;)

	if (rVal == 0)
	{
		// Mark last valid stream pos
		unsigned char end[4] = {0,0,1,0xb9};
		fwrite64(&end, 4, 1, fOut);
		totalOut += 4;
	}

	// Close files
	if (readBuffer->IsFileOpen())
		readBuffer->Close();
//	if (fIn)
//		fclose64(fIn);
	if (fOut)
		fclose64(fOut);

	// Delete video/audio buffers
	if (PIDs.pids[0].buffer)
		delete PIDs.pids[0].buffer;
	if (PIDs.pids[1].buffer)
		delete PIDs.pids[1].buffer;

#ifndef HDTVtoMPEG2
	// All done!
	printf("\n\nNumber of video packets %d audio packets %d\nFinished\n", vidPackets, audPackets);
#endif
	
	return 0;
}

// ------------------------------------------------------------------------------------
//#ifdef VERSION_THAT_CANT_SCAN_MORE_THAN_1_PACKET_FOR_I_FRAMES
#else

#define curbuf (packetbuf[curstream])
#define curpos (packetpos[curstream])
#define curpackets (bufpackets[curstream])
#define curfoundfirst (foundfirst[curstream])
#define curskipbad (skipbad[curstream])
#define curcontinuity (streamcont[curstream])
#define curstreamid (streamid[curstream])

#define vidbuf (packetbuf[0])
#define vidpos (packetpos[0])
#define vidpackets (bufpackets[0])
#define vidfoundfirst (foundfirst[0])
#define vidskipbad (skipbad[0])
#define vidcontinuity (streamcont[0])
#define vidstreamid (streamid[0])

#define audbuf (packetbuf[1])
#define audpos (packetpos[1])
#define audpackets (bufpackets[1])
#define audfoundfirst (foundfirst[1])
#define audskipbad (skipbad[1])
#define audcontinuity (streamcont[1])
#define audstreamid (streamid[1])

int convert_files_to_mpg(TSFileList *ifilelist, char* ofilepattern, 
		int vidpid, int audpid, int maxfilesize, FileReadBuffer *, convert_callback callback, void* data, HWND parentWnd)
{
	char errMsgBuf[256];
	int errMsgVal;

	__int64 file_pos = 0;
	__int64 total_read = 0;
	__int64 total_out = 0;
	__int64 maxfilesize64 = (__int64)maxfilesize * (__int64)(1024 * 1024);
	FILE64 *fin = NULL;
	bool need_to_seek = true;
	int counter = 0;

	// Packet arrays
	unsigned char* packetbuf[2];
	int packetpos[2];
	int bufpackets[2];
	bool foundfirst[2];
	bool skipbad[2];
	int streamcont[2];
	int streamid[2];

	TSFileInfo *fileInfo = ifilelist->GetHead();
	if (!fileInfo)
		return 1;

	TSRegionInfo *region_info = fileInfo->m_head;
	RegionType region_type = rt_include;

	// Open the destination file
	char ofilename[256];
	strcpy(ofilename, ofilepattern);
	FILE64 *fout = fopen64(ofilename, "wb", WRBUFSIZE);
	while (!fout)
	{
#ifndef HDTVtoMPEG2
		printf("ERROR: Couldn't open output file\n%s", ofilename);
#else
		sprintf(errMsgBuf, "Couldn't open output file\n%s", ofilename);
		errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_RETRYCANCEL | MB_ICONWARNING);
		if (errMsgVal == IDRETRY)
		{
			fout = fopen64(ofilename, "wb", WRBUFSIZE);
		}
		else
#endif
			return 1;
	}

	// Get the video PID
	vidbuf = new unsigned char[1024 * 1024];
	vidpos = 0;
	vidfoundfirst = 0;
	vidskipbad = false;
	vidcontinuity = -1;
	vidpackets = 0;
	vidstreamid = 0xE0;	// MPEG2 video stream id

	// Get the audio PID
	if (audpid <= 0)
		audpid = -1;
	audbuf = new unsigned char[1024 * 1024];
	audpos = 0;
	audfoundfirst = false;
	audskipbad = false;
	audcontinuity = -1;
	audpackets = 0;
	audstreamid = 0xBD;	// AC3, etc. streamid

	// Set the starting time value
	__int64 packet_time = 0;
	__int64 file_time = 0;
	__int64 PCR = -1;
	__int64 first_video_PCR;
	bool doing_iframe;
	bool was_doing_iframe = false;
	int frame = 0;

	// Write the first pack header
	if (!write_file_header(fout, vidstreamid, audstreamid, vidpid, audpid))
	{
#ifndef HDTVtoMPEG2
		printf("ERROR: Couldn't write file header to output file\n%s", ofilename);
#else
		sprintf(errMsgBuf, "Couldn't write file header to output file\n%s", ofilename);
		errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
		fclose64(fout);
		delete vidbuf;
		delete audbuf;
		return 1;
	}

#ifndef HDTVtoMPEG2
	// Write out 'processing'..
	printf("Converting file %s to file %s\n"
		"Video stream 0x%x Audio stream 0x%x\n"
		"Please wait", ifilename, ofilename, vidpid, audpid);
#endif

	unsigned char buf[188];
	int curstream = 0;

	// Here's the deal.  We simply read each packet in the transport file, decide if it is one of the PID's
	// we're interested in, detect the 'start' bit and 'random access' bit in the header, then once we find
	// one, simply add the payloads of all subsequent packets with that id into a buffer.  Once we get the
	// next packet for a PID with a 'start bit' for either stream, we dump the previous packet, to the 
	// MPEG2 program file with a pack header, 2048 bytes at a time, then pad the final pack to 2048 bytes.

	// We have to get the PCR timestamps from the transport stream and use them to set the timestamps for
	// the pack headers in the MPEG stream.  We just do a one to one copy when we find one, and use the 
	// number of bytes read times the demux rate to do any pack header that doesn't have a timestamp.

	// As for AC3, we have to add that little 4 byte substreamid, AC3 id value before the AC3 payload for
	// each MPEG2 AC3 2048 byte packet.

	// For demux rate, we just use the ATSC 19.2Mbps rate, since that's pretty close to what the file is.
	// For the decode buffer sizes, we err on the side of caution.  Perhaps they're too big, but I'm not
	// sure.

	// Finally, once we get to the end of the file, we add the MPEG stop packet id, and we're done.
	// Whahla.. we have a nice DVD compatible super high definition standard MPEG video file, all set for
	// compression into DivX or what-have-you.

	int bytesReadInPacket = 0;
	for (;;)
	{
		counter++;

		// Call callback
		if (((counter & 0x3f) == 0) && callback)
		{
			bool cont = callback(fileInfo->m_fileName, ofilename,
					fileInfo->m_totalBytesPrior + file_pos,
					total_read, ifilelist->TotalIncludedBytes(),
					total_out, 0, data);

			// Cancelled?
			if (!cont)
				break;
		}

		// check regions, skip data, open new files
		if (region_info && (file_pos >= region_info->m_offset))
		{
			region_type = region_info->m_type;
			region_info = region_info->m_next;
		}
		if (region_type == rt_exclude)
		{
			if (region_info)
			{
				file_pos = region_info->m_offset;
			}
			else
			{
				if (fin)
				{
					fclose64(fin);
					fin = NULL;
				}
				fileInfo = fileInfo->m_next;
				if (!fileInfo)
					break;	// we're complete!
				region_info = fileInfo->m_head;
				file_pos = 0;
			}
			audskipbad = vidskipbad = true;		// discard any packets in progress!
			need_to_seek = true;
			continue;
		}
		else
		{
			if (!fin)
			{
				fin = fopen64(fileInfo->m_fileName, "rb", RDBUFSIZE);
				while (!fin)
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't open input file\n%s", fileInfo->m_fileName);
#else
					sprintf(errMsgBuf, "Couldn't open input file\n%s", fileInfo->m_fileName);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_RETRYCANCEL | MB_ICONWARNING);
					if (errMsgVal == IDRETRY)
					{
						fin = fopen64(fileInfo->m_fileName, "rb", RDBUFSIZE);
					}
					else
#endif
					{
						if (fout)
							fclose64(fout);
						delete vidbuf;
						delete audbuf;
						return 1;	// where is this file??
					}
				}
				need_to_seek = true;
			}
			if (need_to_seek)
			{
				// check packet alignment of new file
				__int64 offset = (188 - ((file_pos + 188 - fileInfo->m_firstPacketOffset) % 188)) % 188;
				if (offset != ((188 - bytesReadInPacket) % 188))
				{
					// alignment does not match, discard partial packet
					bytesReadInPacket = 0;
					file_pos += offset;
				}
				fseek64(fin, file_pos, SEEK_SET);
				need_to_seek = false;
			}
		}

		// Try to read packet..
		int bytesRead;
		if ((bytesRead = fread64(buf+bytesReadInPacket, 1, 188 - bytesReadInPacket, fin)) != 188 - bytesReadInPacket)
		{
			if (bytesRead < 0)
				bytesRead = 0;
			bytesReadInPacket += bytesRead;
			total_read += bytesRead;

			// Close this file
			fclose64(fin);
			fin = NULL;
			fileInfo = fileInfo->m_next;
			if (!fileInfo)
				break;	// we're complete!
			region_info = fileInfo->m_head;
			file_pos = 0;

			continue;
		}
		else
		{
			file_pos += bytesRead;
			bytesReadInPacket = 0;
			total_read += bytesRead;
		}

		// Check output for max filesize (Only cut files on I-Frames -- i.e. video packets containing Group of Picture id)
		if (ftell64(fout) >= maxfilesize64 && doing_iframe)
		{
			// Mark last valid stream pos
			unsigned char end[4] = {0,0,1,0xb9};
			fwrite64(&end, 4, 1, fout);
			total_out += 4;

			// Close files
			fclose64(fout);
			fout = NULL;

			// Try next file?
			if (inc_file_counter(ofilename))
			{
				// Open the destination file
				fout = fopen64(ofilename, "wb", WRBUFSIZE);
				while (!fout)
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't open output file\n%s", ofilename);
#else
					sprintf(errMsgBuf, "Couldn't open output file\n%s", ofilename);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_RETRYCANCEL | MB_ICONWARNING);
					if (errMsgVal == IDRETRY)
					{
						fout = fopen64(ofilename, "wb", WRBUFSIZE);
					}
					else
#endif
					{
						fclose64(fin);
						delete vidbuf;
						delete audbuf;
						return 1;
					}
				}

				// Set the starting time value
				PCR = -1;
				frame = 0;
				file_time = total_read * SYSTEMCLOCKRATE / STREAMRATE;

				// Write the first pack header
				if (!write_file_header(fout, vidstreamid, audstreamid, vidpid, audpid))
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't write file header to output file\n%s", ofilename);
#else
					sprintf(errMsgBuf, "Couldn't write file header to output file\n%s", ofilename);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
					fclose64(fin);
					fclose64(fout);
					delete vidbuf;
					delete audbuf;
					return 1;
				}

#ifndef HDTVtoMPEG2
				printf("\nWriting next output file %s\n", ofilename);
				printf("Please Wait");
#endif
			}
			else
				break;
		}

		// Check sync byte
		if ((buf[0] != 0x47) && (buf[0] != 0x72) && (buf[0] != 0x29))
		{
			warning("Bad transport packet (no sync byte 0x47)!\n");
			audskipbad = vidskipbad = true;
			continue;
		}

		// Get pid
		int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;

		// Skip this block
		if (pid != audpid && pid != vidpid)
			continue;

		// Get the pos and buf
		if (pid == vidpid)
			curstream = 0;
		else
			curstream = 1;

		// Get start code
		bool start;
		start = (buf[1] & 0x40) != 0;
		if (!start && curskipbad)
			continue;

		// Get error
		bool errorbit = (buf[1] & 0x80) != 0;
		if (errorbit)
		{
			warning("Error bit set in packet\n");
			curskipbad = true;
			continue;
		}

		// Get continuity
		int continuity = (buf[3] & 0xF);
		if (curcontinuity != -1)
		{
			if (continuity != ((curcontinuity + 1) & 0xF))
			{
				warning("Bad continuity code in packet\n");
				curskipbad = true;
				continue;
			}
			curcontinuity = continuity;
		}
			
		// Get adaption header info
		int adaption = (buf[3] & 0x30) >> 4;
		int adapt_len = 0;

		// Get adaption header size
		if (adaption == 0)
		{
			warning("Bad adaption code (code was 0)!\n");
			audskipbad = vidskipbad = true;
			continue;
		}
		else if (adaption == 0x2)
			adapt_len = 184;
		else if (adaption == 0x3)
		{
			adapt_len = buf[4] + 1;
			if (adapt_len > 184)
			{
				warning("Invalid adapt len (was > 183)!\n");
				audskipbad = vidskipbad = true;
			}
		}

		// HBO is slick, it doesn't bother to sync AC3 packets with PES elementary stream packets.. so
		// we have to swizzle them together!  (ARGHH!)
		if (pid == audpid && start)
		{
			// Is there an AC3 packet start 0b77 code in this packet??
			bool sync_found = false;
			unsigned char *p = buf + 4 + adapt_len;
			while (p <= buf + 186)
			{
				if (p[0] == 0x0b && p[1] == 0x77)
				{
					sync_found = true;
					break;
				}
				p++;
			}

			// Couldn't find an AC3 sync start in this packet.. don't make a PES packet!
			if (!sync_found)
			{
//					int pos = ftell(fin);
//					error("AC3 packet sync not found in start frame");
//					return 1;
				adapt_len += 9 + buf[4 + adapt_len + 8];	
				start = false;
			}
		}

		// Get PCR
		if (start && (adaption & 0x2) && (buf[5] & 0x10))
		{
			__int64 PCR_base = ((__int64)buf[6] << 25) | ((__int64)buf[7] << 17) | 
				  ((__int64)buf[8] << 9) | ((__int64)buf[9] << 1) | ((__int64)buf[10] >> 7);
			__int64 PCR_ext = ((__int64)(buf[10] & 0x1) << 8) | ((__int64)buf[11]);
			PCR = PCR_base * 300 + PCR_ext;
		}

		// Get random
		bool random = start;	// Oops.. random bit seems to be used only from some broadcasters???
//		bool random = false;
//		if (start && (adaption & 0x2))
//			random = (buf[5] & 0x40) != 0;		// BUG: SOME TS STREAMS DON'T HAVE THE RANDOM BIT (ABC!! ALIAS)

		// Found a random access point (now we can start a frame/audio packet..)
		if (random)
		{
			// Check to see if this is an i_frame (group of picture start)
			if (pid == vidpid)
			{
				// Look for the Group of Pictures packet.. indicates this is an I-Frame packet..
				was_doing_iframe = doing_iframe;
				doing_iframe = false;
				unsigned int strid = 0;
				for (int i = 4 + adapt_len; i < 188; i++)
				{
					strid = (strid << 8) | buf[i];
					if (strid == 0x000001B8) // group_start_code
					{
						// found a Group of Pictures header, subsequent picture must be an I-frame
						doing_iframe = true;
					}
					else if (strid == 0x00000100) // picture_start_code
					{
						// picture_header, let's see if it's an I-frame
						if (i<186)
						{
							// check if picture_coding_type == 1
							if ((buf[i+2] & (0x7 << 3)) == (1 << 3))
							{
								// found an I-frame picture
								doing_iframe = true;
							}
						}
					}

					if (doing_iframe)
					{
						if (!curfoundfirst)
						{
							curfoundfirst = true;
							first_video_PCR = PCR;
						}
						break;
					}
				}
			}
			else if (pid == audpid && !audfoundfirst && vidfoundfirst)  // Set audio found first ONLY after first video frame found..
				curfoundfirst |= true;

			// If we were skipping a bad packet, start fresh on this new PES packet..
			if (curskipbad == true)
			{
				curskipbad = false;
				curpos = 0;
			}

			// Get the continuity code of this packet
			curcontinuity = continuity;
		}

		// Write a 2048 byte program stream packet..
		if (start && curpos > 0 && curfoundfirst && !curskipbad)
		{
#ifdef HDTVtoMPEG2
			// decode I frames
			if (was_doing_iframe && (pid == vidpid))
			{
				process_PES_video_packet(curbuf, curpos, true);
			}
#endif

			// Save the substream id block so we can added it to subsequent blocks
			unsigned char ac3_substream_id[4];
			int ac3len = 0;
			if (pid == audpid && audstreamid == 0xBD)
			{
				int spos, dpos;

				// Make sure we start with 0x0b77
				if (curbuf[9 + curbuf[8]] != 0x0b || curbuf[9 + curbuf[8] + 1] != 0x77)
				{
					spos = 9 + curbuf[8];
					dpos = 9 + curbuf[8];
					while (spos <= curpos - 2 && !(curbuf[spos] == 0x0b && curbuf[spos + 1] == 0x77))
						spos++;

					if (!(curbuf[spos] == 0x0b && curbuf[spos + 1] == 0x77))
					{
						warning("Couldn't sync AC3 packet!\n");
						curskipbad = true;
						continue;
					}

					while (spos < curpos)
					{
						curbuf[dpos] = curbuf[spos];
						spos++;
						dpos++;
					}
					curpos = dpos;
				}

				// Check the next packet to make sure IT starts with a 0x0b77
				int plen = 0;
//					if (buf[4 + adapt_len] == 0 && buf[4 + adapt_len + 1] == 0 &&		// Starting with an mpeg header?
//						buf[4 + adapt_len + 2] == 1 && buf[4 + adapt_len + 3] == 0xBD)
						plen = 9 + buf[4 + adapt_len + 8];
				int pstart = 4 + adapt_len + plen;
				if (buf[pstart] != 0x0b || buf[pstart + 1] != 0x77)
				{
					spos = pstart;
					while (spos < 188 - 2 && !(buf[spos] == 0x0b && buf[spos + 1] == 0x77))
					{
						curbuf[curpos] = buf[spos];
						curpos++;
						spos++;
					}

					if (!(buf[spos] == 0x0b && buf[spos + 1] == 0x77))
					{
						error("Couldn't sync AC3 packet!\n");
						curskipbad = true;
						continue;
					}

					adapt_len = spos - 4 - plen;

					dpos = spos - 1;
					spos = pstart - 1;
					while (spos >= pstart - plen)
					{
						buf[dpos] = buf[spos];
						spos--;
						dpos--;
					}
				}

				// Make a four byte ac3 streamid
				ac3_substream_id[0] = 0x80;	// Four byte AC3 CODE??
				ac3_substream_id[1] = 0x01;
				ac3_substream_id[2] = 0x00;	// WHY???  OH WHY??
				ac3_substream_id[3] = 0x02;
				ac3len = 4;
			}

			int written = 0;	// Bytes we've written to output file
			int pos = 0;		// Position in PES packet buffer
			for (;;)
			{
				__int64 fpos = ftell64(fout);
				if ((fpos % (__int64)PACKETSIZE) != 0)
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Packets not falling on packetsize boundries";
#else
					errMsgVal = ::MessageBox(parentWnd, "Packets not falling on packetsize boundries", NULL, MB_OK | MB_ICONERROR);
#endif
					fclose64(fin);
					fclose64(fout);
					delete vidbuf;
					delete audbuf;
					return 1;
				}

				// Get total length of this pack
				int len = min(14 + ac3len + curpos - pos, PACKETSIZE);

				// Figure out stuffing (if we have less than 16 bytes left)
				int stuffing = 0;
				if (len < PACKETSIZE && PACKETSIZE - len < 16)
				{
					stuffing = PACKETSIZE - len;
					len += stuffing;
				}

				// Write out pack header
				packet_time = (total_read * SYSTEMCLOCKRATE / STREAMRATE) - file_time;
				if (!write_pack(fout, packet_time))
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't write pack header to output file\n%s", ofilename);
#else
					sprintf(errMsgBuf, "Couldn't write pack header to output file\n%s", ofilename);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
					fclose64(fin);
					fclose64(fout);
					delete vidbuf;
					delete audbuf;
					return 1;
				}

				if (pid == audpid)
					curbuf[pos + 3] = audstreamid;
				else
					curbuf[pos + 3] = vidstreamid;

				// Packet length..
				// Subtract pack size (14) and pes id and len (6) from lenth
				curbuf[pos + 4] = (len - 6 - 14) >> 8; curbuf[pos + 5] = (len - 6 - 14) & 0xFF;

				// Add any stuffing bytes to header extra len
				int hdrsize = 9 + curbuf[pos + 8];
				curbuf[pos + 8] += stuffing;					// Add stuffing to header bytes

				// Write out id, sreamid, len
				if (fwrite64(curbuf + pos, hdrsize, 1, fout) != 1)	// Write pes id, streamid, and len
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't write to output file\n%s", ofilename);
#else
					sprintf(errMsgBuf, "Couldn't write to output file\n%s", ofilename);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
					fclose64(fin);
					fclose64(fout);
					delete vidbuf;
					delete audbuf;
					return 1;
				}
				total_out += hdrsize;

				// Write stuffing
				for (int i = 0; i < stuffing; i++)				// Write any stuffing bytes
				{
					unsigned char stuff = 0xff;
					if (fwrite64(&stuff, 1, 1, fout) != 1)
					{
#ifndef HDTVtoMPEG2
						printf("ERROR: Couldn't write to output file\n%s", ofilename);
#else
						sprintf(errMsgBuf, "Couldn't write to output file\n%s", ofilename);
						errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
						fclose64(fin);
						fclose64(fout);
						delete vidbuf;
						delete audbuf;
						return 1;
					}
				}
				total_out += stuffing;

				// Write ac3 streamid
				if (ac3len != 0)
				{
					if (fwrite64(ac3_substream_id, ac3len, 1, fout) != 1)
					{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't write to output file\n%s", ofilename);
#else
						sprintf(errMsgBuf, "Couldn't write to output file\n%s", ofilename);
						errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
						fclose64(fin);
						fclose64(fout);
						delete vidbuf;
						delete audbuf;
						return 1;
					}
					total_out += ac3len;
				}

				// Write rest of data len minus headersize (9) stuffing, and pack size (14)
				if (fwrite64(curbuf + pos + hdrsize, len - hdrsize - 14 - stuffing - ac3len, 1, fout) != 1)	// Write data bytes
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't write to output file\n%s", ofilename);
#else
					sprintf(errMsgBuf, "Couldn't write to output file\n%s", ofilename);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
					fclose64(fin);
					fclose64(fout);
					delete vidbuf;
					delete audbuf;
					return 1;
				}
				written += len;
				total_out += len;

				// Add len minus stuff we added like the pack (14) and the stuffing.
				pos += len - 14 - stuffing - ac3len;
				if (pos == curpos)
					break;

				// Add pes header for next packet
				pos -= 9;
				make_pes_header(curbuf + pos, (pid == vidpid ? vidstreamid : audstreamid), 0, -1, -1);
			}

			// Write padding
			if ((written % PACKETSIZE) != 0)
			{
				int left = PACKETSIZE - (written % PACKETSIZE);

				// Pad out to PACKETSIZE bytes
				if (!pad_buffer(fout, left))
				{
#ifndef HDTVtoMPEG2
					printf("ERROR: Couldn't write to output file\n%s", ofilename);
#else
					sprintf(errMsgBuf, "Couldn't write to output file\n%s", ofilename);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
#endif
					fclose64(fin);
					fclose64(fout);
					delete vidbuf;
					delete audbuf;
					return 1;
				}
				total_out += left;
			}

			curpos = 0;
			curcontinuity = -1;
			curpackets++;

#ifndef HDTVtoMPEG2
			// Print out a "." every 30 frames
			if (pid == vidpid)
			{
				if (curpackets % 30 == 0)
					printf(".");
			}
#endif
		}

		// Add the payload for this packet to the current buffer
		if (curfoundfirst && (184 - adapt_len) > 0)
		{
			memcpy(curbuf + curpos, buf + 4 + adapt_len, 184 - adapt_len);
			curpos += 184 - adapt_len;
		}
	}

	// Mark last valid stream pos
	unsigned char end[4] = {0,0,1,0xb9};
	fwrite64(&end, 4, 1, fout);
	total_out += 4;

	// Close files
	if (fin)
		fclose64(fin);
	if (fout)
		fclose64(fout);

	// Delete video/audio buffers
	delete vidbuf;
	delete audbuf;

#ifndef HDTVtoMPEG2
	// All done!
	printf("\n\nNumber of video packets %d audio packets %d\nFinished\n", vidpackets, audpackets);
#endif
	
	return 0;
}
#endif

#ifndef HDTVtoMPEG2

int main(int argc, char* argv[])
{
	printf("bcdmux 1.04 - by Ben Cooley\n"
		   "    Converts HiPix or other MPEG2 TS file into standard DVD MPEG2 files.\n"
		   "    usage: bcdmux \"sourcefile\" videopid audiopid \"destfile\" <numfiles>\n"
		   "\n");

	if (argc < 4)
	{
		printf("      sourcefile: First MPEG2 ATSC transport stream file in seq (from HiPix)\n"
			   "        videopid: Usually 0x11 for video program 1, 0x21 for program 2, etc.\n"
			   "         audioid: Usually 0x14 for audio program 1, 0x24 for main audio 2\n"
			   "        destfile: MPEG2 file to output stream to.  Should have an .mpg ext\n"
			   "        numfiles: Num input files for each output file (usually 3-4)\n"
			   "\n"
			   "     note: If output file doesn't have digits in name, only one output\n"
			   "           file will be made. (i.e. needs zeros in Sting0000.mpg)\n"
			   "\n"
			   "  example: bcdmux \"Sting on Leno.ts.0000\" 0x11 0x14 \"Sting0000.mpg\" 3\n"
			   "\n"
			   "  This converts three sequentially numbered files to Sting0000.mpg\n");
		return 0;
	}
	
	// Input files
	int numifiles = -1;				// Indicates we should automatically increment vile number
	char *ifiles[1];
	ifiles[0] = argv[1];

	// Get the video PID
	int vid_pid = get_hex(argv[2]);

	// Get the audio PID
	int aud_pid = get_hex(argv[3]);

	// Get dest file pattern
	char* dfile = argv[4];

	// Get max files
	int maxfiles = 3; // Default three minutes worth of files.. nearly a gig
	if (argc > 4)
		maxfiles = atoi(argv[5]);

	// Convert files
	return convert_files_to_mpg(ifiles, numifiles, dfile, vid_pid, aud_pid, maxfiles, NULL, NULL);
}

#endif

