// ***************************************************************************
// bcdemux.exe - Copyright (C) 2002 - Ben Cooley
//
// bcdemux demultiplexes an ATSC transport stream and converts it to an MPEG2
// DVD compatible video file.  It can select any program in the TS file, and 
// can automatically read a sequence of TS files and convert them to any number
// of MPG files.

#ifndef _BCDMUX_H
#define _BCDMUX_H

#include "stdafx.h"

#include <fcntl.h>
#include <io.h>
#include <errno.h>
//#include <sys/types.h>
#include <sys/stat.h>


#define DO_ASM


#define RDBUFSIZE		(1024 * 1024 * 4)
#define WRBUFSIZE		(1024 * 1024 * 4)

#define STREAMRATE		((__int64)2401587)			// Original HD stream rate 19.2 Mbps
#define SYSTEMCLOCKRATE		((__int64)27000000)			// MPEG System clock rate


// FILE64

struct FILE64
{
	int handle;
	bool read;
	unsigned char* buffer;
	int size, pos, len;
	__int64 filepos, filelen;
	struct _stati64 info;
};

extern int ferror64(void);
extern FILE64* fopen64(const char* name, const char* mode, int buffersize, bool overwrite = false);
extern __int64 ftell64(FILE64* f);
extern __int64 fseek64(FILE64* f, __int64 pos, int set);
extern __int64 filelength64(FILE64* f);
extern int fread64(void* buf, int elsize, int elnum, FILE64* f);
extern int fwrite64(void* buf, int elsize, int elnum, FILE64* f);
extern void fclose64(FILE64* f);

// TSFile list class
// Contains list of files to convert, with additional info,
// such as file length, offset of first TS packet, and
// regions of data to include or exclude from the conversion

typedef enum {
	rt_include,
	rt_exclude,
	rt_cut,
} RegionType;

class TSFileInfo;

struct TSRegionInfo
{
	TSRegionInfo *m_prev;
	TSRegionInfo *m_next;

	TSFileInfo *m_parent;

	__int64 m_offset;
	RegionType m_type;
};

class TSFileInfo
{
public:
	TSFileInfo() :
		m_prev(NULL),
		m_next(NULL),
		m_fileSize(0),
		m_firstPacketOffset(0),
		m_totalBytesPrior(0),
		m_f(INVALID_HANDLE_VALUE),
		m_head(NULL),
		m_tail(NULL),
		m_count(0),
		m_prevIFrameOffset_Info(NULL)
	{
			m_fileName[0] = '\0';
	}

	~TSFileInfo();

	TSFileInfo *m_prev;
	TSFileInfo *m_next;
	TSFileInfo *m_prevIFrameOffset_Info;

	char m_fileName[256];
	__int64 m_fileSize;
	int m_firstPacketOffset;

	__int64 m_totalBytesPrior;
	HANDLE m_f;

	TSRegionInfo *m_head;
	TSRegionInfo *m_tail;
	int m_count;

	TSRegionInfo *Find(__int64 offset);
	TSRegionInfo *Add(RegionType type, __int64 offset);
	void Delete(TSRegionInfo *info);
	void Clear(void);
};

class TSFileList
{
public:
	TSFileList() :
		m_head(NULL),
		m_tail(NULL),
		m_count(0),
		m_totalBytes(0),
		m_totalIncludedBytes(0)
	{
		InitializeCriticalSection(&m_cSec);
	}

	~TSFileList();

	int Count(void) { return m_count; }
	__int64 TotalBytes(void) { return m_totalBytes; }

	TSFileInfo *GetHead(void) { return m_head; }

	TSFileInfo *Find(const char *name);
	TSFileInfo *Find(__int64 bytes);
	TSFileInfo *Add(const char *name, __int64 size, int offset);
	void Delete(const char *name);
	void Delete(TSFileInfo *info);
	void Clear(void);
	void RecomputeBytes(void);
	void RecomputeTotalIncludedBytes(void);
	void CloseFilesOutsideRange(__int64 min, __int64 max);
	void AdjustList( int movingIndex, int moveToIndex );

	TSRegionInfo *FindRegion(__int64 bytes);
	TSRegionInfo *PrevRegion(TSRegionInfo *ri);
	TSRegionInfo *NextRegion(TSRegionInfo *ri);
	TSRegionInfo *AddRegion(RegionType type, __int64 bytes);
	void DeleteRegion(TSRegionInfo *info);
	void ClearRegions(void);
	void ClearPrevIFrameOffsets();
	__int64 TotalIncludedBytes(void) { return m_totalIncludedBytes; }

	void Lock(void);
	void Unlock(void);

private:
	TSFileInfo *m_head;
	TSFileInfo *m_tail;
	int m_count;
	__int64 m_totalBytes;
	__int64 m_totalIncludedBytes;

	CRITICAL_SECTION m_cSec;
};


// bit buffer

extern int bitpos;
extern unsigned int bitval;
extern unsigned char* bitbuf;
extern unsigned int bitmask[];

inline void set_buf(unsigned char* buf, int bufsize, bool clear = true)
{
	bitpos = 0;
	bitbuf = buf;
	bitval = (bitbuf[0] << 24) | (bitbuf[1] << 16) | (bitbuf[2] << 8) | bitbuf[3];
	if (clear)
		memset(bitbuf, 0, bufsize);
}

inline int buf_size()
{
	return bitpos >> 3;
}

inline void set_bits(unsigned int val, int bits)
{
	val &= bitmask[bits];

	if (bits < 1)
		__asm int 3;

	while (bits > 0)
	{
		int bitsleft = (8 - (bitpos & 7));
		if (bits >= bitsleft)
		{
			bitbuf[bitpos >> 3] |= val >> (bits - bitsleft);
			bitpos += bitsleft;
			bits -= bitsleft;
			val &= bitmask[bits];
		}
		else
		{
			bitbuf[bitpos >> 3] |= val << (bitsleft - bits);
			bitpos += bits;
			bits = 0;
		}
	}
}

inline unsigned int get_bit_pos()
{
	return bitpos;
}

inline unsigned int get_byte_pos()
{
	return bitpos >> 3;
}

inline void set_bit_pos(unsigned int newpos)
{
	bitpos = newpos;

	unsigned int pos = (bitpos >> 5) << 3;
	bitval = (bitbuf[pos] << 24) | (bitbuf[pos + 1] << 16) | (bitbuf[pos + 2] << 8) | bitbuf[pos + 3];
}

inline void set_byte_pos(unsigned int newpos)
{
	set_bit_pos(newpos << 3);
}

#ifndef DO_ASM
// original C code

inline unsigned int get_bits(int bits)
{
	if (bits > 32)
		__asm int 3;
	if (bits < 1)
		__asm int 3;

	unsigned int val;
	int left = 32 - (bitpos & 31);

	if (bits < left)
	{
		val = (bitval >> (left - bits)) & bitmask[bits];
		bitpos += bits;
	}
	else
	{
		val = (bitval & bitmask[left]) << (bits - left);
		bitpos += left;
		bits -= left;

		int pos = bitpos >> 3;
		bitval = (bitbuf[pos] << 24) | (bitbuf[pos + 1] << 16) | (bitbuf[pos + 2] << 8) | bitbuf[pos + 3];
		
		if (bits > 0)
		{
			val |= (bitval >> (32 - bits)) & bitmask[bits];
			bitpos += bits;
		}
	}

	return val;
}

/*
inline unsigned int get_bits(int bits)
{
	unsigned int val;
	int left = 32 - (bitpos & 31);

	if (bits < left)
	{
		val = (bitval >> (left - bits)) & bitmask[bits];
		bitpos += bits;
	}
	else
	{
		val = (bitval & (0xffffffff >> (32 - left))) << (bits - left);
		bitpos += left;
		bits -= left;

		__asm {
			mov eax,dword ptr ds:[bitpos]
			mov ecx,dword ptr ds:[bitbuf]
			sar eax,5
			mov ecx,dword ptr [ecx+eax*4]
			bswap ecx
			mov dword ptr ds:[bitval],ecx
		}

		if (bits > 0)
		{
			val |= (bitval >> (32 - bits));
			bitpos += bits;
		}
	}

	return val;
}
*/

#else
#pragma warning(disable:4035)		// "no return value"

inline unsigned int get_bits(int bits)
{
	__asm
	{
		// are these pushes neccessary?
//		push ebx
//		push ecx
//		push edx

		mov eax, dword ptr ds:[bitpos]		// eax = bitpos
		mov cl, 20h			// cl = left = 32 - (bitpos & 31)
		mov dl, al			// cl = left = 32 - (bitpos & 31)
		mov ebx, bits			// ebx = bits
		and dl, 1fh			// cl = left = 32 - (bitpos & 31)
		add eax, ebx							// bitpos += bits;
		sub cl, dl			// cl = left = 32 - (bitpos & 31)
		mov dword ptr ds:[bitpos], eax	// bitpos += bits;

		// if (left > bits)
		sub cl, bl			// cl = left - bits
		jle short part2

	//part1:
		mov eax, dword ptr ds:[bitval]		// eax = bitval
		mov edx, -1				// create mask
		shr eax, cl				// eax = requested bits
		mov cl, bl				// create mask, cl = bits
		shl edx, cl				// create mask
		not edx					// create mask

		and eax, edx	// apply mask

		jmp short end

	part2:
		mov edx, dword ptr ds:[bitbuf]	// edx = bitbuf
		shr eax, 5								// eax = bitpos >> 5
		neg cl					// cl = bits - left
		mov edx, dword ptr [edx+eax*4]	// edx = next dword

		bswap edx				// byte swap next dword

		mov eax, dword ptr ds:[bitval]	// eax = bitval

		shld eax, edx, cl			// eax = requested bits

//		mov dword ptr ds:[bitval], edx		// store new bitval
		mov [bitval], edx		// store new bitval

		mov cl, bl				// create mask, cl = bits
		mov edx, -1				// create mask
		shl edx, cl				// create mask
		not edx					// create mask
			
		and eax, edx	// apply mask

	end:
//		pop edx
//		pop ecx
//		pop ebx

		// return eax
	}
}

#pragma warning(default:4035)
#endif

#ifndef DO_ASM

inline unsigned int peek_bits(int bits)
{
	if (bits > 32)
		__asm int 3;
	if (bits < 1)
		__asm int 3;

	unsigned int val;
	int left = 32 - (bitpos & 31);

	if (bits < left)
	{
		val = (bitval >> (left - bits)) & bitmask[bits];
	}
	else
	{
		val = (bitval & bitmask[left]) << (bits - left);

		int pos = (bitpos+left) >> 3;

		bits -= left;

		unsigned int tmp_bitval = (bitbuf[pos] << 24) | (bitbuf[pos + 1] << 16) | (bitbuf[pos + 2] << 8) | bitbuf[pos + 3];
		
		if (bits > 0)
		{
			val |= (tmp_bitval >> (32 - bits)) & bitmask[bits];
		}
	}

	return val;
}

#else
#pragma warning(disable:4035)		// "no return value"

inline unsigned int peek_bits(int bits)
{
	__asm
	{
		// are these pushes neccessary?
//		push ebx
//		push ecx
//		push edx

		mov eax, dword ptr ds:[bitpos]		// eax = bitpos
		mov cl, 20h			// cl = left = 32 - (bitpos & 31)
		mov dl, al			// cl = left = 32 - (bitpos & 31)
		mov ebx, bits			// ebx = bits
		and dl, 1fh			// cl = left = 32 - (bitpos & 31)
		sub cl, bl			// cl = left - bits
		add eax, ebx							// newbitpos = bitpos + bits;
		sub cl, dl			// cl = left = 32 - (bitpos & 31)

		// if (left - bits > 0)
		jle short part2

	//part1:
		mov eax, dword ptr ds:[bitval]		// eax = bitval
		mov edx, -1				// create mask
		shr eax, cl				// eax = requested bits
		mov cl, bl				// create mask, cl = bits
		shl edx, cl				// create mask
		not edx					// create mask

		and eax, edx	// apply mask

		jmp short end

	part2:
		mov edx, dword ptr ds:[bitbuf]	// edx = bitbuf
		shr eax, 5								// eax = bitpos >> 5
		neg cl					// cl = bits - left
		mov edx, dword ptr [edx+eax*4]	// edx = next dword

		bswap edx				// byte swap next dword

		mov eax, dword ptr ds:[bitval]	// eax = bitval

		shld eax, edx, cl			// eax = requested bits

		mov cl, bl				// create mask, cl = bits
		mov edx, -1				// create mask
		shl edx, cl				// create mask
		not edx					// create mask
			
		and eax, edx	// apply mask

	end:
//		pop edx
//		pop ecx
//		pop ebx

		// return eax
	}
}

#pragma warning(default:4035)
#endif

#ifndef DO_ASM

inline void discard_bits(int bits)
{
	if (bits < 1)
		__asm int 3;
	if (bits > 32)
		__asm int 3;

	int left = 32 - (bitpos & 31);
	bitpos += bits;

	if (left <= bits)
	{
		__asm {
			mov eax,[bitpos]
			mov ecx,dword ptr ds:[bitbuf]
			sar eax,5
			mov ecx,dword ptr [ecx+eax*4]
			bswap ecx
			mov dword ptr ds:[bitval],ecx
		}
	}
}

#else

inline void discard_bits(int bits)
{
	__asm
	{
		// are these pushes neccessary?
//		push ebx
//		push ecx

		mov ecx, dword ptr ds:[bitpos]
		mov eax, 20h
		mov ebx, ecx
		add ecx, bits
		and ebx, 1Fh
		mov dword ptr ds:[bitpos], ecx
		sub eax, ebx
		cmp eax, bits
		jg short end
		mov eax, dword ptr ds:[bitbuf]
		sar ecx, 5
		mov eax, dword ptr [eax+ecx*4]
		bswap eax
		mov dword ptr ds:[bitval], eax
	end:
//		pop ecx
//		pop ebx
	}
}

#endif

// ATSC

#define ATSC_AC3_2CHANNEL 1
#define ATSC_AC3_5CHANNEL 2

struct ATSC_CHANNEL_INFO
{
	char name[8];
	int majorChannel;
	int minorChannel;
	int videopid; // 1st 16 bits = pType, last 16 bits = pid
	int audiopid; // 1st 16 bits = pType, last 16 bits = pid
	unsigned int width;
	unsigned int height;
	unsigned int bitRate; // calculated weighted average during scan_channel_info() call
	unsigned int aspectRatio;
	unsigned int frameRate;
	unsigned int chromaFormat;
	unsigned int progressiveSequence;
	int audio;
	int PMT_pid;
	int PCR_pid;
	int program_number;
};


// function exports
extern int align_to_next_packet(unsigned char *buf, int size);
extern __int64 align_to_next_packet(FILE64* f);
extern bool inc_file_counter(char* fileName);

// Typedef for callback function called during conversion
typedef bool (*convert_callback)(char *inputFileName, char* outputFileName,
	__int64 curTotalPos, __int64 curTotalIn, __int64 totalIn,
	__int64 curTotalOut, long timeDiscrepancy, void* data);

// Convert file function
class FileReadBuffer;

int convert_files_to_mpg(TSFileList *inputList, char *outputFilePattern,
		int vidPid, int audPid, int maxFileSize, FileReadBuffer *readBuffer,
		convert_callback callback, void *data, HWND parentWnd, CPtrArray *pPidInfoList );


#endif