#ifndef _MPEG2STUFF_H
#define _MPEG2STUFF_H


typedef unsigned char uchar;
typedef unsigned int uint;
typedef unsigned __int64 uint64;


// multithreaded buffered 64bit file I/O

class FileReadBuffer
{
public:
	bool Start(int bufferSize, int alignment);	// inits sub thread, allocs buffers
	void Stop(void);

	bool Open(char *fileName);
	void Close(void);

	__int64 Seek(__int64 pos);

	int Read(void *dest, int bytes);

	static unsigned int WINAPI StaticThread(LPVOID data);
	void Thread(void);

	bool IsFileOpen(void) { return m_file != INVALID_HANDLE_VALUE; }
	bool IsError(void) { return m_error; }
	bool IsEOF(void) { return m_EOF; }

private:

	bool m_exit;

	HANDLE m_thread;

	// file variables
	HANDLE m_file;
	__int64 m_filePosCur;
	__int64 m_filePosRead;
	__int64 m_fileSize;
	__int64 m_alignment;
	bool m_error;
	bool m_EOF;

	// and critical section that gives write access to them
	CRITICAL_SECTION m_fileCSec;

	class Buffer
	{
	public:
		// buffer variables
		uchar *m_bufferX;
		uchar *m_buffer;
		int m_size;
		int m_length;
		__int64 m_startOffset;

		bool m_valid;

		// and critical section that gives write access to them
		CRITICAL_SECTION m_cSec;
	};

	Buffer m_buffer[2];

	// buffer use variables
	Buffer *m_using;
	Buffer *m_filling;

	// and critical section that gives write access to them?
	CRITICAL_SECTION m_bufferCSec;
};

// link list used in calculating average stream rates
class StreamRateList
{
public:

	StreamRateList(): m_StreamPidList(NULL){}
	~StreamRateList();	

	__int64 AddNewRate( int pid, __int64 stream_rate );

	// link list of calculated stream rates for a particular PID
	class StreamRateNode
	{
	public:
		StreamRateNode(): m_stream_rate(0), m_weight(0.0), m_prev(NULL), m_next(NULL){}
		__int64 m_stream_rate;
		double m_weight; 
		StreamRateNode *m_prev, *m_next;
	};

	// link list of PID's in the transport stream
	class StreamPidList
	{
	public:
		StreamPidList(): m_pid(-1), m_avg_stream_rate(0), m_prev(NULL), m_next(NULL), m_StreamRateNode(NULL){}

		int m_pid;
		__int64 m_avg_stream_rate;
		StreamPidList  *m_prev, *m_next;
		StreamRateNode *m_StreamRateNode; 
	};

	StreamPidList *m_StreamPidList;
};

// type of output file
typedef enum {
	ft_MIN = 0,

	ft_ts = ft_MIN,
	ft_HiDTV,
	ft_mpg,

	ft_MAX,
} Filetype;

typedef struct
{
	uchar *buffer;
	uint width;
	uint height;
	uint stride;
	uint size;
} bitmap;

typedef struct
{
	bitmap bm;
	volatile uint rendered;
	volatile uint renderedWidth;
	volatile uint renderedHeight;
	volatile uint aspectRatio;
} pdlg_buffer;

extern pdlg_buffer pdlg_buffers[];

// PID stuff

#define MAX_PIDS 100

typedef enum {
	STREAM_UNKNOWN,
	STREAM_TABLE,
	STREAM_PES,
} streamType;

typedef enum {
	PID_UNKNOWN,
	PID_NULL,
	PID_PAT,
	PID_PMT,
	PID_NIT,
	PID_MPEG2_VIDEO,
	PID_USER_PRIVATE,
	PID_AC3_AUDIO,
	PID_AAC_AUDIO,
	PID_MPEG2_AUDIO,
	PID_MPEG1_AUDIO,
	PID_PSIP,
	PID_PCR,
} pidType;

typedef struct
{
	__int64 program_clock_reference_base;		// current PCR_base
	__int64 program_clock_reference_extension;	// current PCR_ext
	__int64 last_pcrb;							// last program_clock_reference_base
	__int64 last_pcre;							// last program_clock_reference_extension
	__int64 new_pcrb;							// new program_clock_reference_base
	__int64 new_pcre;							// new program_clock_reference_extension
	__int64 current_PTS;						// current PTS, video or audio
	__int64 current_DTS;						// current DTS, video only
	__int64 new_PTS;							// new PTS value, video or audio
	__int64 new_DTS;							// new DTS value, video only

	int adaptation_field_length;				// length of the adaptation field
	bool PCR_present;							// indicates PCR is present in PES
	bool PTS_present;							// indicates PTS is present in PES
	bool DTS_present;							// indicates DTS is present in PES

	__int64 PCR_bytes;							// total bytes read since last PCR
	__int64 prev_PCR_bytes;						// last number of bytes read since last PCR
	__int64 TS_rate;							// last calculated weighted average ts rate
	__int64 PCR_value; 
	__int64 last_PCR_value;

	pidType pidType;							// indicates the PID type, video or audio

}timestamp_info;

typedef struct
{
	int pid;
	int count;
	streamType sType;
	pidType pType;
	bool processed;

	uchar *buffer;	// if needed
	int bufferSize;
	int bufferPos;
	int continuity;
	bool reading;
	bool confirmedRead;
	timestamp_info ts_info;
} PIDinfo;

typedef struct
{
	int seen;
	PIDinfo pids[MAX_PIDS];
} PIDinfoGroup;

typedef struct 
{
	uchar *buffer;
	int bytes; 
}payload_buff;


// information about the pids we are manipulating
typedef struct
{
	short original_pid;			// original pid
	short new_pid;				// mapped pid
	pidType type;				// pid type
	bool strip_pid;				// true = strip, false = keep
	bool processed;				// true, pid has been processed
	short continuity_counter;	// current value
	short payload_unit_start_indicator; // for PSIP packets, 1st is 1, rest are 0
	int program_number; // program_number for video or audio pid
}Pid_Info;

// saved copies of the PAT and PMT tables for preprocessing
typedef struct
{
	uchar *pPAT_Packet;

//	uchar *pPMT_Packet[10]; // normally 3 or less, 10 PMT should be generous
	uchar *pPMT_Packet[128]; // 128 PMT max, QAM TS files can contain more than 10
	int numOfPMTPackets;
}PatPmt_Packets;

//#define VIDEO_BUFFER_SIZE (512*1024)
#define VIDEO_BUFFER_SIZE (2048*1024)
#define AUDIO_BUFFER_SIZE (16*1024)
#define TABLE_BUFFER_SIZE (1024)

extern PIDinfo *find_PID(PIDinfoGroup *pg, int pid);
extern PIDinfo *find_PID_no_new(PIDinfoGroup *pg, int pid);

///////////////////////////////////////////////////////////////////////////////
// MPEG2 video structs
///////////////////////////////////////////////////////////////////////////////

// sequence header
typedef struct
{
	uint start_code_identifier;

	uint horizontal_size_value;
	uint vertical_size_value;
	uint bit_rate_value;
	uint vbv_buffer_size_value;
	uint bit_rate_extension;						// sequence extension
	uint lower_layer_prediction_horizontal_size;		// sequence scalable extension
	uint lower_layer_prediction_vertical_size;		// ss ext

	uchar aspect_ratio_information;
	uchar frame_rate_code;
	uchar sequence_extension;						// s ext
	uchar profile_and_level_indication;			// s ext
	uchar chroma_format;								// s ext
	uchar horizontal_size_extension;				// s ext
	uchar vertical_size_extension;				// s ext
	uchar vbv_buffer_size_extension;				// s ext
	uchar frame_rate_extension_n;					// s ext
	uchar frame_rate_extension_d;					// s ext
	uchar sequence_scalable_extension;					// ss ext
	uchar scalable_mode;										// ss ext
	uchar layer_id;											// ss ext
	uchar horizontal_subsampling_factor_m;				// ss ext
	uchar horizontal_subsampling_factor_n;				// ss ext
	uchar vertical_subsampling_factor_m;				// ss ext
	uchar vertical_subsampling_factor_n;				// ss ext
	uchar picture_mux_order;								// ss ext
	uchar picture_mux_factor;								// ss ext

	uint constrained_parameters_flag : 1;
	uint load_intra_quantiser_matrix : 1;
	uint load_non_intra_quantiser_matrix : 1;
	uint progressive_sequence : 1;				// s ext
	uint low_delay : 1;								// s ext
	uint picture_mux_enable : 1;							// ss ext
	uint mux_to_progressive_sequence : 1;				// ss ext

	uchar intra_quantiser_matrix[64];
	uchar non_intra_quantiser_matrix[64];

} sequence_header;

// picture header
typedef struct
{
	uint start_code_identifier;

	uint temporal_reference;
	uint vbv_delay;
	uint lower_layer_temporal_reference;					// picture spatial scalable extension
	int lower_layer_horizontal_offset;						// pss ext
	int lower_layer_vertical_offset;							// pss ext
	uint forward_temporal_reference;							// picture temporal scalable extension
	uint backward_temporal_reference;						// pts ext

	uchar picture_coding_type;
	uchar forward_f_code;
	uchar backward_f_code;
	uchar picture_coding_extension;				// picture coding extension
	uchar intra_dc_precision;						// pc ext
	uchar picture_structure;						// pc ext
	uchar field_sequence;							// pc ext
	uchar burst_amplitude;							// pc ext
	uchar sub_carrier_phase;						// pc ext
	uchar quant_matrix_extension;								// quant matrix extension
	uchar picture_spatial_scalable_extension;				// pss ext
	uchar spatial_temporal_weight_code_table_index;		// pss ext
	uchar picture_temporal_scalable_extension;			// pts ext
	uchar reference_select_code;								// pts ext

	uint full_pel_forward_vector : 1;
	uint full_pel_backward_vector : 1;
	uint top_field_first : 1;						// pc ext
	uint frame_pred_frame_dct : 1;				// pc ext
	uint concealment_motion_vectors : 1;		// pc ext
	uint q_scale_type : 1;							// pc ext
	uint intra_vlc_format : 1;						// pc ext
	uint alternate_scan : 1;						// pc ext
	uint repeat_first_field : 1;					// pc ext
	uint chroma_420_type : 1;						// pc ext
	uint progressive_frame : 1;					// pc ext
	uint composite_display_flag : 1;				// pc ext
	uint v_axis : 1;									// pc ext
	uint sub_carrier : 1;							// pc ext
	uint load_intra_quantiser_matrix : 1;					// qm ext
	uint load_non_intra_quantiser_matrix : 1;				// qm ext
	uint load_chroma_intra_quantiser_matrix : 1;			// qm ext
	uint load_chroma_non_intra_quantiser_matrix : 1;	// qm ext
	uint lower_layer_progressive_frame : 1;				// pss ext
	uint lower_layer_deinterlaced_field_select : 1;		// pss ext

	uchar f_code[2][2];								// pc ext
	uchar intra_quantiser_matrix[64];						// qm ext
	uchar non_intra_quantiser_matrix[64];					// qm ext
	uchar chroma_intra_quantiser_matrix[64];				// qm ext
	uchar chroma_non_intra_quantiser_matrix[64];			// qm ext

} picture_header;

typedef struct
{
	uint macroblock_quant : 1;
	uint macroblock_motion_forward : 1;
	uint macroblock_motion_backward : 1;
	uint macroblock_pattern : 1;
	uint macroblock_intra : 1;
	uint spatial_temporal_weight_code_flag : 1;
	uint permitted_spatial_temporal_weight_classes : 1;

	uchar spatial_temporal_weight_code;
	uchar frame_motion_type;
	uchar field_motion_type;
	uchar prediction_type;							// motion_type_table_entry
	uchar motion_vector_count;						// mtte
	uchar mv_format;									// mtte
	uchar dmv;											// mtte

	uchar motion_vertical_field_select[2][2];
	int motion_code[2][2][2];
	uint motion_residual[2][2][2];
	int dmvector[2];

	uint dct_type : 1;

} macroblock_type;

typedef struct
{
	uchar prediction_type;					// 0 = field-based, 1 = frame-based, 2 = dual-prime, 3 = 16x8 MC
	uchar motion_vector_count;
	uchar mv_format;							// 0 = field, 1 = frame
	uchar dmv;
} motion_type_table_entry;


///////////////////////////////////////////////////////////////////////////////
// AC3 audio structs
///////////////////////////////////////////////////////////////////////////////

typedef struct
{
	uint syncword;
	uint crc1;
	uchar fscod;
	uchar frmsizecod;
} syncinfo;

typedef struct
{
	uchar bsid;
	uchar bsmod;
	uchar acmod;
	uchar cmixlev;
	uchar surmixlev;
	uchar dsurmod;
	uint lfeon : 1;
	uchar dialnorm;
	uint compre : 1;
	uchar compr;
	uint langcode : 1;
	uchar langcod;
	uint audprodie : 1;
	uchar mixlevel5;
	uchar roomtyp;
	uchar dialnorm2;
	uint compr2e : 1;
	uchar compr2;
	uint lngcod2e : 1;
	uchar langcod2;
	uint audprodi2e : 1;
	uchar mixlevel2;
	uchar roomtyp2;
	uint copyrightb : 1;
	uint origbs : 1;
	uint timecod1e : 1;
	uint timecod1;
	uint timecod2e : 1;
	uint timecod2;
	uint addbsie : 1;
	uchar addbsil;
	uchar addbsi[64];	// max 64 bytes
} bsi;

typedef enum
{
	frame_NONE,
	frame_4x3,
	frame_16x9,
}frame_type;

typedef struct
{
	uint macroblock_state[67];			// 68 rows in a 1920x1080 frame
										// 45 rows in a 1280x720 frame
										// left and right "bars"

	uint macroblock_state_middle[67];   // macroblock states for middle of frame

	uint horizontal_size;
	uint vertical_size;
	uint number_of_frame_columns;		// number of columns in the frame
	uint number_of_frame_rows;			// number for rows in the frame
} macroblock_state_info;


// sequence data buffers
typedef struct
{
	unsigned char *pSequenceHeader;		// buffer for sequence_header
	int sh_length;						// length of sequence_header data

	unsigned char *pSequenceExtension; // buffer for sequence_extension

	unsigned char *pSequenceScalableExtension; // buffer for sequence_scalable_extension
	int sse_length;						// length of sequence_scalable_extension

//	unsigned char *pSequenceDisplayExtension; // buffer for sequence_display_extension
//	int sde_length; // length of sequence_display_extension
} sequence_buffers;

///////////////////////////////////////////////////////////////////////////////
// function exports
///////////////////////////////////////////////////////////////////////////////


//#define FINAL_RELEASE
#ifndef FINAL_RELEASE
extern void __cdecl _Error(char *format, ...);
#else
#define _Error {}//
#endif

// Scans the TS file for channel info
extern bool scan_channel_info(const void *inBuffer, int maxScan, int maxChannels, 
					   ATSC_CHANNEL_INFO *channels, int &numChannels, CWordArray *pPIDDeleteList,
					   CWordArray *pHiddedPidDeleteList );

// take a PES video packet and decode it into an info table, and/or into an image buffer
extern bool process_PES_video_packet(uchar *buf, int bufSize,
						bool decodeImage, sequence_header *sh = NULL, pdlg_buffer *target = NULL);

// take an AC3 audio packet and decode it into an info table
extern bool process_PES_AC3_packet(uchar *buf, int bufSize, syncinfo *si = NULL, bsi *bi = NULL);


// when given a pre-filled buffer (packet aligned) with a valid video pid will try to
// extract an I-frame and decode it into the given bitmap buffer
extern bool decode_I_frame(uchar *buffer, int bufferSize, int vidPid, pdlg_buffer *target);



extern void peek_at_file(char *ifilelist, convert_callback callback, void* data);


// conversion functions
extern int convert_files_to_ts(TSFileList *inputList, char *outputFilePattern, Filetype fileType,
					int vidPid, int audPid, int maxFileSize,
					FileReadBuffer *readBuffer, convert_callback callback, void *data, HWND parentWnd,
					bool decodeIFrames, CPtrArray *pPidInfoList, CWordArray *pHiddenPidDeleteList,
					bool bCreateClips, bool bPreserveBitRate, int vertical_size_value, bool bSetVerticalSizeTo1080 );

// scanning functions
extern __int64 scan_for_Iframe_back(unsigned char *buffer, __int64 bufferSize, __int64 buffer_offset,
							 TSFileList *inputList, __int64 totalOffset, __int64 scanRange, int vidPid);
extern __int64 scan_for_Iframe_forward(unsigned char *buffer, __int64 bufferSize, __int64 buffer_offset,
								TSFileList *inputList, __int64 totalOffset, __int64 scanRange, int vidPid);

extern int scan_for_excludes(TSFileList *inputList, int vidPid,
				__int64 minExcludeSize, __int64 responseSize, float maxError, float scanRatio,
				FileReadBuffer *readBuffer, convert_callback callback, void *data, HWND parentWnd);

extern int scan_for_commercials(TSFileList *inputList, int vidPid,
				__int64 minExcludeSize, __int64 responseSize, float maxError, float scanRatio,
				FileReadBuffer *readBuffer, convert_callback callback, void *data, HWND parentWnd,
				uint streamrate, int MaxSeekSize, bool bIncludeBlackFrames );

// image comparison functions
extern float compareFrameEdges(pdlg_buffer *t1, pdlg_buffer *t2, float scanRatio);

extern bool ChooseIDCT(void);

// calculate new CRC_32
extern unsigned int CalculateCRC( unsigned char *p_begin, int i_count );

// is pid a video or audio pid
extern pidType GetPidType( short pid, Pid_Info *pPidArray, int numOfPids, Pid_Info *pPidArrayStart );
// create a new Program Association Table
extern void CreateNewPAT( unsigned char *buf, Pid_Info *pPidArray, int numOfPids, Pid_Info *pPidArrayStart );
// create a new Program Map Table
extern bool	CreateNewPMT( short pid, unsigned char *buf, Pid_Info *pPidArray, int numOfPids, Pid_Info *pPidArrayStart, int audioCount );
// remap the video or audio pid
extern void UpdatePid( short pid, pidType pid_type, unsigned char *buf, int adaptation_field_length, Pid_Info *pPidArray,
					   int numOfPids, Pid_Info *pPidArrayStart );
// create a new VCT packet
extern bool CreateNewVCT( short pid, unsigned char *buf, Pid_Info *pPidArray, int numOfPids, Pid_Info *pPidArrayStart,
						   unsigned char *b2, bool bAddToBuffer );
// create a new MGT packet
extern void CreateNewMGT( short pid, unsigned char *buf, Pid_Info *pPidArray, int numOfPids, Pid_Info *pPidArrayStart );

// convert packet to a NULL packet
extern void ConvertToNULLPacket( unsigned char *buf );

// verify that the PMT defines both a video and an audio pid
extern bool VerifyPMT( unsigned char *buf );
// commercial scan functions
extern void process_frame( uchar *buf, int bufSize, sequence_header *sh, macroblock_state_info *msi );
extern int process_slice( uint slice_vertical_position, bool fullSize, bitmap *bm, sequence_header *sh, picture_header *ph, uint *mb_address, uint *rows, macroblock_state_info *msi );
extern int GetSHLength( uchar *buffer, int bufferPos  );
extern int GetSSELength( uchar *buffer, int bufferPos );
extern bool decode_pattern_bits( sequence_header *sh, int *cbp );
extern void determine_frame_type( frame_type *frameType, macroblock_state_info *msi, RegionType lastRegionType, bool bIncludeBlackFrames );
extern __int64 CalcWeightedAvgTSRate( int pid, StreamRateList *stream_rate_list, __int64 TS_rate_current );
extern void SetVerticalSizeTo1080( unsigned char *buf );

#endif
