#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#include <memory.h>
#include <math.h>

#include "bcdmux.h"
#include "mpeg2stuff.h"
#include "process.h"
#include "idct.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <afxtempl.h>

// The one, the only instance
idct_M64ASM gIDCT_MMX;
idct_M128ASM gIDCT_SSE;
idct *gpIDCT = NULL;

//idct_M128IVEC gIDCT3;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// multithreaded buffered 64bit file I/O

bool ChooseIDCT(void)
{
	bool ret = false;
	if (IsProcessorFeaturePresent(PF_XMMI_INSTRUCTIONS_AVAILABLE))
	{
		gpIDCT = &gIDCT_SSE;
		ret = true;
	}
	else if (IsProcessorFeaturePresent(PF_MMX_INSTRUCTIONS_AVAILABLE))
	{
		gpIDCT = &gIDCT_MMX;
		ret = true;
	}
	return ret;
}
bool FileReadBuffer::Start(int bufferSize, int alignment)
{
	m_exit = false;

	m_file = INVALID_HANDLE_VALUE;
	m_alignment = alignment;

	m_using = NULL;
	m_filling = NULL;

	// alloc buffers
	int bufferSizeX = bufferSize + 3 * alignment;

	m_buffer[0].m_bufferX = new uchar[bufferSizeX];
	m_buffer[0].m_buffer = (uchar*)(((uint)m_buffer[0].m_bufferX + alignment - 1) & (~(alignment - 1)));
	m_buffer[0].m_size = bufferSize;
	m_buffer[0].m_valid = false;

	m_buffer[1].m_bufferX = new uchar[bufferSizeX];
	m_buffer[1].m_buffer = (uchar*)(((uint)m_buffer[1].m_bufferX + alignment - 1) & (~(alignment - 1)));
	m_buffer[1].m_size = bufferSize;
	m_buffer[1].m_valid = false;

	// init critical sections
	InitializeCriticalSection(&m_fileCSec);
	InitializeCriticalSection(&m_bufferCSec);
	InitializeCriticalSection(&m_buffer[0].m_cSec);
	InitializeCriticalSection(&m_buffer[1].m_cSec);

	// init sub thread
	unsigned int threadID;
	m_thread = (HANDLE)_beginthreadex(NULL, 1024 * 16, StaticThread, this, 0, &threadID);

	Sleep(1);

	return true;
}

void FileReadBuffer::Stop(void)
{
	if (m_using)
	{
		// give up access to the currently being used buffer
		LeaveCriticalSection(&m_using->m_cSec);
		m_using = NULL;
	}

	// wait and grab access to the file
	EnterCriticalSection(&m_fileCSec);

	if (m_file != INVALID_HANDLE_VALUE)
	{
		CloseHandle(m_file);
		m_file = INVALID_HANDLE_VALUE;
	}

	m_exit = true;
	// give up access to the file
	LeaveCriticalSection(&m_fileCSec);

	// wait for subthread handle to end
	if (WaitForSingleObject(m_thread, 1000) != WAIT_OBJECT_0)
	{
		// kill the thread?
		// TerminateThread() is kind of nasty though.
	}
	CloseHandle(m_thread);

	// delete critical sections
	DeleteCriticalSection(&m_fileCSec);
	DeleteCriticalSection(&m_bufferCSec);
	DeleteCriticalSection(&m_buffer[0].m_cSec);
	DeleteCriticalSection(&m_buffer[1].m_cSec);

	// free buffers
	if (m_buffer[0].m_bufferX)
		delete []m_buffer[0].m_bufferX;
	if (m_buffer[1].m_bufferX)
		delete []m_buffer[1].m_bufferX;
}

bool FileReadBuffer::Open(char *fileName)
{
	// fail if a file is already open
	if (m_file != INVALID_HANDLE_VALUE)
		return false;

	// grab access to the file
	EnterCriticalSection(&m_fileCSec);

	// clear buffers
	m_buffer[0].m_valid = false;
	m_buffer[1].m_valid = false;

	m_file = CreateFile(fileName, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL | FILE_FLAG_NO_BUFFERING | FILE_FLAG_RANDOM_ACCESS, NULL);

	if (m_file == INVALID_HANDLE_VALUE)
	{
		// give up access to the file
		LeaveCriticalSection(&m_fileCSec);
		return false;	// oops, file open failure
	}

	DWORD hi;
	DWORD lo = GetFileSize(m_file, &hi);
	m_fileSize = (__int64(hi) << 32) | (__int64(lo) & 0xffffffff);

	m_filePosCur = 0;
	m_filePosRead = 0;
	m_error = false;
	m_EOF = false;
	m_using = NULL;
	m_filling = NULL;

	// give up access to the file
	LeaveCriticalSection(&m_fileCSec);
	return true;
}

void FileReadBuffer::Close(void)
{
	if (m_using)
	{
		// give up access to the currently being used buffer
		LeaveCriticalSection(&m_using->m_cSec);
		m_using = NULL;
	}

	// grab access to the file
	EnterCriticalSection(&m_fileCSec);

	if (m_file != INVALID_HANDLE_VALUE)
	{
		CloseHandle(m_file);
		m_file = INVALID_HANDLE_VALUE;
	}

	// give up access to the file
	LeaveCriticalSection(&m_fileCSec);
}

__int64 FileReadBuffer::Seek(__int64 pos)
{
	if (m_using)
	{
		// give up access to any currently used buffer
		LeaveCriticalSection(&m_using->m_cSec);
		m_using = NULL;
	}
	// grab access to the file
	EnterCriticalSection(&m_fileCSec);

	// align down to prior sector size
	__int64 posAligned = pos & (~(m_alignment - 1));

	LONG hi = (LONG)(posAligned >> 32);
	LONG lo = SetFilePointer(m_file, (LONG)(posAligned&0xffffffff), &hi, FILE_BEGIN);
	__int64 newOffset = (__int64(hi) << 32) | (__int64(lo) & 0xffffffff);

	// clear buffers
	m_buffer[0].m_valid = false;
	m_buffer[1].m_valid = false;

	m_filePosCur = pos;
	m_filePosRead = newOffset;
	m_error = false;
	m_EOF = false;
	m_using = NULL;
	m_filling = NULL;

	// give up access to the file
	LeaveCriticalSection(&m_fileCSec);
	return pos;
}

int FileReadBuffer::Read(void *dest, int bytes)
{
	if (m_file == INVALID_HANDLE_VALUE)
		return -1;

	// see if we have a current buffer
	// if not, find a filled buffer or at least a buffer that is filling - wait on buffer's critical section to do so
	int counter = 0;
	while (!m_using)
	{
		Buffer *filling = m_filling;
		Buffer *buf = &m_buffer[0];
		if (buf != filling)
		{
			EnterCriticalSection(&buf->m_cSec);

			if (buf->m_valid &&
				(m_filePosCur >= buf->m_startOffset) &&
				(m_filePosCur < buf->m_startOffset + buf->m_length))
			{
				m_using = buf;
			}
			else
				LeaveCriticalSection(&buf->m_cSec);
		}

		buf = &m_buffer[1];
		if (!m_using && (buf != filling))
		{
			EnterCriticalSection(&buf->m_cSec);

			if (m_buffer[1].m_valid &&
				(m_filePosCur >= buf->m_startOffset) &&
				(m_filePosCur < buf->m_startOffset + buf->m_length))
			{
				m_using = buf;
			}
			else
				LeaveCriticalSection(&buf->m_cSec);
		}

		if (!m_using && filling)
		{
			// let's wait for this one to fill up, and see if it's useful
			EnterCriticalSection(&filling->m_cSec);

			if (filling->m_valid &&
				(m_filePosCur >= filling->m_startOffset) &&
				(m_filePosCur < filling->m_startOffset + filling->m_length))
			{
				m_using = filling;
			}
			else
				LeaveCriticalSection(&filling->m_cSec);
		}

		if (!m_using)
		{
			if (m_error)
				return -1;		// there was a read error
			else if (m_EOF)
				return 0;		// end of file reached
			else if (counter > 100)
				return -1;		// we may have to give the subthread time to start and/or react to a file change
			else
				Sleep(10);
		}
		counter++;
	}

	// read from buffer
	int bytesToCopy = m_using->m_length - (int)(m_filePosCur - m_using->m_startOffset);
	if (bytesToCopy > bytes)
		bytesToCopy = bytes;

	memcpy((uchar*)dest, m_using->m_buffer + (int)(m_filePosCur - m_using->m_startOffset), bytesToCopy);

	m_filePosCur += (__int64)bytesToCopy;

	// end of buffer? free and clear it
	if (m_filePosCur >= m_using->m_startOffset + (__int64)m_using->m_length)
	{
		m_using->m_valid = false;
		LeaveCriticalSection(&m_using->m_cSec);
		m_using = NULL;
	}

	// need to read more bytes? recurse...
	if (bytes > bytesToCopy)
	{
		int r = Read((uchar*)dest + bytesToCopy, bytes - bytesToCopy);
		if (r >= 0)
			bytesToCopy += r;
	}

	return bytesToCopy;
}

unsigned int WINAPI FileReadBuffer::StaticThread(LPVOID data)
{
	FileReadBuffer *fb = (FileReadBuffer *)data;
	fb->Thread();

	return 0;
}

void FileReadBuffer::Thread(void)
{
	while (!m_exit)
	{
		// wait and grab access to the file
		EnterCriticalSection(&m_fileCSec);

		if (m_file == INVALID_HANDLE_VALUE)
		{
			// give up access to the file
			LeaveCriticalSection(&m_fileCSec);
			// wait to see if a file will be opened
			Sleep(1);
			continue;
		}

		// m_filling should always be NULL here
		Buffer *u = m_using;
		Buffer *buf = &m_buffer[0];
		if (buf != u)
		{
			// grab access to this buffer
			EnterCriticalSection(&buf->m_cSec);

			if (!buf->m_valid)
			{
				m_filling = buf;
			}
			else
				// give up access to this buffer
				LeaveCriticalSection(&buf->m_cSec);
		}

		buf = &m_buffer[1];
		if (!m_filling && (buf != u))
		{
			// grab access to this buffer
			EnterCriticalSection(&buf->m_cSec);

			if (!buf->m_valid)
			{
				m_filling = buf;
			}
			else
				// give up access to this buffer
				LeaveCriticalSection(&buf->m_cSec);
		}

		if (!m_filling)
		{
			if (u)
			{
				// give up access to the file
				LeaveCriticalSection(&m_fileCSec);

				// both buffers are full, but this one is being used, let's wait for it to be done with.
				// wait and grab access to this buffer
				EnterCriticalSection(&u->m_cSec);

				// give up access to this buffer
				LeaveCriticalSection(&u->m_cSec);
			}
			else
			{
				// give up access to the file
				LeaveCriticalSection(&m_fileCSec);
				// both buffers are full, and aren't even being used yet. Try again shortly
				Sleep(1);
			}
		}
		else
		{
			// fill this buffer

			// no buffering has some restrictions: all file references must be in sector size
			// alignment, and so must the memory address we're copying into. m_scanBufferAlignment
			// is the "sector size" we're aligning sizes and addresses to. Once we read in the sectors
			// we want, then we copy the actual requested data to the buffer specified.

			int bytesToRead = m_filling->m_size;
			if ((__int64)bytesToRead > m_fileSize - m_filePosRead)
				bytesToRead = (int)(m_fileSize - m_filePosRead);

			if (bytesToRead > 0)
			{
				// align up to next sector size
				int bytesToReadAligned = (int)(((m_filePosRead + (__int64)bytesToRead + m_alignment - 1) & (~(m_alignment - 1))) - m_filePosRead);

				DWORD r;
				if (!ReadFile(m_file, m_filling->m_buffer, (DWORD)bytesToReadAligned, &r, NULL))
					r = 0;

				m_filling->m_length = (int)r;
				if (m_filling->m_length > bytesToRead)
					m_filling->m_length = bytesToRead;
			}
			else
				m_filling->m_length = 0;

			if (m_filling->m_length > 0)
			{
				m_filling->m_valid = true;

				m_filling->m_startOffset = m_filePosRead;

				m_filePosRead += (__int64)m_filling->m_length;
			}

			m_error = (m_filling->m_length < bytesToRead);
			m_EOF = m_error || (m_filePosRead + (__int64)m_filling->m_length == m_fileSize);

			// give up access to this buffer
			LeaveCriticalSection(&m_filling->m_cSec);
			m_filling = NULL;

			// give up access to the file
			LeaveCriticalSection(&m_fileCSec);
		}
	}
}


// enable/disable some code

// use original scan_channel_info(), or advanced one that gets
// information from the sequence headers of PES video streams
//#define ENABLE_ORIGINAL_SCAN_CHANNEL_INFO

// a routine that peeks at a file and dumps to debug out full
// information of all the streams it recognises
#define ENABLE_PEEK_AT_FILE

// switches to enable debug info
//#define SHOW_INFO_SEEN
//#define SHOW_INFO_PAT
//#define SHOW_INFO_PMT
//#define SHOW_INFO_MPEG2_VIDEO
//#define SHOW_INFO_AC3

//#define REPORT_SLICE_DECODE_FAILURES
//#define BREAK_ON_NAUGHTINESS_IN_SLICE
//#define SHOW_DECODE_TIME

// must be defined
#define DECODE_FIRST_I_FRAME

// header for a TGA file, was once used during debug
typedef struct
{
	uint zero0 : 16;
	uint two : 16;
	uint zero1 : 32;
	uint zero2 : 32;
	uint width : 16;
	uint height : 16;
	uchar twentyfour;
	uchar zero3;
} tga_header;

// double buffered bitmaps for use with the progress dialog.
// this mechanism needs to be replaced, having the conversion
// routing passing an I-frame buffer to the UI to decode when
// it wants to, and have the UI do it's own double buffering.
pdlg_buffer pdlg_buffers[2] = { { { NULL, }, }, { { NULL, } } };


// read time stamp counter
typedef union
{
	struct
	{
		unsigned long L;
		unsigned long H;
	};
	uint64 Q;
} UL64;

uint64 rdtsc(void)
{
	UL64 res;
	__asm
	{
		push edx
		_emit 0x0f
		_emit 0x31
		mov res.L, eax
		mov res.H, edx
		pop edx
	}
	return res.Q;
}

//
#ifndef FINAL_RELEASE
#define ERR_C_BUFFER_SIZE	  256
void __cdecl _Error(char *format, ...)
{
	va_list arglist;
	char buffer[ERR_C_BUFFER_SIZE];

	va_start(arglist, format);
	vsprintf(buffer, format, arglist);
	va_end(arglist);

	OutputDebugString(buffer);

	return;
}
#endif

// random tables
char *sample_rate_table[] =
{
	"48",
	"44.1",
	"32",
	"Reserved",
	"48 or 44.1",
	"48 or 32",
	"44.1 or 32",
	"48 or 44.1 or 32",
};

char *bite_rate_table[] =
{
	"32",
	"40",
	"48",
	"56",
	"64",
	"80",
	"96",
	"112",
	"128",
	"160",
	"192",
	"224",
	"256",
	"320",
	"384",
	"448",
	"512",
	"576",
	"640",
};

char *surround_mode_table[] =
{
	"Not indicated",
	"Not Dolby surround encoded",
	"Dolby surround encoded",
	"Reserved",
};

char *num_channels_table[] =
{
	"1+1",
	"1/0",
	"2/0",
	"3/0",
	"2/1",
	"3/1",
	"2/2",
	"3/2",
	"1",
	"<=2",
	"<=3",
	"<=4",
	"<=5",
	"<=6",
	"Reserved",
	"Reserved",
};

// testing stuff

#ifdef SHOW_INFO_SEEN
char *pidTypeName[] = {
	"PID_UNKNOWN",
	"PID_NULL",
	"PID_PAT",
	"PID_PMT",
	"PID_NIT",
	"PID_MPEG2_VIDEO",
	"PID_AC3_AUDIO",
	"PID_MPEG2_AUDIO",
	"PID_MPEG1_AUDIO",
	"PID_PSIP",
};
#endif


//-----------------------------------------------------

PIDinfo *find_PID(PIDinfoGroup *pg, int pid)
{
	PIDinfo *p = pg->pids;
	for (int i=0; i<pg->seen; i++, p++)
		if (p->pid == pid)
			return p;
	if (pg->seen < MAX_PIDS)
	{
		// new!
		pg->seen++;
		p->pid = pid;
		p->count = 0;
		p->sType = STREAM_UNKNOWN;
		p->pType = PID_UNKNOWN;
		p->processed = false;
		p->buffer = NULL;
		p->bufferSize = 0;
		p->bufferPos = 0;
		p->reading = false;
		p->confirmedRead = false;
		return p;
	}
	return NULL;
}

PIDinfo *find_PID_no_new(PIDinfoGroup *pg, int pid)
{
	PIDinfo *p = pg->pids;
	for (int i=0; i<pg->seen; i++, p++)
		if (p->pid == pid)
			return p;
	return NULL;
}

PIDinfo *find_PID_Type( PIDinfoGroup *pg, pidType pt )
{
	PIDinfo *p = pg->pids;
	for( int i = 0; i < pg->seen; i++, p++ )
		if( p->pType == pt )
			return p;
	return NULL;
}

// ------------------------------------------------------------------------------------

#ifdef ENABLE_ORIGINAL_SCAN_CHANNEL_INFO
bool scan_channel_info(const char* fileName, int maxScan, int maxChannels, ATSC_CHANNEL_INFO* channels, int& numChannels)
{
	numChannels = 0;

	memset(channels, 0, sizeof(ATSC_CHANNEL_INFO) * maxChannels);

	FILE64* f = fopen64(fileName, "rb");
	if (!f)
	{
		_Error("Couldn't open file for channel scan!\n");
		return false;
	}

	// align to first packet
	align_to_next_packet(f);

	unsigned char buf[188];

	unsigned char tablebuf[1024];
	unsigned int tablepos = 0;

	bool reading = false;

	while (fread64(buf, 188, 1, f) == 1 && ftell64(f) < maxScan)
	{
		// Check sync byte
		if ((buf[0] != 0x47) && (buf[0] != 0x72) && (buf[0] != 0x29))
			continue;

		// Get error
		bool errorBit = (buf[1] & 0x80) != 0;
		if (errorBit)
			continue;

		// Get pid
		int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;

		// Skip this block
		if (pid < 0x1000 || pid >= 0x1FFF)
			continue;

		// Get adaptation header info
		int adapt_len = 0;
		int adaptation = (buf[3] & 0x30) >> 4;
		if (adaptation == 0)
			continue;
		else if (adaptation == 0x2)
			adapt_len = 184;
		else if (adaptation == 0x3)
			adapt_len = buf[4] + 1;
		if (adapt_len > 184)
			continue;

		// Get pointer length
		int pointer_len = buf[4 + adapt_len] + 1;

		// Get payload start indicator
		bool start;
		start = (buf[1] & 0x40) != 0;

		if (start && reading)
		{
			memcpy(tablebuf + tablepos, buf + 4 + adapt_len + 1, pointer_len - 1);

			unsigned int pos = 0;
			//while (pos < tablepos)
			{
				set_buf(tablebuf + pos, tablepos - pos, false);

				unsigned char section_id	= get_bits(8);
											  get_bits(4);
				unsigned int section_len	= get_bits(12);
				unsigned int transport_id	= get_bits(16);
											  get_bits(2);
				unsigned int version_num	= get_bits(5);
				unsigned int current_next	= get_bits(1);
				unsigned int section_num	= get_bits(8);
				unsigned int last_section	= get_bits(8);
				unsigned int protocol_ver	= get_bits(8);

				switch (section_id)
				{
				  case 0xC7:
					{
						break;
					}
				  case 0xC8:
					{
						int num_ch = get_bits(8);

						for (int i = 0; i < num_ch; i++)
						{
							char *name = channels[numChannels].name;
							name[0] =								(char)get_bits(16);
							name[1] =								(char)get_bits(16);
							name[2] =								(char)get_bits(16);
							name[3] =								(char)get_bits(16);
							name[4] =								(char)get_bits(16);
							name[5] =								(char)get_bits(16);
							name[6] =								(char)get_bits(16);
							name[7] =								'\0';

																	get_bits(4);
							channels[numChannels].majorChannel =	get_bits(10);
							channels[numChannels].minorChannel =	get_bits(10);

							unsigned int modulation =				get_bits(8);
							unsigned int carrier =					get_bits(32);
							unsigned int channel_TSID =				get_bits(16);
							unsigned int program_number =			get_bits(16);

							/* Ignore these bits for now */			get_bits(32);	// Hidden, encrypted, etc. bits

																	get_bits(6);

							unsigned int ch_desclen	=				get_bits(10);

							unsigned int pos = 0;
							while (pos < ch_desclen)
							{
								unsigned int desc_id =				get_bits(8);
								unsigned int desc_len =				get_bits(8);

								switch (desc_id)
								{
								  case 0xA1:	// Stream descriptors
									{
																	get_bits(16); // Ignore PCR_PID
										int num_elem =				get_bits(8);
										for (int j = 0; j < num_elem; j++)
										{
											unsigned int stream_type =	get_bits(8);
																		get_bits(3);
											unsigned int pid =			get_bits(13);
											unsigned int language =		get_bits(24);

											switch (stream_type)	// MPEG-2 video..
											{
											  case 0x02:
												{
													channels[numChannels].videopid = pid;
													break;
												}
											  case 0x81:
												{
													if (channels[numChannels].audiopid != 0 && numChannels + 1 < maxChannels)
													{
														memcpy(channels + numChannels + 1, channels + numChannels, sizeof(ATSC_CHANNEL_INFO));
														numChannels++;
													}
													channels[numChannels].audiopid = pid;
													break;
												}
											}
										}
										break;
									}
								}

								pos += desc_len + 2;
							}

							if (modulation == 4)	// Only if ATSC digital channel.. Ignore ANALOG channels.
								numChannels++;
						}

						fclose64(f);

						return true;
					}
				}

				pos += 3 + section_len;
			}

			tablepos = 0;
		}
		else if (start)
			reading = true;

		// Add the payload for this packet to the current buffer
		if (reading && (184 - adapt_len) > 0)
		{
			if (tablepos + 184 - adapt_len - pointer_len > 1024)
			{
				_Error("Bad program section length (> 1024)\n");
				continue;
			}
			memcpy(tablebuf + tablepos, buf + 4 + adapt_len + pointer_len, 184 - adapt_len - pointer_len);
			tablepos += 184 - adapt_len - pointer_len;
		}
	}

	fclose64(f);

	return true;
}

#else	// ENABLE_ORIGINAL_SCAN_CHANNEL_INFO

// saved PAT and PMT packets
extern PatPmt_Packets PatPmtPackets;
bool scan_channel_info(const void *inBuffer, int maxScan, int maxChannels, ATSC_CHANNEL_INFO* channels,
					   int& numChannels, CWordArray *pPIDDeleteList, CWordArray *pHiddenPidDeleteList )
{

	PIDinfoGroup PIDs = { 4,
		{	{ 0x1fff, 0, STREAM_UNKNOWN, PID_NULL, true, NULL, 0, 0, -1, false, false },	// skip NULL packets
			{ 0x1fef, 0, STREAM_UNKNOWN, PID_NULL, true, NULL, 0, 0, -1, false, false },	// skip NULL packets (HBOHD)
			{ 0, 0, STREAM_TABLE, PID_PAT, false, NULL, 0, 0, -1, false, false }, // PID 0 = PAT tables
			{ 0x1ffb, 0, STREAM_TABLE, PID_PSIP, false, NULL, 0, 0, -1, false, false }// PID 1ffb = PSIP info
		} };

	numChannels = 0;
	memset(channels, 0, sizeof(ATSC_CHANNEL_INFO) * maxChannels);

	uchar *sequenceHeader = NULL;
	int adaptationLength = 0;

	int table_id = 0xff;
	int pointer_length = 0;
	int section_length = 0;
	int section_length_size = 0;
	int section_length_expected = 0;
	bool audioFound = false;

	int offset = align_to_next_packet((uchar*)inBuffer, 20*188);

	int fPos = offset;
	uchar *buf = (uchar*)inBuffer + fPos;

	bool reading = false;
	bool bIgnorePSIP = false;
	int our_videopid = -2;
	__int64 totalRead = 0;
	int video_PID = -1;
	StreamRateList srl;

	while ((fPos += 188) < maxScan)
	{
		totalRead += 188;
		buf = (uchar*)inBuffer + fPos;

		// Check sync byte
		if(( buf[0] != 0x47) && (buf[0] != 0x72) && (buf[0] != 0x29 ))
			continue;

		// Get error
		bool errorBit = (buf[1] & 0x80) != 0;
		if (errorBit)
			continue;

		// Get pid
		int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;

if( pid == 0x31 )
 int foo=0;

		// is this a pid we might be interested in deleteing later?
		// these are "visible" pids that the user can select to keep or strip
		// "Hidden" pids like those added from the Master Guide Table(MGT) can
		//  found in pHiddenPidDeleteList.
		bool found = false;
		if( pid == 0x1fff || pid == 0x1fef || pid == 0x1ffb && pPIDDeleteList->GetCount() < 3 )
		{
			for( int i = 0; i < pPIDDeleteList->GetCount(); i++ )
			{
				if( pPIDDeleteList->GetAt( i ) == pid )
				{
					found = true;
					break;
				}
			}

			if( !found )
				pPIDDeleteList->Add( pid );
		}
			
		// get info for this PID
		PIDinfo *p = find_PID(&PIDs, pid);
		if (!p)
			continue;	// oops, too many pids
		p->count++;

		// we are only done with PID streams after the PMT has been
		//  processed.  The StreamType and pidType are set using the PMT
		//  information
		if( p->sType == STREAM_UNKNOWN && p->pType == PID_UNKNOWN )
		{
			PIDinfo *pt = find_PID_Type( &PIDs, PID_PMT );
			if( pt )
				if( pt->processed )
					continue;
		}

		// Get adaptation header info
		int adaptLength = 0;
		int adaptation = (buf[3] & 0x30) >> 4;
		if (adaptation == 0)
			continue;
		else if (adaptation == 0x2)
			adaptLength = 184;
		else if (adaptation == 0x3)
			adaptLength = buf[4] + 1;
		if (adaptLength > 184)
			continue;

		// Get payload start indicator
		bool start = (buf[1] & 0x40) != 0;

		// If the pid is 0x1ffb and the section_length has not been met yet,
	    //  consider this a "duplicate" of the PSIP information being sent
		// Note: this "fix" has only been tested for PSIP data that only spans 2 packets.
		//  The maximum number of bytes possible is up to 1021 bytes so this may bite me later
		if( start && pid == 0x1ffb && !p->processed )
		{
			pointer_length = buf[ 4 + adaptLength ] + 1;
			table_id = buf[ 4 + pointer_length ];

			if( table_id == 0xc8 || table_id == 0xc9) // if this is a VCT table
			{
				section_length = ((( buf[4 + 1 + pointer_length ] & 0x0f ) << 8 ))| buf[ 4 + 1 + pointer_length + 1 ];

				if( section_length < 180 ) // PSIP data is all in one packet
					section_length_size = section_length_expected = section_length;
				else
				{
					section_length_expected = section_length;
					section_length_size = section_length - 180;
					p->reading = false;
				}
			}
		}

		// special processing for PSIP data
		if( !start && pid == 0x1ffb && !p->processed )
		{
			if( section_length_size != section_length_expected )
			{
				if( table_id == 0xc8 || table_id == 0xc9 ) // if this is a VCT table
					// we are still reading this PES packet
					section_length_size += section_length_expected - section_length_size;
			}
			else // ready to process the data now
				start = true;
		}

		// see if there is a sequence_header hiding in there that we may need later
		if(( buf[1] & 0x40 ) == 0 && adaptation == 0x3 )
		{
			uint start_code = 0;
			for( int i = 0; i < 188 - 4 - adaptLength; i++ )
			{
				start_code = ( start_code << 8 ) | buf[4 + adaptLength + i];
				if( start_code == 0x000001b3 )
				{
					// save this packet, we will use it later
					sequenceHeader = buf;
					adaptationLength = adaptLength;
					break;
				}
			}
		}

		// process content of packet
		switch (p->sType)
		{
		case STREAM_UNKNOWN:
			p->processed = true;
			break;

		case STREAM_TABLE:
			{
				if (!p->buffer)
				{
					// alloc a table buffer
					p->buffer = new uchar[TABLE_BUFFER_SIZE];
					p->bufferSize = TABLE_BUFFER_SIZE;
					p->bufferPos = 0;
					p->reading = false;
				}

				int pointer_length = buf[4 + adaptLength] + 1;

				// if the payload_unit_start_indicator is 0, there is no pointer_field.
				//  special handling for pid 0x1ffb, PSIP data
				if( (buf[1] & 0x40) == 0 )
					pointer_length = 0;

				if (start)
				{
					if (p->reading)
					{
						if( pid != 0x1ffb )
						{
							memcpy(p->buffer + p->bufferPos, buf + 4 + adaptLength + 1, pointer_length - 1);
							p->bufferPos += pointer_length - 1;
						}

						// if this is a PMT PID but the table_id isn't a program_map_table
						if( p->pType == PID_PMT && buf[5] != 0x02 )
							break;

						if (p->bufferPos > 0)
						{
							set_buf(p->buffer, 0, false);

							uchar table_id = get_bits(8);
							discard_bits(4);
							int section_length = get_bits(12);
//							discard_bits(16);		// uint transport_stream_id	= get_bits(16);
							int program_number = get_bits( 16 ); //program_number(PMT), transport_stream_id(PAT\PSIP) 
							discard_bits(2);
							discard_bits(5);		// uint version_number = get_bits(5);
							discard_bits(1);		// uint current_next_indicator = get_bits(1);
							discard_bits(8);		// uint section_number = get_bits(8);
							discard_bits(8);		// uint last_section_number = get_bits(8);

							switch (p->pType)
							{
								case PID_PAT: // table_id = 0x00, program_association_section
								{
									// save copy of the PAT
									if( PatPmtPackets.pPAT_Packet == NULL )
									{
										PatPmtPackets.pPAT_Packet = new uchar[188];
										memcpy( PatPmtPackets.pPAT_Packet, buf, 188 );
									}

									p->processed = true;
									p->reading = false;

									for (int i=5; i<section_length-4; i+=4)
									{
										uint program_number = get_bits(16);
										discard_bits(3);

										if (program_number == 0)
										{
											uint network_PID = get_bits(13);

											PIDinfo *pm = find_PID(&PIDs, network_PID);
											if (pm)
											{
												// if network_PID is assigned to 0x1ffb then
												//  just leave the pm-> settings alone otherwise
												//  we may loose the PSIP data
												if( network_PID != 0x1ffb )
												{
													pm->sType = STREAM_TABLE;
													pm->pType = PID_NIT;
													pm->processed = false; //because I don't care to see a NIT yet
												}
											}
										}

										else
										{
											uint program_map_PID = get_bits(13);

											PIDinfo *pm = find_PID(&PIDs, program_map_PID);
											if (pm)
											{
												pm->sType = STREAM_TABLE;
												pm->pType = PID_PMT;
												pm->processed = false;
											}
										}
									}
									// CRC = get_bits(32);

								}// end case PID_PAT

								break;

								case PID_PMT:	// table_id = 0x02, TS_program_map_section
								{
									// bail if table_id isn't a PMT
									if( table_id != 0x02 )
										break;
									
									p->reading = false;

									discard_bits(3);

									uint PCR_PID = get_bits(13);

									discard_bits(4);
									int program_info_length = get_bits(12);
									int i = 9;

									bool discard_PMT = false;

									// descriptor
									for (int j = 0; j < program_info_length; )
									{
//										discard_bits(8);//uint descriptor_tag = get_bits(8);
										uint descriptor_tag = get_bits(8);

										// ignore PMT's that contain CA_descriptors.  these
										//  are considered "scrambled" channels 
										// Also, set to ignore PSIP info as this is probably
										//  a QAM stream
// commented out because some streams have a CA_descriptor defined but the
//  channel obviously isn't scrambled.
/*
										if( descriptor_tag == 0x09 )
										{
											discard_PMT = true;
											bIgnorePSIP = true;
											break;
										}
*/

										int descriptor_length = get_bits(8);

										for (int k = 0; k < descriptor_length; k++)
											discard_bits(8);

										i += 2 + descriptor_length;
										j += 2 + descriptor_length;
									}

									if( discard_PMT )
										continue;

									// verify that the PMT contains a video and an audio pid
									//  before saving it.  some streams contain incomplete PMT's
									if( !VerifyPMT( buf ))
										continue;

									// save a copy of the PMT
									bool found = false;
									if( PatPmtPackets.numOfPMTPackets == 0 )
									{
										PatPmtPackets.pPMT_Packet[PatPmtPackets.numOfPMTPackets] = new uchar[188];
										memcpy( PatPmtPackets.pPMT_Packet[PatPmtPackets.numOfPMTPackets], buf, 188 );
										PatPmtPackets.numOfPMTPackets++;
									}
									else
									{	// see if this is a new PMT
										for( int n = 0; n < PatPmtPackets.numOfPMTPackets; ++n )
										{
											int pmt_pid = *( PatPmtPackets.pPMT_Packet[n] + 1 ) & 0x1f;
											pmt_pid <<= 8;
											pmt_pid |= *(PatPmtPackets.pPMT_Packet[n] + 2);

											if( pmt_pid == pid )
											{
												found = true;
												break;
											}
										}

										if( !found )
										{
											PatPmtPackets.pPMT_Packet[PatPmtPackets.numOfPMTPackets] = new uchar[188];
											memcpy( PatPmtPackets.pPMT_Packet[PatPmtPackets.numOfPMTPackets], buf, 188 );
											PatPmtPackets.numOfPMTPackets++;
										}
									}


									for (; i < section_length-4; )
									{
										uint stream_type = get_bits(8);
										discard_bits(3);
										uint elementary_PID = get_bits(13);
										discard_bits(4);
										int ES_info_length = get_bits(12);
										i += 5;

										bool bAC3_descriptor = false;
										// descriptor again
										for( int j = 0; j < ES_info_length; )
										{
											//discard_bits(8);
											uint descriptor_tag = get_bits(8);
											int descriptor_length = get_bits(8);

											if( descriptor_tag == 0x6a )
												bAC3_descriptor = true;

											for (int k = 0; k < descriptor_length; k++)
													discard_bits(8);

											i += 2 + descriptor_length;
											j += 2 + descriptor_length;
										}

										PIDinfo *pm = find_PID(&PIDs, elementary_PID);
										if (pm)
										{
											switch (stream_type)
											{
												case 0x02: // ISO 13818-2 (MPEG-2 Video)
												case 0x80: // User Private
												{
													bool videoFound = true;

													pm->sType = STREAM_PES;
													pm->pType = PID_MPEG2_VIDEO;
													// reset to false so that if a PSIP packet is found
													//  it can overwrite these settings
													pm->processed = false;

													// alloc a video buffer
													if( !pm->buffer )
													{
														// just a small "peek" buffer
														pm->buffer = new uchar[1024];
														pm->bufferSize = 1024;
														// a much larger buffer
//														pm->buffer = new uchar[8960];
//														pm->bufferSize = 8960;
													}
													pm->bufferPos = 0;

													// create channel description?
													bool create_new = true;
													for (int k = 0; k < numChannels; k++)
													{
														if( (WORD)channels[k].videopid == (int)elementary_PID &&
															(WORD)channels[k].program_number == program_number )
														{
															// oh, already seen this one
															create_new = false;

															// save the PMT_pid
															channels[k].PMT_pid = pid;
															// save the PCR_PID pid
															channels[k].PCR_pid = PCR_PID;
															// save the program_number
															channels[k].program_number = program_number;

															break;
														}
													}
													if( create_new && ( numChannels < maxChannels ))
													{
														// create a new description
														strcpy(channels[numChannels].name, "NONAME");
														channels[numChannels].majorChannel = numChannels;
														channels[numChannels].minorChannel = 1;
														// videopid = 32 bits, 1st 16 bits = pType, last 16 bits = pid
														channels[numChannels].videopid = pm->pType << 16 | (int)elementary_PID;
														channels[numChannels].audiopid = -1;
														channels[numChannels].PMT_pid = pid;
														channels[numChannels].PCR_pid = PCR_PID;
														channels[numChannels].program_number = program_number;
														numChannels++;
													}

													our_videopid = elementary_PID;

													if( PCR_PID != elementary_PID )
													{
														PIDinfo *pInfo = find_PID( &PIDs, PCR_PID );
														if( pInfo )
														{
															pInfo->sType = STREAM_PES;
															pInfo->pType = PID_PCR;
														}
													}

												}// end cases 0x02

												break;

												case 0x03: // ISO 11172 Audio (MPEG-1 Audio)
												case 0x04: // ISO 13818-3 Audio (MPEG-2 Audio)
												case 0x06: // ISO 13818-1 private data
												case 0x0f: // ISO 13818-7 Audio (AAC Audio )
												case 0x81: // ATSC A/53 Audio (Dolby AC-3 Audio)
												{
													if( stream_type == 0x06 && !bAC3_descriptor )
														break;

													pm->sType = STREAM_PES;

													if( stream_type == 0x03 )
														pm->pType = PID_MPEG1_AUDIO;

													if( stream_type == 0x04 )
														pm->pType = PID_MPEG2_AUDIO;

													if( stream_type == 0x0f )
														pm->pType = PID_AAC_AUDIO;

													if( stream_type == 0x81 || stream_type == 0x06 )
														pm->pType = PID_AC3_AUDIO;

													// if we have seen this pid more than 255 times without
													//  having seen a PSIP packet(PID_PSIP), then assume there are
													//  no PSIP packets available with channel information
													//  and we are done processing this pid
													if( pm->count > 0xff )
														pm->processed = true;
													else // give PSIP packets with channel information a chance
														 //  to overwrite this data
														pm->processed = false;

													// create channel description?
													bool seen_before = false;

													for (int k = 0; k < numChannels; k++)
													{
														// audiopid = 32 bits, 1st 16 bits = pType, last 16 bits = pid
														if( (WORD)channels[k].audiopid == (int)elementary_PID )
														{
															// oh, already seen this one
															seen_before = true;

															// save the PMT_pid
															channels[k].PMT_pid = pid;
															// save the PCR_PID
															channels[k].PCR_pid = PCR_PID;

															break;
														}
													}
													if (!seen_before)
													{
														// look for descriptor with our_videopid but no audiopid,
														int last_k = -1;	// track any with our_videopid already with an audiopid
														bool added = false;
														for( int k = 0; k < numChannels; k++ )
														{   // videopid = 32 bits, 1st 16 bits = pType, last 16 bits = pid
//															if( (WORD)channels[k].videopid == our_videopid )
															if( (WORD)channels[k].PCR_pid == PCR_PID )
															{
																if (channels[k].audiopid < 0)
																{	// add ourselves to this channel
																	// 1st 16 bits are pType, last 16 bits are the pid
																	channels[k].audiopid = pm->pType << 16 | (int)elementary_PID;
																	added = true;

																	// save the PMT_pid
																	channels[k].PMT_pid = pid;
																	// save the PCR_PID
																	channels[k].PCR_pid = PCR_PID;

																	break;
																}
																else
																	last_k = k;
															}
														}
														if( !added && (last_k >= 0) && (numChannels < maxChannels ))
														{
															// make a new entry, copying last_k
															channels[numChannels] = channels[last_k];
															channels[numChannels].minorChannel++;
															// audiopid = 32 bits, 1st 16 bits = pType, last 16 bits = pid
															channels[numChannels].audiopid = pm->pType << 16 | (int)elementary_PID;

															// save the PMT_pid
															channels[numChannels].PMT_pid = pid;
															// save the PCR_PID
															channels[numChannels].PCR_pid = PCR_PID;

															numChannels++;
														}
														// else skip - should be a videopid first?
													}

												}// end case 0x81

												break;

											default:
												pm->sType = STREAM_UNKNOWN;
												pm->pType = PID_UNKNOWN;
												pm->processed = true;	// given that it's unknown, let's ignore this PID
												break;

											}// end switch( stream_type )
										}// end if( pm )
									}// end for (; i < section_length-4; )

									// CRC = get_bits(32);
#ifdef SHOW_INFO_PMT
									_Error("}\n");
#endif
								}// end case PID_PMT

								break;

							case PID_NIT:
								p->processed = true;
								p->reading = false;
								break;

							case PID_PSIP:

								// process at least 1 PMT first to determine if there are
								//  CA_descripters in the stream.  If there are, ingore
								//  the PSIP info as this is probably a QAM stream and
								//  the PSIP data is probably no longer valid. PSIP data is
								//  only useful in an OTA stream.
								if( PatPmtPackets.numOfPMTPackets == 0 || bIgnorePSIP )
									continue;

								switch( table_id )
								{
/* This code can't handle MGT tables that have more than 1 packet worth of table listings
									case 0xc7: // - Master Guide Table(MGT)
									{
										// protocal_version, always 0 until some furture date
										discard_bits( 8 );

										// get tables_definded
										short tables_defined = get_bits( 16 );

										// add all pids defined in the MPG to the "delete" list
										for( int i = 0; i < tables_defined; i++ )
										{
											discard_bits( 16 ); // table_type
											discard_bits( 3 );  // reserved

											short table_type_id = get_bits( 13 );

											// add the table pid to the delete list unless it is
											//  0x1ffb.  this PSIP pid is "visable" to the user
											//  so that they can choose to keep or delete it.
											int found = false;
											if( table_type_id != 0x1ffb )
											{
												for( int i = 0; i < pHiddenPidDeleteList->GetCount(); i++ )
												{
													int pid = pHiddenPidDeleteList->GetAt( i );
													if( table_type_id == pid )
													{
														found = true;
														break;
													}
												}

												if( !found )
													pHiddenPidDeleteList->Add( table_type_id );
											}

											discard_bits( 44 ); // reserved
																// table_type_version_number
																// number_bytes
																// reserved
											short table_type_descriptors_length = get_bits( 12 );

											// skip descriptors
											i += table_type_descriptors_length;
										}

										discard_bits( 4 ); // reserved
										// get the descriptor_length
										short descriptor_length = get_bits( 12 );
										discard_bits( 8 * descriptor_length );

									}// end case 0xc7
									break;
*/

									case 0xc8: // Terrestrial Virtual Channel Table (TVCT)
									case 0xc9: // Cable Virtual Channel Table (CVCT)
									{
										if( audioFound )
											p->processed = true;

										p->reading = false;
										// protocal_version, always 0 until some furture date
										discard_bits( 8 );
										uint num_ch = get_bits( 8 );

										for( uint i = 0; i < num_ch; i++ )
										{
											bool videoFound = false;

											char name[ 8 ];
											name[0] = (char)get_bits( 16 );
											name[1] = (char)get_bits( 16 );
											name[2] = (char)get_bits( 16 );
											name[3] = (char)get_bits( 16 );
											name[4] = (char)get_bits( 16 );
											name[5] = (char)get_bits( 16 );
											name[6] = (char)get_bits( 16 );
											name[7] = '\0';

											discard_bits( 4 );
											int majorChannel = get_bits( 10 );
											int minorChannel = get_bits( 10 );

											uint modulation = get_bits( 8 );
											discard_bits( 32 );		// unsigned int carrier = get_bits(32);
											discard_bits( 16 );		// unsigned int channel_TSID = get_bits(16);
//											discard_bits( 16 );		// unsigned int program_number = get_bits(16);
											int program_number = get_bits( 16 );
											discard_bits( 32 );		// Hidden, encrypted, etc. bits
											discard_bits( 6 );

											uint ch_desclen = get_bits( 10 );


											uint pos = 0;
											while( pos < ch_desclen )
											{
												uint desc_id = get_bits( 8 );
												uint desc_len = get_bits( 8 );

/* Don't ignore non-digital channels so they will display in the Channel listing
												// cheap method to ignore any non-digital channel descriptions
												// while still eating their bits
												if (modulation != 4)
													desc_id = 0xffffff;
*/
												switch( desc_id )
												{
													case 0xA1:	// Service Location Descriptor
													{
//														discard_bits( 16 );		// Ignore PCR_PID
														discard_bits( 3 );
														int PCR_PID = get_bits( 13 );
														int num_elem = get_bits( 8 );
														for( int j = 0; j < num_elem; j++ )
														{
															unsigned int stream_type = get_bits( 8 );
															discard_bits( 3 );
															unsigned int elementary_PID = get_bits( 13 );
															unsigned int language = get_bits( 24 );

															PIDinfo *pm = find_PID( &PIDs, elementary_PID );

															if( pm )
															{
																switch( stream_type )
																{
																	case 0x02: // MPEG-2 Video
																	case 0x80: // user private
																	{
																		pm->sType = STREAM_PES;
																		pm->pType = PID_MPEG2_VIDEO;

																		videoFound = true;

																		// create channel description?
																		bool seen_before = false;
																		for( int k = 0; k < numChannels; k++ )
																		{
																			if( (WORD)channels[ k ].videopid == (int)elementary_PID &&
																				(WORD)channels[k].program_number == program_number )
																			{
																				// fill in info for already existing description
																				strcpy( channels[ k ].name, name );
																				channels[ k ].majorChannel = majorChannel;
																				channels[ k ].minorChannel = minorChannel;
																				channels[ k ].PCR_pid = PCR_PID;

																				seen_before = true;
																			}
																		}
																		if( !seen_before && ( numChannels < maxChannels ) )
																		{
																			// create a new description
																			strcpy( channels[numChannels].name, name );
																			channels[ numChannels ].majorChannel = majorChannel;
																			channels[ numChannels ].minorChannel = minorChannel;
																			channels[numChannels].videopid = pm->pType << 16 | (int)elementary_PID;
																			channels[ numChannels ].audiopid = -1;
																			channels[ numChannels ].program_number = program_number;
																			channels[ numChannels ].PCR_pid = PCR_PID;

																			numChannels++;

																			our_videopid = (int)elementary_PID;
																		}
																	}// end switch(stream_type)
																	break;

																	case 0x03: // MPEG-1 Audio
																	case 0x04: // MPEG-2 Audio
																	case 0x0f: // AAC Audio
																	case 0x81: // AC3 Audio
																	{
																		pm->sType = STREAM_PES;

																		if( stream_type == 0x03 )
																			pm->pType = PID_MPEG1_AUDIO;

																		if( stream_type == 0x04 )
																			pm->pType = PID_MPEG2_AUDIO;

																		if( stream_type == 0x0f )
																			pm->pType = PID_AAC_AUDIO;

																		if( stream_type == 0x81 )
																			pm->pType = PID_AC3_AUDIO;

																		// if a audio descriptor is found before a video descriptor
																		//  then bail.  this can cause bad mapping between PIDs and
																		//  Channel data.
																		if( !videoFound )
																			continue;

																		// create channel description?
																		bool seen_before = false;
																		for( int k = 0; k < numChannels; k++ )
																		{
																			if( (WORD)channels[ k ].audiopid == (int)elementary_PID )
																			{
																				// oh, already seen this one
																				seen_before = true;
																				break;
																			}
																		}
																		if( !seen_before )
																		{
																			// look for descriptor with our_videopid but no audiopid,
																			int last_k = -1;	// track any with our_videopid already with an audiopid
																			bool added = false;

																			audioFound = true;

																			for( int k = 0; k < numChannels; k++ )
																			{
	//																			if( (WORD)channels[ k ].videopid == our_videopid )
																				if( (WORD)channels[ k ].PCR_pid == PCR_PID )
																				{
																					if( channels[ k ].audiopid < 0 )
																					{
																						// add ourselves to this channel
																						channels[ k ].audiopid = pm->pType << 16 | (int)elementary_PID;
																						added = true;
																						break;
																					}
																					else
																						last_k = k;
																				}
																			}
																			if( !added && ( last_k >= 0 ) && ( numChannels < maxChannels ) )
																			{
																				// make a new entry, copying last_k
																				channels[ numChannels ] = channels[ last_k ];
																				channels[ numChannels ].audiopid =  pm->pType << 16 | (int)elementary_PID;
																				numChannels++;
																			}
																			// else skip - should be a videopid first?
																		}// end if(!seen_before)
																	}// end case 0x81
																	break;
																}// end switch(stream_type)
															}// end if( pm )
														}// end for(j < num_elem)
													}// end case 0xA1:
													break;

													// Ignore all other descriptors
													default:
													{
														// skip pass this descriptor
														for (uint i = 0; i < desc_len; i++)
															discard_bits(8);
													}
													break;
												}// end switch(desc_id)

												pos += desc_len + 2;

											}// end while(pos < ch_desclen)
										}// end for(i < num_ch)
									}// end case 0xc8:
									break;
								}// end switch(table_id)
								break;
							}// end switch(p->pType)
						}// end if(p->bufferPos > 0)
					}// end if(p->reading)
					// start a new table packet
					if( p->buffer && !p->processed )
					{
						p->reading = true;
						p->bufferPos = 0;
					}
				}// end if(start)
				if( p->reading )
				{
					// Add the payload for this packet to the current buffer
					int payloadLength = 184 - adaptLength - pointer_length;
					if( payloadLength > 0 )
					{
						if( ( p->bufferPos + payloadLength ) <= p->bufferSize )
						{
							memcpy( p->buffer + p->bufferPos, buf + 4 + adaptLength + pointer_length, payloadLength );
							p->bufferPos += payloadLength;
						}
						else
						{
							// buffer overflow!
							_Error("Error: buffer overflow! PID=0x%02X, Table packet, buffer size = %d\n", pid, p->bufferSize);
							p->reading = false;
						}
					}
				}// end if(p->reading)
			}// end case STREAM_TABLE:
			break;


		case STREAM_PES:
			{
				__int64 TS_rate_current = 0;
				uchar *b; 

				b = buf + 3;
				// if adaptation field
				if( b[0] & 0x20 )
				{
					p->ts_info.PCR_present = false;
					p->ts_info.PTS_present = false;
					p->ts_info.DTS_present = false;
		//			p->ts_info.pidType = p->pType;

					// check PCR stuff here
					b = buf + 5;

						// check the PCR_flag
						if( b[0] & 0x10 )// PCR_flag is set
						{
							p->ts_info.last_pcrb = p->ts_info.program_clock_reference_base;

							// get the program_clock_reference_base
							b = buf + 6;
							p->ts_info.program_clock_reference_base =  (__int64)b[0] << 25;
							p->ts_info.program_clock_reference_base |= (__int64)b[1] << 17;
							p->ts_info.program_clock_reference_base |= (__int64)b[2] << 9;
							p->ts_info.program_clock_reference_base |= (__int64)b[3] << 1;
							p->ts_info.program_clock_reference_base |= (__int64)b[4] >> 7;
//_Error("\nPCR_base:%x\n", p->ts_info.program_clock_reference_base );

							// get the program_clock_reference_extension;
							p->ts_info.program_clock_reference_extension =  (int)( b[4] & 0x01 ) << 8;
							p->ts_info.program_clock_reference_extension |= (int)b[5];

							// calculate the transport stream rate
							p->ts_info.last_PCR_value = p->ts_info.PCR_value;
							p->ts_info.PCR_value = p->ts_info.program_clock_reference_base * 300 + p->ts_info.program_clock_reference_extension;
//_Error("PCR_value:%x\n", p->ts_info.PCR_value );

							p->ts_info.PCR_bytes = totalRead - p->ts_info.prev_PCR_bytes;
//_Error("PCR_bytes:%x\n", p->ts_info.PCR_bytes );

							if( p->ts_info.PCR_value == p->ts_info.last_PCR_value )
								TS_rate_current = 0;
							else
								TS_rate_current = (( p->ts_info.PCR_bytes * SYSTEMCLOCKRATE ) / ( p->ts_info.PCR_value - p->ts_info.last_PCR_value ) * 8 );
//_Error("TS_rate_current,%d\n", TS_rate_current );

							if( TS_rate_current < 0 )
								TS_rate_current = 0;

							p->ts_info.TS_rate = CalcWeightedAvgTSRate( p->pid, &srl, TS_rate_current );
//_Error("Avg TS frame rate:%d\n", p->ts_info.TS_rate_previous );

							p->ts_info.prev_PCR_bytes = totalRead;
							p->ts_info.PCR_present = true;

						}// end if( b[0] & 0x10 )
				}// end if( adaptation == 0x2 || adaptation == 0x3 )
				if( start )
				{
					if (p->reading && (p->bufferPos > 0))
					{
						switch (p->pType)
						{
							case PID_MPEG2_VIDEO:
							case PID_USER_PRIVATE:
							{
								sequence_header sh;
								process_PES_video_packet( buf + 4 + adaptLength, 184 - adaptLength, false, &sh );
	//							process_PES_video_packet( p->buffer, p->bufferSize, false, &sh );
								if( sh.start_code_identifier == 0x000001b3 )// sequence header id code
								{
									p->processed = true;

									for( int k = 0; k < numChannels; k++ )
									{
										if( (WORD)channels[k].videopid == pid )
										{
											channels[k].width = sh.horizontal_size_value;
											channels[k].height = sh.vertical_size_value;

											//channels[k].bitRate = sh.bit_rate_value;
											if( channels[k].PCR_pid != (WORD)channels[k].videopid )
											{
												PIDinfo *pInfo = find_PID( &PIDs, channels[k].PCR_pid );
												channels[k].bitRate = (int)pInfo->ts_info.TS_rate;
											}
											else
												channels[k].bitRate  = (int)p->ts_info.TS_rate;

											channels[k].aspectRatio = sh.aspect_ratio_information;
											channels[k].frameRate = sh.frame_rate_code;
											channels[k].chromaFormat = sh.chroma_format;
											channels[k].progressiveSequence = sh.progressive_sequence;
										}
									}
								}
						}
						break;

						case PID_AC3_AUDIO:
						case PID_AAC_AUDIO:
						case PID_MPEG2_AUDIO:
						case PID_MPEG1_AUDIO:
							p->processed = true;
							break;
						}
					}
					// start a new PES packet
					if (p->buffer && !p->processed)
					{
						p->reading = true;
						p->bufferPos = 0;
					}
				}
				if (p->reading)
				{
					// Add the payload for this packet to the current buffer
					int payloadLength = 184 - adaptLength;
					// for scanning purposes we don't mind if we overflow the buffer,
					// just fill up to the limit and no more.
					if ((p->bufferPos + payloadLength) > p->bufferSize)
						payloadLength = p->bufferSize - p->bufferPos;
					if (payloadLength > 0)
					{
						memcpy(p->buffer + p->bufferPos, buf + 4 + adaptLength, payloadLength);
						p->bufferPos += payloadLength;
					}

#ifndef SHOW_INFO_MPEG2_VIDEO
					if(( p->pType == PID_MPEG2_VIDEO ) && !p->confirmedRead && ( p->bufferPos > 512 ))
//					if(( p->pType == PID_MPEG2_VIDEO ) && !p->confirmedRead && ( p->bufferPos > 8000 ))
					{
						// only want to continue reading this PES packet if it's an I-frame.
						// look for sequence header start code
						uint start_code = 0;
						for (int i = 0; i < p->bufferPos; i++)
						{
							start_code = (start_code << 8) | p->buffer[i];
							if (start_code == 0x000001b3)
							{
								p->confirmedRead = true;
								break;
							}
						}
						if (!p->confirmedRead)
							p->reading = false;
					}
#endif
				}
			}
			break;
		}

		// check if we've completed our reading of information
		// break;
	}

//	fclose64(f);


		// make sure our video data display information got set
		if( sequenceHeader )
		{
			sequence_header sh;
			buf = sequenceHeader;

			int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;

			PIDinfo *p = find_PID( &PIDs, pid );

			process_PES_video_packet( buf + 4 + adaptationLength, 184 - adaptationLength, false, &sh );
			if (sh.start_code_identifier == 0x000001b3)	// sequence header id code
			{
				for( int k = 0; k < numChannels; k++ )
				{
					if( (WORD)channels[k].videopid == pid )
					{
						if( channels[k].width == 0 )
						{
							channels[k].width = sh.horizontal_size_value;
							channels[k].height = sh.vertical_size_value;
	//						channels[k].bitRate = sh.bit_rate_value;
							channels[k].bitRate = (int)p->ts_info.TS_rate;
							channels[k].aspectRatio = sh.aspect_ratio_information;
							channels[k].frameRate = sh.frame_rate_code;
							channels[k].chromaFormat = sh.chroma_format;
							channels[k].progressiveSequence = sh.progressive_sequence;
						}
					}
				}
			}
		}


	// free any buffers
	for (int i = 0; i < PIDs.seen; i++)
	{
		if (PIDs.pids[i].buffer)
		{
			delete []PIDs.pids[i].buffer;
			PIDs.pids[i].buffer = NULL;
			PIDs.pids[i].bufferSize = 0;
			PIDs.pids[i].bufferPos = 0;
		}
	}

	return true;
}
// calculate the weighted average ts rate
//debug
//int count = 0;
__int64 CalcWeightedAvgTSRate( int pid, StreamRateList *srl, __int64 TS_rate_current )
{
	
	__int64 avg_stream_rate = 0;
//debug
/*
count++;
_Error("Calc:%d\n", count );
*/
	avg_stream_rate = srl->AddNewRate( pid, TS_rate_current );

	return avg_stream_rate;
}
// add a new stream rate to the current list
// return weighted average transport rate
__int64 StreamRateList::AddNewRate( int pid, __int64 stream_rate )
{
	StreamPidList *pStreamPidList = m_StreamPidList;
	StreamPidList *pPrevPidList = NULL;

	// first PID
	if( pStreamPidList == NULL )
	{
		pStreamPidList = new StreamPidList();
		m_StreamPidList = pStreamPidList;
		pStreamPidList->m_pid = pid;

		StreamRateNode *pNewNode = new StreamRateNode();

		pNewNode->m_stream_rate = stream_rate;
		pStreamPidList->m_StreamRateNode = pNewNode;
	}
	// find PID
	else 
	{
		while( pStreamPidList )
		{
			if( pStreamPidList->m_pid == pid )
				break;

			pPrevPidList = pStreamPidList;
			pStreamPidList = pStreamPidList->m_next;
		}
		// new PID
		if( pStreamPidList == NULL )
		{
			pStreamPidList = new StreamPidList();
			pPrevPidList->m_next = pStreamPidList;
			pStreamPidList->m_prev = pPrevPidList;
			pStreamPidList->m_pid = pid;
		}

		StreamRateNode *pNewNode = new StreamRateNode();
		pNewNode->m_stream_rate = stream_rate;

		StreamRateNode *pStreamRateNode = pStreamPidList->m_StreamRateNode;
		if( pStreamRateNode == NULL ) 
		{
			pStreamRateNode = new StreamRateNode();
			pStreamPidList->m_StreamRateNode = pStreamRateNode;
		}

		while( pStreamRateNode )
		{
			if( pNewNode->m_stream_rate >= pStreamRateNode->m_stream_rate )
			{	
				// head node
				if( pStreamRateNode == pStreamPidList->m_StreamRateNode  )
				{	
					pNewNode->m_next = pStreamRateNode;
					pStreamRateNode->m_prev = pNewNode;
					pStreamPidList->m_StreamRateNode = pNewNode;
					break;
				}
				// insert into middle
				else 
				{
					pNewNode->m_next = pStreamRateNode;
					pNewNode->m_prev = pStreamRateNode->m_prev;
					pStreamRateNode->m_prev = pNewNode;
					pNewNode->m_prev->m_next = pNewNode;
					break;
				}
			}

			// add to the end of the list
			if( pStreamRateNode->m_next == NULL )
			{
				pNewNode->m_prev = pStreamRateNode;
				pStreamRateNode->m_next = pNewNode;
				break;
			}

           	pStreamRateNode = pStreamRateNode->m_next;
		}
	}

	// calculate new weighted average stream rate
	double weight_total = 0;
	__int64 weighted_ts_rate_total = 0;
	StreamRateNode *pHead = pStreamPidList->m_StreamRateNode;
	StreamRateNode *pStreamRateNode = pHead;

	if( pHead->m_stream_rate != 0 )
	{
		while( pStreamRateNode )
		{
			// calculate weight of stream
			pStreamRateNode->m_weight = (double)pStreamRateNode->m_stream_rate / pHead->m_stream_rate;
			weight_total += pStreamRateNode->m_weight;
			weighted_ts_rate_total += (__int64)( pStreamRateNode->m_stream_rate * pStreamRateNode->m_weight );

			pStreamRateNode = pStreamRateNode->m_next;
		}

		// calculate new weighted average stream rate
		pStreamPidList->m_avg_stream_rate = (__int64)( weighted_ts_rate_total / weight_total );
	}

//_Error("m_avg_stream_rate:%d\n", pStreamPidList->m_avg_stream_rate );
	return pStreamPidList->m_avg_stream_rate;
}

// StreamRateList destructor
StreamRateList::~StreamRateList()
{
	StreamRateNode *pRN = NULL;
	StreamRateNode *pRNTemp = NULL;
	StreamPidList *pPL = NULL;
	StreamPidList *pNextPL = NULL;

	pPL = m_StreamPidList;

	if( pPL )
	{
		while( pPL )
		{
			if( pPL->m_StreamRateNode )
			{	
/*
// debug
pRN = pPL->m_StreamRateNode;
int count = 1;
while( pRN )
{
	_Error("[%x][%d]RN:%x, %d\n", pPL->m_pid, count, pRN, pRN->m_stream_rate );
	pRN = pRN->m_next;
	count++;
}
*/
				pRN = pPL->m_StreamRateNode;
				pRNTemp = pRN;
/*
// debug
count = 1;
*/
				// delete each StreamRateNode
				while( pRNTemp )
				{
					pRN = pRN->m_next;
					delete pRNTemp;
/*
//debug
_Error("[%d]RN:%x - delete\n", count, pRNTemp );
count++;
*/
					pRNTemp = pRN;
				}
			}// end if( pPL->m_StreamRateNode )

			pNextPL = pPL->m_next;
/*
// debug
_Error("PL:%x - delete\n", pPL );
*/
			delete pPL;
			pPL = pNextPL;
		}// end while( pPL )
	}// end if( pPL )
}

#endif // ENABLE_ORIGINAL_SCAN_CHANNEL_INFO

#ifdef ENABLE_PEEK_AT_FILE

uint process_PES_header(uint stream_id)
{
	uint PES_packet_length = get_bits(16);

	_Error("\tPES_packet, stream_id=%0x, PES_packet_length=%d\n", stream_id, PES_packet_length);
	_Error("\t{\n");

	if ((stream_id != 0xBC) &&	//program_stream_map
		(stream_id != 0xBE) &&	//padding_stream
		(stream_id != 0xBF) &&	//private_stream_2
		(stream_id != 0xF0) &&	//ECM
		(stream_id != 0xF1) &&	//EMM
		(stream_id != 0xFF) &&	//program_stream_directory
		(stream_id != 0xF2) &&	//DSMCC_stream
		(stream_id != 0xF8))	//ITU-T Rec. H .222.1 type E_stream
	{
		get_bits(2);		// '10'
		uint PES_scrambling_control = get_bits(2);
		uint PES_priority = get_bits(1);
		uint data_alignment_indicator = get_bits(1);
		uint copyright = get_bits(1);
		uint original_or_copy = get_bits(1);
		uint PTS_DTS_flags = get_bits(2);
		uint ESCR_flag = get_bits(1);
		uint ES_rate_flag = get_bits(1);
		uint DSM_trick_mode_flag = get_bits(1);
		uint additional_copy_info_flag = get_bits(1);
		uint PES_CRC_flag = get_bits(1);
		uint PES_extension_flag = get_bits(1);
		uint PES_header_data_length = get_bits(8);
		__int64 PTS = 0;
		__int64 DTS = 0;
		if (PTS_DTS_flags & 2)
		{
			get_bits(4);	// '0010' or '0011'
			PTS = (__int64)(get_bits(3)) << 30;
			get_bits(1);
			PTS |= (__int64)(get_bits(15)) << 15;
			get_bits(1);
			PTS |= (__int64)(get_bits(15));
			get_bits(1);
		}
		if (PTS_DTS_flags & 1)
		{
			get_bits(4);	// '0001'
			DTS = (__int64)(get_bits(3)) << 30;
			get_bits(1);
			DTS |= (__int64)(get_bits(15)) << 15;
			get_bits(1);
			DTS |= (__int64)(get_bits(15));
			get_bits(1);
		}
		__int64 ESCR_base = 0;
		uint ESCR_extension = 0;
		if (ESCR_flag == 1)
		{
			get_bits(2);
			ESCR_base = (__int64)(get_bits(3)) << 30;
			get_bits(1);
			ESCR_base |= (__int64)(get_bits(15)) << 15;
			get_bits(1);
			ESCR_base |= (__int64)(get_bits(15));
			get_bits(1);
			ESCR_extension = get_bits(9);
			get_bits(1);
		}
		uint ES_rate = 0;
		if (ES_rate_flag == 1)
		{
			get_bits(1);
			ES_rate = get_bits(22);
			get_bits(1);
		}
/*		uint trick_mode_control = 0;
		if (DSM_trick_mode_flag == 1)
		{
			trick_mode_control = get_bits(3);
			if (trick_mode_control == fast_forward)
			{
				field_id 2 bslbf
				intra_slice_refresh 1 bslbf. ISO/IEC ISO/IEC 13818-1: 1994(E)
				ITU-T Rec. H.222.0 (1995 E) 35
				frequency_truncation 2 bslbf
			}
			else if ( trick_mode_control == slow_motion )
			{
				rep_cntrl 5 uimsbf
			}
			else if ( trick_mode_control == freeze_frame)
			{
				field_id 2 uimsbf
				reserved 3 bslbf
			}
			else if ( trick_mode_control == fast_reverse' )
			{
				field_id 2 bslbf
				intra_slice_refresh 1 bslbf
				frequency_truncation 2 bslbf
			}
			else if ( trick_mode_control == slow_reverse )
			{
				rep_cntrl 5 uimsbf
			}
			else
				reserved 5 bslbf
		}
		if ( additional_copy_info_flag == '1' )
		{
			marker_bit 1 bslbf
			additional_copy_info 7 bslbf
		}
		if ( PES_CRC_flag == '1' )
		{
			previous_PES_packet_CRC 16 bslbf
		}
		if ( PES_extension_flag == '1' )
		{
			PES_private_data_flag 1 bslbf
			pack_header_field_flag 1 bslbf
			program_packet_sequence_counter_flag 1 bslbf
			P-STD_buffer_flag 1 bslbf
			reserved 3 bslbf
			PES_extension_flag_2 1 bslbf
			if ( PES_private_data_flag == '1' )
			{
				PES_private_data 128 bslbf
			}
			if (pack_header_field_flag == '1' )
			{
				pack_field_length 8 uimsbf
				pack_header()
			}
			if(program_packet_sequence_counter_flag== '1')
			{
				marker_bit 1 bslbf
				program_packet_sequence_counter 7 uimsbf
				marker_bit 1 bslbf
				MPEG1_MPEG2_identifier 1 bslbf
				original_stuff_length 6 uimsbf
			}
			if ( P-STD_buffer_flag == '1' )
			{
				'01' 2 bslbf
				P-STD_buffer_scale 1 bslbf
				P-STD_buffer_size 13 uimsbf
			}
			if ( PES_extension_flag_2 == '1')
			{
				marker_bit 1 bslbf
				PES_extension_field_length 7 uimsbf.ISO/IEC 13818-1: 1994(E)  ISO/IEC
				for(i=0;i<PES_extension_field_length;i++)
				{
					reserved 8 bslbf
				}
			}
		}
		for (i=0;i<N1;i++)
		{
			stuffing_byte 8 bslbf
		}
		for (i=0;i<N2;i++)
		{
			PES_packet_data_byte 8 bslbf
		}
*/

		_Error("\t\tPES_scrambling_control=%d\n", PES_scrambling_control);
		_Error("\t\tPES_priority=%d\n", PES_priority);
		_Error("\t\tdata_alignment_indicator=%d\n", data_alignment_indicator);
		_Error("\t\tcopyright=%d\n", copyright);
		_Error("\t\toriginal_or_copy=%d\n", original_or_copy);
		_Error("\t\tPTS_DTS_flags=%d\n", PTS_DTS_flags);
		_Error("\t\tESCR_flag=%d\n", ESCR_flag);
		_Error("\t\tES_rate_flag=%d\n", ES_rate_flag);
		_Error("\t\tDSM_trick_mode_flag=%d\n", DSM_trick_mode_flag);
		_Error("\t\tadditional_copy_info_flag=%d\n", additional_copy_info_flag);
		_Error("\t\tPES_CRC_flag=%d\n", PES_CRC_flag);
		_Error("\t\tPES_extension_flag=%d\n", PES_extension_flag);
		_Error("\t\tPES_header_data_length=%d\n", PES_header_data_length);

		if (PTS_DTS_flags & 2)
			_Error("\t\tPTS=%03X%06X, %.4f\n", (int)(PTS>>24), (int)(PTS&0xffffff), (double)PTS * 300 / SYSTEMCLOCKRATE);
		if (PTS_DTS_flags & 1)
			_Error("\t\tDTS=%03X%06X, %.4f\n", (int)(DTS>>24), (int)(DTS&0xffffff), (double)DTS * 300 / SYSTEMCLOCKRATE);
	}
/*	else if ((stream_id == 0xBC) ||	//program_stream_map
			(stream_id == 0xBF) ||	//private_stream_2
			(stream_id == 0xF0) ||	//ECM
			(stream_id == 0xF1) ||	//EMM
			(stream_id == 0xFF) ||	//program_stream_directory
			(stream_id == 0xF2) ||	//DSMCC_stream
			(stream_id == 0xF8))	//ITU-T Rec. H .222.1 type E_stream
	{
		for ( i=0;i<PES_packet_length;i++)
		{
			PES_packet_data_byte 8 bslbf
		}
	}
	else if (stream_id == 0xBE)		//padding_stream
	{
		for ( i=0;i<PES_packet_length;i++)
		{
			padding_byte 8 bslbf
		}
	}
*/

	_Error("\t}\n");

	return PES_packet_length;
}

int process_descriptor(char *indent)
{
	uint descriptor_tag = get_bits(8);
	int descriptor_length = get_bits(8);
	int k;

#ifdef SHOW_INFO_PMT
	_Error("%s{\tdescriptor_tag=0x%02X\n", indent, descriptor_tag);
	_Error("%s\tdescriptor_length=%d\n", indent, descriptor_length);
#endif

	switch (descriptor_tag)
	{
	case 0x05:	// registration descriptor (text?)
#ifdef SHOW_INFO_PMT
		_Error("%s\t", indent);
#endif
		for (k = 0; k < descriptor_length; k++)
		{
			uchar tmp = get_bits(8);
#ifdef SHOW_INFO_PMT
			_Error("%c", tmp);
#endif
		}
#ifdef SHOW_INFO_PMT
		_Error("\n");
#endif
		break;

	case 0x81:	// AC3 audio_stream_descriptor
		{
			uchar sample_rate_code = get_bits(3);
			uchar bsid = get_bits(5);
			uchar bit_rate_code = get_bits(6);
			uchar surround_mode = get_bits(2);
			uchar bsmod = get_bits(3);
			uchar num_channels = get_bits(4);
			uchar full_svc = get_bits(1);

#ifdef SHOW_INFO_PMT
			_Error("%s\tsample_rate_code=%d, %s\n", indent, sample_rate_code, sample_rate_table[sample_rate_code]);
			_Error("%s\tbsid=%d\n", indent, bsid);
			_Error("%s\tbit_rate_code=%d, %s %s\n", indent, bit_rate_code, (bit_rate_code&32)?"exact":"max", bite_rate_table[bit_rate_code%31]);
			_Error("%s\tsurround_mode=%d, %s\n", indent, surround_mode, surround_mode_table[surround_mode]);
			_Error("%s\tbsmod=%d\n", indent, bsmod);
			_Error("%s\tnum_channels=%d, %s\n", indent, num_channels, num_channels_table[num_channels]);
			_Error("%s\tfull_svc=%d\n", indent, full_svc);

			if (descriptor_length > 3)
				_Error("\n%s\t0x", indent);
#endif
			for (k = 3; k < descriptor_length; k++)
			{
				uint tmp = get_bits(8);
#ifdef SHOW_INFO_PMT
				_Error("%02X ", tmp);
#endif
			}
#ifdef SHOW_INFO_PMT
			_Error("\n");
#endif
		}
		break;

	default:
#ifdef SHOW_INFO_PMT
		if (descriptor_length > 0)
			_Error("%s\t0x", indent);
#endif
		for (k = 0; k < descriptor_length; k++)
		{
			uint tmp = get_bits(8);
#ifdef SHOW_INFO_PMT
			_Error("%02X ", tmp);
#endif
		}
#ifdef SHOW_INFO_PMT
		_Error("\n");
#endif
		break;
	}

#ifdef SHOW_INFO_PMT
	_Error("%s}\n", indent);
#endif

	return descriptor_length;
}

void peek_at_file(char *ifilelist, convert_callback callback, void* data)
{
//	for (int filei = 0; filei < numifiles; filei++)
	{
		_Error("file: %s\n", ifilelist);

uint64 time_fileinfo_start = rdtsc();

		PIDinfoGroup PIDs = { 3,
			{	{ 0x1fff, 0, STREAM_UNKNOWN, PID_NULL, true, NULL, 0, 0, -1, false, false },	// skip NULL packets
				{ 0, 0, STREAM_TABLE, PID_PAT, false, NULL, 0, 0, -1, false, false },	// PID 0 = PAT tables
				{ 0x1ffb, 0, STREAM_UNKNOWN, PID_PSIP, false, NULL, 0, 0, -1, false, false },
			} };

		__int64 curfilesize = 0;
		__int64 curfilepos = 0;
		char ifilename[256];
		strcpy(ifilename, ifilelist);
		FILE64 *fIn = fopen64(ifilename, "rb", 4*1024*1024);
		if (!fIn)
		{
			_Error("Couldn't open source file %s!\n", ifilename);
			return;// 1;
		}
		curfilesize = filelength64(fIn);		// These files are less than 1GB I think..

		curfilepos = align_to_next_packet(fIn);

		uchar buf[188];

		bool cont = true;
		int counter = 0;
		for (;;)
		{
			if ((counter & 0x3f) == 0)
			{
				// Call callback
				if (callback)
				{
					cont = callback(ifilename, "",
							-1, curfilepos, curfilesize,
							0, 0, data);
				}
				// check if we've processed all streams
				if (cont)
				{
					cont = false;
					for (int i=0; i<PIDs.seen; i++)
						if (!PIDs.pids[i].processed)
							cont = true;
#ifdef SHOW_INFO_SEEN
					if (!cont)
						_Error("\nProcessed all target streams, %d bytes\n", (int)curfilepos);
#endif
				}
			}

			counter++;

			// Cancelled?
			if (!cont)
			{
				break;
			}

			// Try to read packet..
			if (fread64(buf, 188, 1, fIn) != 1)
			{
				break;
			} else
				curfilepos += 188;

			// Check sync byte
			if ((buf[0] != 0x47) && (buf[0] != 0x72) && (buf[0] != 0x29))
			{
				__int64 pos = ftell64(fIn);
				_Error("Bad transport packet (no sync byte 0x47)!\n");
				continue;
			}

			// Get pid
			int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;

			// get info for this PID
			PIDinfo *p = find_PID(&PIDs, pid);
			if (!p)
				continue;	// oops, too many pids
			p->count++;

			// skip this packet if we're done with this PID stream
			if (p->processed)
				continue;

			// Get error
			bool errorBit = (buf[1] & 0x80) != 0;
			if (errorBit)
			{
				// invalidate accumulated content of this stream
				p->reading = false;
				continue;
			}

			// Get adaptation header info
			int adaptLength = 0;
			int adaptation = (buf[3] & 0x30) >> 4;
			if (adaptation == 0)
				continue;
			else if (adaptation == 0x2)
				adaptLength = 184;
			else if (adaptation == 0x3)
				adaptLength = buf[4] + 1;
			if (adaptLength > 184)
				continue;

if (adaptation & 2)
{
	static int count = 0;
	count++;
	if (count < 20)
	{
		set_buf(&buf[5], 0, false);

		uint adaptation_field_length = get_bits(8);
		if (adaptation_field_length > 0)
		{
			uint discontinuity_indicator = get_bits(1);
			uint random_access_indicator = get_bits(1);
			uint elementary_stream_priority_indicator = get_bits(1);
			uint PCR_flag = get_bits(1);
			uint OPCR_flag = get_bits(1);
			uint splicing_point_flag = get_bits(1);
			uint transport_private_data_flag = get_bits(1);
			uint adaptation_field_extension_flag = get_bits(1);
			__int64 program_clock_reference_base = 0;
			uint program_clock_reference_extension = 0;
			if (PCR_flag == 1)
			{
				program_clock_reference_base = (__int64)get_bits(33);
				get_bits(6);
				program_clock_reference_extension = get_bits(9);
			}
			__int64 original_program_clock_reference_base = 0;
			uint original_program_clock_reference_extension = 0;
			if(OPCR_flag == 1)
			{
				original_program_clock_reference_base = (__int64)get_bits(33);
				get_bits(6);
				original_program_clock_reference_extension = get_bits(9);
			}

			_Error("\tAdaptation header, PID = 0x%x\n", pid);
			_Error("{\tadaptation_field_length = %d\n", adaptation_field_length);
			_Error("\tdiscontinuity_indicator = %d\n", discontinuity_indicator);
			_Error("\trandom_access_indicator = %d\n", random_access_indicator);
			_Error("\telementary_stream_priority_indicator = %d\n", elementary_stream_priority_indicator);
			_Error("\tPCR_flag = %d\n", PCR_flag);
			_Error("\tOPCR_flag = %d\n", OPCR_flag);
			_Error("\tsplicing_point_flag = %d\n", splicing_point_flag);
			_Error("\ttransport_private_data_flag = %d\n", transport_private_data_flag);
			_Error("\tadaptation_field_extension_flag = %d\n", adaptation_field_extension_flag);
			if (PCR_flag == 1)
			{
//PCR_base(i) = ((SYSTEMCLOCKRATE*t(i)) / 300) % (1<<33)
//PCR_ext(i) = (SYSTEMCLOCKRATE*t(i)) % 300
				_Error("\tprogram_clock_reference_base = %03X%06X, %.4f\n", (int)(program_clock_reference_base>>24), (int)(program_clock_reference_base&0xffffff), (double)program_clock_reference_base * 300 / SYSTEMCLOCKRATE);
				_Error("\tprogram_clock_reference_extension = %d\n", program_clock_reference_extension);
			}
			if(OPCR_flag == 1)
			{
				_Error("\toriginal_program_clock_reference_base = %03X%06X, %.4f\n", (int)(original_program_clock_reference_base>>24), (int)(original_program_clock_reference_base&0xffffff), (double)original_program_clock_reference_base * 300 / SYSTEMCLOCKRATE);
				_Error("\toriginal_program_clock_reference_extension = %d\n", original_program_clock_reference_extension);
			}
			_Error("\t}\n");
/*			if (splicing_point_flag == 1)
			{
				splice_countdown = get_bits(8);
			}
			if(transport_private_data_flag == 1)
			{
				transport_private_data_length = get_bits(8);
				for (i=0; i<transport_private_data_length;i++)
				{
					private_data_byte 8 bslbf
				}
			}
			if (adaptation_field_extension_flag == '1' )
			{
				adaptation_field_extension_length 8 uimsbf
				ltw_flag 1 bslbf
				piecewise_rate_flag 1 bslbf
				seamless_splice_flag 1 bslbf
				reserved 5 bslbf
				if (ltw_flag == '1')
				{
					ltw_valid_flag 1 bslbf
					ltw_offset
				}
				if (piecewise_rate_flag == '1')
				{
					reserved 2 bslbf
					piecewise_rate 22 uimsbf
				}
				if (seamless_splice_flag == '1')
				{
					splice_type 4 bslbf
					DTS_next_AU[32..30] 3 bslbf
					marker_bit 1 bslbf
					DTS_next_AU[29..15] 15 bslbf
					marker_bit 1 bslbf
					DTS_next_AU[14..0] 15 bslbf
					marker_bit 1 bslbf
				}
				for ( i=0;i<N;i++)
				{
					reserved 8 bslbf
				}
			}
*/
		}
	}
}

			// Get payload start indicator
			bool start = (buf[1] & 0x40) != 0;

			// process content of packet
			switch (p->sType)
			{
			case STREAM_UNKNOWN:
				p->processed = true;
				break;

			case STREAM_TABLE:
				{
					if (!p->buffer)
					{
						// alloc a table buffer
						p->buffer = new uchar[TABLE_BUFFER_SIZE];
						p->bufferSize = TABLE_BUFFER_SIZE;
						p->bufferPos = 0;
						p->reading = false;
					}

					int pointer_length = buf[4 + adaptLength] + 1;
					if (start)
					{
						if (p->reading)
						{
							memcpy(p->buffer + p->bufferPos, buf + 4 + adaptLength + 1, pointer_length - 1);
							p->bufferPos += pointer_length - 1;

							if (p->bufferPos > 0)
							{
								set_buf(p->buffer, 0, false);

								uchar table_id = get_bits(8);	// = 0 = program_association_section
								discard_bits(4);
								int section_length = get_bits(12);
								uint transport_stream_id	= get_bits(16);
								discard_bits(2);
								uint version_number = get_bits(5);
								uint current_next_indicator = get_bits(1);
								uint section_number = get_bits(8);
								uint last_section_number = get_bits(8);

								switch (p->pType)
								{
								case PID_PAT:
									{
										p->processed = true;
										p->reading = false;

										// table_id == 0; program_association_section

#ifdef SHOW_INFO_PAT
										_Error("\nPID 0x%X, PAT packet, size %d\n", pid, p->bufferPos);

										_Error("{\ttable_id=%d\n", table_id);
										_Error("\tsection_length=%d\n", section_length);
										_Error("\ttransport_stream_id=%d\n", transport_stream_id);
										_Error("\tversion_number=%d\n", version_number);
										_Error("\tcurrent_next_indicator=%d\n", current_next_indicator);
										_Error("\tsection_number=%d\n", section_number);
										_Error("\tlast_section_number=%d\n", last_section_number);
#endif

										for (int i=5; i<section_length-4; i+=4)
										{
											uint program_number = get_bits(16);
											discard_bits(3);
#ifdef SHOW_INFO_PAT
											_Error("\t{\tprogram_number=%d\n", program_number);
#endif

											if (program_number == 0)
											{
												uint network_PID = get_bits(13);
#ifdef SHOW_INFO_PAT
												_Error("\t\tnetwork_PID=0x%X\n", network_PID);
												_Error("\t}\n");
#endif

												PIDinfo *pm = find_PID(&PIDs, network_PID);
												if (pm)
												{
													pm->sType = STREAM_TABLE;
													pm->pType = PID_NIT;
													pm->processed = true; //false;	because I don't care to see a NIT yet
												}
											} else
											{
												uint program_map_PID = get_bits(13);
#ifdef SHOW_INFO_PAT
												_Error("\t\tprogram_map_PID=0x%X\n", program_map_PID);
												_Error("\t}\n");
#endif

												PIDinfo *pm = find_PID(&PIDs, program_map_PID);
												if (pm)
												{
													pm->sType = STREAM_TABLE;
													pm->pType = PID_PMT;
													pm->processed = false;
												}
											}
										}
										// CRC = get_bits(32);
#ifdef SHOW_INFO_PAT
										_Error("}\n");
#endif
									}
									break;

								case PID_PMT:
									{
										p->processed = true;
										p->reading = false;

										// table_id == 2; TS_program_map_section

#ifdef SHOW_INFO_PMT
										_Error("\nPID 0x%X, PMT packet, size %d\n", pid, p->bufferPos);
#endif

										discard_bits(3);
										uint PCR_PID = get_bits(13);
										discard_bits(4);
										int program_info_length = get_bits(12);
										int i = 9;

#ifdef SHOW_INFO_PMT
										_Error("{\ttable_id=%d\n", table_id);
										_Error("\tsection_length=%d\n", section_length);
										_Error("\ttransport_stream_id=%d\n", transport_stream_id);
										_Error("\tversion_number=%d\n", version_number);
										_Error("\tcurrent_next_indicator=%d\n", current_next_indicator);
										_Error("\tsection_number=%d\n", section_number);
										_Error("\tlast_section_number=%d\n", last_section_number);
										_Error("\tPCR_PID=0x%X\n", PCR_PID);
										_Error("\tprogram_info_length=%d\n", program_info_length);
#endif

										// descriptor
										for (int j = 0; j < program_info_length; )
										{
											int descriptor_length = process_descriptor("\t");
/*											uint descriptor_tag = get_bits(8);
											int descriptor_length = get_bits(8);

#ifdef SHOW_INFO_PMT
											_Error("\t{\tdescriptor_tag=0x%02X\n", descriptor_tag);
											_Error("\t\tdescriptor_length=%d\n", descriptor_length);

											if (descriptor_length > 0)
												_Error("\t\t0x");
#endif
											for (int k = 0; k < descriptor_length; k++)
											{
												uint tmp = get_bits(8);
#ifdef SHOW_INFO_PMT
												_Error("%02X ", tmp);
#endif
											}
#ifdef SHOW_INFO_PMT
											_Error("\n");
											_Error("\t}\n");
#endif
*/											i += 2 + descriptor_length;
											j += 2 + descriptor_length;
										}

										for (; i < section_length-4; )
										{
											uint stream_type = get_bits(8);
											discard_bits(3);
											uint elementary_PID = get_bits(13);
											discard_bits(4);
											int ES_info_length = get_bits(12);
											i += 5;

#ifdef SHOW_INFO_PMT
											_Error("\t{\tstream_type=0x%02X\n", stream_type);
											_Error("\t\telementary_PID=0x%02X\n", elementary_PID);
											_Error("\t\tES_info_length=%d\n", ES_info_length);
#endif

											// descriptor again
											for( int j = 0; j < ES_info_length; )
											{
												int descriptor_length = process_descriptor("\t\t");
/*												uint descriptor_tag = get_bits(8);
												int descriptor_length = get_bits(8);

#ifdef SHOW_INFO_PMT
												_Error("\t\t{\tdescriptor_tag=0x%02X\n", descriptor_tag);
												_Error("\t\t\tdescriptor_length=%d\n", descriptor_length);
												if (descriptor_length > 0)
													_Error("\t\t\t0x");
#endif
												for (int k = 0; k < descriptor_length; k++)
												{
													uint tmp = get_bits(8);
#ifdef SHOW_INFO_PMT
													_Error("%02X ", tmp);
#endif
												}
#ifdef SHOW_INFO_PMT
												_Error("\n");
												_Error("\t\t}\n");
#endif
*/
												i += 2 + descriptor_length;
												j += 2 + descriptor_length;
											}

#ifdef SHOW_INFO_PMT
											_Error("\t}\n");
#endif

											PIDinfo *pm = find_PID(&PIDs, elementary_PID);
											if (pm)
											{
												switch (stream_type)
												{
												case 0x02:
													{
														pm->sType = STREAM_PES;
														pm->pType = PID_MPEG2_VIDEO;
														pm->processed = false;

														// alloc a video buffer
														pm->buffer = new uchar[VIDEO_BUFFER_SIZE];
														pm->bufferSize = VIDEO_BUFFER_SIZE;
														pm->bufferPos = 0;
													}
													break;

												case 0x81:
													pm->sType = STREAM_PES;
													pm->pType = PID_AC3_AUDIO;
													pm->processed = false;

													// alloc an audio buffer
													pm->buffer = new uchar[AUDIO_BUFFER_SIZE];
													pm->bufferSize = AUDIO_BUFFER_SIZE;
													pm->bufferPos = 0;
													break;

												default:
													pm->sType = STREAM_UNKNOWN;
													pm->pType = PID_UNKNOWN;
													pm->processed = true;	// given that it's unknown, let's ignore this PID
													break;
												}
											}
										}
										// CRC = get_bits(32);
	#ifdef SHOW_INFO_PMT
										_Error("}\n");
	#endif
									}
									break;

								case PID_NIT:
									p->processed = true;
									p->reading = false;
									break;
								}
							}
						}
						// start a new table packet
						if (p->buffer && !p->processed)
						{
							p->reading = true;
							p->bufferPos = 0;
						}
					}
					if (p->reading)
					{
						// Add the payload for this packet to the current buffer
						int payloadLength = 184 - adaptLength - pointer_length;
						if (payloadLength > 0)
						{
							if ((p->bufferPos + payloadLength) <= p->bufferSize)
							{
								memcpy(p->buffer + p->bufferPos, buf + 4 + adaptLength + pointer_length, payloadLength);
								p->bufferPos += payloadLength;
							}
							else
							{
								// buffer overflow!
								_Error("Error: buffer overflow! PID=0x%02X, Table packet, buffer size = %d\n", pid, p->bufferSize);
								p->reading = false;
							}
						}
					}
				}
				break;


			case STREAM_PES:
				{
					if (start)
					{
						if (p->reading && (p->bufferPos > 0))
						{
							switch (p->pType)
							{
							case PID_MPEG2_VIDEO:
#ifdef SHOW_INFO_MPEG2_VIDEO
								_Error("\nPID 0x%X, PES MPEG2 Video packet, size %d\n", pid, p->bufferPos);
#endif

{
process_PES_video_packet(p->buffer, p->bufferPos, false);
static int count = 0;
count++;
if (count >= 20)
	p->processed = true;

/*								if (process_PES_video_packet(p->buffer, p->bufferPos, false))
								{
									// found and processed an I-frame
									p->processed = true;
									p->reading = false;

									// free buffer, as we don't need it any more
									delete p->buffer;
									p->buffer = NULL;
									p->bufferSize = 0;
									p->bufferPos = 0;
*/								}
								break;

							case PID_AC3_AUDIO:
#ifdef SHOW_INFO_AC3
								_Error("\nPID 0x%X, PES AC3 Audio packet, size %d\n", pid, p->bufferPos);
#endif

{
process_PES_AC3_packet(p->buffer, p->bufferPos);
static int count = 0;
count++;
if (count >= 20)
	p->processed = true;

/*								if (process_PES_AC3_packet(p->buffer, p->bufferPos))
								{
									// found and processed an AC3 packet
									p->processed = true;
									p->reading = false;

									// free buffer, as we don't need it any more
									delete p->buffer;
									p->buffer = NULL;
									p->bufferSize = 0;
									p->bufferPos = 0;
*/								}
								break;
							}
						}
						// start a new PES packet
						if (p->buffer && !p->processed)
						{
							p->reading = true;
							p->bufferPos = 0;
						}
					}
					if (p->reading)
					{
						// Add the payload for this packet to the current buffer
						int payloadLength = 184 - adaptLength;
						if (payloadLength > 0)
						{
							if ((p->bufferPos + payloadLength) <= p->bufferSize)
							{
								memcpy(p->buffer + p->bufferPos, buf + 4 + adaptLength, payloadLength);
								p->bufferPos += payloadLength;
							}
							else
							{
								// buffer overflow!
								_Error("Error: buffer overflow! PID=0x%02X, PES packet, buffer size = %d\n", pid, p->bufferSize);
								p->reading = false;
							}
						}

#ifndef SHOW_INFO_MPEG2_VIDEO
						if ((p->pType == PID_MPEG2_VIDEO) &&
							!p->confirmedRead && (p->bufferPos > 512))
						{
							// only want to continue reading this PES packet if it's an I-frame.
							// look for sequence header start code
							uint start_code = 0;
							for (int i = 0; i < p->bufferPos; i++)
							{
								start_code = (start_code << 8) | p->buffer[i];
								if (start_code == 0x000001b3)
								{
									p->confirmedRead = true;
									break;
								}
							}
							if (!p->confirmedRead)
								p->reading = false;
						}
#endif
					}
				}
				break;
			}
		}

		// Close this file
		fclose64(fIn);

uint64 time_fileinfo_end = rdtsc();
_Error("\t\ttime_fileinfo_taken=%I64d\n",time_fileinfo_end-time_fileinfo_start);

		// free any buffers
		for (int i=0; i<PIDs.seen; i++)
		{
			if (PIDs.pids[i].buffer)
			{
				delete []PIDs.pids[i].buffer;
				PIDs.pids[i].buffer = NULL;
				PIDs.pids[i].bufferSize = 0;
				PIDs.pids[i].bufferPos = 0;
			}
		}

		// log info out
#ifdef SHOW_INFO_SEEN
		Sleep(1000);
		_Error("\nsaw %d pids\n{\n", PIDs.seen);
		int total = 0;
		for (i=0; i<PIDs.seen; i++)
		{
			total += PIDs.pids[i].count;
		}
		for (i=0; i<PIDs.seen; i++)
		{
			PIDinfo *p = &PIDs.pids[i];
			_Error("\t0x%X  %d / %.2f%% %s\n", p->pid, p->count, 100.0f*(float)p->count/(float)total, pidTypeName[p->pType]);
		}
		_Error("}\n");
#endif
	}
}

#endif	// ENABLE_PEEK_AT_FILE

// not part of decode_I_frame() because it has to stick
//  around for the whole translation
extern payload_buff Stream_Header; // saved copy of stream_header

bool decode_I_frame(uchar *buffer, int bufferSize, int vidPid, pdlg_buffer *target)
{
	// assume buffer is packet aligned already

	uchar *buf = buffer;

	PIDinfo pp;
	PIDinfo *p = &pp;
	p->pid = vidPid;
	p->count = 0;
	p->sType = STREAM_PES;			// assumption
	p->pType = PID_MPEG2_VIDEO;	// assumption
	p->processed = false;
	p->buffer = new uchar[VIDEO_BUFFER_SIZE];
	p->bufferSize = VIDEO_BUFFER_SIZE;
	p->bufferPos = 0;
	p->reading = false;
	p->confirmedRead = false;

	payload_buff pb;
	pb.buffer = NULL;
	pb.bytes = 0;

	if (!p->buffer)
		return false;

	for (; !p->processed && (buf <= (buffer + bufferSize - 188)); buf += 188)
	{
		// Check sync byte
		if ((buf[0] != 0x47) && (buf[0] != 0x72) && (buf[0] != 0x29))
			continue;

		// Get pid
		int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;
		if (pid != p->pid)
			continue;

		p->count++;

		// Get error
		bool errorBit = (buf[1] & 0x80) != 0;
		if (errorBit)
		{
			// invalidate accumulated content of this stream
			p->reading = false;
			continue;
		}

		// Get adaptation header info
		int adaptLength = 0;
		int adaptation = (buf[3] & 0x30) >> 4;
		if (adaptation == 0)
			continue;
		else if (adaptation == 0x2)
			adaptLength = 184;
		else if (adaptation == 0x3)
			adaptLength = buf[4] + 1;
		if (adaptLength > 184)
			continue;

		// Get payload start indicator
		bool start = (buf[1] & 0x40) != 0;

		// make sure we have a sequence_header
		// note this only looks in the current packet for a sequence header
		if(( buf[1] & 0x40) == 0 && adaptation == 0x3 )
		{
			uint start_code = -1; // -1 to keep from getting a false positive find
			for( int i = 0; i < 188 - 4 - adaptLength; i++ )
			{
				start_code = ( start_code << 8 ) | buf[4 + adaptLength + i];
				if( start_code == 0x000001b3 )
				{
					pb.buffer = new uchar[184];

					int payloadLength = 184 - adaptLength;
					memcpy(pb.buffer, buf + 4 + adaptLength, payloadLength);
					pb.bytes = payloadLength;

					// just in case we don't have one while processing
					if( !Stream_Header.buffer )
					{
						Stream_Header.buffer = new uchar[184];
						Stream_Header.bytes = 0;
					}

					memcpy( Stream_Header.buffer, pb.buffer, pb.bytes );
					Stream_Header.bytes = pb.bytes;

					break;
				}
			}
		}

		// process content of packet
		// assuming is a valid PES stream of MPEG2 video
		if (start)
		{
			if (p->reading && (p->bufferPos > 0))
			{
				if (process_PES_video_packet(p->buffer, p->bufferPos, true, NULL, target))
				{
					// found and processed an I-frame
					p->processed = true;
					p->reading = false;
				}
			}
			// start a new PES packet
			if (p->buffer && !p->processed)
			{
				p->reading = true;

				p->bufferPos = 0;
			}
		}
		if (p->reading)
		{
			// Add the payload for this packet to the current buffer
			int payloadLength = 184 - adaptLength;
			if (payloadLength > 0)
			{
				if ((p->bufferPos + payloadLength) <= p->bufferSize)
				{
					memcpy(p->buffer + p->bufferPos, buf + 4 + adaptLength, payloadLength);
					p->bufferPos += payloadLength;
				}
				else
				{
					// buffer overflow!
					_Error("Error: buffer overflow! PID=0x%02X, PES packet, buffer size = %d\n", pid, p->bufferSize);
					p->reading = false;
				}
			}

			// only want to continue reading this PES packet if it's and i-frame
			// check for a sequence_header_code, if we have it were good to go
			// check for a group_start_code, if we have it were good to go
			// check for a picture_start_code, if it's an i-frame, add a
			//  sequence_header to the beginning of the buffer
			if( !p->confirmedRead && ( p->bufferPos > 512 ))
			{
				uint start_code = 0;
				for( int i = 0; i < p->bufferPos; i++ )
				{
					start_code = ( start_code << 8 ) | p->buffer[i];

					// sequence_header_code || group_start_code
					if( start_code == 0x000001b3 || start_code == 0x000001b8 )
					{
						// this is all we need to continue
						p->confirmedRead = true;
						break;
					}

					// picture_start_code
					if( start_code == 0x00000100 )
					{
						// check if picture_coding_type == 1
						if(( p->buffer[i + 2] & ( 0x7 << 3 )) == ( 1 << 3 ))
						{
							// add a sequence_header to the beginning of the buffer
							if( Stream_Header.buffer )
							{
								unsigned char *pTemp = new unsigned char[ p->bufferPos ];
								memcpy( pTemp, p->buffer, p->bufferPos );
								memcpy( p->buffer, Stream_Header.buffer, Stream_Header.bytes );
								memcpy( p->buffer + Stream_Header.bytes, pTemp, p->bufferPos );
								p->bufferPos += Stream_Header.bytes;

								delete pTemp;

								p->confirmedRead = true; // i-frame found
								break;
							}
							else // can't use this i-frame, no sequence_header found yet
								break;
						}
					}// end if( start_code == 0x00000100 ) // picture_start_code
				}// end for( int i = 0; i < p->bufferPos; i++ )

				if( !p->confirmedRead )
					p->reading = false;

			}// end if( !p->confirmedRead && ( p->bufferPos > 512 ))
		}// end if( p->reading )
	}// end for( ; !p->processed && ( buf <= ( buffer + bufferSize - 188 )); buf += 188 )

	if( p )
	delete []p->buffer;
	if( pb.buffer )
		delete pb.buffer;

	return p->processed;
}

bool next_start_code(uchar *buf, int &bufOffset, int bufSize)
{
	uint start_code = 0xffffffff;
	for (int i = bufOffset + get_byte_pos(); i < bufSize; i++)
	{
		start_code = (start_code << 8) | buf[i];
		if ((start_code & 0x00ffffff) == 0x000001)
		{
			bufOffset = i + 1;
			set_buf(buf + bufOffset, 0, false);
			return true;
		}
	}
	return false;
}

bool decode_slice(uint slice_vertical_position, bool fullSize, bitmap *bm, sequence_header *sh, picture_header *ph);

bool process_PES_video_packet(uchar *buf, int bufSize, bool decodeImage, sequence_header *sh, pdlg_buffer *target)
{
	bool foundIFrame = false;
	bool slices_done = false;
	sequence_header m_sh;
	picture_header m_ph;
	int vertical_size = 0;

	if (!sh)
		sh = &m_sh;
	picture_header *ph = &m_ph;

	pdlg_buffer *activeBuffer = NULL;
	uint maxRendered = 0;
	if (decodeImage)
	{
		if (target)
		{
			activeBuffer = target;
		}
		else
		{
			activeBuffer = &pdlg_buffers[0];
			if ((activeBuffer->rendered > 0) || !activeBuffer->bm.buffer)
			{
				activeBuffer = &pdlg_buffers[1];
				if ((activeBuffer->rendered > 0) || !activeBuffer->bm.buffer)
				{
					return false;	// no rendering this frame!
				}
			}
			maxRendered = max(pdlg_buffers[0].rendered, pdlg_buffers[1].rendered);
		}
		activeBuffer->renderedWidth = 0;	// signal
	}

/*
static DRAW_ONCE_ONLY = false;
if (decodeImage)
{
	if (DRAW_ONCE_ONLY)
		return false;
	DRAW_ONCE_ONLY = true;
}
*/

	memset(sh, 0, sizeof(sequence_header));
	memset(ph, 0, sizeof(picture_header));

#ifdef SHOW_DECODE_TIME
	uint64 decode_time = 0;
#endif

#ifdef SHOW_INFO_MPEG2_VIDEO
	_Error("{\n");
#endif

	int bufOffset = 0;
	set_buf(buf, 0, false);

	while (next_start_code(buf, bufOffset, bufSize))
	{
		// found a start code
		uint stream_id = get_bits(8);
		switch (stream_id)
		{
		case 0x00:	// picture_start_code
			{
#ifdef SHOW_INFO_MPEG2_VIDEO
				_Error("\t0x00 picture_start_code, %d\n", bufOffset-3);
uint64 time_start = rdtsc();
#endif
				ph->start_code_identifier = 0x00000100;


				ph->temporal_reference = get_bits(10);
				ph->picture_coding_type = get_bits(3);
				ph->vbv_delay = get_bits(16);

				if ((ph->picture_coding_type == 2) || (ph->picture_coding_type == 3))
				{
					ph->full_pel_forward_vector = get_bits(1);
					ph->forward_f_code = get_bits(3);
				}
				if (ph->picture_coding_type == 3)
				{
					ph->full_pel_backward_vector = get_bits(1);
					ph->backward_f_code = get_bits(3);
				}
				while (get_bits(1) == 1)
				{
					discard_bits(8);	//extra_information_picture
				}

				bool searchForExtensions = true;
				while (searchForExtensions &&
					next_start_code(buf, bufOffset, bufSize))
				{
					if (get_bits(8) == 0xb5)	// extension_start_code
					{
						// process sequence_extension
						uint extension_start_code_identifier = get_bits(4);
						switch (extension_start_code_identifier)
						{
						case 8:	// Picture Coding Extension ID
							ph->picture_coding_extension = extension_start_code_identifier;
							ph->f_code[0][0] = get_bits(4);
							ph->f_code[0][1] = get_bits(4);
							ph->f_code[1][0] = get_bits(4);
							ph->f_code[1][1] = get_bits(4);
							ph->intra_dc_precision = get_bits(2);
							ph->picture_structure = get_bits(2);
							ph->top_field_first = get_bits(1);
							ph->frame_pred_frame_dct = get_bits(1);
							ph->concealment_motion_vectors = get_bits(1);
							ph->q_scale_type = get_bits(1);
							ph->intra_vlc_format = get_bits(1);
							ph->alternate_scan = get_bits(1);
							ph->repeat_first_field = get_bits(1);
							ph->chroma_420_type = get_bits(1);
							ph->progressive_frame = get_bits(1);
							ph->composite_display_flag = get_bits(1);
							if (ph->composite_display_flag)
							{
								ph->v_axis = get_bits(1);
								ph->field_sequence = get_bits(3);
								ph->sub_carrier = get_bits(1);
								ph->burst_amplitude = get_bits(7);
								ph->sub_carrier_phase = get_bits(8);
							}
							break;

						case 3: // Quant Matrix Extension ID
							ph->quant_matrix_extension = extension_start_code_identifier;
							ph->load_intra_quantiser_matrix = get_bits(1);
							if (ph->load_intra_quantiser_matrix)
							{
								for (int i = 0; i < 64; i++)
									ph->intra_quantiser_matrix[i] = get_bits(8);
							}
							ph->load_non_intra_quantiser_matrix = get_bits(1);
							if (ph->load_non_intra_quantiser_matrix)
							{
								for (int i = 0; i < 64; i++)
									ph->non_intra_quantiser_matrix[i] = get_bits(8);
							}
							ph->load_chroma_intra_quantiser_matrix = get_bits(1);
							if (ph->load_chroma_intra_quantiser_matrix)
							{
								for (int i = 0; i < 64; i++)
									ph->chroma_intra_quantiser_matrix[i] = get_bits(8);
							}
							ph->load_chroma_non_intra_quantiser_matrix = get_bits(1);
							if (ph->load_chroma_non_intra_quantiser_matrix)
							{
								for (int i = 0; i < 64; i++)
									ph->chroma_non_intra_quantiser_matrix[i] = get_bits(8);
							}
							break;

						case 9: // Picture Spatial Scalable Extension ID
							ph->picture_spatial_scalable_extension = extension_start_code_identifier;
							ph->lower_layer_temporal_reference = get_bits(10);
							discard_bits(1);
							ph->lower_layer_horizontal_offset = get_bits(15);
							if (ph->lower_layer_horizontal_offset & 0x4000)	// sign extend
								ph->lower_layer_horizontal_offset |= 0xffff8000;
							discard_bits(1);
							ph->lower_layer_vertical_offset = get_bits(15);
							if (ph->lower_layer_vertical_offset & 0x4000)	// sign extend
								ph->lower_layer_vertical_offset |= 0xffff8000;
							ph->spatial_temporal_weight_code_table_index = get_bits(2);
							ph->lower_layer_progressive_frame = get_bits(1);
							ph->lower_layer_deinterlaced_field_select = get_bits(1);
							break;

						case 10: // Picture Temporal Scalable Extension ID
							ph->picture_temporal_scalable_extension = extension_start_code_identifier;
							ph->reference_select_code = get_bits(2);
							ph->forward_temporal_reference = get_bits(10);
							discard_bits(1);
							ph->backward_temporal_reference = get_bits(10);
							break;

						default:
							searchForExtensions = false;
							// back up to start of start code for someone else to process
							bufOffset -= 4;
							set_buf(buf + bufOffset, 0, false);
							break;
						}
					} else
					{
						searchForExtensions = false;
						// back up to start of start code for someone else to process
						bufOffset -= 4;
						set_buf(buf + bufOffset, 0, false);
					}
				}

				if (ph->picture_coding_type == 1)
					foundIFrame = true;

#ifdef SHOW_INFO_MPEG2_VIDEO
uint64 time_end = rdtsc();

				// info display
				_Error("\t{\ttemporal_reference=%d\n", ph->temporal_reference);
				char *s;
				switch (ph->picture_coding_type)
				{
				case 1: s = "I"; break;
				case 2: s = "P"; break;
				case 3: s = "B"; break;
				default: s = "?"; break;
				}
				_Error("\t\tpicture_coding_type=%d, %s frame\n", (uint)ph->picture_coding_type, s);
				_Error("\t\tvbv_delay=%d\n", ph->vbv_delay);
				if ((ph->picture_coding_type == 2) || (ph->picture_coding_type == 3))
				{
					_Error("\t\tfull_pel_forward_vector=%d\n", ph->full_pel_forward_vector);
					_Error("\t\tforward_f_code=%d\n", (uint)ph->forward_f_code);
				}
				if (ph->picture_coding_type == 3)
				{
					_Error("\t\tfull_pel_backward_vector=%d\n", ph->full_pel_backward_vector);
					_Error("\t\tbackward_f_code=%d\n", (uint)ph->backward_f_code);
				}

				if (ph->picture_coding_extension)
				{
					_Error("\t\tf_code[][]={{%d,%d},{%d,%d}}\n", (uint)ph->f_code[0][0],
							(uint)ph->f_code[0][1], (uint)ph->f_code[1][0], (uint)ph->f_code[1][1]);
					_Error("\t\tintra_dc_precision=%d, %d bits\n", (uint)ph->intra_dc_precision, 8+(uint)ph->intra_dc_precision);
					switch (ph->picture_structure)
					{
					case 1: s = "top field"; break;
					case 2: s = "bottom field"; break;
					case 3: s = "frame picture"; break;
					default: s = "?"; break;
					}
					_Error("\t\tpicture_structure=%d, %s\n", (uint)ph->picture_structure, s);
					_Error("\t\ttop_field_first=%d\n", ph->top_field_first);
					_Error("\t\tframe_pred_frame_dct=%d\n", ph->frame_pred_frame_dct);
					_Error("\t\tconcealment_motion_vectors=%d\n", ph->concealment_motion_vectors);
					_Error("\t\tq_scale_type=%d\n", ph->q_scale_type);
					_Error("\t\tintra_vlc_format=%d\n", ph->intra_vlc_format);
					_Error("\t\talternate_scan=%d\n", ph->alternate_scan);
					_Error("\t\trepeat_first_field=%d\n", ph->repeat_first_field);
					_Error("\t\tchroma_420_type=%d\n", ph->chroma_420_type);
					_Error("\t\tprogressive_frame=%d\n", ph->progressive_frame);
					_Error("\t\tcomposite_display_flag=%d\n", ph->composite_display_flag);
					if (ph->composite_display_flag)
					{
						_Error("\t\tv_axis=%d\n", ph->v_axis);
						_Error("\t\tfield_sequence=%d\n", (uint)ph->field_sequence);
						_Error("\t\tsub_carrier=%d\n", ph->sub_carrier);
						_Error("\t\tburst_amplitude=%d\n", (uint)ph->burst_amplitude);
						_Error("\t\tsub_carrier_phase=%d\n", (uint)ph->sub_carrier_phase);
					}
				}

				if (ph->quant_matrix_extension)
				{
					_Error("\t\tload_intra_quantiser_matrix=%d\n", ph->load_intra_quantiser_matrix);
					if (ph->load_intra_quantiser_matrix)
					{
						for (int i = 0; i < 64; i++)
						{
							if ((i%32) == 0)
								_Error("\t\t\t");
							_Error("%02X ", (uint)ph->intra_quantiser_matrix[i]);
							if ((i%32) == 31)
								_Error("\n");
						}
					}
					_Error("\t\tload_non_intra_quantiser_matrix=%d\n", ph->load_non_intra_quantiser_matrix);
					if (ph->load_non_intra_quantiser_matrix)
					{
						for (int i = 0; i < 64; i++)
						{
							if ((i%32) == 0)
								_Error("\t\t\t");
							_Error("%02X ", (uint)ph->non_intra_quantiser_matrix[i]);
							if ((i%32) == 31)
								_Error("\n");
						}
					}
					_Error("\t\tload_chroma_intra_quantiser_matrix=%d\n", ph->load_chroma_intra_quantiser_matrix);
					if (ph->load_chroma_intra_quantiser_matrix)
					{
						for (int i = 0; i < 64; i++)
						{
							if ((i%32) == 0)
								_Error("\t\t\t");
							_Error("%02X ", (uint)ph->chroma_intra_quantiser_matrix[i]);
							if ((i%32) == 31)
								_Error("\n");
						}
					}
					_Error("\t\tload_chroma_non_intra_quantiser_matrix=%d\n", ph->load_chroma_non_intra_quantiser_matrix);
					if (ph->load_chroma_non_intra_quantiser_matrix)
					{
						for (int i = 0; i < 64; i++)
						{
							if ((i%32) == 0)
								_Error("\t\t\t");
							_Error("%02X ", (uint)ph->chroma_non_intra_quantiser_matrix[i]);
							if ((i%32) == 31)
								_Error("\n");
						}
					}
				}

_Error("\t\ttime_taken=%I64d\n",time_end-time_start);

//
#ifdef TESTING_BITS
uint val1, val2, val3, val4, v;

time_start = rdtsc();

for (int i=1000; i>0; i--)
{
	set_buf(buf + bufOffset, 0, false);
	v = 0;

	for (int j=10000; j>0; j--)
	{
#if 1
		val1 = get_bits(3);
		val2 = get_bits(15);
		val3 = get_bits(9);
		val4 = get_bits(2);
		v += val1 + val2 + val3 + val4;
		val1 = get_bits(7);
		val2 = get_bits(23);
		val3 = get_bits(13);
		val4 = get_bits(4);
		v += val1 + val2 + val3 + val4;
		val1 = get_bits(1);
		val2 = get_bits(11);
		val3 = get_bits(5);
		val4 = get_bits(8);
		v += val1 + val2 + val3 + val4;
		val2 = get_bits(6);
		v += val2;
#else
		discard_bits(3);
		discard_bits(15);
		discard_bits(9);
		discard_bits(2);
		discard_bits(7);
		discard_bits(23);
		discard_bits(13);
		discard_bits(4);
		discard_bits(1);
		discard_bits(11);
		discard_bits(5);
		discard_bits(8);
		discard_bits(6);
#endif
	}
}

time_end = rdtsc();
_Error("\t\tTEST time=%I64d\n",time_end-time_start, v);
#endif
//

				_Error("\t}\n");
				Sleep(100);
#endif
			}
			break;

#ifdef SHOW_INFO_MPEG2_VIDEO
		case 0xb2:	// user_data_start_code
			_Error("\t0xb2 user_data_start_code, %d\n", bufOffset-3);
			break;
#endif

		case 0xb3:	// sequence_header_code
			{
#ifdef SHOW_INFO_MPEG2_VIDEO
				_Error("\t0xb3 sequence_header_code, %d\n", bufOffset-3);
#endif
				sh->start_code_identifier = 0x000001b3;

				sh->horizontal_size_value = get_bits(12);
				sh->vertical_size_value = get_bits(12);

				if( sh->vertical_size_value == 1080 )
					vertical_size = 1088; // evenly divisible by 16
				else
					vertical_size = sh->vertical_size_value;

				sh->aspect_ratio_information = get_bits(4);
				sh->frame_rate_code = get_bits(4);
				sh->bit_rate_value = get_bits(18);
				discard_bits(1);
				sh->vbv_buffer_size_value = get_bits(10);
				sh->constrained_parameters_flag = get_bits(1);
				sh->load_intra_quantiser_matrix = get_bits(1);
				sh->intra_quantiser_matrix[64];
				if (sh->load_intra_quantiser_matrix)
				{
					for (int i = 0; i < 64; i++)
						sh->intra_quantiser_matrix[i] = get_bits(8);
				}
				sh->load_non_intra_quantiser_matrix = get_bits(1);
				sh->non_intra_quantiser_matrix[64];
				if (sh->load_non_intra_quantiser_matrix)
				{
					for (int i = 0; i < 64; i++)
						sh->non_intra_quantiser_matrix[i] = get_bits(8);
				}

				bool searchForExtensions = true;
				while (searchForExtensions &&
					next_start_code(buf, bufOffset, bufSize))
				{
					if (get_bits(8) == 0xb5)	// extension_start_code
					{
						// process sequence_extension
						uint extension_start_code_identifier = get_bits(4);
						switch (extension_start_code_identifier)
						{
						case 1:	// sequence extension ID
							sh->sequence_extension = extension_start_code_identifier;

							sh->profile_and_level_indication = get_bits(8);
							sh->progressive_sequence = get_bits(1);
							sh->chroma_format = get_bits(2);
							sh->horizontal_size_extension = get_bits(2);
							sh->vertical_size_extension = get_bits(2);
							sh->bit_rate_extension = get_bits(12);
							discard_bits(1);
							sh->vbv_buffer_size_extension = get_bits(8);
							sh->low_delay = get_bits(1);
							sh->frame_rate_extension_n = get_bits(2);
							sh->frame_rate_extension_d = get_bits(5);

							// apply extension
							sh->horizontal_size_value |= sh->horizontal_size_extension << 12;
							sh->vertical_size_value |= sh->vertical_size_extension << 12;
							sh->bit_rate_value |= sh->bit_rate_extension << 18;
							sh->vbv_buffer_size_value |= sh->vbv_buffer_size_extension << 10;
							break;

						case 5:	// Sequence Scalable Extension ID
							sh->sequence_scalable_extension = extension_start_code_identifier;

							sh->scalable_mode = get_bits(2);
							sh->layer_id  = get_bits(4);
							if (sh->scalable_mode == 1)		// spatial scalability
							{
								sh->lower_layer_prediction_horizontal_size = get_bits(14);
								discard_bits(1);
								sh->lower_layer_prediction_vertical_size = get_bits(14);
								sh->horizontal_subsampling_factor_m = get_bits(5);
								sh->horizontal_subsampling_factor_n = get_bits(5);
								sh->vertical_subsampling_factor_m = get_bits(5);
								sh->vertical_subsampling_factor_n = get_bits(5);
							}
							if (sh->scalable_mode == 3)		// temporal scalability
							{
								sh->picture_mux_enable = get_bits(1);
								if (sh->picture_mux_enable)
									sh->mux_to_progressive_sequence = get_bits(1);
								sh->picture_mux_order = get_bits(3);
								sh->picture_mux_factor = get_bits(3);
							}
							break;

						default:
							searchForExtensions = false;
							// back up to start of start code for someone else to process
							bufOffset -= 4;
							set_buf(buf + bufOffset, 0, false);
							break;
						}
					} else
					{
						searchForExtensions = false;
						// back up to start of start code for someone else to process
						bufOffset -= 4;
						set_buf(buf + bufOffset, 0, false);
					}
				}

				// set up bitmap buffer
				if (activeBuffer)
				{
					// presume bitmap to have already been set up by progress dialog
					activeBuffer->renderedWidth = sh->horizontal_size_value>>3;
					if (activeBuffer->renderedWidth > activeBuffer->bm.width)
						activeBuffer->renderedWidth = activeBuffer->bm.width;
					activeBuffer->renderedHeight = sh->vertical_size_value>>3;
					if (activeBuffer->renderedHeight > activeBuffer->bm.height)
						activeBuffer->renderedHeight = activeBuffer->bm.height;
					activeBuffer->aspectRatio = sh->aspect_ratio_information;

					// full sized image
//					uint w = sh->horizontal_size_value;
//					uint h = sh->vertical_size_value;
					// preview sized image
//					uint w = 240; //sh->horizontal_size_value>>3;
//					uint h = 135; //sh->vertical_size_value>>3;
//					uint stride = w*3;
//					uint size = h*stride;
//					bm = (bitmap*)(new char[sizeof(bitmap) +
//							size]);
//					bm->width = w;
//					bm->height = h;
//					bm->stride = stride;
//					bm->size = size;
				}

#ifdef SHOW_INFO_MPEG2_VIDEO
				// info display
				_Error("\t{\thorizontal_size_value=%d\n", sh->horizontal_size_value);
				_Error("\t\tvertical_size_value=%d\n", sh->vertical_size_value);
				char *s;
				switch (sh->aspect_ratio_information)
				{
				case 1: s = "1:1"; break;
				case 2: s = "4:3"; break;
				case 3: s = "16:9"; break;
				case 4: s = "2.21:1"; break;
				default: s = "?";
				}
				_Error("\t\taspect_ratio_information=%d, %s\n", (uint)sh->aspect_ratio_information, s);
				switch (sh->frame_rate_code)
				{
				case 1: s = "23.976"; break;
				case 2: s = "24"; break;
				case 3: s = "25"; break;
				case 4: s = "29.97"; break;
				case 5: s = "30"; break;
				case 6: s = "50"; break;
				case 7: s = "59.94"; break;
				case 8: s = "60"; break;
				default:
//						float frameRate = (float)frame_rate_value * (frame_rate_extension_n + 1)  (frame_rate_extension_d + 1)
					s = "?"; break;
				}
				_Error("\t\tframe_rate_code=%d, %s\n", (uint)sh->frame_rate_code, s);
				_Error("\t\tbit_rate_value=%d, %d bits/s\n", sh->bit_rate_value, sh->bit_rate_value*400);
				_Error("\t\tvbv_buffer_size_value=%d, %d bytes\n", sh->vbv_buffer_size_value, 2*1024 * sh->vbv_buffer_size_value);
				_Error("\t\tconstrained_parameters_flag=%d\n", (uint)sh->constrained_parameters_flag);
				_Error("\t\tload_intra_quantiser_matrix=%d\n", (uint)sh->load_intra_quantiser_matrix);
				if (sh->load_intra_quantiser_matrix)
				{
					for (int i = 0; i < 64; i++)
					{
						if ((i%32) == 0)
							_Error("\t\t\t");
						_Error("%02X ", (uint)sh->intra_quantiser_matrix[i]);
						if ((i%32) == 31)
							_Error("\n");
					}
				}
				_Error("\t\tload_non_intra_quantiser_matrix=%d\n", (uint)sh->load_non_intra_quantiser_matrix);
				if (sh->load_non_intra_quantiser_matrix)
				{
					for (int i = 0; i < 64; i++)
					{
						if ((i%32) == 0)
							_Error("\t\t\t");
						_Error("%02X ", (uint)sh->non_intra_quantiser_matrix[i]);
						if ((i%32) == 31)
							_Error("\n");
					}
				}

				if (sh->sequence_extension == 1)
				{
					_Error("\t\tprofile_and_level_indication=%d\n", (uint)sh->profile_and_level_indication);
					_Error("\t\tprogressive_sequence=%d\n", (uint)sh->progressive_sequence);
					switch (sh->chroma_format)
					{
					case 1: s = "4:2:0"; break;
					case 2: s = "4:2:2"; break;
					case 3: s = "4:4:4"; break;
					default: s = "?"; break;
					}
					_Error("\t\tchroma_format=%d, %s\n", (uint)sh->chroma_format, s);
					_Error("\t\tlow_delay=%d\n", (uint)sh->low_delay);
				}

				_Error("\t}\n");
				Sleep(100);
#endif
			}
			break;

#ifdef SHOW_INFO_MPEG2_VIDEO
		case 0xb4:	// sequence_error_code
			_Error("\t0xb4 sequence_error_code, %d\n", bufOffset-3);
			break;

		case 0xb5:	// extension_start_code
			{
				_Error("\t0xb5 extension_start_code\n");
				uint extension_start_code_identifier = get_bits(4);
				char *s;
				switch (extension_start_code_identifier)
				{
				case 1: s = "Sequence Extension ID"; break;
				case 2: s = "Sequence Display Extension ID"; break;
				case 3: s = "Quant Matrix Extension ID"; break;
				case 4: s = "Copyright Extension ID"; break;
				case 5: s = "Sequence Scalable Extension ID"; break;
				case 7: s = "Picture Display Extension ID"; break;
				case 8: s = "Picture Coding Extension ID"; break;
				case 9: s = "Picture Spatial Scalable Extension ID"; break;
				case 10: s = "Picture Temporal Scalable Extension ID"; break;
				default: s = "??"; break;
				}
				_Error("\t{\textension_start_code_identifier=%d, %s\n", extension_start_code_identifier, s);
				_Error("\t}\n");
			}
			break;

		case 0xb7:	// sequence_end_code
			_Error("\t0xb7 sequence_end_code, %d\n", bufOffset-3);
			break;

		case 0xB8:	// group_start_code
			{
				_Error("\t0xb8 group_start_code, %d\n", bufOffset-3);

				// uint time_code = get_bits(25);
				uint drop_frame_flag = get_bits(1);
				uint time_code_hours = get_bits(5);
				uint time_code_minutes = get_bits(6);
				discard_bits(1);
				uint time_code_seconds = get_bits(6);
				uint time_code_pictures = get_bits(6);

				uint closed_gop = get_bits(1);
				uint broken_link = get_bits(1);

				// info display
				//_Error("\t{\ttime_code=%d\n", time_code);
				_Error("\t\ttime_code=%ddf %dh%02dm%02ds %dp\n", drop_frame_flag,
					time_code_hours, time_code_minutes, time_code_seconds, time_code_pictures);
				_Error("\t\tclosed_gop=%d\n", closed_gop);
				_Error("\t\tbroken_link=%d\n", broken_link);

				_Error("\t}\n");
			}
			break;
#endif

		default:
			if (stream_id < 0xB0)
			{
#ifdef DECODE_FIRST_I_FRAME
				//	slice_start_code 01 through AF (page 70)
				if (foundIFrame && activeBuffer && (sh->start_code_identifier == 0x000001b3) &&
					(ph->start_code_identifier == 0x00000100))
				{
#ifdef SHOW_DECODE_TIME
#ifdef SHOW_INFO_MPEG2_VIDEO
if (decode_time == 0)
	Sleep(100);
#endif
uint64 time_start = rdtsc();
#endif
					// pass slice through for decoding
					// XXX testing full frame decode
//					if (!decode_slice(stream_id, true, &activeBuffer->bm, sh, ph))
					if (!decode_slice(stream_id, false, &activeBuffer->bm, sh, ph))
					{
#ifdef REPORT_SLICE_DECODE_FAILURES
						_Error("slice 0x%X (%d) decode failure\n", stream_id-1, (stream_id-1)<<4);
#endif
					}
#ifdef SHOW_DECODE_TIME
uint64 time_end = rdtsc();
decode_time += time_end-time_start;
#endif
				}
#endif
			}
			else
			{
#ifdef SHOW_INFO_MPEG2_VIDEO
				if ((stream_id >= 0xE0) && (stream_id <= 0xEF))
				{
					_Error("\t0x%02X MPEG2 Video stream start code, %d\n", stream_id, bufOffset-3);
					process_PES_header(stream_id);
				} else if ((stream_id >= 0xC0) && (stream_id <= 0xDF))
				{
					_Error("\t0x%02X MPEG2 Audio stream start code, %d\n", stream_id, bufOffset-3);
				} else
				{
					_Error("\t0x%02X unknown start code, %d\n", stream_id, bufOffset-3);
				}
#endif
			}
			break;
		}

		// done processing the I frame
		if( vertical_size != 0 )
		{
			if( slices_done )
				break;

			if( stream_id == vertical_size / 16 )
				slices_done = true;
		}
	}

#ifdef SHOW_DECODE_TIME
if (decode_time > 0)
	_Error("\tdecode_time=%I64d, %dx%d, %d\n", decode_time, sh->horizontal_size_value, sh->vertical_size_value, bufSize);
#endif

#ifdef SHOW_INFO_MPEG2_VIDEO
	_Error("}\n");
#endif

	if (activeBuffer && activeBuffer->renderedWidth)
	{
		// signal progress dialog this buffer is ready to display
		activeBuffer->rendered = maxRendered + 1;
	}

	return foundIFrame;
}


bool process_PES_AC3_packet(uchar *buf, int bufSize, syncinfo *si, bsi *bi)
{
	bool foundAC3Frame = false;

	syncinfo m_si;
	bsi m_bi;

	if (!si)
		si = &m_si;
	if (!bi)
		bi = &m_bi;

	memset(si, 0, sizeof(syncinfo));
	memset(bi, 0, sizeof(bsi));

#ifdef SHOW_INFO_AC3
	_Error("{\n");
#endif

	int bufOffset = 0;
	set_buf(buf, 0, false);

	if (next_start_code(buf, bufOffset, bufSize))
	{
		// found a start code
		uint stream_id = get_bits(8);
		process_PES_header(stream_id);


		uint syncword = get_bits(16);

		if (syncword == 0x0b77)
		{
#ifdef SHOW_INFO_AC3
			_Error("\tsyncword=0x0b77\n");
#endif
			si->syncword = syncword;
		}
	}

#ifdef SHOW_INFO_AC3
	_Error("}\n");
#endif

foundAC3Frame = true;

	return foundAC3Frame;
}

bool decode_macroblock_address_increment(uint *val);
bool decode_macroblock_type(macroblock_type *mbt, uchar picture_coding_type, uchar scalable_mode);
uint decode_dct_dc_size_luminance(void);
uint decode_dct_dc_size_chrominance(void);
bool decode_DCT_coefficient(int n, uint intra_vlc_format, bool *end_of_block, uint *run, int *signed_level);
bool discard_decode_DCT_coefficients(int n, uint intra_vlc_format, macroblock_type *mbt );
bool decode_motion_code(int *val);
bool motion_vectors(int s, picture_header *ph, macroblock_type *mbt);
bool motion_vector(int r, int s, picture_header *ph, macroblock_type *mbt);

int scan_QFS_QF[2][64] =
{
	{	 0,  1,  5,  6, 14, 15, 27, 28,
		 2,  4,  7, 13, 16, 26, 29, 42,
		 3,  8, 12, 17, 25, 30, 41, 43,
		 9, 11, 18, 24, 31, 40, 44, 53,
		10, 19, 23, 32, 39, 45, 52, 54,
		20, 22, 33, 38, 46, 51, 55, 60,
		21, 34, 37, 47, 50, 56, 59, 61,
		35, 36, 48, 49, 57, 58, 62, 63,
	},
	{	 0,  4,  6, 20, 22, 36, 38, 52,
		 1,  5,  7, 21, 23, 37, 39, 53,
		 2,  8, 19, 24, 34, 40, 50, 54,
		 3,  9, 18, 25, 35, 41, 51, 55,
		10, 17, 26, 30, 42, 46, 56, 60,
		11, 16, 27, 31, 43, 47, 57, 61,
		12, 15, 28, 32, 44, 48, 58, 62,
		13, 14, 29, 33, 45, 49, 59, 63,
	},
};

int scan_QF_QFS[2][64] =
{
	{	 0,  1,  8, 16,  9,  2,  3, 10,
		17, 24, 32, 25, 18, 11,  4,  5,
		12, 19, 26, 33, 40, 48, 41, 34,
		27, 20, 13,  6,  7, 14, 21, 28,
		35, 42, 49, 56, 57, 50, 43, 36,
		29, 22, 15, 23, 30, 37, 44, 51,
		58, 59, 52, 45, 38, 31, 39, 46,
		53, 60, 61, 54, 47, 55, 62, 63,
	},
	{	 0,  8, 16, 24,  1,  9,  2, 10,
		17, 25, 32, 40, 48, 56, 57, 49,
		41, 33, 26, 18,  3, 11,  4, 12,
		19, 27, 34, 42, 50, 58, 35, 43,
		51, 59, 20, 28,  5, 13,  6, 14,
		21, 29, 36, 44, 52, 60, 37, 45,
		53, 61, 22, 30,  7, 15, 23, 31,
		38, 46, 54, 62, 39, 47, 55, 63,
	},
};

motion_type_table_entry table_6_17[] =
{						// frame_motion_type  spatial_temporal_weight_class
	{ 0, 2, 0, 0, },	// 1				  0, 1
	{ 0, 1, 0, 0, },	// 1				  2, 3
	{ 1, 1,	1, 0, },	// 2				  0, 1, 2, 3
	{ 2, 1, 0, 1, },	// 3				  0, 2, 3
};

motion_type_table_entry table_6_18[] =
{						// field_motion_type  spatial_temporal_weight_class
	{ 0, 1, 0, 0, },	// 1				  0, 1
	{ 3, 2, 0, 0, },	// 2				  0, 1
	{ 2, 1, 0, 1, },	// 3				  0
};

/* Table 7-6. Relation between quantiser_scale and quantiser_scale_code */
const unsigned char q_scale[2][32] = {
  { 255,  2,  4,  6,  8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30,
     32, 34, 36, 38, 40, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 62 },

  { 255,   1, 2,  3,  4,  5,  6,  7,  8, 10, 12, 14, 16, 18, 20, 22,
     24, 28, 32, 36, 40, 44, 48, 52, 56, 64, 72, 80, 88, 96,104,112 }
};

extern float g_brightness, g_contrast;
extern bool g_fBrightness, g_fContrast;

__declspec(align(16)) short QF[12][8][8];
__declspec(align(16)) short dst[64];
bool decode_slice(uint slice_vertical_position, bool fullSize, bitmap *bm, sequence_header *sh, picture_header *ph)
{
	uchar pattern_code[12];
	uint slice_vertical_position_extension = 0;
	uint priority_breakpoint = 0;
	uchar quantiser_scale_code = 0;
	uint intra_slice = 0;


	int dc_dct_pred[3];
	dc_dct_pred[0] = 128 << ph->intra_dc_precision;
	dc_dct_pred[1] = 128 << ph->intra_dc_precision;
	dc_dct_pred[2] = 128 << ph->intra_dc_precision;

	if (sh->vertical_size_value > 2800)
	{
		slice_vertical_position_extension = get_bits(3);
	}
	if ((sh->sequence_scalable_extension == 5) &&
		(sh->scalable_mode == 0))
	{
		priority_breakpoint = get_bits(7);
	}
	quantiser_scale_code = get_bits(5);
	if (get_bits(1) == 1)
	{
		intra_slice = get_bits(1);
		discard_bits(7);
		while (get_bits(1) == 1)
		{
			discard_bits(8);	// extra_information_slice
		}
	}

	uint mb_row = (slice_vertical_position_extension << 7) + slice_vertical_position - 1;
	uint mb_width = sh->horizontal_size_value >> 4;
	uint macroblock_address = (mb_row * mb_width) - 1;
//debug
//_Error("\nds %-2d:", slice_vertical_position );

#ifdef REPORT_SLICE_DECODE_FAILURES
	uint start_mb_add = macroblock_address + 1;
#endif

	// loop through macroblocks
	do
	{
		// address
		uint inc = 0;
		while (peek_bits(11) == 0x008)		// macroblock_escape
		{
			discard_bits(11);
			inc += 33;
		}
		uint inc2;
		if (!decode_macroblock_address_increment(&inc2))
		{
#ifdef REPORT_SLICE_DECODE_FAILURES
			_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
			return false;
		}
		inc += inc2;
		macroblock_address += inc;

		uint mb_column = macroblock_address % mb_width;

		// modes	(macroblock_modes()) page 72, tables B-2 to B-8
		macroblock_type mbt;
		memset(&mbt, 0, sizeof(mbt));

		if (!decode_macroblock_type(&mbt, ph->picture_coding_type, sh->scalable_mode))
		{
#ifdef REPORT_SLICE_DECODE_FAILURES
			_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
			return false;
		}
		if (mbt.spatial_temporal_weight_code_flag &&
			(ph->spatial_temporal_weight_code_table_index != 0))
		{
			mbt.spatial_temporal_weight_code = get_bits(2);
		}
		if (mbt.macroblock_motion_forward ||
			mbt.macroblock_motion_backward)
		{
			if (ph->picture_structure == 3)	// frame
			{
				if (ph->frame_pred_frame_dct == 0)
					mbt.frame_motion_type = get_bits(2);
			}
			else
			{
				mbt.field_motion_type = get_bits(2);
			}
		}
		if ((ph->picture_structure == 3) &&		// frame
			(ph->frame_pred_frame_dct == 0) &&
			(mbt.macroblock_intra || mbt.macroblock_pattern))
		{
			mbt.dct_type = get_bits(1);
		}

		// reset predictors if a macroblock is skipped, or is non-intra
		if (!mbt.macroblock_intra || (inc > 1))
		{
			dc_dct_pred[0] = 128 << ph->intra_dc_precision;
			dc_dct_pred[1] = 128 << ph->intra_dc_precision;
			dc_dct_pred[2] = 128 << ph->intra_dc_precision;
		}

		// stuff
		if (mbt.macroblock_quant)
			quantiser_scale_code = get_bits(5);

		if (mbt.macroblock_motion_forward ||
			mbt.macroblock_motion_backward ||
			(mbt.macroblock_intra && ph->concealment_motion_vectors))
		{
			// fill in derived motion vector info
			motion_type_table_entry *mtte;
			if (ph->picture_structure == 3)	// frame
			{
				if (mbt.macroblock_intra && ph->concealment_motion_vectors)
					mbt.frame_motion_type = 2;
				// table 6-17 based on frame_motion_type
				// note: not properly looking up values for other than frame_motion_type == 2
				mtte = &table_6_17[mbt.frame_motion_type];
			}
			else
			{
				if (mbt.macroblock_intra && ph->concealment_motion_vectors)
					mbt.field_motion_type = 2;
				// table 6-18 based on field_motion_type
				mtte = &table_6_18[mbt.field_motion_type - 1];
			}

			mbt.prediction_type = mtte->prediction_type;
			mbt.motion_vector_count = mtte->motion_vector_count;
			mbt.mv_format = mtte->mv_format;
			mbt.dmv = mtte->dmv;
		}

		if (mbt.macroblock_motion_forward ||
			(mbt.macroblock_intra && ph->concealment_motion_vectors) )
		{
			// see 7.6.3, page 92, for motion vector use
			// 7.6.3.9 for concealment_motion_vectors specifically
			if (!motion_vectors(0, ph, &mbt))
			{
#ifdef REPORT_SLICE_DECODE_FAILURES
				_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
				return false;
			}
		}
		if (mbt.macroblock_motion_backward)
		{
			if (!motion_vectors(1, ph, &mbt))
			{
#ifdef REPORT_SLICE_DECODE_FAILURES
				_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
				return false;
			}
		}
		if (mbt.macroblock_intra && ph->concealment_motion_vectors)
			discard_bits(1);
		if (mbt.macroblock_pattern)
			__asm int 3		//coded_block_pattern(); // never for I frames

		// blocks
		int block_count = 0;
		switch (sh->chroma_format)
		{
		case 1:	// 4:2:0
			block_count = 6;
			break;
		case 2:	// 4:2:2
			block_count = 8;
			break;
		case 3:	// 4:4:4
			block_count = 12;
			break;
		default:
			__asm int 3		// bleargh!
		}

		for (int i = block_count - 1; i >= 0; i--)
		{
			if (mbt.macroblock_intra)
				pattern_code[i] = 1;
			else
				pattern_code[i] = 0;
		}
		if (mbt.macroblock_pattern)
		{
			__asm int 3		// never for I frames
/*			for (i=0; i<6; i++)
				if ( cbp & (1<<(5-i)) )
					pattern_code[i] = 1;
			if (chromaFormat == 4:2:2)
				for (i=6; i<8; i++)
					if ( coded_block_pattern_1 & (1<<(7-i)) )
						pattern_code[i] = 1;
			if (chromaFormat == 4:4:4)
				for (i=8; i<12; i++)
					if ( coded_block_pattern_2 & (1<<(11-i)) )
						pattern_code[i] = 1;
*/		}

		uint dct_dc_size;
		uint dct_dc_differential;

		// loop through blocks (7.2.1, page 77)
		for( int i = 0; i < block_count; i++)
		{
			if (pattern_code[i])
			{
				uint cc = 0;			// 0 = Y
				if (i >= 4)
					cc = 1 + (i&1);	// 1 = Cb, 2 = Cr

				short *Qd = &QF[i][0][0];
				if (mbt.macroblock_intra)
				{
					if (i < 4)
					{
						dct_dc_size = decode_dct_dc_size_luminance();
					}
					else
					{
						dct_dc_size = decode_dct_dc_size_chrominance();
					}
					dct_dc_differential = 0;
					if (dct_dc_size != 0)
						dct_dc_differential = get_bits(dct_dc_size);

					// calculate QFS[0] (*Qd)
					int dct_diff;
					if (dct_dc_size == 0)
						dct_diff = 0;
					else
					{
						uint half_range = 1 << (dct_dc_size - 1);
						if (dct_dc_differential >= half_range)
							dct_diff = dct_dc_differential;
						else
							dct_diff = (int)(dct_dc_differential + 1) - (int)(2 * half_range);
					}
					*Qd = dc_dct_pred[cc] + dct_diff;
					dc_dct_pred[cc] = *Qd;

					// QFS[0] should have range of 0 to ((256<<ph->intra_dc_precision)-1)
				}
				else
				{
					__asm int 3		// not required for basic I frames (?)
//					First DCT coefficient 2-24
					//??
//					if (!decode_DCT_coefficient(0, ph->intra_vlc_format, &eob_read, &run, Qd))
//						return false;
				}

				if (fullSize)
				{
					// convert QFS[n] to QF[i][v][u] via inverse scan, 7.3, page 80
					int *scan = &scan_QF_QFS[ph->alternate_scan][1];
					uint n = 1;
					bool eob_read = false;
					unsigned char sf = q_scale[ph->q_scale_type][quantiser_scale_code];
					unsigned int quant_sum = *Qd + 1;

					while (!eob_read)
					{
						uint run;
						int signed_level;
						if (!decode_DCT_coefficient(n, ph->intra_vlc_format, &eob_read, &run, &signed_level))
						{
#ifdef REPORT_SLICE_DECODE_FAILURES
							_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
							return false;
						}
						if (eob_read)
						{
							// end of block
							while (n++ < 64)
								Qd[*scan++] = 0;
							Qd[63] ^= quant_sum & 1;

							gpIDCT->transform(Qd, dst);

							// XXX Temp - recalc average to see if it displays correctly.
							int sum = 0;
							for (int idst = 0; idst < 64; idst++)
								sum += dst[idst];
							*Qd = sum / 8;
							if ((sum % 8) & 0x4)
								(*Qd)++;
						}
						else
						{
							for (uint m = 0; (m < run) && (n++ < 64); m++)
								Qd[*scan++] = 0;
							if (n++ < 64)
							{
								int scanCur = *scan++;
								int f = (signed_level * sf * ph->intra_quantiser_matrix[scanCur]) / 16;
								Qd[scanCur] = f;
								quant_sum += f;
							}
						}
					}
				}// end if(fullSize)
				else
				{
					// preview size, we don't actually need to keep the rest of the QFS elements,
					// so just process the bitstream only.
#if 1
					if (!discard_decode_DCT_coefficients( 1, ph->intra_vlc_format, &mbt ))
					{
#ifdef REPORT_SLICE_DECODE_FAILURES
						_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
						return false;
					}
#else
					uint n = 1;
					bool eob_read = false;
					while (!eob_read)
					{
						uint run;
						int signed_level;
						if (!decode_DCT_coefficient(n++, ph->intra_vlc_format, &eob_read, &run, &signed_level))
						{
#ifdef REPORT_SLICE_DECODE_FAILURES
							_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
							return false;
						}
					}
#endif
				}
			}
		}

		// fill block in! yay!
		if (bm)
		{
//			uint x_offset = (bm->width - (sh->horizontal_size_value>>3))>>1;
//			if (x_offset < 0)
//				x_offset = 0;
//			uint y_offset = (bm->height - (sh->vertical_size_value>>3))>>1;
//			if (y_offset < 0)
//				y_offset = 0;

			float invYdiv = 1.0f / (219.0f * (float)(1 << ph->intra_dc_precision));
			float invCdiv = 1.0f / (224.0f * (float)(1 << ph->intra_dc_precision));

//int R_color = 0;
//int G_color = 0;
//int B_color = 0;

			for( int i = 0; i < 4; i++)
			{
//				uint x = x_offset + ((mb_column<<1) + (i&1));
//				uint y = y_offset + ((mb_row<<1) + (i>>1));
				uint x = ((mb_column<<1) + (i&1));
				uint y = ((mb_row<<1) + (i>>1));

				if ((x < bm->width) && (y < bm->height))
				{
					int Y = QF[i][0][0];
					int Cb = 0;
					int Cr = 0;
					switch (sh->chroma_format)
					{
					case 1:	// 4:2:0
						Cb = QF[4][0][0];
						Cr = QF[5][0][0];
						break;
					case 2:	// 4:2:2
						Cb = QF[4][0][0];
						Cr = QF[5][0][0];
						break;
					case 3:	// 4:4:4
						Cb = QF[4+(i<<1)][0][0];
						Cr = QF[5+(i<<1)][0][0];
						break;
					}

					float EY = (float)(Y - (16 << ph->intra_dc_precision));
					float EPb = (float)(Cb - (128 << ph->intra_dc_precision));
					float EPr = (float)(Cr - (128 << ph->intra_dc_precision));

					if (g_fBrightness || g_fContrast)
					{
						EY *= g_contrast + g_brightness;
						EPb *= g_contrast;
						EPr *= g_contrast;
					}

					EY *= invYdiv;
					EPb *= invCdiv;
					EPr *= invCdiv;

					// see page 56 for matrix coefficients. Unfortunately, they
					// only list the RGB -> YUV matrix, so have to invert it.
					float EG = 0.9998f * EY - 0.1873f * EPb - 0.4580f * EPr;
					float EB = 1.0018f * EY + 1.8556f * EPb + 0.0010f * EPr;
					float ER = 1.0000f * EY + 0.0007f * EPb + 1.5751f * EPr;

#ifndef DO_ASM
					int R = (int)(255.0f * ER);
					if (R > 255) R = 255;
					if (R < 0) R = 0;
					int G = (int)(255.0f * EG);
					if (G > 255) G = 255;
					if (G < 0) G = 0;
					int B = (int)(255.0f * EB);
					if (B > 255) B = 255;
					if (B < 0) B = 0;

					uchar *p = bm->buf + (bm->height-y-1)*bm->stride + x*3;
					*p++ = (uchar)B;
					*p++ = (uchar)G;
					*p = (uchar)R;
#else
					// Branchless & pipelined clamping and writing.
					// not sure if this is much faster for just a preview size, but
					// it definitely helps for full size pics
					uchar *p = bm->buffer + (bm->height-y-1)*bm->stride + x*3;

					int R = (int)(255.0f * ER);
					int G = (int)(255.0f * EG);
					int B = (int)(255.0f * EB);

//R_color |= R;
//G_color |= G;
//B_color |= B;

					__asm
					{
						mov eax, R
						xor edx, edx		// = 0
						mov ebx, G
						mov ecx, B
						cmp eax, edx		// if (R < 0)
						cmovl eax, edx		//		R = 0;
						cmp ebx, edx		// if (G < 0)
						cmovl ebx, edx		// 	G = 0;
						cmp ecx, edx		// if (B < 0)
						cmovl ecx, edx		//		B = 0;
						mov edx, 255		// = 255
						mov esi, p
						cmp eax, edx		// if (R > 255)
						cmovg eax, edx		//		R = 255;
						cmp ebx, edx		// if (G > 255)
						cmovg ebx, edx		//		G = 255;
						mov [esi+2], al	// p[2] = R
						cmp ecx, edx		// if (B > 255)
						cmovg ecx, edx		//		B = 255;
						mov [esi+1], bl	// p[1] = G
						mov [esi], cl		// p[0] = B
					}
#endif
				}
			}

//int color = 1;
//if(( R_color & G_color & B_color ) == 0 )
//	color = 0;
//_Error("%d", color );
		}

	} while (peek_bits(23) != 0);

	return true;
}

bool decode_macroblock_address_increment(uint *val)
{
	uint bits = peek_bits(11);
	uint size;

	if (bits & 0x400)
	{
		size = 1;
		// 1-- ---- ---- = 1
		*val = 1;
	}
	else if (bits & 0x200)
	{
		size = 3;
		// 011 ---- ---- = 2
		// 010 ---- ---- = 3
		*val = 3 - ((bits & 0x100) >> 8);
	}
	else if (bits & 0x100)
	{
		size = 4;
		// 001 1--- ---- = 4
		// 001 0--- ---- = 5
		*val = 5 - ((bits & 0x80) >> 7);
	}
	else if (bits & 0x80)
	{
		size = 5;
		// 000 11-- ---- = 6
		// 000 10-- ---- = 7
		*val = 7 - ((bits & 0x40) >> 6);
	}
	else if (bits & 0x40)
	{
		if (bits & 0x20)
		{
			size = 7;
			// 000 0111 ---- = 8
			// 000 0110 ---- = 9
			*val = 9 - ((bits & 0x10) >> 4);
		}
		else
		{
			size = 8;
			// 000 0101 1--- = 10
			// 000 0101 0--- = 11
			// 000 0100 1--- = 12
			// 000 0100 0--- = 13
			*val = 13 - ((bits & 0x18) >> 3);
		}
	}
	else if (bits & 0x20)
	{
		if (bits & 0x10)
		{
			size = 8;
			// 000 0011 1--- = 14
			// 000 0011 0--- = 15
			*val = 15 - ((bits & 0x8) >> 3);
		}
		else
		{
			if (bits & 0xc)
			{
				size = 10;
				// 000 0010 111- = 16
				// 000 0010 110- = 17
				// 000 0010 101- = 18
				// 000 0010 100- = 19
				// 000 0010 011- = 20
				// 000 0010 010- = 21
				*val = 23 - ((bits & 0xe) >> 1);
			}
			else
			{
				size = 11;
				// 000 0010 0011 = 22
				// 000 0010 0010 = 23
				// 000 0010 0001 = 24
				// 000 0010 0000 = 25
				*val = 25 - (bits & 0x3);
			}
		}
	}
	else if (bits & 0x10)	// bit 5
	{
		size = 11;
		if (bits & 0x8)
		{
			// 000 0001 1111 = 26
			// 000 0001 1110 = 27
			// 000 0001 1101 = 28
			// 000 0001 1100 = 29
			// 000 0001 1011 = 30
			// 000 0001 1010 = 31
			// 000 0001 1001 = 32
			// 000 0001 1000 = 33
			*val = 33 - (bits & 0x7);
		}
		else
		{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
			__asm int 3		// ??????
#endif
			return false;
		}
	} else if ((bits & 0xf) == 0x8)
	{
		size = 11;
		*val = 33;	// 000 0000 1000 = 33 macroblock_escape
	} else
	{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
		__asm int 3		// ??????
#endif
		return false;
	}

	discard_bits(size);

	return true;
}

bool decode_macroblock_type(macroblock_type *mbt, uchar picture_coding_type, uchar scalable_mode)
{
	uint bits = peek_bits(9);
	uint size = 1;

	switch( scalable_mode )
	{
		case 2:	// SNR scalability
			__asm int 3		// not doing SNR support yet
			// all picture_coding_types
			return false;
			break;

		case 1:	// spatial scalability (picture_spatial_scalable_extension() -> present/not present????)

		switch( picture_coding_type )
		{
			case 1:	// I frame
			{
				if (bits & 0x100)
				{
					size = 1;
					// 1 ---- ----  Coded, Compatible
					mbt->macroblock_quant = 0;
					mbt->macroblock_motion_forward = 0;
					mbt->macroblock_motion_backward = 0;
					mbt->macroblock_pattern = 1;
					mbt->macroblock_intra = 0;
					mbt->spatial_temporal_weight_code_flag = 0;
				}
				else if (bits & 0x80)
				{
					size = 2;
					// 0 1--- ----  Coded, Compatible, Quant
					mbt->macroblock_quant = 1;
					mbt->macroblock_motion_forward = 0;
					mbt->macroblock_motion_backward = 0;
					mbt->macroblock_pattern = 1;
					mbt->macroblock_intra = 0;
					mbt->spatial_temporal_weight_code_flag = 0;
				}
				else if (bits & 0x40)
				{
					size = 4;
					if (bits & 0x20)
					{
						// 0 011- ----  Intra
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 1;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
					else
					{
						// 0 010- ----  Intra, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 1;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else if (bits & 0x20)
				{
					size = 4;
					// 0 001- ----  Not Coded, Compatible
					mbt->macroblock_quant = 0;
					mbt->macroblock_motion_forward = 0;
					mbt->macroblock_motion_backward = 0;
					mbt->macroblock_pattern = 0;
					mbt->macroblock_intra = 0;
					mbt->spatial_temporal_weight_code_flag = 0;
				}
				else
				{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
					__asm int 3		// invalid??
#endif
					return false;
				}

			}// end case 1:
			break;

			case 2:	// P frame
			{
				if( bits & 0x100 )
				{
					size = 2;
					if( bits & 0x80 )
					{
						// 1 1--- ---- MC, Coded, Compatible, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 1;
					}
					else 
					{
						// 1 0--- ---- MC, Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else if( bits & 0x80 )
				{
					size = 3;
					if( bits & 0x40 )
					{
						// 0 11-- ---- MC, Coded, Compatible
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 1;
					}
					else
					{
						// 0 10-- ---- MC, Coded, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else if( bits & 0x40 )
				{
					size = 4;
					if( bits & 0x20 )
					{
						// 0 011- ---- MC, Not Coded, Compatible
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 1;
					}
					else
					{
						// 0 010- ---- MC, Not Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else if( bits & 0x20 )
				{
					size = 6;
					if( bits & 0x10 )
					{ 
						if( bits & 0x08 )
						{
							// 0 0011 1--- No MC, Coded, Compatible
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 01;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 1;
						}
						else
						{
							// 0 0011 0--- No MC, Not Coded, Compatible
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 01;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 1;
						}
					}

					if( bits & 0x08 )
					{
						// 0 0010 1--- No MC, Code, Compatible, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 1;
					}
					else
					{
						// 0 0010 0--- No MC, Coded, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else if( bits & 0x10 )
				{
					size = 7;
					if( bits & 0x08 )
					{
						if( bits & 0x01 )
						{
							// 0 0001 11-- Intra
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 1;
							mbt->spatial_temporal_weight_code_flag = 0;
						}
						else
						{
							// 0 0001 10-- Intra, Quant
							mbt->macroblock_quant = 1;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 1;
							mbt->spatial_temporal_weight_code_flag = 0;
						}
					}

					if( bits & 0x01 )
					{
						// 0 0001 01-- Coded, Compatible
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;


					}
					else
					{
						// 0 0001 00-- No MC, Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else if( bits & 0x80 )
				{
					size = 7;
					if( bits & 0x04 )
					{
						// 0 0000 11-- Not Coded, Compatible
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
					else
					{
						// 0 0000 10-- Coded, Compatible, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else
				{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
					__asm int 3		// invalid??
#endif
					return false;
				}
			}// end case 2:
			break;

			case 3:	// B frame
			{
				if( bits & 0x100 )
				{
					size = 2;
					if( bits & 0x80 )
					{
						// 1 1--- ---- Interp, Coded
						mbt->macroblock_quant = 0; 
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 1;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
					else
					{
						// 1 0--- ---- Interp, Not Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 1;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else if( bits & 0x80 )
				{
					size = 3;
					if( bits & 0x40 )
					{
						// 0 11-- ---- Back, Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 1;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
					else
					{
						// 0 10-- ---- Back, Not Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 1;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else if( bits & 0x40 )
				{
					size = 4;
					if( bits & 0x20 )
					{
						// 0 011- ---- For, Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
					else
					{
						// 0 010- ---- For, Not Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 1;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else if( bits & 0x20 )
				{
					size = 6;
					if( bits & 0x10 )
					{
						if( bits & 0x08 )
						{
							// 0 0011 1--- Back, Coded, Compatible
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 1;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 1;
						}
						else
						{
							// 0 0011 0--- Back, Not Coded, Compatible
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 1;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 1;
						}
					}
					else if( bits & 0x08 )
					{
						// 0 0010 1--- For, Coded, Compatible
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 1;

					}
					else
					{
						// 0 0010 0--- For, Not Coded, Compatible
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 1;

					}
				}
				else if( bits & 0x10 )
				{
					size = 7;
					if( bits & 0x08 )
					{
						if( bits & 0x04 )
						{
							// 0 0001 11-- Interp, Coded, Quant
							mbt->macroblock_quant = 1;
							mbt->macroblock_motion_forward = 1;
							mbt->macroblock_motion_backward = 1;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
						}
						else
						{
							// 0 0001 10-- Intra
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 1;
							mbt->spatial_temporal_weight_code_flag = 0;
						}
					}
					else if( bits & 0x04 )
					{
						// 0 0001 01-- Back, Coded, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 1;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
					else
					{
						// 0 0001 00-- For, Coded, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
					}
				}
				else if( bits & 0x08 )
				{
					if( bits & 0x40 )
					{
						size = 9;
						if( bits & 0x02 )
						{
							// 0 0000 1111 Coded, Compatible
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
						}
						else
						{
							// 0 0000 1110 Not Coded, Compatible
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
						}
						if( bits & 0x01 )
						{
							// 0 0000 1101 Coded, Compatible, Quant
							mbt->macroblock_quant = 1;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
						}
						else
						{
							// 0 0000 1100 Back, Coded, Compatible, Quant
							mbt->macroblock_quant = 1;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 1;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 1;
						}
					}
					else
					{
						size = 8;
						if( bits & 0x02 )
						{
							// 0 0000 101- For, Coded, Compatible, Quant
							mbt->macroblock_quant = 1;
							mbt->macroblock_motion_forward = 1;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 1;
						}
						else
						{
							// 0 0000 100- Intra, Quant
							mbt->macroblock_quant = 1;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 1;
							mbt->spatial_temporal_weight_code_flag = 0;
						}
					}
				}
				else
				{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
					__asm int 3		// invalid??
#endif
					return false;
				}
			}// end case 3:
			break;
		}// end switch( picture_coding_type )

		default: // 0=data partitioning & 3=temporal scalability
		{
			switch( picture_coding_type )
			{
				case 1:	// I frame
				{
					// Table B-2, page 154
					if( bits & 0x100 )
					{
						size = 1;
						// 1 ---- ----  Intra
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 1;
						mbt->spatial_temporal_weight_code_flag = 0;
						mbt->permitted_spatial_temporal_weight_classes = 0;
					}
					else if( bits & 0x80 )
					{
						size = 2;
						// 0 1--- ----  Intra, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 1;
						mbt->spatial_temporal_weight_code_flag = 0;
						mbt->permitted_spatial_temporal_weight_classes = 0;
					}
					else
					{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
						__asm int 3		// invalid??
#endif
						return false;
					}
				}// end case 1:
				break;

				case 2:	// P frame
				{
					// Table B-3, page 154

					if( bits & 0x100 )
					{
						size = 1;
						// 1 ---- ---- MC, Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
						mbt->permitted_spatial_temporal_weight_classes = 0;
					}
					else if( bits & 0x80 )
					{
						size = 2;
						// 0 1--- ---- No MC, Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
						mbt->permitted_spatial_temporal_weight_classes = 0;
					}
					else if( bits & 0x40 )
					{
						size = 3;
						// 0 01-- ---- MC, Not Coded
						mbt->macroblock_quant = 0;
						mbt->macroblock_motion_forward = 1;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
						mbt->permitted_spatial_temporal_weight_classes = 0;
					}
					else if( bits & 0x20 )
					{
						size = 5;
						if( bits & 0x10 )
						{
							// 0 0011 ---- Intra
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 1;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
						else
						{
							// 0 0010 ---- MC, Coded, Quant
							mbt->macroblock_quant = 1;
							mbt->macroblock_motion_forward = 1;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
					}
					else if( bits & 0x10 )
					{
						size = 5;
						// 0 0001 ---- No MC, Coded, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 1;
						mbt->macroblock_intra = 0;
						mbt->spatial_temporal_weight_code_flag = 0;
						mbt->permitted_spatial_temporal_weight_classes = 0;
					}
					else if( bits & 0x08 )
					{
						size = 6;
						// 0 0000 1--- Intra, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 1;
						mbt->spatial_temporal_weight_code_flag = 0;
						mbt->permitted_spatial_temporal_weight_classes = 0;
					}
					else
					{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
						__asm int 3		// invalid??
#endif
						return false;
					}
				}// end case 2:
				break;

				case 3:	// B frame
				{
					// Table B-4, page 155
					if( bits & 0x100 )
					{
						size = 2;
						if( bits & 0x80 )
						{
							// 1 1--- ---- Interp, Coded
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 1;
							mbt->macroblock_motion_backward = 1;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
						else
						{
							// 1 0--- ---- Interp, Not Coded
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 1;
							mbt->macroblock_motion_backward = 1;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
					}
					else if( bits & 0x80 ) 
					{
						size = 3;
						if( bits & 0x40 )
						{
							// 0 11-- ---- Bwd, Coded
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 1;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;

						}
						else
						{
							// 0 10-- ---- Bwd, Not Coded
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 1;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
					}
					else if( bits & 0x40 )
					{	
						size = 4;
						if( bits & 0x20 )
						{
							// 0 011- ---- Fwd, Coded
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 1;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
						else
						{
							// 0 010- ---- Fwd, Not Coded
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 1;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
					}
					else if( bits & 0x20 )
					{
						size = 5;
						if( bits & 0x10 )
						{
							// 0 0011 ---- Intra
							mbt->macroblock_quant = 0;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 0;
							mbt->macroblock_intra = 1;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
						else
						{
							// 0 0010 ---- Interp, Coded, Quant
							mbt->macroblock_quant = 1;
							mbt->macroblock_motion_forward = 1;
							mbt->macroblock_motion_backward = 1;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
					}
					else if( bits & 0x10 )
					{
						size = 6;
						if( bits & 0x08 )
						{
							// 0 0001 1--- Fwd, Coded, Quant
							mbt->macroblock_quant = 1;
							mbt->macroblock_motion_forward = 1;
							mbt->macroblock_motion_backward = 0;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
						else
						{
							// 0 0001 0--- Bwd, Coded, Quant
							mbt->macroblock_quant = 1;
							mbt->macroblock_motion_forward = 0;
							mbt->macroblock_motion_backward = 1;
							mbt->macroblock_pattern = 1;
							mbt->macroblock_intra = 0;
							mbt->spatial_temporal_weight_code_flag = 0;
							mbt->permitted_spatial_temporal_weight_classes = 0;
						}
					}
					else if( bits & 0x08 )
					{
						size = 6;
						// 0 0000 1--- Intra, Quant
						mbt->macroblock_quant = 1;
						mbt->macroblock_motion_forward = 0;
						mbt->macroblock_motion_backward = 0;
						mbt->macroblock_pattern = 0;
						mbt->macroblock_intra = 1;
						mbt->spatial_temporal_weight_code_flag = 0;
						mbt->permitted_spatial_temporal_weight_classes = 0;
					}
					else
					{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
						__asm int 3		// invalid??
#endif
						return false;
					}
				}// end case 3:
			}// end switch( picture_coding_type )
		}// end default:
		break;
	}// end switch( scalable_mode )

 	discard_bits(size);

	return true;
}

// Table B-12
uint decode_dct_dc_size_luminance(void)
{
	// can't fail
	uint bits = peek_bits(9);
	uint size;
	uint val;

	if (bits & 0x100)
	{
		if (bits & 0x80)
		{
			if (bits & 0x40)
			{
				if (bits & 0x20)
				{
					if (bits & 0x10)
					{
						if (bits & 0x8)
						{
							if (bits & 0x4)
							{
								if (bits & 0x2)
								{
									size = 9;
									// 1 1111 1111 = 11
									// 1 1111 1110 = 10
									val = 10 + (bits & 0x1);
								}
								else
								{
									size = 8;
									// 1 1111 110- = 9
									val = 9;
								}
							}
							else
							{
								size = 7;
								// 1 1111 10-- = 8
								val = 8;
							}
						}
						else
						{
							size = 6;
							// 1 1111 0--- = 7
							val = 7;
						}
					}
					else
					{
						size = 5;
						// 1 1110 ---- = 6
						val = 6;
					}
				}
				else
				{
					size = 4;
					// 1 110- ---- = 5
					val = 5;
				}
			}
			else
			{
				size = 3;
				// 1 10-- ---- = 4
				val = 4;
			}
		}
		else
		{
			size = 3;
			if (bits & 0x40)
				// 1 01-- ---- = 3
				val = 3;
			else
				// 1 00-- ---- = 0
				val = 0;
		}
	}
	else
	{
		size = 2;
		// 0 1--- ---- = 2
		// 0 0--- ---- = 1
		val = 1 + ((bits & 0x80) >> 7);
	}

	discard_bits(size);

	return val;
}

// Table B-13
uint decode_dct_dc_size_chrominance(void)
{
	// can't fail
	uint bits = peek_bits(10);
	uint size;
	uint val;

	if (bits & 0x200)
	{
		if (bits & 0x100)
		{
			if (bits & 0x80)
			{
				if (bits & 0x40)
				{
					if (bits & 0x20)
					{
						if (bits & 0x10)
						{
							if (bits & 0x8)
							{
								if (bits & 0x4)
								{
									if (bits & 0x2)
									{
										size = 10;
										// 11 1111 1111 = 11
										// 11 1111 1110 = 10
										val = 10 + (bits & 0x1);
									}
									else
									{
										size = 9;
										val = 9;		// 11 1111 110- = 9
									}
								}
								else
								{
									size = 8;
									val = 8;		// 11 1111 10-- = 8
								}
							}
							else
							{
								size = 7;
								val = 7;		// 11 1111 0--- = 7
							}
						}
						else
						{
							size = 6;
							val = 6;		// 11 1110 ---- = 6
						}
					}
					else
					{
						size = 5;
						val = 5;		// 11 110- ---- = 5
					}
				}
				else
				{
					size = 4;
					val = 4;		// 11 10-- ---- = 4
				}
			}
			else
			{
				size = 3;
				val = 3;		// 11 0--- ---- = 3
			}
		}
		else
		{
			size = 2;
			val = 2;		// 10 ---- ---- = 2
		}
	}
	else
	{
		size = 2;
		// 01 ---- ---- = 1
		// 00 ---- ---- = 0
		val = ((bits & 0x100) >> 8);
	}

	discard_bits(size);

	return val;
}

// tables B-14, B-15, B16
bool decode_DCT_coefficient(int n, uint intra_vlc_format, bool *end_of_block, uint *run, int *signed_level)
{
	uint bits = peek_bits(17);
	uint size;
	*end_of_block = false;
	uint signed_bit = 0;

	if (intra_vlc_format == 0)
	{
		// Table B-14 --- DCT coefficients Table zero
		if (bits & 0x10000)
		{
			if (n == 0)
			{
				size = 2;
				signed_bit = (bits & 0x8000);
				// 1 s--- ---- ---- ---- = 0 1 (NOTE 3) This code shall be used for the first (DC) coefficient in the block
				*run = 0;
				*signed_level = 1;
			}
			else if (bits & 0x8000)
			{
				size = 3;
				signed_bit = (bits & 0x4000);
				// 1 1s-- ---- ---- ---- = 0 1 (NOTE 4) This code shall be used for all other coefficients
				*run = 0;
				*signed_level = 1;
			}
			else
			{
				size = 2;
				// 1 0--- ---- ---- ---- = End of Block
				*end_of_block = true;
			}
		}
		else if (bits & 0x8000)
		{
			if (bits & 0x4000)
			{
				size = 4;
				signed_bit = (bits & 0x2000);
				// 0 11s- ---- ---- ---- = 1 1
				*run = 1;
				*signed_level = 1;
			}
			else
			{
				size = 5;
				signed_bit = (bits & 0x1000);
				// 0 101s ---- ---- ---- = 2 1
				// 0 100s ---- ---- ---- = 0 2
				if (bits & 0x2000)
				{
					*run = 2;
					*signed_level = 1;
				}
				else
				{
					*run = 0;
					*signed_level = 2;
				}

			}
		}
		else if (bits & 0x4000)
		{
			if (bits & 0x3000)
			{
				size = 6;
				signed_bit = (bits & 0x800);
				// 0 0111 s--- ---- ---- = 3 1
				// 0 0110 s--- ---- ---- = 4 1
				// 0 0101 s--- ---- ---- = 0 3
				if (bits & 0x2000)
				{
					if (bits & 0x1000)
						*run = 3;
					else
						*run = 4;
					*signed_level = 1;
				}
				else
				{
					*run = 0;
					*signed_level = 3;
				}
			}
			else
			{
				size = 9;
				signed_bit = (bits & 0x100);
				if (bits & 0x800)
				{
					if (bits & 0x400)
					{
					// 0 0100 111s ---- ---- = 10 1
					// 0 0100 110s ---- ---- = 0 5
						if (bits & 0x200)
						{
							*run = 10;
							*signed_level = 1;
						}
						else
						{
							*run = 0;
							*signed_level = 5;
						}
					}
					else
					{
					// 0 0100 101s ---- ---- = 1 3
					// 0 0100 100s ---- ---- = 3 2
						if (bits & 0x200)
						{
							*run = 1;
							*signed_level = 3;
						}
						else
						{
							*run = 3;
							*signed_level = 2;
						}
					}
				}
				else
				{
					if (bits & 0x400)
					{
					// 0 0100 011s ---- ---- = 11 1
					// 0 0100 010s ---- ---- = 12 1
						if (bits & 0x200)
						{
							*run = 11;
							*signed_level = 1;
						}
						else
						{
							*run = 12;
							*signed_level = 1;
						}
					}
					else
					{
					// 0 0100 001s ---- ---- = 0 6
					// 0 0100 000s ---- ---- = 13 1
						if (bits & 0x200)
						{
							*run = 0;
							*signed_level = 6;
						}
						else
						{
							*run = 13;
							*signed_level = 1;
						}
					}
				}
			}
		}
		else if (bits & 0x2000)
		{
			size = 7;
			signed_bit = (bits & 0x400);
			if (bits & 0x1000)
			{
				// 0 0011 1s-- ---- ---- = 5 1
				// 0 0011 0s-- ---- ---- = 1 2
				if (bits & 0x800)
				{
					*run = 5;
					*signed_level = 1;
				}
				else
				{
					*run = 1;
					*signed_level = 2;
				}
			}
			else
			{
				// 0 0010 1s-- ---- ---- = 6 1
				// 0 0010 0s-- ---- ---- = 7 1
				if (bits & 0x800)
				{
					*run = 6;
					*signed_level = 1;
				}
				else
				{
					*run = 7;
					*signed_level = 1;
				}
			}
		}
		else if (bits & 0x1000)
		{
			size = 8;
			signed_bit = (bits & 0x200);
			if (bits & 0x800)
			{
				// 0 0001 11s- ---- ---- = 8 1
				// 0 0001 10s- ---- ---- = 0 4
				if (bits & 0x400)
				{
					*run = 8;
					*signed_level = 1;
				}
				else
				{
					*run = 0;
					*signed_level = 4;
				}
			}
			else
			{
				// 0 0001 01s- ---- ---- = 9 1
				// 0 0001 00s- ---- ---- = 2 2
				if (bits & 0x400)
				{
					*run = 9;
					*signed_level = 1;
				}
				else
				{
					*run = 2;
					*signed_level = 2;
				}
			}
		}
		else if (bits & 0x800)
		{
			// size = 6;
			discard_bits(6);
			// 0 0000 1--- ---- ---- = Escape
			*run = get_bits(6);
			*signed_level = get_bits(12);
			if ((*signed_level) & 0x800)
					*signed_level |= 0xfffff000;	// sign extend
			return true;
		}
		else if (bits & 0x400)
		{
			size = 11;
			signed_bit = (bits & 0x40);
			if (bits & 0x200)
			{
			// 0 0000 0111 1s-- ---- = 4 2
			// 0 0000 0111 0s-- ---- = 14 1
			// 0 0000 0110 1s-- ---- = 15 1
			// 0 0000 0110 0s-- ---- = 1 4
				if (bits & 0x100)
				{
					if (bits & 0x80)
					{
						*run = 4;
						*signed_level = 2;
					}
					else
					{
						*run = 14;
						*signed_level = 1;
					}
				}
				else
				{
					if (bits & 0x80)
					{
						*run = 15;
						*signed_level = 1;
					}
					else
					{
						*run = 1;
						*signed_level = 4;
					}
				}
			}
			else
			{
			// 0 0000 0101 1s-- ---- = 2 3
			// 0 0000 0101 0s-- ---- = 0 7
			// 0 0000 0100 1s-- ---- = 5 2
			// 0 0000 0100 0s-- ---- = 16 1
				if (bits & 0x100)
				{
					if (bits & 0x80)
					{
						*run = 2;
						*signed_level = 3;
					}
					else
					{
						*run = 0;
						*signed_level = 7;
					}
				}
				else
				{
					if (bits & 0x80)
					{
						*run = 5;
						*signed_level = 2;
					}
					else
					{
						*run = 16;
						*signed_level = 1;
					}
				}
			}
		}
		else if (bits & 0x200)
		{
			size = 13;
			signed_bit = (bits & 0x10);
			if (bits & 0x100)
			{
				if (bits & 0x80)
				{
				// 0 0000 0011 111s ---- = 17 1
				// 0 0000 0011 110s ---- = 6 2
				// 0 0000 0011 101s ---- = 0 8
				// 0 0000 0011 100s ---- = 3 3
					if (bits & 0x40)
					{
						if (bits & 0x20)
						{
							*run = 17;
							*signed_level = 1;
						}
						else
						{
							*run = 6;
							*signed_level = 2;
						}
					}
					else
					{
						if (bits & 0x20)
						{
							*run = 0;
							*signed_level = 8;
						}
						else
						{
							*run = 3;
							*signed_level = 3;
						}
					}
				}
				else
				{
				// 0 0000 0011 011s ---- = 1 5
				// 0 0000 0011 010s ---- = 18 1
				// 0 0000 0011 001s ---- = 19 1
				// 0 0000 0011 000s ---- = 0 9
					if (bits & 0x40)
					{
						if (bits & 0x20)
						{
							*run = 1;
							*signed_level = 5;
						}
						else
						{
							*run = 18;
							*signed_level = 1;
						}
					}
					else
					{
						if (bits & 0x20)
						{
							*run = 19;
							*signed_level = 1;
						}
						else
						{
							*run = 0;
							*signed_level = 9;
						}
					}
				}
			}
			else
			{
				if (bits & 0x80)
				{
				// 0 0000 0010 111s ---- = 20 1
				// 0 0000 0010 110s ---- = 21 1
				// 0 0000 0010 101s ---- = 7 2
				// 0 0000 0010 100s ---- = 2 4
					if (bits & 0x40)
					{
						if (bits & 0x20)
						{
							*run = 20;
							*signed_level = 1;
						}
						else
						{
							*run = 21;
							*signed_level = 1;
						}
					}
					else
					{
						if (bits & 0x20)
						{
							*run = 7;
							*signed_level = 2;
						}
						else
						{
							*run = 2;
							*signed_level = 4;
						}
					}
				}
				else
				{
				// 0 0000 0010 011s ---- = 0 10
				// 0 0000 0010 010s ---- = 4 3
				// 0 0000 0010 001s ---- = 8 2
				// 0 0000 0010 000s ---- = 0 11
					if (bits & 0x40)
					{
						if (bits & 0x20)
						{
							*run = 0;
							*signed_level = 10;
						}
						else
						{
							*run = 4;
							*signed_level = 3;
						}
					}
					else
					{
						if (bits & 0x20)
						{
							*run = 8;
							*signed_level = 2;
						}
						else
						{
							*run = 0;
							*signed_level = 11;
						}
					}
				}
			}
		}
		else if (bits & 0x100)
		{
			size = 14;
			signed_bit = (bits & 0x4);
			if (bits & 0x80)
			{
				if (bits & 0x40)
				{
			// 0 0000 0001 1111 s--- = 22 1
			// 0 0000 0001 1110 s--- = 23 1
			// 0 0000 0001 1101 s--- = 24 1
			// 0 0000 0001 1100 s--- = 25 1
					*run = 25 - (int)((bits & 0x30) >> 4);
					*signed_level = 1;
				}
				else
				{
					if ((bits & 0x30) == 0x30)
					{
			// 0 0000 0001 1011 s--- = 26 1
						*run = 26;
						*signed_level = 1;
					}
					else
					{
			// 0 0000 0001 1010 s--- = 0 12
			// 0 0000 0001 1001 s--- = 0 13
			// 0 0000 0001 1000 s--- = 0 14
						*run = 0;
						*signed_level = 14 - (int)((bits & 0x30) >> 4);
					}
				}
			}
			else
			{
				if (bits & 0x40)
				{
				// 0 0000 0001 0111 s--- = 0 15
				// 0 0000 0001 0110 s--- = 1 6
				// 0 0000 0001 0101 s--- = 1 7
				// 0 0000 0001 0100 s--- = 2 5
					if (bits & 0x20)
					{
						if (bits & 0x10)
						{
							*run = 0;
							*signed_level = 15;
						}
						else
						{
							*run = 1;
							*signed_level = 6;
						}
					}
					else
					{
						if (bits & 0x10)
						{
							*run = 1;
							*signed_level = 7;
						}
						else
						{
							*run = 2;
							*signed_level = 5;
						}
					}
				}
				else
				{
				// 0 0000 0001 0011 s--- = 3 4
				// 0 0000 0001 0010 s--- = 5 3
				// 0 0000 0001 0001 s--- = 9 2
				// 0 0000 0001 0000 s--- = 10 2
					if (bits & 0x20)
					{
						if (bits & 0x10)
						{
							*run = 3;
							*signed_level = 4;
						}
						else
						{
							*run = 5;
							*signed_level = 3;
						}
					}
					else
					{
						if (bits & 0x10)
						{
							*run = 9;
							*signed_level = 2;
						}
						else
						{
							*run = 10;
							*signed_level = 2;
						}
					}
				}
			}
		}
		else if (bits & 0x80)
		{
			size = 15;
			signed_bit = (bits & 0x4);
			// 0 0000 0000 1111 1s-- = 0 16
			// 0 0000 0000 1111 0s-- = 0 17
			// 0 0000 0000 1110 1s-- = 0 18
			// 0 0000 0000 1110 0s-- = 0 19
			// 0 0000 0000 1101 1s-- = 0 20
			// 0 0000 0000 1101 0s-- = 0 21
			// 0 0000 0000 1100 1s-- = 0 22
			// 0 0000 0000 1100 0s-- = 0 23
			// 0 0000 0000 1011 1s-- = 0 24
			// 0 0000 0000 1011 0s-- = 0 25
			// 0 0000 0000 1010 1s-- = 0 26
			// 0 0000 0000 1010 0s-- = 0 27
			// 0 0000 0000 1001 1s-- = 0 28
			// 0 0000 0000 1001 0s-- = 0 29
			// 0 0000 0000 1000 1s-- = 0 30
			// 0 0000 0000 1000 0s-- = 0 31
			*run = 0;
			*signed_level = 31 - (int)((bits & 0x78) >> 3);
		}
		else if (bits & 0x40)
		{
			size = 16;
			signed_bit = (bits & 0x2);
			if (bits & 0x20)
			{
				if (bits & 0x1c)
				{
				// 0 0000 0000 0111 11s- = 1 8
				// 0 0000 0000 0111 10s- = 1 9
				// 0 0000 0000 0111 01s- = 1 10
				// 0 0000 0000 0111 00s- = 1 11
				// 0 0000 0000 0110 11s- = 1 12
				// 0 0000 0000 0110 10s- = 1 13
				// 0 0000 0000 0110 01s- = 1 14
					*run = 1;
					*signed_level = 15 - (int)((bits & 0x1c) >> 2);
				}
				else
				{
				// 0 0000 0000 0110 00s- = 0 32
					*run = 0;
					*signed_level = 32;
				}
			}
			else
			{
			// 0 0000 0000 0101 11s- = 0 33
			// 0 0000 0000 0101 10s- = 0 34
			// 0 0000 0000 0101 01s- = 0 35
			// 0 0000 0000 0101 00s- = 0 36
			// 0 0000 0000 0100 11s- = 0 37
			// 0 0000 0000 0100 10s- = 0 38
			// 0 0000 0000 0100 01s- = 0 39
			// 0 0000 0000 0100 00s- = 0 40
				*run = 0;
				*signed_level = 40 - (int)((bits & 0x1c) >> 2);
			}
		}
		else if (bits & 0x20)
		{
			size = 17;
			signed_bit = (bits & 0x1);
			if (bits & 0x10)
			{
				if (bits & 0x8)
				{
			// 0 0000 0000 0011 111s = 27 1
			// 0 0000 0000 0011 110s = 28 1
			// 0 0000 0000 0011 101s = 29 1
			// 0 0000 0000 0011 100s = 30 1
					*run = 30 - (int)((bits & 0x6) >> 1);
					*signed_level = 1;
				}
				else
				{
					if (bits & 0x4)
					{
			// 0 0000 0000 0011 011s = 31 1
			// 0 0000 0000 0011 010s = 11 2
						if (bits & 0x2)
						{
							*run = 31;
							*signed_level = 1;
						}
						else
						{
							*run = 11;
							*signed_level = 2;
						}
					}
					else
					{
			// 0 0000 0000 0011 001s = 12 2
			// 0 0000 0000 0011 000s = 13 2
						*run = 13 - (int)((bits & 0x2) >> 1);
						*signed_level = 2;
					}
				}
			}
			else
			{
				if (bits & 0x8)
				{
					if (bits & 0x6)
					{
					// 0 0000 0000 0010 111s = 14 2
					// 0 0000 0000 0010 110s = 15 2
					// 0 0000 0000 0010 101s = 16 2
						*run = 17 - (int)((bits & 0x6) >> 1);
						*signed_level = 2;
					}
					else
					{
					// 0 0000 0000 0010 100s = 6 3
						*run = 6;
						*signed_level = 3;
					}
				}
				else
				{
				// 0 0000 0000 0010 011s = 1 15
				// 0 0000 0000 0010 010s = 1 16
				// 0 0000 0000 0010 001s = 1 17
				// 0 0000 0000 0010 000s = 1 18
					*run = 1;
					*signed_level = 18 - (int)((bits & 0x6) >> 1);
				}
			}
		}
		else
		{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
			__asm int 3		// huh???
#endif
			return false;
		}
	}
	else
	{
		// table B-15
		if (bits & 0x10000)
		{
			if (bits & 0x8000)
			{
				if (bits & 0x4000)
				{
					if (bits & 0x2000)
					{
						if (bits & 0x1000)
						{
							if (bits & 0x800)
							{
								size = 9;
								signed_bit = (bits & 0x100);
								if (bits & 0x400)
								{
									// 1 1111 111s ---- ---- = 0 15
									// 1 1111 110s ---- ---- = 0 14
									*run = 0;
									*signed_level = 14 + ((bits & 0x200) >> 9);
								}
								else if (bits & 0x200)
								{
									// 1 1111 101s ---- ---- = 4 2
									*run = 4;
									*signed_level = 2;
								}
								else
								{
									// 1 1111 100s ---- ---- = 2 3
									*run = 2;
									*signed_level = 3;
								}
							}
							else if (bits & 0x400)
							{
								size = 9;
								signed_bit = (bits & 0x100);
								// 1 1111 011s ---- ---- = 0 13
								// 1 1111 010s ---- ---- = 0 12
								*run = 0;
								*signed_level = 12 + ((bits & 0x200) >> 9);
							}
							else
							{
								size = 8;
								signed_bit = (bits & 0x200);
								// 1 1111 00s- ---- ---- = 0 9
								*run = 0;
								*signed_level = 9;
							}
						}
						else
						{
							size = 8;
							signed_bit = (bits & 0x200);
							if (bits & 0x800)
							{
								if (bits & 0x400)
								{
									// 1 1110 11s- ---- ---- = 0 8
									*run = 0;
									*signed_level = 8;
								}
								else
								{
									// 1 1110 10s- ---- ---- = 10 1
									*run = 10;
									*signed_level = 1;
								}
							}
							else if (bits & 0x400)
							{
								// 1 1110 01s- ---- ---- = 1 3
								*run = 1;
								*signed_level = 3;
							}
							else
							{
								// 1 1110 00s- ---- ---- = 9 1
								*run = 9;
								*signed_level = 1;
							}
						}
					}
					else
					{
						size = 6;
						signed_bit = (bits & 0x800);
						// 1 1101 s--- ---- ---- = 0 5
						// 1 1100 s--- ---- ---- = 0 4
						*run = 0;
						*signed_level = 4 + ((bits & 0x1000) >> 12);
					}
				}
				else
				{
					size = 4;
					signed_bit = (bits & 0x2000);
					// 1 10s- ---- ---- ---- = 0 2
					*run = 0;
					*signed_level = 2;
				}
			}
			else
			{
				size = 3;
				signed_bit = (bits & 0x4000);
				// 1 0s-- ---- ---- ---- = 0 1
				*run = 0;
				*signed_level = 1;
			}
		}
		else if (bits & 0x8000)
		{
			if (bits & 0x4000)
			{
				if (bits & 0x2000)
				{
					size = 5;
					signed_bit = (bits & 0x1000);
					// 0 111s ---- ---- ---- = 0 3
					*run = 0;
					*signed_level = 3;
				}
				else
				{
					size = 4;
					// 0 110- ---- ---- ---- = End of Block
					*end_of_block = true;
				}
			}
			else
			{
				size = 4;
				signed_bit = (bits & 0x2000);
				// 0 10s- ---- ---- ---- = 1 1
				*run = 1;
				*signed_level = 1;
			}
		}
		else if (bits & 0x4000)
		{
			if (bits & 0x2000)
			{
				size = 6;
				signed_bit = (bits & 0x800);
				if (bits & 0x1000)
				{
					// 0 0111 s--- ---- ---- = 3 1
					*run = 3;
					*signed_level = 1;
				}
				else
				{
					// 0 0110 s--- ---- ---- = 1 2
					*run = 1;
					*signed_level = 2;
				}
			}
			else if (bits & 0x1000)
			{
				size = 6;
				signed_bit = (bits & 0x800);
				// 0 0101 s--- ---- ---- = 2 1
				*run = 2;
				*signed_level = 1;
			}
			else
			{
				size = 9;
				signed_bit = (bits & 0x100);
				if (bits & 0x800)
				{
					if (bits & 0x400)
					{
						if (bits & 0x200)
						{
							// 0 0100 111s ---- ---- = 1 4
							*run = 1;
							*signed_level = 4;
						}
						else
						{
							// 0 0100 110s ---- ---- = 3 2
							*run = 3;
							*signed_level = 2;
						}
					}
					else
					{
						// 0 0100 101s ---- ---- = 12 1
						// 0 0100 100s ---- ---- = 13 1
						*run = 13 - ((bits & 0x200) >> 9);
						*signed_level = 1;
					}
				}
				else if (bits & 0x400)
				{
					// 0 0100 011s ---- ---- = 0 10
					// 0 0100 010s ---- ---- = 0 11
					*run = 0;
					*signed_level = 11 - ((bits & 0x200) >> 9);
				}
				else if (bits & 0x200)
				{
					// 0 0100 001s ---- ---- = 11 1
					*run = 11;
					*signed_level = 1;
				}
				else
				{
					// 0 0100 000s ---- ---- = 1 5
					*run = 1;
					*signed_level = 5;
				}
			}
		}
		else if (bits & 0x2000)
		{
			size = 7;
			signed_bit = (bits & 0x400);
			if (bits & 0x1000)
			{
				// 0 0011 1s-- ---- ---- = 5 1
				// 0 0011 0s-- ---- ---- = 4 1
				*run = 4 + ((bits & 0x800) >> 11);
				*signed_level = 1;
			}
			else
			{
				// 0 0010 1s-- ---- ---- = 0 6
				// 0 0010 0s-- ---- ---- = 0 7
				*run = 0;
				*signed_level = 7 - ((bits & 0x800) >> 11);
			}
		}
		else if (bits & 0x1000)
		{
			size = 8;
			signed_bit = (bits & 0x200);
			if (bits & 0x800)
			{
				if (bits & 0x400)
				{
					// 0 0001 11s- ---- ---- = 2 2
					*run = 2;
					*signed_level = 2;
				}
				else
				{
					// 0 0001 10s- ---- ---- = 6 1
					*run = 6;
					*signed_level = 1;
				}
			} else
			{
				// 0 0001 01s- ---- ---- = 8 1
				// 0 0001 00s- ---- ---- = 7 1
				*run = 7 + ((bits & 0x400) >> 10);
				*signed_level = 1;
			}
		}
		else if (bits & 0x800)
		{
			// size = 6;
			discard_bits(6);
			// 0 0000 1--- ---- ---- = Escape
			*run = get_bits(6);
			*signed_level = get_bits(12);
			if ((*signed_level) & 0x800)
					*signed_level |= 0xfffff000;	// sign extend
			return true;
		}
		else if (bits & 0x400)
		{
			if (bits & 0x200)
			{
				if (bits & 0x100)
				{
					size = 10;
					signed_bit = (bits & 0x80);
					// 0 0000 0111 s--- ---- = 15 1
					*run = 15;
					*signed_level = 1;
				}
				else
				{
					size = 11;
					signed_bit = (bits & 0x40);
					if (bits & 0x80)
					{
						// 0 0000 0110 1s-- ---- = 16 1
						*run = 16;
						*signed_level = 1;
					}
					else
					{
						// 0 0000 0110 0s-- ---- = 2 4
						*run = 2;
						*signed_level = 4;
					}
				}
			}
			else
			{
				size = 10;
				signed_bit = (bits & 0x80);
				if (bits & 0x100)
				{
					// 0 0000 0101 s--- ---- = 14 1
					*run = 14;
					*signed_level = 1;
				}
				else
				{
					// 0 0000 0100 s--- ---- = 5 2
					*run = 5;
					*signed_level = 2;
				}
			}
		}
		else if (bits & 0x200)
		{
			size = 13;
			signed_bit = (bits & 0x10);
			switch (bits & 0x1e0)
			{
			case 0x1e0:
				// 0 0000 0011 111s ---- = 17 1
				*run = 17;
				*signed_level = 1;
				break;

			case 0x1c0:
				// 0 0000 0011 110s ---- = 6 2
				*run = 6;
				*signed_level = 2;
				break;

			case 0x180:
				// 0 0000 0011 100s ---- = 3 3
				*run = 3;
				*signed_level = 3;
				break;

			case 0x140:
				// 0 0000 0011 010s ---- = 18 1
				*run = 18;
				*signed_level = 1;
				break;

			case 0x120:
				// 0 0000 0011 001s ---- = 19 1
				*run = 19;
				*signed_level = 1;
				break;

			case 0x0e0:
				// 0 0000 0010 111s ---- = 20 1
				*run = 20;
				*signed_level = 1;
				break;

			case 0x0c0:
				// 0 0000 0010 110s ---- = 21 1
				*run = 21;
				*signed_level = 1;
				break;

			case 0x0a0:
				// 0 0000 0010 101s ---- = 7 2
				*run = 7;
				*signed_level = 2;
				break;

			case 0x040:
				// 0 0000 0010 010s ---- = 4 3
				*run = 4;
				*signed_level = 3;
				break;

			case 0x020:
				// 0 0000 0010 001s ---- = 8 2
				*run = 8;
				*signed_level = 2;
				break;

			default:
				// 0 0000 0011 101x ---- = ??????
				// 0 0000 0011 011x ---- = ??????
				// 0 0000 0011 000x ---- = ??????
				// 0 0000 0010 100x ---- = ??????
				// 0 0000 0010 011x ---- = ??????
				// 0 0000 0010 000x ---- = ??????
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
				__asm int 3		// huh?
#endif
				return false;
			}
		}
		else if (bits & 0x100)
		{
			size = 14;
			signed_bit = (bits & 0x8);
			switch (bits & 0xf0)
			{
			case 0xf0:
				// 0 0000 0001 1111 s--- = 22 1
				*run = 22;
				*signed_level = 1;
				break;

			case 0xe0:
				// 0 0000 0001 1110 s--- = 23 1
				*run = 23;
				*signed_level = 1;
				break;

			case 0xd0:
				// 0 0000 0001 1101 s--- = 24 1
				*run = 24;
				*signed_level = 1;
				break;

			case 0xc0:
				// 0 0000 0001 1100 s--- = 25 1
				*run = 25;
				*signed_level = 1;
				break;

			case 0xb0:
				// 0 0000 0001 1011 s--- = 26 1
				*run = 26;
				*signed_level = 1;
				break;

			case 0x60:
				// 0 0000 0001 0110 s--- = 1 6
				*run = 1;
				*signed_level = 6;
				break;

			case 0x50:
				// 0 0000 0001 0101 s--- = 1 7
				*run = 1;
				*signed_level = 7;
				break;

			case 0x40:
				// 0 0000 0001 0100 s--- = 2 5
				*run = 2;
				*signed_level = 5;
				break;

			case 0x30:
				// 0 0000 0001 0011 s--- = 3 4
				*run = 3;
				*signed_level = 4;
				break;

			case 0x20:
				// 0 0000 0001 0010 s--- = 5 3
				*run = 5;
				*signed_level = 3;
				break;

			case 0x10:
				// 0 0000 0001 0001 s--- = 9 2
				*run = 9;
				*signed_level = 2;
				break;

			case 0x00:
				// 0 0000 0001 0000 s--- = 10 2
				*run = 10;
				*signed_level = 2;
				break;

			default:
				// 0 0000 0001 1010 x--- = ??????
				// 0 0000 0001 100x x--- = ??????
				// 0 0000 0001 0111 x--- = ??????
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
				__asm int 3		// huh???
#endif
				return false;
			}
		}
		else if (bits & 0x80)
		{
			size = 15;
			signed_bit = (bits & 0x4);
			// 0 0000 0000 1111 1s-- = 0 16
			// 0 0000 0000 1111 0s-- = 0 17
			// 0 0000 0000 1110 1s-- = 0 18
			// 0 0000 0000 1110 0s-- = 0 19
			// 0 0000 0000 1101 1s-- = 0 20
			// 0 0000 0000 1101 0s-- = 0 21
			// 0 0000 0000 1100 1s-- = 0 22
			// 0 0000 0000 1100 0s-- = 0 23
			// 0 0000 0000 1011 1s-- = 0 24
			// 0 0000 0000 1011 0s-- = 0 25
			// 0 0000 0000 1010 1s-- = 0 26
			// 0 0000 0000 1010 0s-- = 0 27
			// 0 0000 0000 1001 1s-- = 0 28
			// 0 0000 0000 1001 0s-- = 0 29
			// 0 0000 0000 1000 1s-- = 0 30
			// 0 0000 0000 1000 0s-- = 0 31
			*run = 0;
			*signed_level = 31 - ((bits & 0x78) >> 3);
		}
		else if (bits & 0x40)
		{
			size = 16;
			signed_bit = (bits & 0x2);
			if ((bits & 0x3c) > 0x20)
			{
				// 0 0000 0000 0111 11s- = 1 8
				// 0 0000 0000 0111 10s- = 1 9
				// 0 0000 0000 0111 01s- = 1 10
				// 0 0000 0000 0111 00s- = 1 11
				// 0 0000 0000 0110 11s- = 1 12
				// 0 0000 0000 0110 10s- = 1 13
				// 0 0000 0000 0110 01s- = 1 14
				*run = 1;
				*signed_level = 23 - ((bits & 0x3c) >> 2);
			}
			else
			{
				// 0 0000 0000 0110 00s- = 0 32
				// 0 0000 0000 0101 11s- = 0 33
				// 0 0000 0000 0101 10s- = 0 34
				// 0 0000 0000 0101 01s- = 0 35
				// 0 0000 0000 0101 00s- = 0 36
				// 0 0000 0000 0100 11s- = 0 37
				// 0 0000 0000 0100 10s- = 0 38
				// 0 0000 0000 0100 01s- = 0 39
				// 0 0000 0000 0100 00s- = 0 40
				*run = 0;
				*signed_level = 40 - ((bits & 0x3c) >> 2);
			}
		}
		else if (bits & 0x20)
		{
			size = 17;
			signed_bit = (bits & 0x1);
			if ((bits & 0x1e) > 0x14)
			{
				// 0 0000 0000 0011 111s = 27 1
				// 0 0000 0000 0011 110s = 28 1
				// 0 0000 0000 0011 101s = 29 1
				// 0 0000 0000 0011 100s = 30 1
				// 0 0000 0000 0011 011s = 31 1
				*run = 34 - ((bits & 0xe) >> 1);
				*signed_level = 1;
			}
			else if ((bits & 0x1e) > 0x8)
			{
				// 0 0000 0000 0011 010s = 11 2
				// 0 0000 0000 0011 001s = 12 2
				// 0 0000 0000 0011 000s = 13 2
				// 0 0000 0000 0010 111s = 14 2
				// 0 0000 0000 0010 110s = 15 2
				// 0 0000 0000 0010 101s = 16 2
				*run = 21 - ((bits & 0x1e) >> 1);
				*signed_level = 2;
			}
			else if ((bits & 0x1e) == 0x8)
			{
				// 0 0000 0000 0010 100s = 6 3
				*run = 6;
				*signed_level = 3;
			}
			else
			{
				// 0 0000 0000 0010 011s = 1 15
				// 0 0000 0000 0010 010s = 1 16
				// 0 0000 0000 0010 001s = 1 17
				// 0 0000 0000 0010 000s = 1 18
				*run = 1;
				*signed_level = 18 - ((bits & 0x6) >> 1);
			}
		}
		else
		{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
			__asm int 3		// huh???
#endif
			return false;
		}
	}

	if (signed_bit)
		*signed_level = -(*signed_level);

	discard_bits(size);

	return true;
}

bool discard_decode_DCT_coefficients(int n, uint intra_vlc_format, macroblock_type *mbt )
{
	if( intra_vlc_format == 0 || mbt->macroblock_intra == 0 )
	{
		bool end_of_block = false;
		uint size;
	
		uint bits;

		while (!end_of_block)
		{
			bits = peek_bits(17);

			// Table B-14 --- DCT coefficients Table zero
			if (bits & 0x10000)
			{
				if (n == 0)
				{
					size = 2;
					// 1 s--- ---- ---- ---- = 0 1 (NOTE 3) This code shall be used for the first (DC) coefficient in the block
				}
				else if (bits & 0x8000)
				{
					size = 3;
					// 1 1s-- ---- ---- ---- = 0 1 (NOTE 4) This code shall be used for all other coefficients
				}
				else
				{
					size = 2;
					// 1 0--- ---- ---- ---- = End of Block
					end_of_block = true;
				}
			}
			else if (bits & 0x8000)
			{
				if (bits & 0x4000)
				{
					size = 4;
					// 0 11s- ---- ---- ---- = 1 1
				}
				else
				{
					size = 5;
					// 0 101s ---- ---- ---- = 2 1
					// 0 100s ---- ---- ---- = 0 2
				}
			}
			else if (bits & 0x4000)
			{
				if (bits & 0x3000)
				{
					size = 6;
					// 0 0111 s--- ---- ---- = 3 1
					// 0 0110 s--- ---- ---- = 4 1
					// 0 0101 s--- ---- ---- = 0 3
				}
				else
				{
					size = 9;
					// 0 0100 111s ---- ---- = 10 1
					// 0 0100 110s ---- ---- = 0 5
					// 0 0100 101s ---- ---- = 1 3
					// 0 0100 100s ---- ---- = 3 2
					// 0 0100 011s ---- ---- = 11 1
					// 0 0100 010s ---- ---- = 12 1
					// 0 0100 001s ---- ---- = 0 6
					// 0 0100 000s ---- ---- = 13 1
				}
			}
			else if (bits & 0x2000)
			{
 				size = 7;
				// 0 0011 1s-- ---- ---- = 5 1
				// 0 0011 0s-- ---- ---- = 1 2
				// 0 0010 1s-- ---- ---- = 6 1
				// 0 0010 0s-- ---- ---- = 7 1
			}
			else if (bits & 0x1000)
			{
				size = 8;
				// 0 0001 11s- ---- ---- = 8 1
				// 0 0001 10s- ---- ---- = 0 4
				// 0 0001 01s- ---- ---- = 9 1
				// 0 0001 00s- ---- ---- = 2 2
			}
			else if (bits & 0x800)
			{
				size = 6 + 6 + 12;
				// 0 0000 1--- ---- ---- = Escape
			}
			else if (bits & 0x400)
			{
				size = 11;
				// 0 0000 0111 1s-- ---- = 4 2
				// 0 0000 0111 0s-- ---- = 14 1
				// 0 0000 0110 1s-- ---- = 15 1
				// 0 0000 0110 0s-- ---- = 1 4
				// 0 0000 0101 1s-- ---- = 2 3
				// 0 0000 0101 0s-- ---- = 0 7
				// 0 0000 0100 1s-- ---- = 5 2
				// 0 0000 0100 0s-- ---- = 16 1
			}
			else if (bits & 0x200)
			{
				size = 13;
				// 0 0000 0011 111s ---- = 17 1
				// 0 0000 0011 110s ---- = 6 2
				// 0 0000 0011 101s ---- = 0 8
				// 0 0000 0011 100s ---- = 3 3
				// 0 0000 0011 011s ---- = 1 5
				// 0 0000 0011 010s ---- = 18 1
				// 0 0000 0011 001s ---- = 19 1
				// 0 0000 0011 000s ---- = 0 9
				// 0 0000 0010 111s ---- = 20 1
				// 0 0000 0010 110s ---- = 21 1
				// 0 0000 0010 101s ---- = 7 2
				// 0 0000 0010 100s ---- = 2 4
				// 0 0000 0010 011s ---- = 0 10
				// 0 0000 0010 010s ---- = 4 3
				// 0 0000 0010 001s ---- = 8 2
				// 0 0000 0010 000s ---- = 0 11
			}
			else if (bits & 0x100)
			{
				size = 14;
				// 0 0000 0001 1111 s--- = 22 1
				// 0 0000 0001 1110 s--- = 23 1
				// 0 0000 0001 1101 s--- = 24 1
				// 0 0000 0001 1100 s--- = 25 1
				// 0 0000 0001 1011 s--- = 26 1
				// 0 0000 0001 1010 s--- = 0 12
				// 0 0000 0001 1001 s--- = 0 13
				// 0 0000 0001 1000 s--- = 0 14
				// 0 0000 0001 0111 s--- = 0 15
				// 0 0000 0001 0110 s--- = 1 6
				// 0 0000 0001 0101 s--- = 1 7
				// 0 0000 0001 0100 s--- = 2 5
				// 0 0000 0001 0011 s--- = 3 4
				// 0 0000 0001 0010 s--- = 5 3
				// 0 0000 0001 0001 s--- = 9 2
				// 0 0000 0001 0000 s--- = 10 2
			}
			else if (bits & 0x80)
			{
				size = 15;
				// 0 0000 0000 1111 1s-- = 0 16
				// 0 0000 0000 1111 0s-- = 0 17
				// 0 0000 0000 1110 1s-- = 0 18
				// 0 0000 0000 1110 0s-- = 0 19
				// 0 0000 0000 1101 1s-- = 0 20
				// 0 0000 0000 1101 0s-- = 0 21
				// 0 0000 0000 1100 1s-- = 0 22
				// 0 0000 0000 1100 0s-- = 0 23
				// 0 0000 0000 1011 1s-- = 0 24
				// 0 0000 0000 1011 0s-- = 0 25
				// 0 0000 0000 1010 1s-- = 0 26
				// 0 0000 0000 1010 0s-- = 0 27
				// 0 0000 0000 1001 1s-- = 0 28
				// 0 0000 0000 1001 0s-- = 0 29
				// 0 0000 0000 1000 1s-- = 0 30
				// 0 0000 0000 1000 0s-- = 0 31
			}
			else if (bits & 0x40)
			{
				size = 16;
				// 0 0000 0000 0111 11s- = 1 8
				// 0 0000 0000 0111 10s- = 1 9
				// 0 0000 0000 0111 01s- = 1 10
				// 0 0000 0000 0111 00s- = 1 11
				// 0 0000 0000 0110 11s- = 1 12
				// 0 0000 0000 0110 10s- = 1 13
				// 0 0000 0000 0110 01s- = 1 14
				// 0 0000 0000 0110 00s- = 0 32
				// 0 0000 0000 0101 11s- = 0 33
				// 0 0000 0000 0101 10s- = 0 34
				// 0 0000 0000 0101 01s- = 0 35
				// 0 0000 0000 0101 00s- = 0 36
				// 0 0000 0000 0100 11s- = 0 37
				// 0 0000 0000 0100 10s- = 0 38
				// 0 0000 0000 0100 01s- = 0 39
				// 0 0000 0000 0100 00s- = 0 40
			}
			else if (bits & 0x20)
			{
				size = 17;
				// 0 0000 0000 0011 111s = 27 1
				// 0 0000 0000 0011 110s = 28 1
				// 0 0000 0000 0011 101s = 29 1
				// 0 0000 0000 0011 100s = 30 1
				// 0 0000 0000 0011 011s = 31 1
				// 0 0000 0000 0011 010s = 11 2
				// 0 0000 0000 0011 001s = 12 2
				// 0 0000 0000 0011 000s = 13 2
				// 0 0000 0000 0010 111s = 14 2
				// 0 0000 0000 0010 110s = 15 2
				// 0 0000 0000 0010 101s = 16 2
				// 0 0000 0000 0010 100s = 6 3
				// 0 0000 0000 0010 011s = 1 15
				// 0 0000 0000 0010 010s = 1 16
				// 0 0000 0000 0010 001s = 1 17
				// 0 0000 0000 0010 000s = 1 18
			}
			else
			{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
				__asm int 3		// huh???
#endif
				return false;
			}

			discard_bits(size);
			n = 1;
		}
	}
	else // 
	{
		bool end_of_block = false;
		uint size;
		while (!end_of_block)
		{
			uint bits = peek_bits(17);

			// table B-15
			if (bits & 0x10000)
			{
				if (bits & 0x8000)
				{
					if (bits & 0x4000)
					{
						if (bits & 0x2000)
						{
							if (bits & 0x1000)
							{
								if (bits & 0xc00)
								{
									size = 9;
									// 1 1111 111s ---- ---- = 0 15
									// 1 1111 110s ---- ---- = 0 14
									// 1 1111 101s ---- ---- = 4 2
									// 1 1111 100s ---- ---- = 2 3
									// 1 1111 011s ---- ---- = 0 13
									// 1 1111 010s ---- ---- = 0 12
								}
								else
								{
									size = 8;
									// 1 1111 00s- ---- ---- = 0 9
								}
							}
							else
							{
								size = 8;
								// 1 1110 11s- ---- ---- = 0 8
								// 1 1110 10s- ---- ---- = 10 1
								// 1 1110 01s- ---- ---- = 1 3
								// 1 1110 00s- ---- ---- = 9 1
							}
						}
						else
						{
							size = 6;
							// 1 1101 s--- ---- ---- = 0 5
							// 1 1100 s--- ---- ---- = 0 4
						}
					}
					else
					{
						size = 4;
						// 1 10s- ---- ---- ---- = 0 2
					}
				}
				else
				{
					size = 3;
					// 1 0s-- ---- ---- ---- = 0 1
				}
			}
			else if (bits & 0x8000)
			{
				if (bits & 0x4000)
				{
					if (bits & 0x2000)
					{
						size = 5;
						// 0 111s ---- ---- ---- = 0 3
					}
					else
					{
						size = 4;
						// 0 110- ---- ---- ---- = End of Block
						end_of_block = true;
					}
				}
				else
				{
					size = 4;
					// 0 10s- ---- ---- ---- = 1 1
				}
			}
			else if (bits & 0x4000)
			{
				if (bits & 0x3000)
				{
					size = 6;
					// 0 0111 s--- ---- ---- = 3 1
					// 0 0110 s--- ---- ---- = 1 2
					// 0 0101 s--- ---- ---- = 2 1
				}
				else
				{
					size = 9;
					// 0 0100 111s ---- ---- = 1 4
					// 0 0100 110s ---- ---- = 3 2
					// 0 0100 101s ---- ---- = 12 1
					// 0 0100 100s ---- ---- = 13 1
					// 0 0100 011s ---- ---- = 0 10
					// 0 0100 010s ---- ---- = 0 11
					// 0 0100 001s ---- ---- = 11 1
					// 0 0100 000s ---- ---- = 1 5
				}
			}
			else if (bits & 0x2000)
			{
				size = 7;
				// 0 0011 1s-- ---- ---- = 5 1
				// 0 0011 0s-- ---- ---- = 4 1
				// 0 0010 1s-- ---- ---- = 0 6
				// 0 0010 0s-- ---- ---- = 0 7
			}
			else if (bits & 0x1000)
			{
				size = 8;
				// 0 0001 11s- ---- ---- = 2 2
				// 0 0001 10s- ---- ---- = 6 1
				// 0 0001 01s- ---- ---- = 8 1
				// 0 0001 00s- ---- ---- = 7 1
			}
			else if (bits & 0x800)
			{
				size = 6 + 6 + 12;
				// 0 0000 1--- ---- ---- = Escape
			}
			else if (bits & 0x400)
			{
				if (bits & 0x200)
				{
					if (bits & 0x100)
					{
						size = 10;
						// 0 0000 0111 s--- ---- = 15 1
					}
					else
					{
						size = 11;
						// 0 0000 0110 1s-- ---- = 16 1
						// 0 0000 0110 0s-- ---- = 2 4
					}
				}
				else
				{
					size = 10;
					// 0 0000 0101 s--- ---- = 14 1
					// 0 0000 0100 s--- ---- = 5 2
				}
			}
			else if (bits & 0x200)
			{
				size = 13;
				switch (bits & 0x1e0)
				{
				case 0x1e0:
					// 0 0000 0011 111s ---- = 17 1
				case 0x1c0:
					// 0 0000 0011 110s ---- = 6 2
				case 0x180:
					// 0 0000 0011 100s ---- = 3 3
				case 0x140:
					// 0 0000 0011 010s ---- = 18 1
				case 0x120:
					// 0 0000 0011 001s ---- = 19 1
				case 0x0e0:
					// 0 0000 0010 111s ---- = 20 1
				case 0x0c0:
					// 0 0000 0010 110s ---- = 21 1
				case 0x0a0:
					// 0 0000 0010 101s ---- = 7 2
				case 0x040:
					// 0 0000 0010 010s ---- = 4 3
				case 0x020:
					// 0 0000 0010 001s ---- = 8 2
					break;

				default:
					// 0 0000 0011 101x ---- = ??????
					// 0 0000 0011 011x ---- = ??????
					// 0 0000 0011 000x ---- = ??????
					// 0 0000 0010 100x ---- = ??????
					// 0 0000 0010 011x ---- = ??????
					// 0 0000 0010 000x ---- = ??????
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
					__asm int 3		// huh?
#endif
					return false;
				}
			}
			else if (bits & 0x100)
			{
				size = 14;
				switch (bits & 0xf0)
				{
				case 0xf0:
					// 0 0000 0001 1111 s--- = 22 1
				case 0xe0:
					// 0 0000 0001 1110 s--- = 23 1
				case 0xd0:
					// 0 0000 0001 1101 s--- = 24 1
				case 0xc0:
					// 0 0000 0001 1100 s--- = 25 1
				case 0xb0:
					// 0 0000 0001 1011 s--- = 26 1
				case 0x60:
					// 0 0000 0001 0110 s--- = 1 6
				case 0x50:
					// 0 0000 0001 0101 s--- = 1 7
				case 0x40:
					// 0 0000 0001 0100 s--- = 2 5
				case 0x30:
					// 0 0000 0001 0011 s--- = 3 4
				case 0x20:
					// 0 0000 0001 0010 s--- = 5 3
				case 0x10:
					// 0 0000 0001 0001 s--- = 9 2
				case 0x00:
					// 0 0000 0001 0000 s--- = 10 2
					break;

				default:
					// 0 0000 0001 1010 x--- = ??????
					// 0 0000 0001 100x x--- = ??????
					// 0 0000 0001 0111 x--- = ??????
	#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
					__asm int 3		// huh???
	#endif
					return false;
				}
			}
			else if (bits & 0x80)
			{
				size = 15;
				// 0 0000 0000 1111 1s-- = 0 16
				// 0 0000 0000 1111 0s-- = 0 17
				// 0 0000 0000 1110 1s-- = 0 18
				// 0 0000 0000 1110 0s-- = 0 19
				// 0 0000 0000 1101 1s-- = 0 20
				// 0 0000 0000 1101 0s-- = 0 21
				// 0 0000 0000 1100 1s-- = 0 22
				// 0 0000 0000 1100 0s-- = 0 23
				// 0 0000 0000 1011 1s-- = 0 24
				// 0 0000 0000 1011 0s-- = 0 25
				// 0 0000 0000 1010 1s-- = 0 26
				// 0 0000 0000 1010 0s-- = 0 27
				// 0 0000 0000 1001 1s-- = 0 28
				// 0 0000 0000 1001 0s-- = 0 29
				// 0 0000 0000 1000 1s-- = 0 30
				// 0 0000 0000 1000 0s-- = 0 31
			}
			else if (bits & 0x40)
			{
				size = 16;
				// 0 0000 0000 0111 11s- = 1 8
				// 0 0000 0000 0111 10s- = 1 9
				// 0 0000 0000 0111 01s- = 1 10
				// 0 0000 0000 0111 00s- = 1 11
				// 0 0000 0000 0110 11s- = 1 12
				// 0 0000 0000 0110 10s- = 1 13
				// 0 0000 0000 0110 01s- = 1 14
				// 0 0000 0000 0110 00s- = 0 32
				// 0 0000 0000 0101 11s- = 0 33
				// 0 0000 0000 0101 10s- = 0 34
				// 0 0000 0000 0101 01s- = 0 35
				// 0 0000 0000 0101 00s- = 0 36
				// 0 0000 0000 0100 11s- = 0 37
				// 0 0000 0000 0100 10s- = 0 38
				// 0 0000 0000 0100 01s- = 0 39
				// 0 0000 0000 0100 00s- = 0 40
			}
			else if (bits & 0x20)
			{
				size = 17;
				// 0 0000 0000 0011 111s = 27 1
				// 0 0000 0000 0011 110s = 28 1
				// 0 0000 0000 0011 101s = 29 1
				// 0 0000 0000 0011 100s = 30 1
				// 0 0000 0000 0011 011s = 31 1
				// 0 0000 0000 0011 010s = 11 2
				// 0 0000 0000 0011 001s = 12 2
				// 0 0000 0000 0011 000s = 13 2
				// 0 0000 0000 0010 111s = 14 2
				// 0 0000 0000 0010 110s = 15 2
				// 0 0000 0000 0010 101s = 16 2
				// 0 0000 0000 0010 100s = 6 3
				// 0 0000 0000 0010 011s = 1 15
				// 0 0000 0000 0010 010s = 1 16
				// 0 0000 0000 0010 001s = 1 17
				// 0 0000 0000 0010 000s = 1 18
			}
			else
			{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
				__asm int 3		// huh???
#endif
				return false;
			}

			discard_bits(size);
		}
	}

	return true;
}

// Table B-10 --- Variable length codes for motion_code
bool decode_motion_code(int *val)
{
	uint bits = peek_bits(11);
	uint size;
	uint signed_bit = 0;

	if (bits & 0x400)
	{
		size = 1;
		// 1xxx xxxx xxx 0
		*val = 0;
	}
	else if (bits & 0x200)
	{
		size = 3;
		signed_bit = (bits & 0x100);
// 011 xxxx xxxx -1
// 010 xxxx xxxx 1
		*val = 1;
	}
	else if (bits & 0x100)
	{
		size = 4;
		signed_bit = (bits & 0x80);
// 001 1xxx xxxx -2
// 001 0xxx xxxx 2
		*val = 2;
	}
	else if (bits & 0x80)
	{
		size = 5;
		signed_bit = (bits & 0x40);
// 000 11xx xxxx -3
// 000 10xx xxxx 3
		*val = 3;
	}
	else if (bits & 0x40)
	{
		if (bits & 0x20)
		{
			size = 7;
			signed_bit = (bits & 0x10);
// 000 0111 xxxx -4
// 000 0110 xxxx 4
			*val = 4;
		}
		else
		{
			size = 8;
			signed_bit = (bits & 0x8);
// 000 0101 1xxx -5
// 000 0101 0xxx 5
// 000 0100 1xxx -6
// 000 0100 0xxx 6
			*val = 6 - ((bits & 0x10) >> 4);
		}
	}
	else if (bits & 0x20)
	{
		if (bits & 0x10)
		{
			size = 8;
			signed_bit = (bits & 0x8);
// 000 0011 1xxx -7
// 000 0011 0xxx 7
			*val = 7;
		}
		else if (bits & 0xc)
		{
			size = 10;
			signed_bit = (bits & 0x2);
// 000 0010 111x -8
// 000 0010 110x 8
// 000 0010 101x -9
// 000 0010 100x 9
// 000 0010 011x -10
// 000 0010 010x 10
			*val = 11 - ((bits & 0xc) >> 2);
		} else
		{
			size = 11;
			signed_bit = (bits & 0x1);
// 000 0010 0011 -11
// 000 0010 0010 11
// 000 0010 0001 -12
// 000 0010 0000 12
			*val = 12 - ((bits & 0x2) >> 1);
		}
	}
	else if (bits & 0x10)
	{
		size = 11;
		signed_bit = (bits & 0x1);
// 000 0001 1111 -13
// 000 0001 1110 13
// 000 0001 1101 -14
// 000 0001 1100 14
// 000 0001 1011 -15
// 000 0001 1010 15
// 000 0001 1001 -16
// 000 0001 1000 16
		*val = 16 - ((bits & 0x6) >> 1);
	}
	else
	{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
		__asm int 3		// ??????
#endif
		return false;
	}

	if (signed_bit)
		*val = -(*val);

	discard_bits(size);

	return true;
}

bool motion_vectors(int s, picture_header *ph, macroblock_type *mbt)
{
	if (mbt->motion_vector_count == 1)
	{
		if ((mbt->mv_format == 0/*field*/) && (mbt->dmv != 1))
			mbt->motion_vertical_field_select[0][s] = get_bits(1);
		if (!motion_vector(0, s, ph, mbt))
			return false;
	}
	else
	{
		mbt->motion_vertical_field_select[0][s] = get_bits(1);
		if (!motion_vector(0, s, ph, mbt))
			return false;
		mbt->motion_vertical_field_select[1][s] = get_bits(1);
		if (!motion_vector(1, s, ph, mbt))
			return false;
	}
	return true;
}

bool motion_vector(int r, int s, picture_header *ph, macroblock_type *mbt)
{
	if (!decode_motion_code(&mbt->motion_code[r][s][0]))
		return false;

	if ((ph->f_code[s][0] != 1) && (mbt->motion_code[r][s][0] != 0))
	{
		uint r_size = ph->f_code[s][0] - 1;
		if ((r_size < 1) || (r_size > 8))
		{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
			__asm int 3		// ??????
#endif
			return false;
		}
		mbt->motion_residual[r][s][0] = get_bits(r_size);
	}

	if (mbt->dmv == 1)
	{
		// Table B-11
		uint bits = peek_bits(2);
		if (bits & 0x2)
		{
			if (bits & 0x1)
				mbt->dmvector[0] = -1;
			else
				mbt->dmvector[0] = 1;
			discard_bits(2);
		}
		else
		{
			mbt->dmvector[0] = 0;
			discard_bits(1);
		}
	}

	if (!decode_motion_code(&mbt->motion_code[r][s][1]))
		return false;

	if ((ph->f_code[s][1] != 1) && (mbt->motion_code[r][s][1] != 0))
	{
		uint r_size = ph->f_code[s][1] - 1;
		if ((r_size < 1) || (r_size > 8))
		{
#ifdef BREAK_ON_NAUGHTINESS_IN_SLICE
			__asm int 3		// ??????
#endif
			return false;
		}
		mbt->motion_residual[r][s][1] = get_bits(r_size);
	}

	if (mbt->dmv == 1)
	{
		// Table B-11
		uint bits = peek_bits(2);
		if (bits & 0x2)
		{
			if (bits & 0x1)
				mbt->dmvector[1] = -1;
			else
				mbt->dmvector[1] = 1;
			discard_bits(2);
		}
		else
		{
			mbt->dmvector[1] = 0;
			discard_bits(1);
		}
	}

	return true;
}

// scan backwards through files from given position to find a valid GOP/Iframe start
__int64 scan_for_Iframe_back(unsigned char *buffer, __int64 bufferSize, __int64 bufferStartOffset,
							 TSFileList *inputList, __int64 totalOffset, __int64 scanRange, int vidPid)
{
	__int64 currentOffset = totalOffset + 188;	// want to try to look at current packet
	__int64 bufferOffset = currentOffset - bufferStartOffset;
	if (bufferOffset > bufferSize)
	{
		bufferOffset = bufferSize;
		currentOffset = bufferOffset + bufferStartOffset;
	}

	TSFileInfo *ti = inputList->Find(currentOffset);
	if (ti)
	{
		// now scan backwards through buffer looking for an I-frame
		__int64 bufferStart = bufferOffset - scanRange;
		if (bufferStart < 0)
			bufferStart = 0;

		while (bufferStart <= bufferOffset - 188)
		{
			// confirm correct file
			if (currentOffset <= ti->m_totalBytesPrior)
				ti = ti->m_prev;
			if (!ti)
				break;

			// make sure this is aligned to this file's ts packets
			__int64 packetOffset = (currentOffset - ti->m_totalBytesPrior - ti->m_firstPacketOffset) % 188;
			currentOffset -= packetOffset;
			bufferOffset -= packetOffset;

			// now back up 1 packet
			currentOffset -= 188;
			bufferOffset -= 188;

			// make sure we're in the right file now
			while (ti && (currentOffset < ti->m_totalBytesPrior))
				ti = ti->m_prev;
			if (!ti)
				break;

			// confirm packet offset in case we backed up a file
			packetOffset = (currentOffset - ti->m_totalBytesPrior - ti->m_firstPacketOffset) % 188;
			if (packetOffset > 0)
			{
				// packet offset is wrong, must be a new differently aligned file, skip this packet
				continue;
			}

			// bail if we have overshot the buffer
			if (bufferOffset < bufferStart)
				break;

			unsigned char *buf = &buffer[bufferOffset];

			// Check sync byte
			if ((buf[0] == 0x47) || (buf[0] == 0x72) || (buf[0] == 0x29))
			{
				// Get pid
				int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;
				if (pid == vidPid)
				{
					// Get error
					bool errorBit = (buf[1] & 0x80) != 0;
					if (!errorBit)
					{
						// Get adaptation header info
						int adaptation = (buf[3] & 0x30) >> 4;
						int adaptLength = 0;
						if (adaptation == 0)
							adaptLength = -1;	// bad!
						else if (adaptation == 0x2)
							adaptLength = 184;
						else if (adaptation == 0x3)
							adaptLength = buf[4] + 1;
						if (adaptLength > 184)
							adaptLength = -1;	// too long!

						if (adaptLength >= 0)
						{
							// Get start code
							bool start = (buf[1] & 0x40) != 0;
							if (start && (currentOffset <= totalOffset))
							{
								// Check to see if this is an i_frame (group of picture start)
								__int64 r = scan_for_Iframe_forward(buffer, bufferSize, bufferStartOffset,
												inputList, currentOffset - 1, 188*100, vidPid);
								if (r == currentOffset)
								{
#ifdef _DEBUG
static int maxoffset = 0;
if (maxoffset < int(totalOffset - currentOffset))
{
	maxoffset = int(totalOffset - currentOffset);
	_Error("Max RW: %d\n", maxoffset);
}
#endif
									return currentOffset;
								}
							}
						}
					}
				}
			}
		}
	}

	return -1;
}
#if 0
__int64 scan_for_Iframe_back(unsigned char *buffer, __int64 bufferSize, __int64 bufferStartOffset,
							 TSFileList *inputList, __int64 totalOffset, __int64 scanRange, int vidPid)
{
	uchar b[1024];
	int bSize = 1024;
	int bUsed = 0;

	// identifying an I-frame sometimes requires 2-3 packets ahead, in our requested vidPid stream!
	// so we start ahead a good bit to make sure we've read in enough packets by the time we get
	// back down to the requested starting offset.
	__int64 currentOffset = totalOffset + 32768;	// ~200ish packets should be enough
	__int64 bufferOffset = currentOffset - bufferStartOffset;
	if (bufferOffset > bufferSize)
	{
		bufferOffset = bufferSize;
		currentOffset = bufferOffset + bufferStartOffset;
	}

	TSFileInfo *ti = inputList->Find(currentOffset);
	if (ti)
	{
		// now scan backwards through buffer looking for an I-frame
		__int64 bufferStart = bufferOffset - scanRange;
		if (bufferStart < 0)
			bufferStart = 0;

		while (bufferStart <= bufferOffset - 188)
		{
			// confirm correct file
			if (currentOffset <= ti->m_totalBytesPrior)
				ti = ti->m_prev;
			if (!ti)
				break;

			// make sure this is aligned to this file's ts packets
			__int64 packetOffset = (currentOffset - ti->m_totalBytesPrior - ti->m_firstPacketOffset) % 188;
			currentOffset -= packetOffset;
			bufferOffset -= packetOffset;

			// now back up 1 packet
			currentOffset -= 188;
			bufferOffset -= 188;

			// make sure we're in the right file now
			while (ti && (currentOffset < ti->m_totalBytesPrior))
				ti = ti->m_prev;
			if (!ti)
				break;

			// confirm packet offset in case we backed up a file
			packetOffset = (currentOffset - ti->m_totalBytesPrior - ti->m_firstPacketOffset) % 188;
			if (packetOffset > 0)
			{
				// packet offset is wrong, must be a new differently aligned file, skip this packet
				continue;
			}

			// bail if we have overshot the buffer
			if (bufferOffset < bufferStart)
				break;

			unsigned char *buf = &buffer[bufferOffset];

			// Check sync byte
			if ((buf[0] == 0x47) || (buf[0] == 0x72) || (buf[0] == 0x29))
			{
				// Get pid
				int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;
				if (pid == vidPid)
				{
					// Get error
					bool errorBit = (buf[1] & 0x80) != 0;
					if (errorBit)
					{
						// invalidate accumulated content of this stream
						bUsed = 0;
					}
					else
					{
						// Get adaptation header info
						int adaptation = (buf[3] & 0x30) >> 4;
						int adaptLength = 0;
						if (adaptation == 0)
							adaptLength = -1;	// bad!
						else if (adaptation == 0x2)
							adaptLength = 184;
						else if (adaptation == 0x3)
							adaptLength = buf[4] + 1;
						if (adaptLength > 184)
							adaptLength = -1;	// too long!

						if (adaptLength >= 0)
						{
							// Add the payload for this packet to the start of the current buffer
							int payloadLength = 184 - adaptLength;
							if (payloadLength > 0)
							{
								int oldSize = min(bUsed, bSize - payloadLength);
								if (oldSize > 0)
									memmove(&b[payloadLength], b, oldSize);
								memcpy(b, buf + 4 + adaptLength, payloadLength);
								bUsed = oldSize + payloadLength;
							}

							// Get start code
							bool start = (buf[1] & 0x40) != 0;
							if (start && (currentOffset <= totalOffset))
							{
								// Check to see if this is an i_frame (group of picture start)
								bool doingIFrame = false;
								unsigned int strid = 0;
								for (int i = 0; (i < bSize) && !doingIFrame; i++)
								{
									strid = (strid << 8) | b[i];
									if (strid == 0x000001B8) // group_start_code
									{
										// found a Group of Pictures header, subsequent picture must be an I-frame
										doingIFrame = true;
									}
									else if (strid == 0x00000100) // picture_start_code
									{
										// picture_header, let's see if it's an I-frame
										if (i < bSize-2)
										{
											// check if picture_coding_type == 1
											if ((b[i+2] & (0x7 << 3)) == (1 << 3))
											{
												// found an I-frame picture
												doingIFrame = true;
											}
										}
									}
								}

								if (doingIFrame)
								{
#ifdef _DEBUG
static int maxoffset = 0;
if (maxoffset < int(totalOffset - currentOffset))
{
	maxoffset = int(totalOffset - currentOffset);
	_Error("Max RW: %d\n", maxoffset);
}
#endif
									return currentOffset;
								}
							}
						}
					}
				}
			}
		}
	}

	return -1;
}
#endif

__int64 scan_for_Iframe_forward(unsigned char *buffer, __int64 bufferSize, __int64 bufferStartOffset,
								TSFileList *inputList, __int64 totalOffset, __int64 scanRange, int vidPid)
{
	uchar b[1024];
	int bSize = 1024;
	__int64 bStartOffset = -1;
	int bPos = 0;
	bool bReading = false;

	__int64 currentOffset = totalOffset + 1;	// totalOffset is probably already an I-frame, look for *next* one
	TSFileInfo *ti = inputList->Find(currentOffset);
	if (ti)
	{
		// now scan forwards through buffer looking for an I-frame
		__int64 bufferOffset = currentOffset - bufferStartOffset;
		__int64 bufferEnd = bufferOffset + scanRange;
		if (bufferEnd > bufferSize)
			bufferEnd = bufferSize;

		while (bufferOffset <= bufferEnd - 188)
		{
			// confirm correct file
			if (currentOffset >= ti->m_totalBytesPrior + ti->m_fileSize)
				ti = ti->m_next;
			if (!ti)
				break;

			// make sure this is aligned to this file's ts packets
			__int64 packetOffset = (currentOffset - ti->m_totalBytesPrior - ti->m_firstPacketOffset) % 188;
			if (packetOffset > 0)
			{
				currentOffset += 188 - packetOffset;
				bufferOffset += 188 - packetOffset;
			}

			// now advance 1 packet
			__int64 futureCurrentOffset = currentOffset + 188;
			__int64 futureBufferOffset = bufferOffset + 188;

			// make sure we're in the right file now
			while (ti && (futureCurrentOffset > (ti->m_totalBytesPrior + ti->m_fileSize)))
				ti = ti->m_next;
			if (!ti)
				break;

			// confirm packet offset in case we advanced a file
			packetOffset = (futureCurrentOffset - ti->m_totalBytesPrior - ti->m_firstPacketOffset) % 188;
			if (packetOffset > 0)
			{
				// packet offset is wrong, must be a new differently aligned file, skip this packet
				currentOffset = futureCurrentOffset;
				bufferOffset = futureBufferOffset;
				continue;
			}

			// bail if we have overshot the buffer
			if (futureBufferOffset > bufferEnd)
				break;

			unsigned char *buf = &buffer[bufferOffset];

			// Check sync byte
			if ((buf[0] == 0x47) || (buf[0] == 0x72) || (buf[0] == 0x29))
			{
				// Get pid
				int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;
				if (pid == vidPid)
				{
					// Get error
					bool errorBit = (buf[1] & 0x80) != 0;
					if (errorBit)
					{
						// invalidate accumulated content of this stream
						bReading = false;
					}
					else
					{
						// Get adaptation header info
						int adaptation = (buf[3] & 0x30) >> 4;
						int adaptLength = 0;

						// Get adaptation header size
						if (adaptation == 0)
						{
							adaptLength = -1;	// bad!
						}
						else if (adaptation == 0x2)
							adaptLength = 184;
						else if (adaptation == 0x3)
						{
							adaptLength = buf[4] + 1;
							if (adaptLength > 184)
							{
								adaptLength = -1;	// too long!
							}
						}

						if (adaptLength >= 0)
						{
							// Get start code
							bool start = (buf[1] & 0x40) != 0;
							if (start)
							{
								// begin storing this PES packet
								bStartOffset = currentOffset;
								bPos = 0;
								bReading = true;
							}
							if (bReading)
							{
								// Add the payload for this packet to the current buffer
								int payloadLength = 184 - adaptLength;
								if (payloadLength > 0)
								{
									if ((bPos + payloadLength) <= bSize)
									{
										memcpy(b + bPos, buf + 4 + adaptLength, payloadLength);
										bPos += payloadLength;
									}
									else
									{
										// buffer overflow!
										bReading = false;
									}
								}

								if (bPos > 512)
								{
									// Check to see if this is an i_frame (group of picture start)
									bool doingIFrame = false;
									unsigned int strid = 0;
									for (int i = 0; (i < bPos) && !doingIFrame; i++)
									{
										strid = (strid << 8) | b[i];
										if (strid == 0x000001B8) // group_start_code
										{
											// found a Group of Pictures header, subsequent picture must be an I-frame
											doingIFrame = true;
										}
										else if (strid == 0x00000100) // picture_start_code
										{
											// picture_header, let's see if it's an I-frame
											if (i < bPos-2)
											{
												// check if picture_coding_type == 1
												if ((b[i+2] & (0x7 << 3)) == (1 << 3))
												{
													// found an I-frame picture
													doingIFrame = true;
												}
											}
										}
									}

									if (doingIFrame)
									{
#ifdef _DEBUG
static int maxoffset = 0;
if (maxoffset < int(currentOffset - totalOffset))
{
	maxoffset = int(currentOffset - totalOffset);
	_Error("Max FF: %d\n", maxoffset);
}
#endif
										return bStartOffset;
									}
									bReading = false;
								}
							}
						}
					}
				}
			}

			currentOffset = futureCurrentOffset;
			bufferOffset = futureBufferOffset;
		}
	}

	return -1;
}


char *HiDTVHeader = "HiDiTV List File.\x0a""\0""\0""\0""\0""\0""\0""\0""\0";
int HiDTVHeaderLength = 26;

int convert_files_to_ts( TSFileList *inputList,
						 char *outputFilePattern,
						 Filetype fileType,
						 int vidPid,
						 int audPid,
						 int maxFileSize,
						 FileReadBuffer *readBuffer,
						 convert_callback callback,
						 void *data, HWND parentWnd,
						 bool decodeIFrames,
						 CPtrArray *pPidInfoList,
						 CWordArray *pHiddenPidDeleteList,
						 bool bCreateClipFiles,
						 bool bPreserveBitRate,
						 int vertical_size_value, 
						 bool bSetVerticalSizeTo1080 )
{
	int rVal = 0;
	char errMsgBuf[256];
	int errMsgVal;
	long timeDiscrepancy = 0;

	__int64 filePos = 0;
	__int64 totalRead = 0;
	__int64 totalOut = 0;
	__int64 maxFileSize64 = 0;
//	FILE64 *fIn = NULL;
	bool needToSeek = true;
	int counter = 0;
	__int64 TSOffset = 0;
	__int64 TSPerFrame = 0;
	__int64 PTSLast = 0;
	bool calcNewTSOffset = false;

	TSFileInfo *fileInfo = inputList->GetHead();
	if (!fileInfo)
		return 1;

	TSRegionInfo *regionInfo = fileInfo->m_head;
	RegionType regionType = rt_include;

	if( bCreateClipFiles )
		maxFileSize64 = 0;
	else
		maxFileSize64 = (__int64)maxFileSize * (__int64)(1024 * 1024);

	// Open the destination file
	char outputFileName[256];
	strcpy(outputFileName, outputFilePattern);
	if (fileType == ft_HiDTV)
	{
		// remove .tls extension, and append _000.tp
		char *t = outputFileName + strlen(outputFileName) - 4;
//		if (!stricmp(t, ".tls"))
		if( !_stricmp(t, ".tls" ))
		{
			*t = '\0';
		}
		strcat(outputFileName, "_000.tp");
	}

	bool overwrite = false;
	FILE64 *fOut = fopen64(outputFileName, "wb", WRBUFSIZE, overwrite);
	while (!fOut)
	{
		char *s = "Couldn't open output file\n%s";
		if (ferror64() == EEXIST)
			s = "File already exists!\n%s\nRetry and overwrite anyway?";
		sprintf(errMsgBuf, s, outputFileName);
//		long tmp1, tmp2;
		time_t tmp1, tmp2;
		time( &tmp1 );
		errMsgVal = ::MessageBox(parentWnd, errMsgBuf, "Warning!", MB_RETRYCANCEL | MB_ICONWARNING);
		time( &tmp2 );
		timeDiscrepancy += tmp2 - tmp1;
		if (errMsgVal == IDRETRY)
		{
			overwrite = (ferror64() == EEXIST);
			fOut = fopen64(outputFileName, "wb", WRBUFSIZE, overwrite);
		}
		else
			return 1;
	}

	FILE64 *fList = NULL;
	if (fileType == ft_HiDTV)
	{
		// create list file
		overwrite = false;
		fList = fopen64(outputFilePattern, "wb", WRBUFSIZE, overwrite);
		while (!fList)
		{
			char *s = "Couldn't open output list file\n%s";
			if (ferror64() == EEXIST)
				s = "List file already exists\n%s\nRetry and overwrite anyway?";
			sprintf(errMsgBuf, s, outputFilePattern);
			long tmp1, tmp2;
			time((time_t*)&tmp1);
			errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_RETRYCANCEL | MB_ICONWARNING);
			time((time_t*)&tmp2);
			timeDiscrepancy += tmp2 - tmp1;
			if (errMsgVal == IDRETRY)
			{
				overwrite = (ferror64() == EEXIST);
				fList = fopen64(outputFilePattern, "wb", WRBUFSIZE, overwrite);
			}
			else
			{
				if (fOut)
					fclose64(fOut);
				return 1;
			}
		}

		if ((fwrite64(HiDTVHeader, HiDTVHeaderLength, 1, fList) != 1) ||
			(fwrite64(outputFileName, strlen(outputFileName), 1, fList) != 1) ||
			(fwrite64("||", 2, 1, fList) != 1))
		{
			sprintf(errMsgBuf, "Couldn't write to output list file\n%s", outputFilePattern);
			errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
			if (fOut)
				fclose64(fOut);
			if (fList)
				fclose64(fList);
			return 1;
		}
	}

	// stream info
/*
	PIDinfoGroup PIDs = { 2,
		{	{ vidPid, 0, STREAM_PES, PID_MPEG2_VIDEO, false, NULL, 0, 0, -1, false, false },
			{ audPid, 0, STREAM_PES, PID_AC3_AUDIO, false, NULL, 0, 0, -1, false, false } } } ;
*/
// Note: currently the only use for this is to filter video for the preview display so we don't
//       need to filter audio pids
	PIDinfoGroup PIDs = { 1,
		{	{ vidPid, 0, STREAM_PES, PID_MPEG2_VIDEO, false, NULL, 0, 0, -1, false, false }	} };


	PIDs.pids[0].buffer = new uchar[VIDEO_BUFFER_SIZE];
	PIDs.pids[0].bufferSize = VIDEO_BUFFER_SIZE;
//	PIDs.pids[1].buffer = new uchar[AUDIO_BUFFER_SIZE];
//	PIDs.pids[1].bufferSize = AUDIO_BUFFER_SIZE;

//	if (!PIDs.pids[0].buffer || !PIDs.pids[1].buffer)
	if (!PIDs.pids[0].buffer )
	{
		if (PIDs.pids[0].buffer)
			delete []PIDs.pids[0].buffer;
//		if (PIDs.pids[1].buffer)
//			delete []PIDs.pids[1].buffer;
		if (fOut)
			fclose64(fOut);
		if (fList)
			fclose64(fList);

		::MessageBox(parentWnd, "Out of memory", NULL, MB_OK | MB_ICONERROR);
		return 1;
	}

	unsigned char buf[188];
	int bytesReadInPacket = 0;

	short pid = -1;
	int adaptation_field_length = 0;
	bool newPAT = false;	// set when the first PAT is found and processed
	bool newPMT = false;	// set when the first PMT is found and processed
	bool newVCT = false;	// set when the first VCT is found and processed

	// WHY is this defined here instead of in CreateNewPMT()???
	// This is too simple of a way to do this.  If multiple audio channels
	//  are keep as well as multiple channels, the audio PIDs will probably
	//  get messed up.  Fortunatly, most people only save one channel, the
	//  HD channel.
	int audioCount = 0;		// keep track of the number of audio sub-channels

	// create a buffer to hold our VCT data. the amount of data may
	//  exceed 1 packet, but probably not more than two. this buffer
	//  is big enough to hold around 5 packets.
	unsigned char *vctBuf = new unsigned char[1024]; 
	unsigned char *b2 = vctBuf;  
	bool bAddToBuffer = false;  

	// hopefully a faster way to walk through the pid information without having
	//  to use the CPtrArray object pPidInfoList since this is done A LOT!
	short numOfPids = pPidInfoList->GetCount() + pHiddenPidDeleteList->GetCount();
	Pid_Info *pPidArray = new Pid_Info[ numOfPids ];
	Pid_Info *pPidArrayStart = pPidArray;

	// copy the data from the pPidInfoList to the pid info array
	Pid_Info *pPidInfo;
	for( int i = 0; i < pPidInfoList->GetCount(); ++i )
	{
		pPidInfo = (Pid_Info*)pPidInfoList->GetAt( i );
		memcpy( pPidArray, pPidInfo, sizeof( Pid_Info ) );
		*pPidArray++;
	}

	pPidInfo = new Pid_Info;
	// copy the data from the pHiddenDeleteList to the pid info array
	for( int i = 0; i < pHiddenPidDeleteList->GetCount(); i++ )
	{
		int pid = pHiddenPidDeleteList->GetAt( i );

		pPidInfo->new_pid = -1;
		pPidInfo->original_pid = pid;
		pPidInfo->processed = false;
		pPidInfo->strip_pid = true;
		pPidInfo->type = PID_UNKNOWN;
		pPidInfo->continuity_counter = -1;

		memcpy( pPidArray, pPidInfo, sizeof( Pid_Info ) );

		*pPidArray++;
	}
	delete pPidInfo;
	pPidArray = pPidArrayStart;

	// since the PAT and PMT(s) need to be processed BEFORE the video and audio
	//  pids can be processed, lets pre-process the PAT and PMT(s) so that remapping
	//  of the video and audio pids can begin on the first packets encountered
	//  in the transport stream. If we don't do this, then the first video and
	//  audio packets will likely be stripped because that is usually the first PIDs
	//  encountered after editing before a PAT or PMT is encountered.

	// make copies of our original PAT and PMT's to be modified so that on subsequent runs 
	//   of H2 without quiting or using "Clear All" the PIDs will be remapped properly again.
	PatPmt_Packets copyPatPmtPackets;
	copyPatPmtPackets.numOfPMTPackets = 0;

	copyPatPmtPackets.pPAT_Packet = new uchar[188];
	memcpy( copyPatPmtPackets.pPAT_Packet, PatPmtPackets.pPAT_Packet, 188 );

	for( int i = 0; i < PatPmtPackets.numOfPMTPackets; i++ )
		copyPatPmtPackets.pPMT_Packet[i] = new uchar[188];

	for( int i = 0; i < PatPmtPackets.numOfPMTPackets; ++i )
	{
		memcpy( copyPatPmtPackets.pPMT_Packet[i], PatPmtPackets.pPMT_Packet[i], 188 );
		++copyPatPmtPackets.numOfPMTPackets;
	}
	
	// write out 3 PATs and 2 PMTs. A small number of ts files don't have enough PAT and PMT data
	//  at the beginning of the file after editing with H2 to display the channel information.
	//  Having 3 PATs and 2 PMTs at the beginning of the file assures that H2 will be able
	//  to at least display NONAME for the channel info instead it being dimmed, which is bad.

	// PAT #1
	CreateNewPAT( copyPatPmtPackets.pPAT_Packet, pPidArray, numOfPids, pPidArrayStart );
	fwrite64( copyPatPmtPackets.pPAT_Packet, 188, 1, fOut );

	// PAT #2
	memcpy( copyPatPmtPackets.pPAT_Packet, PatPmtPackets.pPAT_Packet, 188 );
	CreateNewPAT( copyPatPmtPackets.pPAT_Packet, pPidArray, numOfPids, pPidArrayStart );
	fwrite64( copyPatPmtPackets.pPAT_Packet, 188, 1, fOut );

	// PAT #3
	memcpy( copyPatPmtPackets.pPAT_Packet, PatPmtPackets.pPAT_Packet, 188 );
	CreateNewPAT( copyPatPmtPackets.pPAT_Packet, pPidArray, numOfPids, pPidArrayStart );
	fwrite64( copyPatPmtPackets.pPAT_Packet, 188, 1, fOut );

	for( int i = 0; i < copyPatPmtPackets.numOfPMTPackets; i++ )
	{
		int pmt_pid = *(copyPatPmtPackets.pPMT_Packet[i] + 1) & 0x1f;
		pmt_pid <<= 8;
		pmt_pid |= *(copyPatPmtPackets.pPMT_Packet[i] + 2);

		for( int j = 0; j < numOfPids; j++ )
		{
			if( pPidArray->type == PID_PMT )
			{
				if( pPidArray->original_pid == pmt_pid && pPidArray->strip_pid != true )
				{
					// PMT #1
					CreateNewPMT( pmt_pid, copyPatPmtPackets.pPMT_Packet[i], pPidArray, numOfPids, pPidArrayStart, audioCount );
					fwrite64( copyPatPmtPackets.pPMT_Packet[i], 188, 1, fOut );

					// PMT #2
					memcpy( copyPatPmtPackets.pPMT_Packet[i], PatPmtPackets.pPMT_Packet[i], 188 );
					CreateNewPMT( pmt_pid, copyPatPmtPackets.pPMT_Packet[i], pPidArray, numOfPids, pPidArrayStart, audioCount );
					fwrite64( copyPatPmtPackets.pPMT_Packet[i], 188, 1, fOut );

					break;
				}
			}
			*pPidArray++;
		}
		pPidArray = pPidArrayStart;
	}
	
	newPAT = true;
	newPMT = true; 


	for (;;)
	{
		counter++;

		bool newFile = false;

		// Call callback
		if (((counter & 0x3f) == 0) && callback && fileInfo)
		{
			bool cont = callback(fileInfo->m_fileName, outputFileName,
					fileInfo->m_totalBytesPrior + filePos,
					totalRead, inputList->TotalIncludedBytes(),
					totalOut, timeDiscrepancy, data);
			timeDiscrepancy = 0;

			// Cancelled?
			if (!cont)
				break;
		}

		// check regions, skip data, open new files
		if (regionInfo && (filePos >= regionInfo->m_offset))
		{
			if ((totalOut > 0) &&
				(regionType == rt_exclude) && (regionInfo->m_type == rt_include))
			{
				// about to come back from an exclude!
				// gotta set flags to compensate for TS discontinuities

//				calcNewTSOffset = true;

				if( bCreateClipFiles )
						newFile = true;
			}


			regionType = regionInfo->m_type;
			regionInfo = regionInfo->m_next;

			if( regionType == rt_cut )
				newFile = true;
		}
		
		if (regionType == rt_exclude)
		{
			if (regionInfo)
			{
				filePos = regionInfo->m_offset;
			}
			else
			{
				if (readBuffer->IsFileOpen())
//				if (fIn)
				{
					readBuffer->Close();
//					fclose64(fIn);
//					fIn = NULL;
				}
				fileInfo = fileInfo->m_next;
				if (!fileInfo)
					break;	// we're complete!
				regionInfo = fileInfo->m_head;
				filePos = 0;
			}
			needToSeek = true;
			continue;
		}
		else
		{
			if (!readBuffer->IsFileOpen())
//			if (!fIn)
			{
				readBuffer->Open(fileInfo->m_fileName);
//				fIn = fopen64(fileInfo->m_fileName, "rb", RDBUFSIZE);
				while (!readBuffer->IsFileOpen())
//				while (!fIn)
				{
					sprintf(errMsgBuf, "Couldn't open input file\n%s", fileInfo->m_fileName);
					long tmp1, tmp2;
					time((time_t*)&tmp1);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_RETRYCANCEL | MB_ICONWARNING);
					time((time_t*)&tmp2);
					timeDiscrepancy += tmp2 - tmp1;
					if (errMsgVal == IDRETRY)
					{
						readBuffer->Open(fileInfo->m_fileName);
//						fIn = fopen64(fileInfo->m_fileName, "rb", RDBUFSIZE);
					}
					else
					{
						rVal = 1;
						break;	// where is this file??
					}
				}
				if (rVal)
					break;
				needToSeek = true;
			}
			if (needToSeek)
			{
				// check packet alignment of new file
				__int64 offset = (188 - ((filePos + 188 - fileInfo->m_firstPacketOffset) % 188)) % 188;
				if (offset != ((188 - bytesReadInPacket) % 188))
				{
					// alignment does not match, discard partial packet
					bytesReadInPacket = 0;
					filePos += offset;
				}
				readBuffer->Seek(filePos);
//				fseek64(fIn, filePos, SEEK_SET);
				needToSeek = false;
			}
		}

		// Try to read packet..
		int bytesRead;
		if ((bytesRead = readBuffer->Read(buf + bytesReadInPacket, 188 - bytesReadInPacket)) != 188 - bytesReadInPacket)
//		if ((bytesRead = fread64(buf+bytesReadInPacket, 1, 188 - bytesReadInPacket, fIn)) != 188 - bytesReadInPacket)
		{
			if (bytesRead < 0)
				bytesRead = 0;
			bytesReadInPacket += bytesRead;
			totalRead += bytesRead;

			// Close this file
			readBuffer->Close();
//			fclose64(fIn);
//			fIn = NULL;
			fileInfo = fileInfo->m_next;
			if (!fileInfo)
				break;	// we're complete!
			regionInfo = fileInfo->m_head;
			filePos = 0;

			continue;
		}
		else
		{
			filePos += bytesRead;
			bytesReadInPacket = 0;
			totalRead += bytesRead;
		}

		if( !bCreateClipFiles )
			if( ftell64(fOut) + 188 >= maxFileSize64 )
				newFile = true;

		// Check output for max filesize
//		if (ftell64(fOut) + 188 >= maxFileSize64)
		if( newFile )
		{
			// Close files
			fclose64(fOut);
			fOut = NULL;

			// Try next file?
			if (inc_file_counter(outputFileName))
			{
				// Open the destination file
				overwrite = false;
				fOut = fopen64(outputFileName, "wb", WRBUFSIZE, overwrite);
				while (!fOut)
				{
					char *s = "Couldn't open output file\n%s";
					if (ferror64() == EEXIST)
						s = "File already exists\n%s\nRetry and overwrite anyway?";
					sprintf(errMsgBuf, s, outputFileName);
					long tmp1, tmp2;
					time((time_t*)&tmp1);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_RETRYCANCEL | MB_ICONWARNING);
					time((time_t*)&tmp2);
					timeDiscrepancy += tmp2 - tmp1;
					if (errMsgVal == IDRETRY)
					{
						overwrite = (ferror64() == EEXIST);
						fOut = fopen64(outputFileName, "wb", WRBUFSIZE, overwrite);
					}
					else
					{
						rVal = 1;
						break;
					}
				}
				if (rVal)
					break;

				if (fList &&
					((fwrite64(outputFileName, strlen(outputFileName), 1, fList) != 1) ||
					(fwrite64("||", 2, 1, fList) != 1)))
				{
					sprintf(errMsgBuf, "Couldn't write to output list file\n%s", outputFilePattern);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
					rVal = 1;
					break;
				}
			}
			else
				break;
		}

		// Convert sync byte if converting from an AccessDTV file
		if ((buf[0] == 0x72) || (buf[0] == 0x29))
		{
			buf[0] = 0x47;
		}

		// exclude this packet?
		bool exclude = false;

		// Check sync byte
		if( buf[0] == 0x47 )
		{
			// Get pid
			pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;

			for( int i = 0; i < numOfPids; i++  )
			{
				if( pid == pPidArray->original_pid )
				{
					if( pPidArray->strip_pid  )
					{
						exclude = true;
						break;
					}
					else
						break;
				}
				*pPidArray++;
			}

			pPidArray = pPidArrayStart;

			if( exclude && !bPreserveBitRate )
				continue;

			// need to get the adaptation_field_length for pids that are not
			//  video and audio pids
			int adaptation = ( buf[3] & 0x30 ) >> 4;
			int adaptLength = 0;
			if( adaptation == 0 )
				adaptLength = -1;	// bad!
			else if( adaptation == 0x2 )
				adaptLength = 184;
			else if( adaptation == 0x3 )
				adaptLength = buf[4] + 1;
			if( adaptLength > 184 )
			adaptLength = -1;	// too long!

			adaptation_field_length = adaptLength;

			PIDinfo *p = find_PID_no_new(&PIDs, pid);

			if (p) // only looking for video and audio pids here
			{
				p->count++;

				// Get error
				bool errorBit = (buf[1] & 0x80) != 0;
				if( errorBit )
				{
					// invalidate accumulated content of this stream
					p->reading = false;
				}
				else
				{
					// Get adaptation header info
					int adaptation = (buf[3] & 0x30) >> 4;
					int adaptLength = 0;
					if (adaptation == 0)
						adaptLength = -1;	// bad!
					else if (adaptation == 0x2)
						adaptLength = 184;
					else if (adaptation == 0x3)
						adaptLength = buf[4] + 1;
					if (adaptLength > 184)
						adaptLength = -1;	// too long!

					adaptation_field_length = adaptLength;
					if( adaptLength >= 0 )
					{
						// Get payload start indicator
						bool start = ( buf[1] & 0x40 ) != 0;

						// process content of packet assuming is a valid PES stream
						if( start )
						{
							if( p->reading && ( p->bufferPos > 0 ) )
							{
								if(( p->pType == PID_MPEG2_VIDEO || p->pType == PID_USER_PRIVATE ) && decodeIFrames )
									process_PES_video_packet( p->buffer, p->bufferPos, true );
							}
							// start a new PES packet
							if( p->buffer && !p->processed )
							{
								p->reading = true;
								p->bufferPos = 0;
								p->confirmedRead = false;
/*
								// check PTS/DTS stuff here
								uchar *b = buf + 4 + adaptLength;
								if(( b[0] == 0 ) && ( b[1] == 0 ) && ( b[2] == 1 ))
								{
									// PES packet header
									__int64 PTS = -1;
									__int64 DTS = -1;
									if (b[7] & 0x80)
									{
										// extract PTS
										PTS = (__int64)( b[9] & 0x0e ) << 29;
										PTS |= (__int64)b[10] << 22;
										PTS |= (__int64)( b[11] & 0xfe ) << 14;
										PTS |= (__int64)b[12] << 7;
										PTS |= (__int64)( b[13] & 0xfe ) >> 1;
// NOTE: this code never gets excecuted, calcNewTSOffset is always false. the code that sets it to true
//        is commented out above.
										if( calcNewTSOffset && ( p->pType == PID_MPEG2_VIDEO ))
										{
//_Error("CUT\n");
											// (PTS + TSOffset)&((1<<33)-1) = PTSLast + TSPerFrame
											TSOffset = PTSLast + TSPerFrame*4 - PTS;
											if( TSOffset < 0 )
												TSOffset += 1 << 33;
											calcNewTSOffset = false;
										}

										if( TSOffset != 0 )
										{
											// new PTS with offset
											PTS += TSOffset;

											// apply new PTS back into buffer so it will get written out to disk
											b[9] = ( b[9] & ~0x0e ) | (uchar)(( PTS >> 29 ) & 0x0e );
											b[10] = (uchar)(( PTS >> 22 ) & 0xff );
											b[11] = ( b[11] & ~0xfe ) | (uchar)(( PTS >> 14 ) & 0xfe );
											b[12] = (uchar)(( PTS >> 7 ) & 0xff );
											b[13] = ( b[13] & ~0xfe ) | (uchar)(( PTS << 1 ) & 0xfe );
										}
									}
									if( b[7] & 0x40 )
									{
										// extract DTS
										DTS = (__int64)( b[14] & 0x0e ) << 29;
										DTS |= (__int64)b[15] << 22;
										DTS |= (__int64)( b[16] & 0xfe ) << 14;
										DTS |= (__int64)b[17] << 7;
										DTS |= (__int64)( b[18] & 0xfe ) >> 1;

										if( TSOffset != 0 )
										{
											// new DTS with offset
											DTS += TSOffset;

											// apply new DTS back into buffer so it will get written out to disk
											b[9] = ( b[14] & ~0x0e ) | (uchar)(( DTS >> 29 ) & 0x0e );
											b[10] = (uchar)(( DTS >> 22 ) & 0xff );
											b[11] = ( b[16] & ~0xfe ) | (uchar)(( DTS >> 14 ) & 0xfe );
											b[12] = (uchar)(( DTS >> 7 ) & 0xff );
											b[13] = ( b[18] & ~0xfe) | (uchar)(( DTS << 1 ) & 0xfe );
										}
									}
/*
char *s = "- ";
if (p->pType == PID_MPEG2_VIDEO)
{
	uint start_code = 0;
	for (int i = 4; i < 186; i++)
	{
		start_code = (start_code << 8) | buf[i];
		if (start_code == 0x00000100)
		{
			// check if picture_coding_type == 1
			uint picture_coding_type = (uint)((buf[i+2] >> 3) & 0x7);
			switch (picture_coding_type)
			{
			case 1: s = "vI"; break;
			case 2: s = "vP"; break;
			case 3: s = "vB"; break;
			default: s = "v?"; break;
			}
			break;
		}
	}
}
else if (p->pType == PID_AC3_AUDIO)
{
	s = "a ";
}

if (PTS >= 0)
	_Error("%s PTS=%03X%06X, %.4f\n", s, (int)(PTS>>24), (int)(PTS&0xffffff), (double)PTS * 300 / SYSTEMCLOCKRATE);
else
	_Error("%s No PTS\n", s);
if (DTS >= 0)
	_Error("   DTS=%03X%06X, %.4f\n", (int)(DTS>>24), (int)(DTS&0xffffff), (double)DTS * 300 / SYSTEMCLOCKRATE);
*/
/*
									if (p->pType == PID_MPEG2_VIDEO)
									{
										__int64 t = PTS - PTSLast;
										if ((t > 0) && ((TSPerFrame <= 0) || (TSPerFrame > t)))
										{
											TSPerFrame = t;
										}
										PTSLast = PTS;
									}
								}// end if ((b[0] == 0) && (b[1] == 0) && (b[2] == 1))
*/

							}// end if( p->buffer && !p->processed )
						}// end if( start )

						if( p->reading )
						{
							// Add the payload for this packet to the current buffer
							int payloadLength = 184 - adaptLength;
							if( payloadLength > 0 )
							{
								if(( p->bufferPos + payloadLength ) <= p->bufferSize )
								{
									memcpy( p->buffer + p->bufferPos, buf + 4 + adaptLength, payloadLength );
									p->bufferPos += payloadLength;
								}
								else
								{
									// buffer overflow!
									_Error("Error: buffer overflow! PID=0x%02X, PES packet, buffer size = %d\n", pid, p->bufferSize);
									p->reading = false;
								}
							}

							// only want to continue reading this PES packet if it's an i-frame
							// check for a sequence_header_code, if we have it were good to go
							// check for a group_start_code, if we have it were good to go
							// check for a picture_start_code, if it's an i-frame, add a
							//  sequence_header to the beginning of the buffer

							if( !p->confirmedRead && ( p->bufferPos > 512 ))
							{
								uint start_code = 0;
								for( int i = 0; i < p->bufferPos; i++ )
								{
									start_code = ( start_code << 8 ) | p->buffer[i];
									// sequence_header_code || group_start_code
									if( start_code == 0x000001b3 || start_code == 0x000001b8  )
									{
										// this is all we need to continue
										p->confirmedRead = true;
										break;
									}

									// picture_start_code

									if( start_code == 0x00000100 ) // picture_start_code
									{
										// check if picture_coding_type == 1
										if(( p->buffer[i + 2] & ( 0x7 << 3 )) == ( 1 << 3 ))
										{
											// add a sequence_header to the beginning of the buffer

											unsigned char *pTemp = new unsigned char[ p->bufferPos ];
											memcpy( pTemp, p->buffer, p->bufferPos );
											memcpy( p->buffer, Stream_Header.buffer, Stream_Header.bytes );
											memcpy( p->buffer + Stream_Header.bytes, pTemp, p->bufferPos );
											p->bufferPos += Stream_Header.bytes;

											delete pTemp;

											p->confirmedRead = true; // i-frame found
											break;
										}
									}// end if( start_code == 0x00000100 ) // picture_start_code
								}// end for( int i = 0; i < p->bufferPos; i++ )

								if( !p->confirmedRead )
									p->reading = false;

							}// end if( !p->confirmedRead && ( p->bufferPos > 512 ))
						}// end if (p->reading)
					}// end if( adaptLength >= 0 )
				}// end else for if (errorBit)
			}// end if (p)
		}// end if (buf[0] == 0x47)

		bool errorBit = ( buf[1] & 0x80 ) != 0;
		if( errorBit )
			continue;

		// update the PAT data
		if( pid == 0x00 ) // 0x00 = PAT
		{
			// verify PAT before processing
			uchar *pPat, *pBuf;
			pPat = PatPmtPackets.pPAT_Packet + 4;
			pBuf = buf + 4;
			bool match = true;

			// quick compare
			// checks: pointer_field to program_number
			for( int i = 0; i < 6; i++ )
			{
				if( *pBuf != *pPat )
				{
					match = false;
					break;
				}
				else
				{
					pBuf++;
					pPat++;
				}
			}
			if( !match ) // strip this packet
				continue; 

			// create a new Program Association Table
			CreateNewPAT( buf, pPidArray, numOfPids, pPidArrayStart );
			newPAT = true;
		}

		// 0x1ffb is PSIP, i.e. contains the following:
		//  MGT - Master Guide Table ( we're interested in keeping this )
		//  VCT - Virtual Channel Table ( we're interested in keeping this )
		//  RRT - Region Table
		//  EIT - Event Information Table
		//  ETT - Extended Text Table
		//  DCCT - Directed Channel Change Table(optional)
		//  DCCSCT - Directed Channel Change Selection Code Table(optional)
		else if( pid == 0x1ffb )
		{
			if( !exclude )
			{
				// don't process PSIP packets until new PAT and PMT have been created
				if( !newPAT || !newPMT )
					continue;

				// if this is true, let PSIP data through whos payload_unit_start_indicator is set to 0
				//  so that we can get the rest of the packet data we need to create a new PSIP packet
				if( bAddToBuffer )
				{
					// get the payload_unit_start_indicator
					short payload_unit_start_indicator = ( buf[1] >> 6 ) & 0x01;
					if( payload_unit_start_indicator == 0 )
					{
						if(! CreateNewVCT( pid, buf, pPidArray, numOfPids, pPidArrayStart, b2, bAddToBuffer ))
						{
							if( bPreserveBitRate )
							{
								ConvertToNULLPacket( buf );
								break;
							}
							else
								continue; // strip this packet, we need the next packet to finish the new PSIP
						}
						else
						{
							bAddToBuffer = false;
							newVCT = true;
						}
					}
					else
					{
						if( bPreserveBitRate )
						{
							ConvertToNULLPacket( buf );
							break;
						}
						else
							continue; // strip this packet
					}
				}// end if(bAddToBuffer)
				else
				{
					// only interested in table_id 0x07(MGT), 0xc8,0xc9(VCTs)
					uchar *b;
					// get the table_id
					// 5 = sync_byte to table_id
					b = buf + 5;
					short table_id = b[0];

					// if not a MGT or VCT, strip it
					if( table_id == 0xc7 || table_id == 0xc8 || table_id == 0xc9)
					{
						// create new Virtual Channel Table
						if( table_id == 0xc8 || table_id == 0xc9 )
						{
							if(! CreateNewVCT( pid, buf, pPidArray, numOfPids, pPidArrayStart, b2, bAddToBuffer ))
							{
								bAddToBuffer = true;

								if( bPreserveBitRate )
								{
									ConvertToNULLPacket( buf );
									break;
								}
								else
									continue; // strip this packet, we need the next packet to finish new VCT
							}
							else
								newVCT = true;
						}

						// create new Master Guide Table
						if( table_id == 0xc7 )
						{
							if( newVCT )// can't create a new MGT until the new VCT has been created
							{
								CreateNewMGT( pid, buf, pPidArray, numOfPids, pPidArrayStart ); // create a new MGT packet
							}
							else // strip this packet
							{
								if( bPreserveBitRate )
								{
									ConvertToNULLPacket( buf );
									break;
								}
								else
									continue;
							}
						}
					}// end if( table_id == 0xc7 || table_id == 0xc8 || table_id == 0xc9)
					else
					{
						if( bPreserveBitRate )
						{
							ConvertToNULLPacket( buf );
							break;
						}
						else
							continue; // strip this packet
					}

				}// end else for if(bAddToBuffer)
			}// end if( !exclude )

			else if( bPreserveBitRate && exclude )
				ConvertToNULLPacket( buf );

		}//end else if( pid == 0x1ffb )
		else
		{	// is this a video, audio, PMT, PCR pid?
			pidType pid_Type = GetPidType( pid, pPidArray, numOfPids, pPidArrayStart );

			switch( pid_Type )
			{
				case PID_PMT:
					if( newPAT )
					{
						// verify PMT before processing
						uchar *pPmt, *pBuf;

						bool found = false;
						int i;
						for( i = 0; i < PatPmtPackets.numOfPMTPackets; i++ )
						{
							int pmt_pid = *(PatPmtPackets.pPMT_Packet[i] + 1) & 0x1f;
							pmt_pid <<= 8;
							pmt_pid |= *(PatPmtPackets.pPMT_Packet[i] + 2);

							if( pmt_pid == pid )
							{
								found = true;
								break;
							}
						}
						if( !found ) 
						{
							if( bPreserveBitRate )
							{
								ConvertToNULLPacket( buf );
								break;
							}
							else // strip this packet
								continue;
						}

						// quick compare
						// check the program_number and the PCR_PID

						// check program_number 
						pPmt = PatPmtPackets.pPMT_Packet[i] + 8;
						pBuf = buf + 8;

						int pPmt_program_number = pPmt[0] << 8 | pPmt[1];
						int pBuf_program_number = pBuf[0] << 8 | pPmt[1];

						if( pPmt_program_number != pBuf_program_number )
						{
							if( bPreserveBitRate )
							{
								ConvertToNULLPacket( buf );
								break;
							}
							else
								continue;
						}

						// check PCR_PID
						pPmt += 5;
						pBuf += 5;

						int pPmt_PCR_PID = (pPmt[0] & 0x0f) << 8 | pPmt[1];
						int pBuf_PCR_PID = (pBuf[0] & 0x0f) << 8 | pBuf[1];

						if( pPmt_PCR_PID != pBuf_PCR_PID )
						{
							if( bPreserveBitRate )
							{
								ConvertToNULLPacket( buf );
								break;
							}
							else
								continue;
						}

						if( CreateNewPMT( pid, buf, pPidArray, numOfPids,
									  pPidArrayStart, audioCount ))
						{
							// ALL PMT pids must be process before newPMT is true
							newPMT = true;
						}
						else
						{
							if( bPreserveBitRate )
							{
								ConvertToNULLPacket( buf );
								break;
							}
							else
								continue;
						}
					}
					else
					{
						if( bPreserveBitRate )
						{
							ConvertToNULLPacket( buf );
							break;
						}
						else
							continue;
					}
					break;

				// remap the pid
				// Note: Any video or audio pid that is encountered before a new PMT is created will be stripped
				case PID_MPEG2_VIDEO:
					if( newPMT )
						if( bSetVerticalSizeTo1080 )
							if( vertical_size_value == 0x440 ) //0x440(1088)
								SetVerticalSizeTo1080( buf );
					// let fall through to UpdatePid(...)
				case PID_USER_PRIVATE:
				case PID_AC3_AUDIO:
				case PID_AAC_AUDIO:
				case PID_MPEG2_AUDIO:
				case PID_MPEG1_AUDIO:
				case PID_PCR:
					if( newPMT )
						UpdatePid( pid, pid_Type, buf, adaptation_field_length, pPidArray, numOfPids, pPidArrayStart );

					else
					{
						if( bPreserveBitRate )
						{
							ConvertToNULLPacket( buf );
							break;
						}
						else
							continue;
					}

					break;

				case PID_NULL:
					break;

				// strip this pid
				case PID_UNKNOWN:
					if( bPreserveBitRate )
					{
						ConvertToNULLPacket( buf );
						break;
					}
					else
						continue;
				break;

			}// end switch( pid_Type )
		}// end else pidType

		// write packet to output file
		if( fwrite64(buf, 188, 1, fOut ) != 1 )
		{
			sprintf(errMsgBuf, "Couldn't write to output file\n%s", outputFileName);
			errMsgVal = ::MessageBox(parentWnd, errMsgBuf, "Error!", MB_OK | MB_ICONERROR);
			rVal = 1;
			break;
		}
		totalOut += 188;
	}

	// Close files

	if (readBuffer->IsFileOpen())
		readBuffer->Close();
//	if (fIn)
//		fclose64(fIn);
	if (fOut)
		fclose64(fOut);
	if (fList)
		fclose64(fList);
	if (PIDs.pids[0].buffer)
		delete []PIDs.pids[0].buffer;
//	if (PIDs.pids[1].buffer)
//		delete []PIDs.pids[1].buffer;

	// free buffers
	if( pPidArray )
		delete []pPidArray;
	if( b2 )
		delete []b2;
	if( copyPatPmtPackets.pPAT_Packet )
		delete []copyPatPmtPackets.pPAT_Packet;
	if( copyPatPmtPackets.pPMT_Packet )
		for( int i = 0; i < copyPatPmtPackets.numOfPMTPackets; i++ )
			delete []copyPatPmtPackets.pPMT_Packet[i];

	return rVal;
}

// if there is a sequence_header in this packet
//  update the vertical_size_value to be 1080 
//  instead of 1088
void SetVerticalSizeTo1080( unsigned char *buf )
{
	uint start_code = 0;

	for( int i = 0; i < 188; i++ )
	{
		start_code = ( start_code << 8 ) | buf[i];

		// sequence_header_code
		if( start_code == 0x000001b3  )
		{
			if( buf[i+3] == 0x40 )
			{
				buf[i+3] = 0x38;
				// assume there is only one sequence_header in a packet
				break;
			}
		}
	}
}

// verify that the PMT defines both a video and an audio pid
bool VerifyPMT( unsigned char *buf )
{
	uchar *b;
	int section_length;
	int program_info_length;
	int stream_type;
	int ES_info_length;
	int pointer_field = 0;

	bool videoPID = false;
	bool audioPID = false;

    // if adaptation_field_length == 3, adjust offset to data
	if(( buf[3] & 0x30 ) >> 4 == 3 )
		pointer_field = buf[4] + 1;
	
	// 6 - sync_byte to section_length + pointer_offset
	b = buf + 6 + pointer_field;

	section_length = ( b[0] & 0x0f ) << 4;
	section_length |= b[1];

	// 15 - sync_byte to program_info_length + pointer_offset
	b = buf + 15 + pointer_field;

	program_info_length = ( b[0]& 0x0f ) << 4;
	program_info_length |= b[1];

	// skip descriptors
	b += 2 + program_info_length;

	// 9 - program_number to program_info_length
	// 4 - CRC_32
	for( int i = 0; i < section_length - 9 - program_info_length - 4; i++ )
	{
		stream_type = b[0];

		if( stream_type == 0x02 || stream_type == 0x80 )
			videoPID = true;

		else if( stream_type == 0x03 || stream_type == 0x04 || stream_type == 0x06 || 
			     stream_type == 0x0f || stream_type == 0x81 )
	 		audioPID = true;

		// skip reserved, elementary_PID, and reserved
		b += 3;

		ES_info_length = ( b[0] & 0x0f ) << 4;
		ES_info_length |= b[1];

		// skip descriptors
		b += ES_info_length + 2;

		i += 5 + ES_info_length;
	}

	return videoPID && audioPID ? true : false;
}

// create a new Program Association Table (PAT)
void CreateNewPAT( unsigned char *buf, Pid_Info *pPidArray, int numOfPids, Pid_Info *pPidArrayStart )
{
	uchar *b;
	int pointer_field = 0;

	short adaptation_field_control;
	// 3 = sync_byte to continuity_counter
	b = buf + 3;
	// check adaptation_field_control
	adaptation_field_control = ( b[0] >> 4 ) & 0x03;

	// if adaptation_field_control == 3, pointer_field is not zero
	if( adaptation_field_control == 3 )
		pointer_field = buf[4] + 1;

	// get the continuity_counter
    for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->type == PID_PAT )
		{
			if( pPidArray->processed )
			{
				// if the adaptation_field_control is 0x2, Adaptation_field_only, no payload,
				//  don't increment the continuity_counter
				if( adaptation_field_control != 0x2 )
				{
					pPidArray->continuity_counter++;
					if( pPidArray->continuity_counter > 0xf ) // 0x0 - 0xf
						pPidArray->continuity_counter = 0;
				}
			}
			else
			{
				pPidArray->continuity_counter = 0;
				pPidArray->processed = true;
			}
			break;
		}
		*pPidArray++;
	}

	// update the continuity_counter
	// 3 = syc_byte to continuity_counter
	b = buf + 3;
	b[0] = ( b[0] | 0x0f ) & (uchar)( pPidArray->continuity_counter | 0xf0 );

	// get the transport_stream_id
	// 8 = sync_byte to transport_stream_id
	b = buf + 8 + pointer_field;
	short transport_stream_id = b[0] << 8;
	transport_stream_id |= b[1];

	pPidArray = pPidArrayStart;

	// 4 = sync_byte to continuity_counter
	// 9 = pointer_field + table_id to last_section_number
	b = buf + 4 + pointer_field + 9;

	short program_number = 1;
	short base_pid = program_number * 0x10;

	// update the program_map_PID's
	for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->type == PID_PMT )
		{
			// keep track of the new pid mappings
			pPidArray->new_pid = base_pid;

			// write in the new bytes
			b[0] = (uchar)program_number >> 8;
			b[1] = (uchar)program_number;
			b[2] = b[2] & (uchar)( base_pid >> 8 | 0xe0 );
			b[3] = (uchar)base_pid;

			program_number++;
			base_pid = program_number * 0x10;
			b += 4;
		}

		*pPidArray++;
	}

	pPidArray = pPidArrayStart;

	// set the new section_length
	// 5 = transport_stream_id to last_section_number
	// 4 = program_number to program_map_PID
	// 4 = CRC_32
	short section_length = 5 + ( 4 * ( program_number - 1 )) + 4;

	// write out the new bytes
	// 4 = sync_byte to continuity_counter
	// 2 = pointer_field + table_id
	b = buf + 4 + pointer_field + 2;
	b[0] = b[0] & (uchar)( section_length >> 8 | 0xf0 );
	b[1] = (uchar)section_length;

	// re-calculate the CRC_32
	// 3 = table_id to section_length
	// -4 = CRC_32
	int length = 3 + section_length - 4;
	unsigned char *pPayLoadData = new unsigned char[ length ];

	// 4 = sync_byte to continuity_counter
	// 1 = pointer_field
	b = buf + 4 + pointer_field + 1;
	memcpy( pPayLoadData, b, length );
	int CRC_32 = CalculateCRC( pPayLoadData, length );
	delete []pPayLoadData;

	// update the CRC_32 data
	// 4 = sync_byte to continuity_counter
	// 4 = pointer_field + table_id to section_length
	// -4 = CRC_32
	b = buf + 4 + pointer_field + 4 + section_length - 4;
	b[0] = CRC_32 >> 24;
	b[1] = CRC_32 >> 16;
	b[2] = CRC_32 >> 8;
	b[3] = CRC_32;

	// fill in the rest of the packet with 0xff
	// -8 = sync_byte to continuity_counter
	int fillLength = 188 - pointer_field - section_length - 8;

	// 4 = sync_byte to continuity_counter
	// 4 = pointer_field + table_id to section_length
	b = buf + 4 + pointer_field + section_length + 4;

	for( int i = 0; i < fillLength; i++ )
		b[i] = 0xff;
}

// returns the pid type for a pid
pidType GetPidType( short pid, Pid_Info *pPidArray, int numOfPids, Pid_Info *pPidArrayStart )
{
	// search though our pids and return its pid_type
	for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->original_pid == pid  )
			return pPidArray->type;

		*pPidArray++;
	}

	pPidArray = pPidArrayStart;

	return PID_UNKNOWN;
}

// create a new Program Map Table (PMT)
bool CreateNewPMT( short pid, unsigned char *buf, Pid_Info *pPidArray, int numOfPids, Pid_Info *pPidArrayStart, int audioCount )
{
	uchar *b;
	int pointer_field = 0;

	short adaptation_field_control;
	// 3 = sync_byte to continuity_counter
	b = buf + 3;
	// check adaptation_field_control
	adaptation_field_control = ( b[0] >> 4 ) & 0x03;

	// if adaptation_field_control == 3, pointer_field is not zero
	if( adaptation_field_control == 3 )
		pointer_field = buf[4] + 1;

	// get last continuity_counter
    for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->original_pid == pid )
		{
			if( pPidArray->processed )
			{
				// if the adaptation_field_control is 0x2, Adaptation_field_only, no payload
				//  don't increment the continuity_counter
				if( adaptation_field_control != 0x2 )
				{
					pPidArray->continuity_counter++;
					if( pPidArray->continuity_counter > 0xf )// 0x0 - 0xf
						pPidArray->continuity_counter = 0;
				}
			}
			else
			{
				pPidArray->continuity_counter = 0;
				pPidArray->processed = true;
			}
			break;
		}
		*pPidArray++;
	}

	// update the continuity_counter
	// 3 = syc_byte to PID
	b = buf + 3;
	b[0] = ( b[0] | 0x0f ) & (uchar)( pPidArray->continuity_counter | 0xf0 );

	pPidArray = pPidArrayStart;

	// 1 = sync_byte
	b = buf + 1;

	short newPid;
	// re-map old pid with new pid
	for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->original_pid == pid )
		{
			newPid = pPidArray->new_pid;
			break;
		}

		*pPidArray++;
	}

	pPidArray = pPidArrayStart;

	// write in the PID
	b[0] = b[0] & (uchar)( newPid >> 12 | 0xe0 );
	b[1] = (uchar)newPid;

	// update the program_number
	// 8 = sync_byte to continuity_counter + table_id to section_length 
	b = buf + pointer_field + 8;
	// the program_number follows the base pid
	short program_number = newPid / 0x16 + 1;

	// write in the program_number
	b[0] = (uchar)program_number >> 8;
	b[1] = (uchar)program_number;

	// if the PCR_PID is + 1 of the PMT PID(BASE pid) then remap as usual.
	// else remap the PCR_PID to + 0xf of the BASE PID because the 
	//  PCR info is not included in the Video PID but is in its own pid
	b = buf + pointer_field + 13;
	short PCR_PID = ( b[0] & 0x0f ) << 8 | b[1];
	bool found = false;
	for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->type == PID_PCR )
		{
			found = true;

			// the PCR info is not in the video pid, but is in its own pid
			PCR_PID = newPid + 0xf;
			pPidArray->new_pid = PCR_PID;
			break;
		}

		*pPidArray++;
	}
	pPidArray = pPidArrayStart;

	if( !found ) // PCR_PID is the same as the Video PID
	{
		PCR_PID = newPid + 1;
		found = false;
	}

	// update the PCR_PID
	// 13 = sync_byte to last_section_number
	b = buf + pointer_field + 13;
	// write new PCR_PID
	b[0] = b[0] & (uchar)( PCR_PID >> 12 | 0xe0 );
	b[1] = (uchar)PCR_PID;

	// get the section_length
	// 6 = sync_byte to table_id
	b = buf + pointer_field + 6;
	short section_length = ( b[0] & 0x0f ) << 8;
	section_length |= b[1];

	// check stream_type data and see if we will be deleting any of them
	// 15 = sync_byte to PCR_PID + pointer_field
	b = buf + pointer_field + 15;
	short program_info_length;
	program_info_length = ( b[0] & 0x0f ) << 8;
	program_info_length |=  b[1];

	// skip over the descriptors(), we will just keep these
	// 17 = sync_byte to program_info_length
	b = buf + pointer_field + 17 + program_info_length;

	// loop through the stream_types
	short stream_type; // for debugging
	short elementary_PID;
	short ES_info_length;
	unsigned char buf2[188]; // temp buff for holding the "stream_type" data
	unsigned char *b2;
	int stream_type_data_length = 0; // length of new stream_type data

	// copy the "stream_type" data from the "buf" buffer to the "buf2" buffer
	// 9 = program_number to program_info_length
	// 4 = CRC_32
	memcpy( buf2, b, section_length - 9 - program_info_length - 4 );

	b2 = buf2;
	// 9 = program_number to program_info_length
	// -4 = CRC_32
	for( int i = 9 + program_info_length; i < section_length - 4; )
	{
		stream_type = b2[0];
		elementary_PID = ( b2[1] & 0x1f ) << 8;
		elementary_PID |= b2[2];
		ES_info_length = ( b2[3] & 0x0f ) << 8;
		ES_info_length |= b2[4];

		// are we keeping this elementary_PID?
		bool exclude = false;
		bool found = false;
		for( int j = 0; j < numOfPids; j++  )
		{
			if( pPidArray->original_pid == elementary_PID )
			{
				if( pPidArray->strip_pid  )
				{
					exclude = true;
					break;
				}
				else // we are keeping this pid so remap the elementary_PID
				{
					found = true;

					if( stream_type == 0x02 ) // MPEG-2 Video
						pPidArray->type = PID_MPEG2_VIDEO;
					else if( stream_type == 0x80 ) // User private
						pPidArray->type = PID_USER_PRIVATE;
					else if( stream_type == 0x81 || stream_type == 0x06 ) // Dolby AC3-Audio
						pPidArray->type = PID_AC3_AUDIO;
					else if( stream_type == 0x0f ) // AAC Audio
						pPidArray->type = PID_AAC_AUDIO;
					else if( stream_type == 0x03 ) // MPEG-1 Audio
						pPidArray->type = PID_MPEG1_AUDIO;
					else if( stream_type == 0x04 ) // MPEG-2 Audio
						pPidArray->type = PID_MPEG2_AUDIO;
					else
						pPidArray->type = PID_UNKNOWN;

					// will handle this as I come across it
					ASSERT( pPidArray->type != PID_UNKNOWN );

					// video elementary pids are base_pid + 1
					if( pPidArray->type == PID_MPEG2_VIDEO || pPidArray->type == PID_USER_PRIVATE )
					{
						short newVideoPid = newPid + 1;
						pPidArray->new_pid = newVideoPid;

						// update the new pid
						b2[1] = b2[1] & ( newVideoPid >> 8  | 0xe0 );
						b2[2] = (uchar)newVideoPid;
					}

					// audio elementary pids are base_pid + 4
					if(( pPidArray->type == PID_AC3_AUDIO )   ||
					   ( pPidArray->type == PID_AAC_AUDIO )   ||
					   ( pPidArray->type == PID_MPEG2_AUDIO ) ||
					   ( pPidArray->type == PID_MPEG1_AUDIO ))
					{
						short newAudioPid;

						if( !pPidArray->processed )
						{
							newAudioPid = newPid + 4 + audioCount;
                            pPidArray->new_pid = newAudioPid;
							pPidArray->processed = true;

							audioCount++;
						}

						// update the new pid
						b2[1] = b2[1] & (uchar)( pPidArray->new_pid >> 8  | 0xe0 );
						b2[2] = (uchar)pPidArray->new_pid;
					}

					break;
				}
			}

			*pPidArray++;
		}

		pPidArray = pPidArrayStart;

		if( exclude || !found )
		{	// don't copy this data back to "buf"
			i += 5 + ES_info_length;
			b2 += 5 + ES_info_length;
		}
		else
		{	// copy the data back to "buf"
			memcpy( b, b2, 5 + ES_info_length );

			stream_type_data_length += 5 + ES_info_length;

			// 5 - stream_type to ES_info_length
			i += 5 + ES_info_length;
			b2 += 5 + ES_info_length;
			b += 5 + ES_info_length;
		}
	}

    // update the section_length
	// 9 = program_number to program_info_length
	// 4 = CRC_32
	section_length = 9 + program_info_length + stream_type_data_length + 4;
	// 6 = sync_byte to table_id
	b = buf + 6 + pointer_field;
	b[0] = b[0] & (uchar)( section_length >> 8 | 0xf0 );
	b[1] = (uchar)section_length;

	// re-calculate the CRC_32
	// 3 = table_id to section_length 
	// -4 = CRC_32
	int length = 3 + section_length - 4;
	unsigned char *pPayLoadData = new unsigned char[ length ];

	// 4 = sync_byte to continuity_counter
	// 1 = pointer_field
	b = buf + 4 + pointer_field + 1;
	memcpy( pPayLoadData, b, length );
	int CRC_32 = CalculateCRC( pPayLoadData, length );
	delete []pPayLoadData;

	// write the CRC_32 data
	// 4 = sync_byte to continuity_counter
	// 4 = pointer_field + table_id to section_length
	// -4 = CRC_32
	b = buf + 4 + pointer_field + 4 + section_length - 4;
	b[0] = CRC_32 >> 24;
	b[1] = CRC_32 >> 16;
	b[2] = CRC_32 >> 8;
	b[3] = CRC_32;

	// fill in the rest of the packet with 0xff
	// -8 = sync_byte to continuity_counter
	int fillLength = 188 - pointer_field - section_length - 8;

	// 4 = sync_byte to continuity_counter
	// 4 = pointer_field + table_id to section_length
	b = buf + 4 + pointer_field + section_length + 4;

	for( int i = 0; i < fillLength; i++ )
		b[i] = 0xff;

	// check the PMT pids, if all PMT pids have been processed, then return true,
	//  else not finished processing all PMT's yet, return false
	bool bAllPMTProcessed = true;
	for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->type == PID_PMT )
			if( !pPidArray->processed )
			{
				bAllPMTProcessed = false;
				break;
			}

		*pPidArray++;
	}

	pPidArray = pPidArrayStart;

	return bAllPMTProcessed;

}

// remap the pid
void UpdatePid( short pid, pidType pid_type, unsigned char *buf, int adaptation_field_length, Pid_Info *pPidArray,
				int numOfPids, Pid_Info *pPidArrayStart )
{
	uchar *b;

	short adaptation_field_control;
	// 3 = sync_byte to continuity_counter
	b = buf + 3;
	// check adaptation_field_control
	adaptation_field_control = ( b[0] >> 4 ) & 0x03;

	// get last continuity_counter
    for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->original_pid == pid )
		{
			if( pPidArray->processed )
			{
				// if the adaptation_field_control is 0x2, Adaptation_field_only, no payload
				//  don't increment the continuity_counter

				if( adaptation_field_control != 0x2 )
				{
					pPidArray->continuity_counter++;
					if( pPidArray->continuity_counter > 0xf )// 0x0 - 0xf
						pPidArray->continuity_counter = 0;
				}
			}
			else
			{
				pPidArray->continuity_counter = 0;
				pPidArray->processed = true;
			}

			break;
		}
		*pPidArray++;
	}

	pPidArray->type = pid_type;

	// update the continuity_counter
	// 3 = syc_byte to PID
	b = buf + 3;
	b[0] = ( b[0] | 0x0f ) & (uchar)( pPidArray->continuity_counter | 0xf0 );

	pPidArray = pPidArrayStart;

	// 1 = sync_byte
	b = buf + 1;

	// remap pid
	for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->original_pid == pid )
		{
			// assign the new pid
			short newPid = pPidArray->new_pid;

			// write the new pid
			b[0] = b[0] & (uchar)( newPid >> 8 | 0xe0 );
			b[1] = (uchar)newPid;
			break;
		}

		*pPidArray++;
	}

	pPidArray = pPidArrayStart;
}

// Virtual Channel Table number_bytes value for when the Master Guide Table is created
static int VCT_number_bytes = 0;
// create a new VCT packet
// this function creates a very small PSIP TVCT section based on the current packet that
//  contains just enough data for channel information display
bool CreateNewVCT( short pid, unsigned char *buf, Pid_Info *pPidArray, int numOfPids,
				    Pid_Info *pPidArrayStart, unsigned char *b2, bool bAddToBuffer )
{
	// NOTE: pointer_field is assumed to be 0

	uchar *psipBuf = b2; // the beginning of our psipBuf
	uchar *b = buf; // beginning of the current buffer

	// get the payload_start_indicator
	short payload_start_indicator = ( buf[1] >> 6 ) & 0x01;

	if( bAddToBuffer && !payload_start_indicator )
	{ // add this packets packet data to our psipBuf
		b = buf + 4;
		b2 += 188;
		// 4 = sync_byte
		memcpy( b2, b, 188 );
	}

	// first pass through
	else
	{
		// copy all of the data from buf to b3(our psipBuf)
		// 188 = everything in buf
		memcpy( b2, b, 188 );
	}

	// get the section_length
	// 6 = sync_byte to table_id
	b2 = psipBuf + 6;
	short section_length = ( b2[0] & 0x0f ) << 8 | b2[1];

	// PSIP data extends into additional packets
	if( section_length > 180 && bAddToBuffer == false )
		return false; // we need more data

	// copy sync_byte to num_channels_in_section to the "buf" buffer
	// 15 = sync_byte to num_channels_in_section
	b2 = psipBuf;
	b = buf;
	memcpy( b, b2, 15 );
	b += 15;

	// get the transport_stream_id
	// 8 = sync_byte to transport_stream_id
	b2 = psipBuf + 8;
	short transport_stream_id = b2[0] << 8 | b2[1];

	// get number of channels
	// 14 = sync_byte to num_channels_in_section
	b2 = psipBuf + 14;
	short num_channels_in_section = b2[0];

	// loop through the channels, keeping only the channels we selected

	short num_of_channels = 0; // number of channels left after modification

	// create a temp buffer to store out channel data in
	unsigned char *channel_info = new unsigned char[1024];
	unsigned char *pCI = channel_info;

	b2 += 1; // next byte starts channels section
	for( int z = 0; z < num_channels_in_section; z++ )
	{
		// do we want to keep this channel information?
		// if we can't find a service location descriptor(0xa1) then we don't want the channel
		// if we find that the PCR_PID in the service location descriptor isn't based on one of
		//  our PM pids OR there isn't an independant PCR_PID, we don't want the channel either

		// get the channel descriptor_length
		short ch_descriptor_length = ( b2[30] & 0x02 ) << 8 | b2[31];
 		short new_ch_descriptor_length = 0;
		short total_descriptor_length = 0;

		short program_number = ( b2[24] << 16 ) | b2[25];

		int bAlreadyCopied = false;

		// copy from short_name to descriptor_length to pCI
		memcpy( pCI, b2, 32 );
		pCI += 32;
		b2 += 32;

		// find a sevice location descriptor - 0xa1
		for( int k = 0; k < ch_descriptor_length; k ++ )
		{
			// get descriptor_tag
			short descriptor_tag = b2[0];
			short tag_descriptor_length = b2[1];

			if( descriptor_tag != 0xa1 ) // service location descriptor
			{
				// get tag descriptor_length
				k += tag_descriptor_length + 2;
				b2 += tag_descriptor_length + 2 ;
			}
			else // process the location descriptor
			{
				// service location descriptor found
				// does its PCR_PID match one of the PM base pids

				// get the PCR_PID
				short PCR_PID = ( b2[2] & 0x0f ) << 8 | b2[3];

				// does this PCR_PID need re-mapping?
				int found = false;
				for( int l = 0; l < numOfPids; l++ )
				{
					// if PID_PCR, then PCR info is in its own PID
					if( pPidArray->type == PID_PCR || pPidArray->type == PID_PMT )
					{
						if( pPidArray->type == PID_PCR )
						{
							PCR_PID = pPidArray->new_pid;
							found = true;

							// update new PCR_PID
							b2[2] = ( b2[2] & 0xe0 ) | PCR_PID >> 8;
							b2[3] = ( b2[3] ^ b2[3] ) | PCR_PID;
							break;
						}

						// if PCR_PID-1, then PCR info is in Video PID
						if( pPidArray->original_pid == PCR_PID - 1 && pPidArray->program_number == program_number )
						{
							PCR_PID = pPidArray->new_pid + 1;
							found = true;

							// update new PCR_PID
							b2[2] = ( b2[2] & 0xe0 ) | PCR_PID >> 8;
							b2[3] = ( b2[3] ^ b2[3] ) | PCR_PID;
							break;
						}
					}

					if( ( pPidArray->type == PID_MPEG2_VIDEO || pPidArray->type == PID_USER_PRIVATE ) && pPidArray->program_number == program_number )
					{
						if( pPidArray->original_pid == PCR_PID )
						{
							PCR_PID = pPidArray->new_pid;
							found = true;

							// update new PCR_PID
							b2[2] = ( b2[2] & 0xe0 ) | PCR_PID >> 8;
							b2[3] = ( b2[3] ^ b2[3] ) | PCR_PID;
							break;
						}
					}
				*pPidArray++;
				}
				pPidArray = pPidArrayStart;

				// could not find an PM that matched so strip this channel
				if( !found )
				{
					k += tag_descriptor_length + 2;
					b2 += tag_descriptor_length + 2 ;
					continue;
				}

				num_of_channels++;

				// get this before we increment passed it
				short number_elements = b2[4];

				// now that we have found a service location descriptor(0xa),
				//  copy from short_name to descriptor_length, but only copy it once!
				// 32 = short_name to descriptor_length
				if( !bAlreadyCopied )
				{
					memcpy( b, channel_info, 32 );

					// update the program_number
					b[24] = (uchar)num_of_channels >> 4;
					b[25] = (uchar)num_of_channels;

					b += 32;
					bAlreadyCopied = true;
				}

				// copy descriptor_tag to number_element to pCI
				// 5 = descriptor_tag to number_element
				memcpy( pCI, b2, 5 );
				pCI += 5;
				b2 += 5;
				k += 5;

				// loop through elements and remap pids or remove them

				short new_number_elements = 0;
				short new_descriptor_length = 0;
				short elementary_PID;

				// get number_elements
				for( int l = 0; l < number_elements; l++ )
				{
					// get elmentary_PID
					elementary_PID = ( b2[1] & 0x0f ) << 8 | b2[2];

					found = false;
					short new_pid = -1;
					// are we stripping this one?
					for( int m = 0; m < numOfPids; m++ )
					{
						if( pPidArray->original_pid == elementary_PID )
						{
							if( pPidArray->strip_pid )
							{
								found = true;
								break;
							}
							else
							{ // remap pid
								new_pid = pPidArray->new_pid;

								// update pid
								b2[1] = ( b2[1] & 0xe0 ) | new_pid >> 8;
								b2[2] = ( b2[2] ^ b2[2] ) | new_pid;

								break;
							}
						}
						*pPidArray++;
					}
					pPidArray = pPidArrayStart;

					if( found ) // strip this pid
					{
						// 6 = stream_type to ISO_639_language_code
						b2 += 6;
						k += 6;
						continue;
					}

					// add this element to pCI
					// 6 = stream_type to ISO_639_language_code
					memcpy( pCI, b2, 6 );
					new_number_elements++;
					new_descriptor_length += 6;

					// 6 = stream_type to ISO_639_language_code
					pCI += 6;
					b2 += 6;
					k+= 6;

				}// end for( l < number_elements )

				// 3 = reserved to number_elements
				new_descriptor_length += 3;

				unsigned char *s = pCI; // save pCI position
				// update the descriptor_length
				// +1 = descriptor_length
				pCI -= new_descriptor_length + 1;
				pCI[0] = (uchar)new_descriptor_length;
				total_descriptor_length += new_descriptor_length;

				// update the number_elements
				pCI += 3;
				pCI[0] = (uchar)new_number_elements;
				pCI = s; // restore pCI

				// copy pCI data to buf
				// 2 = descriptor_tag + descriptor_length
				pCI -= new_descriptor_length + 2;
				memcpy( b, pCI, new_descriptor_length + 2 );
				b += new_descriptor_length + 2;

				pCI = channel_info;

				// update the ch_descriptor_length
				// Note: the descriptor_length is updated each time through the loop
				//       because we don't know if this will be the last time through
				//       the loop or not
				// 2 = descriptor_tag and descriptor_length
				total_descriptor_length += 2;
				unsigned char *s2 = b; // save b position
				// +4 bytes = descriptor_tag, descriptor_length, reserved, descriptor_length
				b -= new_descriptor_length + 4;
				b[0] = ( b[0] & 0xfc ) | total_descriptor_length >> 8;
				b[1] = ( b[1] ^ b[1] ) | total_descriptor_length;
				b = s2; // restore b

				// subtract off reserved to number_elements incase
				//  we go through the loop again
				new_descriptor_length -= 3;

			}// end else for if( descriptor_tag )
		}// end for( k < descriptor_length )
		pCI = channel_info;
	}// end for( z < num_channels_in_section )

	// add reserved bits
	// Note: additional_descriptors_length = 0
	b[0] = 0xfc;
	b[1] = 0x00;

	section_length = ( b += 3 ) - ( buf + 5 );
	// update the section_length
	// 6 = sync_byte to table_id
	b = buf + 6;
	b[0] = b[0] & (uchar)( section_length >> 8 | 0xf0 );
	b[1] = (uchar)section_length;

	// save VCT section length for MGT
	// +3 = table_id to section_length
	VCT_number_bytes = section_length + 3;

	// update num_channels_in_section
	b = buf + 14;
	b[0] = (uchar)num_of_channels;

	short adaptation_field_control;
	// 3 = sync_byte to continuity_counter
	b = buf + 3;	// check adaptation_field_control
	adaptation_field_control = ( b[0] >> 4 ) & 0x03;

	// get last continuity_counter
    for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->original_pid == pid )
		{
			if( pPidArray->processed )
			{
				// if the adaptation_field_control is 0x2, Adaptation_field_only, no payload
				//  don't increment the continuity_counter
				if( adaptation_field_control != 0x2 )
				{
					pPidArray->continuity_counter++;
					if( pPidArray->continuity_counter > 0xf )// 0x0 - 0xf
						pPidArray->continuity_counter = 0;
				}
			}
			else
				pPidArray->continuity_counter = 0;

			break;
		}
		*pPidArray++;
	}

	// update the payload_unit_start_indicator
	// 1 for the first packet, 0 for all of the rest
	// VCT will always have its payload_unit_start_indicator set to 1
	short payload_unit_start_indicator = 1;
//	if( !pPidArray->processed )
//	{
		payload_unit_start_indicator = 1;
		pPidArray->processed = true;
//	}

	// write payload_unit_start_indicator
	// 1 = sync_byte
	b = buf + 1;
	b[0] = ( b[0] & 0xbf ) | payload_unit_start_indicator << 6;

	// update the continuity_counter
	// 3 = syc_byte to PID
	b = buf + 3;
	b[0] = ( b[0] | 0x0f ) & (uchar)( pPidArray->continuity_counter | 0xf0 );

	pPidArray = pPidArrayStart;

	// re-calculate the CRC_32
	// 3 = table_id to section_length
	// -4 = CRC_32
	int length = 3 + section_length - 4;
	unsigned char *pPayLoadData = new unsigned char[ length ];

	// 4 = sync_byte to continuity_counter
	// 1 = pointer_field
	b = buf + 4 + 1;
	memcpy( pPayLoadData, b, length );
	int CRC_32 = CalculateCRC( pPayLoadData, length );
	delete []pPayLoadData;

	// write the CRC_32 data
	// 4 = sync_byte to continuity_counter
	// 4 = pointer_field + table_id to section_length
	// -4 = CRC_32
	b = buf + 4 + 4 + section_length - 4;
	b[0] = CRC_32 >> 24;
	b[1] = CRC_32 >> 16;
	b[2] = CRC_32 >> 8;
	b[3] = CRC_32;

	// fill in the rest of the packet with 0xff
	// -8 = sync_byte to continuity_counter + pointer_field
	int fillLength = 188 - section_length - 8;

	// 4 = sync_byte to continuity_counter
	// 4 = pointer_field + table_id to section_length
	b = buf + 4 + 4 + section_length;

	for( int i = 0; i < fillLength; i++ )
		b[i] = 0xff;

	if( channel_info )
		delete []channel_info;

	return true; // new PSIP successfully created
}

// creat new Master Guide Table that references our newly created PSIP VCT.  strip all of the other referenced
//  PIDs
void CreateNewMGT( short pid, unsigned char *buf, Pid_Info *pPidArray, int numOfPids, Pid_Info *pPidArrayStart )
{
	// NOTE: pointer_field is assumed to be 0

	uchar *b;

	// walk all of the tables. when we find the VCT table, move it to be the first
	//  table listed.

	// 4 = sync_byte to pointer_field
	// 10 = table_id to tables_defined
	b = buf + 4 + 10;
	int tables_defined = b[0] << 8 | b[1];

	b += 2;
	uchar *pFirstTablePos = b;
	short table_type_descriptors_length = 0;
	bool bVCTFound = false;
	for( int i = 0; i < tables_defined; i++ )
	{
		// get the table_type
		int table_type = b[0] << 8 | b[1];

		// 0x0000 = Terrestrial VCT with current_next_indicator=1
		// 0x0001 = Terrestrial VCT with current_next_indicator=0
		// 0x0002 = Cable VCT with current_next_indicator=1
		// 0x0003 = Cable VCT with current_next_indicator=0
		if( table_type == 0x0000 ||
			table_type == 0x0001 ||
			table_type == 0x0002 ||
			table_type == 0x0003 )
		{
			// get the table_type_descriptors_length
			table_type_descriptors_length = ( b[9] & 0x0f ) << 4 | b[10];

			// move the VCT table to be the first table
			uchar *pStart = b;
			int End = ( b + 12 + table_type_descriptors_length + 1 ) - pStart;
			memcpy( pFirstTablePos, pStart, End );

			bVCTFound = true;

			break;
		}
		else // next table
		{
			// get the table_type_descriptors_length for this table
			// 9 = table_type to number_bytes
			table_type_descriptors_length = ( b[9] & 0x0f ) << 4 | b[10];
			b += 11 + table_type_descriptors_length;
		}
	}// end for( int i = 0; i < tables_defined; i++ )

	ASSERT( bVCTFound );
	if( !bVCTFound )// No VCT table was found!
		return;

	// add the reserved bits and descriptors_length
	// Note: additional_descriptors_length = 0
	// 4 = sync_byte to continuity_counter
	// 1 = pointer_field
	// 22 = table_id to table_type_descriptors_length
	b = buf + 4 + 1 + 22 + table_type_descriptors_length;
	b[0] = 0xf0;
	b[1] = 0x00;

	// update number_bytes
	// 4 = sync_byte to pointer_field
	// 10 = table_id to tables_defined
	// 5 = table_type to table_type_version_number
	b = buf + 4 + 10 + 7;
	b[0] = (byte)( VCT_number_bytes >> 24 & 0xff );
	b[1] = (byte)( VCT_number_bytes >> 16 & 0xff );
	b[2] = (byte)( VCT_number_bytes >> 8 & 0xff );
	b[3] = (byte)( VCT_number_bytes & 0xff );

	// update tables_defined
	// 4 = sync_byte to pointer_field
	// 10 = table_id to tables_defined
	b = buf + 4 + 10;
	b[0] = 0x00; // the only table that there will be
	b[1] = 0x01; //  is the new VCT that we created

	// update the section_length
	// 19 = table_id_extension to table_type_descriptors_length
	// 2 = reserved + descriptor_length
	// 4 = CRC_32
	int section_length = 19 + 2 + 4 + table_type_descriptors_length;
	// 4 = sync_byte to continuity_counter
	// 1 = pointer_field
	// 1 = table_id
	b = buf + 4 + 1 + 1;
	b[0] = ( b[0] & 0xf0 ) | ( section_length >> 8 );
	b[1] = section_length;

	short adaptation_field_control;
	// 3 = sync_byte to continuity_counter
	b = buf + 3;	// check adaptation_field_control
	adaptation_field_control = ( b[0] >> 4 ) & 0x03;

	// get last continuity_counter
    for( int i = 0; i < numOfPids; i++ )
	{
		if( pPidArray->original_pid == pid )
		{
			if( pPidArray->processed )
			{
				// if the adaptation_field_control is 0x2, Adaptation_field_only, no payload
				//  don't increment the continuity_counter
				if( adaptation_field_control != 0x2 )
				{
					pPidArray->continuity_counter++;
					if( pPidArray->continuity_counter > 0xf )// 0x0 - 0xf
						pPidArray->continuity_counter = 0;
				}
			}
			else
				pPidArray->continuity_counter = 0;

			break;
		}
		*pPidArray++;
	}

	// update the payload_unit_start_indicator
	// 1 for the first packet, 0 for all of the rest
	// MGT always has the payload_unit_start_indicator set to 1
	short payload_unit_start_indicator = 0;
//	if( !pPidArray->processed )
//	{
		payload_unit_start_indicator = 1;
		pPidArray->processed = true;
//	}

	// write payload_unit_start_indicator
	// 1 = sync_byte
	b = buf + 1;
	b[0] = ( b[0] & 0xbf ) | payload_unit_start_indicator << 6;

	// update the continuity_counter
	// 3 = syc_byte to PID
	b = buf + 3;
	b[0] = ( b[0] | 0x0f ) & (uchar)( pPidArray->continuity_counter | 0xf0 );

	pPidArray = pPidArrayStart;

	// re-calculate the CRC_32
	// 3 = table_id to section_length
	// -4 = CRC_32
	int length = 3 + section_length - 4;
	unsigned char *pPayLoadData = new unsigned char[ length ];

	// 4 = sync_byte to continuity_counter
	// 1 = pointer_field
	b = buf + 4 + 1;
	memcpy( pPayLoadData, b, length );
	int CRC_32 = CalculateCRC( pPayLoadData, length );
	delete []pPayLoadData;

	// write the CRC_32 data
	// 4 = sync_byte to continuity_counter
	// 4 = pointer_field + table_id to section_length
	// -4 = CRC_32
	b = buf + 4 + 4 + section_length - 4;
	b[0] = CRC_32 >> 24;
	b[1] = CRC_32 >> 16;
	b[2] = CRC_32 >> 8;
	b[3] = CRC_32;

	// fill in the rest of the packet with 0xff
	// -8 = sync_byte to continuity_counter + pointer_field
	int fillLength = 188 - section_length - 8;

	// 4 = sync_byte to continuity_counter
	// 4 = pointer_field + table_id to section_length
	b = buf + 4 + 4 + section_length;

	for( int i = 0; i < fillLength; i++ )
		b[i] = 0xff;
}



// calculate a CRC_32 value
// code gleaned from the ts muxer from the VideoLAN project
static unsigned int CalculateCRC( unsigned char *p_begin, int i_count )
{
    static unsigned int CRC32[256] =
    {
        0x00000000, 0x04c11db7, 0x09823b6e, 0x0d4326d9,
        0x130476dc, 0x17c56b6b, 0x1a864db2, 0x1e475005,
        0x2608edb8, 0x22c9f00f, 0x2f8ad6d6, 0x2b4bcb61,
        0x350c9b64, 0x31cd86d3, 0x3c8ea00a, 0x384fbdbd,
        0x4c11db70, 0x48d0c6c7, 0x4593e01e, 0x4152fda9,
        0x5f15adac, 0x5bd4b01b, 0x569796c2, 0x52568b75,
        0x6a1936c8, 0x6ed82b7f, 0x639b0da6, 0x675a1011,
        0x791d4014, 0x7ddc5da3, 0x709f7b7a, 0x745e66cd,
        0x9823b6e0, 0x9ce2ab57, 0x91a18d8e, 0x95609039,
        0x8b27c03c, 0x8fe6dd8b, 0x82a5fb52, 0x8664e6e5,
        0xbe2b5b58, 0xbaea46ef, 0xb7a96036, 0xb3687d81,
        0xad2f2d84, 0xa9ee3033, 0xa4ad16ea, 0xa06c0b5d,
        0xd4326d90, 0xd0f37027, 0xddb056fe, 0xd9714b49,
        0xc7361b4c, 0xc3f706fb, 0xceb42022, 0xca753d95,
        0xf23a8028, 0xf6fb9d9f, 0xfbb8bb46, 0xff79a6f1,
        0xe13ef6f4, 0xe5ffeb43, 0xe8bccd9a, 0xec7dd02d,
        0x34867077, 0x30476dc0, 0x3d044b19, 0x39c556ae,
        0x278206ab, 0x23431b1c, 0x2e003dc5, 0x2ac12072,
        0x128e9dcf, 0x164f8078, 0x1b0ca6a1, 0x1fcdbb16,
        0x018aeb13, 0x054bf6a4, 0x0808d07d, 0x0cc9cdca,
        0x7897ab07, 0x7c56b6b0, 0x71159069, 0x75d48dde,
        0x6b93dddb, 0x6f52c06c, 0x6211e6b5, 0x66d0fb02,
        0x5e9f46bf, 0x5a5e5b08, 0x571d7dd1, 0x53dc6066,
        0x4d9b3063, 0x495a2dd4, 0x44190b0d, 0x40d816ba,
        0xaca5c697, 0xa864db20, 0xa527fdf9, 0xa1e6e04e,
        0xbfa1b04b, 0xbb60adfc, 0xb6238b25, 0xb2e29692,
        0x8aad2b2f, 0x8e6c3698, 0x832f1041, 0x87ee0df6,
        0x99a95df3, 0x9d684044, 0x902b669d, 0x94ea7b2a,
        0xe0b41de7, 0xe4750050, 0xe9362689, 0xedf73b3e,
        0xf3b06b3b, 0xf771768c, 0xfa325055, 0xfef34de2,
        0xc6bcf05f, 0xc27dede8, 0xcf3ecb31, 0xcbffd686,
        0xd5b88683, 0xd1799b34, 0xdc3abded, 0xd8fba05a,
        0x690ce0ee, 0x6dcdfd59, 0x608edb80, 0x644fc637,
        0x7a089632, 0x7ec98b85, 0x738aad5c, 0x774bb0eb,
        0x4f040d56, 0x4bc510e1, 0x46863638, 0x42472b8f,
        0x5c007b8a, 0x58c1663d, 0x558240e4, 0x51435d53,
        0x251d3b9e, 0x21dc2629, 0x2c9f00f0, 0x285e1d47,
        0x36194d42, 0x32d850f5, 0x3f9b762c, 0x3b5a6b9b,
        0x0315d626, 0x07d4cb91, 0x0a97ed48, 0x0e56f0ff,
        0x1011a0fa, 0x14d0bd4d, 0x19939b94, 0x1d528623,
        0xf12f560e, 0xf5ee4bb9, 0xf8ad6d60, 0xfc6c70d7,
        0xe22b20d2, 0xe6ea3d65, 0xeba91bbc, 0xef68060b,
        0xd727bbb6, 0xd3e6a601, 0xdea580d8, 0xda649d6f,
        0xc423cd6a, 0xc0e2d0dd, 0xcda1f604, 0xc960ebb3,
        0xbd3e8d7e, 0xb9ff90c9, 0xb4bcb610, 0xb07daba7,
        0xae3afba2, 0xaafbe615, 0xa7b8c0cc, 0xa379dd7b,
        0x9b3660c6, 0x9ff77d71, 0x92b45ba8, 0x9675461f,
        0x8832161a, 0x8cf30bad, 0x81b02d74, 0x857130c3,
        0x5d8a9099, 0x594b8d2e, 0x5408abf7, 0x50c9b640,
        0x4e8ee645, 0x4a4ffbf2, 0x470cdd2b, 0x43cdc09c,
        0x7b827d21, 0x7f436096, 0x7200464f, 0x76c15bf8,
        0x68860bfd, 0x6c47164a, 0x61043093, 0x65c52d24,
        0x119b4be9, 0x155a565e, 0x18197087, 0x1cd86d30,
        0x029f3d35, 0x065e2082, 0x0b1d065b, 0x0fdc1bec,
        0x3793a651, 0x3352bbe6, 0x3e119d3f, 0x3ad08088,
        0x2497d08d, 0x2056cd3a, 0x2d15ebe3, 0x29d4f654,
        0xc5a92679, 0xc1683bce, 0xcc2b1d17, 0xc8ea00a0,
        0xd6ad50a5, 0xd26c4d12, 0xdf2f6bcb, 0xdbee767c,
        0xe3a1cbc1, 0xe760d676, 0xea23f0af, 0xeee2ed18,
        0xf0a5bd1d, 0xf464a0aa, 0xf9278673, 0xfde69bc4,
        0x89b8fd09, 0x8d79e0be, 0x803ac667, 0x84fbdbd0,
        0x9abc8bd5, 0x9e7d9662, 0x933eb0bb, 0x97ffad0c,
        0xafb010b1, 0xab710d06, 0xa6322bdf, 0xa2f33668,
        0xbcb4666d, 0xb8757bda, 0xb5365d03, 0xb1f740b4
    };

    unsigned int i_crc = 0xffffffff;

    /* Calculate the CRC */
    while( i_count > 0 )
    {
        i_crc = (i_crc<<8) ^ CRC32[ (i_crc>>24) ^ ((unsigned int)*p_begin) ];
        p_begin++;
        i_count--;
    }

    return( i_crc );
}

// convert packet to a NULL packet
void ConvertToNULLPacket( unsigned char *buf )
{
	uchar *b;

	b = buf + 1;

	// 4 byte header for a NULL packet is 47 1f ff 10
	b[0] = 0x1f;
	b[1] = 0xff;
	b[2] = 0x10;

	// fill the rest with 0xff
	// 4 = sync_byte to continuity_counter
	b = buf + 4;
	for( int i = 0; i < 184; i++ )
		b[i] = 0xff;
}


// new channel scanning routing
// determines 16x9\4x3 frame transitions by analyzing I-frame macroblocks
int scan_for_commercials( TSFileList *inputList, int vidPid, __int64 minExcludeSize, __int64 responseSize,
						  float maxError, float scanRatio, FileReadBuffer *readBuffer, convert_callback callback,
						  void *data, HWND parentWnd, uint streamrate, int MaxSeekSize, bool bIncludeBlackFrames )
{
	int rVal = 0;
	char errMsgBuf[256];
	int errMsgVal;
	long timeDiscrepancy = 0;

	__int64 filePos = 0;
	__int64 totalRead = 0;
	__int64 lastIndicated = 0;

	TSFileInfo *fileInfo = inputList->GetHead();
	if (!fileInfo)
		return 1;

	PIDinfoGroup PIDs = { 1, { { vidPid, 0, STREAM_PES, PID_MPEG2_VIDEO, false, NULL, 0, 0, -1, false, false } } } ;

	PIDs.pids[0].buffer = new uchar[VIDEO_BUFFER_SIZE];
	PIDs.pids[0].bufferSize = VIDEO_BUFFER_SIZE;

	if (!PIDs.pids[0].buffer)
	{
		::MessageBox(parentWnd, "Out of memory", NULL, MB_OK | MB_ICONERROR);
		return 1;
	}

	TSRegionInfo *regionInfo = fileInfo->m_head;
	RegionType regionType = rt_include;

	__int64 cOffset = 0;
	__int64 originalTotalIncludedBytes = inputList->TotalIncludedBytes();
	__int64 IFrameOffset = 0;
	__int64 prevIFrameOffset = 0;
	__int64 totalBytes = inputList->TotalBytes();


	bool bIFrameFound = false;
	bool startOfFile = true;
	bool bSeekBack = false;
	bool needToSeek = true;

	unsigned char buf[188];

	int seek0 = 0; // by i-frame
	int seek10 = 10 * streamrate; // by 10 seconds
	int seek60 = 60 * streamrate; // by 60 seconds
	int seek = seek0; // initial seek is by i-frame

	int maxSeekSize = 0;
	if( MaxSeekSize == 10 )
		maxSeekSize = 10 * streamrate;
	else
		maxSeekSize = 60 * streamrate;

	int bytesReadInPacket = 0;

	frame_type frameType;
	macroblock_state_info msi;

	// assume we are going to start out by excluding stuff
	RegionType lastRegionType = rt_exclude; 

/*
	// sequence data
	sequence_buffers seq_bufs;
	seq_bufs.pSequenceHeader = NULL;
	seq_bufs.sh_length = 0;
	seq_bufs.pSequenceExtension = NULL;
//	seq_bufs.pSequenceDisplayExtension = NULL;
//	seq_bufs.sde_length = 0;
	seq_bufs.pSequenceScalableExtension = NULL;
	seq_bufs.sse_length = 0;
*/

// debug
//	uint frame_count = 0;
	
	for (;;)
	{
		// seeking back into previous file
		if( bSeekBack )
		{
			fileInfo = fileInfo->m_prev;
			regionInfo = NULL;
			readBuffer->Close();
			bSeekBack = false;
		}

		__int64 indicated = fileInfo->m_totalBytesPrior + filePos;
		if( callback && ( indicated > lastIndicated + 20000000 ))
		{
			lastIndicated = indicated;


			bool cont = callback( fileInfo->m_fileName,
								  NULL,
								  fileInfo->m_totalBytesPrior + filePos, // green bar
								  fileInfo->m_totalBytesPrior + filePos,
								  totalBytes,
								  -1,
								  timeDiscrepancy,
								  data );
            timeDiscrepancy = 0;

			// Cancelled?
			if (!cont)
				break;
		}

		// check regions, skip data, open new files
		if( regionInfo && ( filePos >= regionInfo->m_offset ))
		{
			regionType = regionInfo->m_type;
			regionInfo = regionInfo->m_next;
			lastRegionType = regionType;
		}


		// an Exclude region
		if( regionType == rt_exclude )
		{
			if( regionInfo )
			{
				filePos = regionInfo->m_offset;
				prevIFrameOffset = filePos;
				fileInfo->m_prevIFrameOffset_Info = fileInfo;
			}
			else
			{
				if( readBuffer->IsFileOpen() )
					readBuffer->Close();

				fileInfo = fileInfo->m_next;

				if( !fileInfo )
					break;	// we're complete!

				regionInfo = fileInfo->m_head;
				filePos = 0;
			}

			needToSeek = true;
			continue;
		}
		else
		{
			if (!readBuffer->IsFileOpen())
			{
				readBuffer->Open(fileInfo->m_fileName);

				while (!readBuffer->IsFileOpen())
				{
					sprintf(errMsgBuf, "Couldn't open input file\n%s", fileInfo->m_fileName);
					long tmp1, tmp2;
					time((time_t*)&tmp1);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_RETRYCANCEL | MB_ICONWARNING);
					time((time_t*)&tmp2);
					timeDiscrepancy += tmp2 - tmp1;
					if (errMsgVal == IDRETRY)
					{
						readBuffer->Open(fileInfo->m_fileName);
					}
					else
					{
						rVal = 1;
						break;	// where is this file??
					}
				}
				if (rVal)
					break;
				needToSeek = true;
			}
			if (needToSeek)
			{
				// check packet alignment of new file
				__int64 offset = (188 - ((filePos + 188 - fileInfo->m_firstPacketOffset) % 188)) % 188;
				if (offset != ((188 - bytesReadInPacket) % 188))
				{
					// alignment does not match, discard partial packet
					bytesReadInPacket = 0;
					filePos += offset;
				}

				readBuffer->Seek(filePos);
				needToSeek = false;
			}
		}

		// Try to read packet..
		int bytesRead;
		if ((bytesRead = readBuffer->Read(buf + bytesReadInPacket, 188 - bytesReadInPacket)) != 188 - bytesReadInPacket)
		{
			if (bytesRead < 0)
				bytesRead = 0;
			bytesReadInPacket += bytesRead;

			if (readBuffer->IsError())
			{
				// say something to the user, and exit!
				sprintf(errMsgBuf, "Error reading input file\n%s", fileInfo->m_fileName);
				::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
				rVal = 1;
				break;
			}

			// Close this file
			readBuffer->Close();

			fileInfo = fileInfo->m_next;

			if( !fileInfo )
				break;	// we're complete!

			regionInfo = fileInfo->m_head;
			filePos = 0;

			continue;
		}
		else
		{
			filePos += bytesRead;
			bytesReadInPacket = 0;
		}

		// Check sync byte
		if(( buf[0] == 0x47 ) || ( buf[0] == 0x72 ) || ( buf[0] == 0x29 ))
		{
			// Get pid
			int pid = ((( buf[1] & 0x1F ) << 8 ) | buf[2] ) & 0x1FFF;

			PIDinfo *p = find_PID_no_new(&PIDs, pid);
			if( p )
			{
				p->count++;

				// Get error
				bool errorBit = (buf[1] & 0x80) != 0;
				if( errorBit )
				{
					// invalidate accumulated content of this stream
					p->reading = false;
				}
				else
				{
					// Get adaptation header info
					int adaptLength = 0;
					int adaptation = ( buf[3] & 0x30 ) >> 4;
					if( adaptation != 0 )
					{
						if( adaptation == 0x2 )
							adaptLength = 184;
						else if( adaptation == 0x3 )
							adaptLength = buf[4] + 1;
						if( adaptLength <= 184 )
						{
							// Get payload start indicator
							bool start = ( buf[1] & 0x40 ) != 0;

							// process content of packet
							// assuming is a valid PES stream
							if( start )
							{
								if( p->reading && ( p->bufferPos > 0 ))
								{
									// only want to process B and P frames not I frames
									if( p->pType == PID_MPEG2_VIDEO )
									{
										// reset out macroblock state info before we loop throught the macroblocks
										memset( msi.macroblock_state, 0xffffffff, sizeof( msi.macroblock_state ));
										memset( msi.macroblock_state_middle, 0xffffffff, sizeof( msi.macroblock_state_middle ));

// debug
//frame_count++;
//_Error("IFrameOffset[%d], frame[%d]\n", IFrameOffset, frame_count );
//_Error("frame[%x] IFrameOffset[%x]", frame_count, IFrameOffset );
//if( frame_count == 0x40 )
//	int foo = 0;

										// process the B frame
										process_frame( p->buffer, p->bufferPos, NULL, &msi );

										// determine the frame type from the macroblock info
										determine_frame_type( &frameType, &msi, lastRegionType, bIncludeBlackFrames  );
// debug

//CString sFrameType;
//if( frameType == frame_4x3 )
//	sFrameType = "4x3";
//if( frameType == frame_16x9 )
//	sFrameType = "16x9";
//_Error(" type [%s]\n", sFrameType );


										// determine "Include"\"Exclude" and next jump point
										if( frameType == frame_4x3 )
										{	
											if( lastRegionType == rt_exclude )
											{
												if( seek == seek0 && startOfFile )
												{
													startOfFile = false;
													inputList->AddRegion( rt_exclude, fileInfo->m_totalBytesPrior );
													prevIFrameOffset = cOffset;
													fileInfo->m_prevIFrameOffset_Info = fileInfo;
													
													// are we going to seek passed the end of our files?
//													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek60 > totalBytes )
													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + maxSeekSize > totalBytes )
													{
														// if +60 was too much, try +10
														if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek10 > totalBytes )
														{
															seek = 0; // by I Frame
														}
														else
															seek = seek10; // +10 is OK
													}
													else
//														seek = seek60; // +60 is OK
														seek = maxSeekSize; // +60 is OK

													filePos = prevIFrameOffset + seek;
												}
												else
												{
													prevIFrameOffset = filePos;
													fileInfo->m_prevIFrameOffset_Info = fileInfo;

													// are we going to seek passed the end of our file list?
													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek > totalBytes )
													{
														if( seek == seek60 ) 
														{
															if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek10 > totalBytes )
																seek = seek0; 
															else
																seek = seek10; 
														}
														else if( seek == seek10 )
														{
															seek = seek0; 
														}

														//else if( seek == seek0 )
														//  do nothing...
													}

													filePos = prevIFrameOffset + seek;
												}
											}
											else // lastRegionType == rt_include 
											{
												// too far,
												// go back to the previous I Frame offset +10 sec
												//  set seek to +10
												if( seek == seek60 )
												{
													seek = seek10;
													filePos = prevIFrameOffset + seek10;

													if( fileInfo->m_prevIFrameOffset_Info != fileInfo )
														bSeekBack = true;
												}
												// too far,
												// go back to the previous I Frame offset 
												// set seek to I Frame seeking
												else if( seek == seek10 )
												{
													seek = seek0;
													filePos = prevIFrameOffset;

													if( fileInfo->m_prevIFrameOffset_Info != fileInfo )
														bSeekBack = true;
												}
												else if( seek == seek0 )
												{
													inputList->AddRegion( rt_exclude, fileInfo->m_totalBytesPrior + IFrameOffset );
													lastRegionType = rt_exclude;
													prevIFrameOffset = cOffset;
													fileInfo->m_prevIFrameOffset_Info = fileInfo;

//													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek60 > totalBytes )
													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + maxSeekSize > totalBytes )
													{
														if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek10 > totalBytes )
														{
															seek = 0;
														}
														else
															seek = seek10;
													}
													else
//														seek = seek60;
														seek = maxSeekSize;

													filePos = prevIFrameOffset + seek;
												}
											}
										}// end if(frame_type == frame_4x3)

										if( frameType == frame_16x9 )
										{
											if( lastRegionType == rt_exclude )
											{
												// went too far,
												// backup to previous I Frame +10 sec
												if( seek == seek60 )
												{
													seek = seek10;
													filePos = prevIFrameOffset + seek10;

													if( fileInfo->m_prevIFrameOffset_Info != fileInfo )
														bSeekBack = true;
												}
												// went too far,
												// backup to previous I Frame and seek foward by I Frame
												else if( seek == seek10 )
												{
													seek = seek0;
													filePos = prevIFrameOffset;

													if( fileInfo->m_prevIFrameOffset_Info != fileInfo )
														bSeekBack = true;
												}
												else if( seek == seek0 && startOfFile )
												{
													startOfFile = false;
													inputList->AddRegion( rt_include, fileInfo->m_totalBytesPrior + 0x0 );
													lastRegionType = rt_include;
													prevIFrameOffset = cOffset;
													fileInfo->m_prevIFrameOffset_Info = fileInfo;

//													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek60 > totalBytes )
													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + maxSeekSize > totalBytes )
													{
														if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek10 > totalBytes )
														{
															seek = 0;
														}
														else
															seek = seek10;
													}
													else
//														seek = seek60;
														seek = maxSeekSize;

													filePos = prevIFrameOffset + seek;
												}
												else if( seek == seek0 && !startOfFile )
												{
													inputList->AddRegion( rt_include, fileInfo->m_totalBytesPrior + IFrameOffset );
													lastRegionType = rt_include;
													prevIFrameOffset = filePos;
													fileInfo->m_prevIFrameOffset_Info = fileInfo;

//													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek60 > totalBytes )
													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + maxSeekSize > totalBytes )
													{
														if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek10 > totalBytes )
														{
															seek = 0;
														}
														else
															seek = seek10;
													}
													else
//														seek = seek60;
														seek = maxSeekSize;

													filePos = prevIFrameOffset + seek;
												}
											}
											else // lastRegionType == rt_include
											{
												if( seek == seek0 && startOfFile )
												{
													startOfFile = false;
													prevIFrameOffset = cOffset;
													fileInfo->m_prevIFrameOffset_Info = fileInfo;

//													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek60 > totalBytes )
													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + maxSeekSize > totalBytes )
													{
														if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek10 > totalBytes )
														{
															seek = 0;
														}
														else
															seek = seek10;
													}
													else
//														seek = seek60;
														seek = maxSeekSize;

													filePos = prevIFrameOffset + seek;
												}
												else if( seek == seek0 && !startOfFile )
												{	
													prevIFrameOffset = filePos;
													fileInfo->m_prevIFrameOffset_Info = fileInfo;
													filePos = prevIFrameOffset;
												}
												else
												{
													prevIFrameOffset = cOffset;
													fileInfo->m_prevIFrameOffset_Info = fileInfo;

													if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek > totalBytes )
													{
														if( seek == seek60 )
														{
															if( fileInfo->m_totalBytesPrior + prevIFrameOffset + seek10 > totalBytes )
																seek = seek0; 
															else
																seek = seek10;
														}
														else if( seek == seek10 )
														{
															seek = seek0;
														}
													}

													filePos = prevIFrameOffset + seek;
												}
											}
										}// end if(frameType == frame_16x9)
									}// end if( p->pType == PID_MPEG2_VIDEO )

									readBuffer->Seek( filePos );
									needToSeek = true;
									bIFrameFound = false;
									p->reading = false;
									p->confirmedRead = false;
									continue;

								}// end if( p->reading && ( p->bufferPos > 0 ))

								// start a new PES packet
								if( p->buffer && !p->processed )
								{
									p->reading = true;
									p->bufferPos = 0;
									p->confirmedRead = false;

									// cOffset is the start of a new PES packet
									cOffset = filePos - 188;
								}
							}
							if( p->reading )
							{
								// Add the payload for this packet to the current buffer
								int payloadLength = 184 - adaptLength;
								if( payloadLength > 0 )
								{
									if(( p->bufferPos + payloadLength ) <= p->bufferSize )
									{
										memcpy( p->buffer + p->bufferPos, buf + 4 + adaptLength, payloadLength );
										p->bufferPos += payloadLength;
									}
									else
									{
										// buffer overflow!
										_Error( "Error: buffer overflow! PID=0x%02X, PES packet, buffer size = %d\n", pid, p->bufferSize );
										p->reading = false;
									}
								}

								// only want to continue reading this PES packet if it's an i-frame
								// check for a sequence_header_code, if we have it were good to go
								// check for a group_start_code, if we have it were good to go
								// check for a picture_start_code, if it's an i-frame, add a
								//  sequence_header to the beginning of the buffer

								if( !p->confirmedRead && ( p->bufferPos > 512 ))
								{
									uint start_code = 0;
									for( int i = 0; i < p->bufferPos; i++ )
									{
										start_code = ( start_code << 8 ) | p->buffer[i];

/* 
										//  sequence_header_code
										if( start_code == 0x000001b3 ) 
										{  
											// save the sequence_header info.
											//  we will stuff this info into the beginning of the buffer if
											//  we find an I frame by picture_start_code instead of by
											//  sequence_header_code or group_start_code
											if( !seq_bufs.pSequenceHeader )// only need to do this once
												seq_bufs.pSequenceHeader = new unsigned char[184]; // ~ a data paket

											// copy just the sequence_header info
											seq_bufs.sh_length = GetSHLength( p->buffer, i );
											memcpy( seq_bufs.pSequenceHeader, &p->buffer[i-3], seq_bufs.sh_length );
										}// end if( start_code == 0x000001b3 )

										// sequence_extension_code
										if( start_code == 0x000001b5 )
										{
											// pg 46 Table 6-2.
											// only want Sequence Extension ID
											// sequence extension ID		  = 0x0001
											// sequence display extension ID  = 0x0010
											// sequence scalable extension ID = 0x0101

											// save the sequence extension ID 
											//  we will stuff this info into the beginning of the buffer if
											//  we find an I frame by picture_start_code instead of by
											//  sequence_header_code or group_start_code
											if(( p->buffer[i+1] >> 4 ) == 0x1 )
											{
												if( !seq_bufs.pSequenceExtension ) // only need to do this once
													seq_bufs.pSequenceExtension = new unsigned char[184]; // ~ a data packet

												// copy just the sequence_externsion info
												memcpy( seq_bufs.pSequenceExtension, &p->buffer[i-3], 9 );
											}
										}// end if( start_code == 0x000001b5 )
*/

										// sequence_header_code || group_start_code
										if( start_code == 0x000001b3 || start_code == 0x000001b8  )
										{
											// this is all we need to continue
											//  next frame will be an I-Frame
											p->confirmedRead = true;
											IFrameOffset = cOffset;
											break;
										}

/* if using picture_start_code to trigger an I-Frame hit then will need to check that a
   a sequence_extension and sequence_header were found first before being able to use it.
										// picture_start_code
										if( start_code == 0x00000100 )
										{	// check if picture_coding_type == 1 (I-Frame)
											if(( p->buffer[i + 2] & ( 0x7 << 3 )) == ( 1 << 3 ))
											{
												// NOTE: order is important here.  the sequence_header
												//  must be appear in the buffer before the
												//  sequence_extension 
												
												// add a sequence_extension to the beginning of the buffer
												if( seq_bufs.pSequenceExtension )
												{
													unsigned char *pTemp = new unsigned char[ p->bufferPos ];
													memcpy( pTemp, p->buffer, p->bufferPos );
													memcpy( p->buffer, seq_bufs.pSequenceExtension, 9 );
													memcpy( p->buffer + 9, pTemp, p->bufferPos );
													p->bufferPos += 9;

													delete pTemp;
												}

												// add a sequence_header to the beginning of the buffer
												if( seq_bufs.pSequenceHeader )
												{
													unsigned char *pTemp = new unsigned char[ p->bufferPos ];
													memcpy( pTemp, p->buffer, p->bufferPos );
													memcpy( p->buffer, seq_bufs.pSequenceHeader, seq_bufs.sh_length );
													memcpy( p->buffer + seq_bufs.sh_length, pTemp, p->bufferPos );
													p->bufferPos += seq_bufs.sh_length;

													delete pTemp;
												}
												
												p->confirmedRead = true;
												IFrameOffset = cOffset;
												break;
											}// end if(( p->buffer[i + 2] & ( 0x7 << 3 )) == ( 1 << 3 ))
										}// end	if( start_code == 0x00000100 ) // picture_start_code
*/
									}// end for( int i = 0; i < p->bufferPos; i++ )

									if( !p->confirmedRead )
										p->reading = false;

								}// end if( !p->confirmedRead && ( p->bufferPos > 512 ))
							}// end if( p->reading )
						}// end if( adaptLength <= 184 )
					}// end if( adaptation != 0 )
				}// end else for if( errorBit )
			}// end if( p )
		}// end if(( buf[0] == 0x47 ) || ( buf[0] == 0x72 ) || ( buf[0] == 0x29 ))
	}// end for (;;)

	// Close files
	if (readBuffer->IsFileOpen())
		readBuffer->Close();

	if (PIDs.pids[0].buffer)
		delete []PIDs.pids[0].buffer;
/*
	if( seq_bufs.pSequenceHeader )
	{
		delete []seq_bufs.pSequenceHeader;
		seq_bufs.pSequenceHeader = NULL;
	}

	if( seq_bufs.pSequenceExtension )
	{
		delete []seq_bufs.pSequenceExtension;
		seq_bufs.pSequenceExtension = NULL;
	}


	if( seq_bufs.pSequenceScalableExtension )
	{
		delete []seq_bufs.pSequenceScalableExtension;
		seq_bufs.pSequenceScalableExtension = NULL;
	}

	if( seq_bufs.pSequenceDisplayExtension )
	{
		delete []seq_bufs.pSequenceDisplayExtension;
		seq_bufs.pSequenceDisplayExtension = NULL;
	}
*/

	return rVal;
}

#define max_1920_column 0xb // 11 wide
#define max_1280_column  0x9 // 9 wide

// determine the frame type based on 4x3 bars
void determine_frame_type( frame_type *frameType, macroblock_state_info *msi, RegionType lastRegionType, bool bIncludeBlackFrames )
{
	int left_column_start = -1;
	int left_column_end = -1;
	int right_column_start = -1;
	int right_column_end = -1;
	int max_column = -1;
	int bit_shift = 16;
	bool _continue_ = false;

	*frameType = frame_NONE;

	if( msi->horizontal_size == 0x780 )
	{
		max_column = max_1920_column;
		// 19 - skip 3 columns at 4x3 boundary to avoid macroblock "bleeding"
		bit_shift = 19;
	}
	else if( msi->horizontal_size == 0x500 )
	{
		max_column = max_1280_column;
		// not necessary yet with 1280x720 stuff
//		bit_shift = ??; 			
	}

	else
		ASSERT( true ); // !

	// seach for a left column
	for( int i = 0; i < max_column; i++ )
	{
		_continue_ = false;

		// didn't find a start column within 5 columns OR were done processing
		if( i == 5 && left_column_start == -1 || left_column_end != -1 )
			break;

		// ignore the first and last row. sometimes its not "clean". i.e. FOX
		for( uint row = 1; row < msi->number_of_frame_rows; row++ )
		{
			// if 1 then macroblock was not black
			if( msi->macroblock_state[row] >> i & 0x1 )
			{
				if( left_column_start == -1 )
				{
					_continue_ = true;
					break;
				}
				else
				{
					left_column_end = i;
					break;
				}
			}
		}

		if( _continue_ )
			continue;

		// made it to the end of the column consisting of all black macroblocks, mark as a start
		if( left_column_start == -1 )
			left_column_start = i;

		// we reached our max column length so this column is our end
		if( i == max_column - 1 && left_column_end == -1 )
			left_column_end = i + 1;
	}

	// search for a right column
	for( uint i = msi->number_of_frame_columns - max_column - 1; i < msi->number_of_frame_columns; i++ )
	{
		_continue_ = false;

		// didn't find a start column within 5 columns OR were are done processing
		if( i == msi->number_of_frame_columns - max_column + 5 && right_column_start == -1 || right_column_end != -1 )
			break;

		// ignore the first row
		for( uint row = 1; row < msi->number_of_frame_rows; row++ )
		{
			// if 1 then macroblock was not black
			if(( msi->macroblock_state[row] >> (bit_shift + i -( msi->number_of_frame_columns - max_column - 1 )) & 0x1 ))
			{
				if( right_column_start == -1 )
				{
					_continue_ = true;
					break;
				}
				else
				{
					right_column_end = i;
					break;
				}
			}
		}

		if( _continue_ )
			continue;

		// made it to the end of the column consisting of all black macroblocks, mark as a start
		if( right_column_start == -1 )
			right_column_start = i;
		
		// we reached our max column length so this column is our end
		if( i == msi->number_of_frame_columns - 1 && right_column_end == -1 )
			right_column_end = i;
	}

	int min_column_size = 0;
	if( msi->horizontal_size == 0x780 )
		min_column_size = 7;
	else
		min_column_size = 4;

	// must have a left and right column in order to be considered a 4x3 frame
	if( left_column_start != -1 && left_column_end != -1 && right_column_start != -1 && right_column_end != -1 )
	{	// and columns must not be smaller than min_column_size
		if( left_column_end - left_column_start >= min_column_size && right_column_end - right_column_start >= min_column_size )
			*frameType = frame_4x3;
		else
			*frameType = frame_16x9;
	}
	else // must be a 16x9 frame
		*frameType = frame_16x9;

	if( bIncludeBlackFrames )
	{	// check for a 16x9 black frame
		if( *frameType == frame_4x3 )
		{
			*frameType = frame_16x9;

			for( int i = 0; i < 32; i++ )
			{
				// ignore the first row. sometimes its not "clean". i.e. FOX
				for( uint row = 1; row < msi->number_of_frame_rows; row++ )
				{
					// if 1 then macroblock was not black
					if( msi->macroblock_state_middle[row] >> i & 0x1 )
						*frameType = frame_4x3;

					if( *frameType == frame_4x3 )
						break;
				}

				if( *frameType == frame_4x3 )
					break;
			}// end for( int i = 0; i < 33; i++ )
		}// end	if( *frameType == frame_4x3 )
	}// end if( bIncludeBlackFrames )

	// blew it
	ASSERT( *frameType == frame_4x3 || *frameType == frame_16x9 );
}

// return the length of the sequence header
// NOT GENERIC!!
// Funtion call for scan_for_commercials() only!
int GetSHLength( uchar *buffer, int bufferPos  )
{
	int length = 0;

	// sequence_header_code to vbv_buffer_size_value
	length += 12;
	bufferPos += 8;

	// load_intra_quantiser_maxtix
	if( buffer[bufferPos] & 0x02 )
	{
		length += 64;
		bufferPos += 64;
	}

	// load_non_intra_quantiser_matrix
	if( buffer[bufferPos] & 0x01 )
	{
		length += 64;
		bufferPos += 64;
	}

	return length;
}

/*
// return the length of the sequence scalable extension
// NOT GENERIC!!
// Funtion call for scan_for_commercials() only!
int GetSSELength( uchar *buffer, int bufferPos )
{
	int length = 0;

	// extension_start_code_indentifier
	ASSERT( true );

	return length;
}
*/


// process the I-frame and determine if there are left and right "bars" in the frame.
// this would indicate a 4x3 frame
void process_frame( uchar *buf, int bufSize, sequence_header *sh, macroblock_state_info *msi )
{
	// loop through video slices
	bool foundIFrame = false;
	sequence_header m_sh;
	picture_header m_ph;

	if( !sh )
		sh = &m_sh;
	picture_header *ph = &m_ph;

	memset( sh, 0, sizeof( sequence_header ));
	memset( ph, 0, sizeof( picture_header ));

	int bufOffset = 0;
	set_buf( buf, 0, false );

	while( next_start_code( buf, bufOffset, bufSize ))
	{
		// found a start code
		uint stream_id = get_bits( 8 );
		
		switch( stream_id )
		{

			case 0x00:	// picture_start_code
			{
#ifdef SHOW_INFO_MPEG2_VIDEO
				_Error("\t0x00 picture_start_code, %d\n", bufOffset-3);
uint64 time_start = rdtsc();
#endif
				ph->start_code_identifier = 0x00000100;


				ph->temporal_reference = get_bits(10);
				ph->picture_coding_type = get_bits(3);
				ph->vbv_delay = get_bits(16);

				if ((ph->picture_coding_type == 2) || (ph->picture_coding_type == 3))
				{
					ph->full_pel_forward_vector = get_bits(1);
					ph->forward_f_code = get_bits(3);
				}
				if (ph->picture_coding_type == 3)
				{
					ph->full_pel_backward_vector = get_bits(1);
					ph->backward_f_code = get_bits(3);
				}
				while (get_bits(1) == 1)
				{
					discard_bits(8);	//extra_information_picture
				}

				bool searchForExtensions = true;
				while( searchForExtensions &&
					   next_start_code(buf, bufOffset, bufSize))
				{
					if (get_bits(8) == 0xb5)	// extension_start_code
					{
						// process sequence_extension
						uint extension_start_code_identifier = get_bits(4);
						switch (extension_start_code_identifier)
						{
						case 8:	// Picture Coding Extension ID
							ph->picture_coding_extension = extension_start_code_identifier;
							ph->f_code[0][0] = get_bits(4);
							ph->f_code[0][1] = get_bits(4);
							ph->f_code[1][0] = get_bits(4);
							ph->f_code[1][1] = get_bits(4);
							ph->intra_dc_precision = get_bits(2);
							ph->picture_structure = get_bits(2);
							ph->top_field_first = get_bits(1);
							ph->frame_pred_frame_dct = get_bits(1);
							ph->concealment_motion_vectors = get_bits(1);
							ph->q_scale_type = get_bits(1);
							ph->intra_vlc_format = get_bits(1);
							ph->alternate_scan = get_bits(1);
							ph->repeat_first_field = get_bits(1);
							ph->chroma_420_type = get_bits(1);
							ph->progressive_frame = get_bits(1);
							ph->composite_display_flag = get_bits(1);
							if (ph->composite_display_flag)
							{
								ph->v_axis = get_bits(1);
								ph->field_sequence = get_bits(3);
								ph->sub_carrier = get_bits(1);
								ph->burst_amplitude = get_bits(7);
								ph->sub_carrier_phase = get_bits(8);
							}
							break;

						case 3: // Quant Matrix Extension ID
							ph->quant_matrix_extension = extension_start_code_identifier;
							ph->load_intra_quantiser_matrix = get_bits(1);
							if (ph->load_intra_quantiser_matrix)
							{
								for (int i = 0; i < 64; i++)
									ph->intra_quantiser_matrix[i] = get_bits(8);
							}
							ph->load_non_intra_quantiser_matrix = get_bits(1);
							if (ph->load_non_intra_quantiser_matrix)
							{
								for (int i = 0; i < 64; i++)
									ph->non_intra_quantiser_matrix[i] = get_bits(8);
							}
							ph->load_chroma_intra_quantiser_matrix = get_bits(1);
							if (ph->load_chroma_intra_quantiser_matrix)
							{
								for (int i = 0; i < 64; i++)
									ph->chroma_intra_quantiser_matrix[i] = get_bits(8);
							}
							ph->load_chroma_non_intra_quantiser_matrix = get_bits(1);
							if (ph->load_chroma_non_intra_quantiser_matrix)
							{
								for (int i = 0; i < 64; i++)
									ph->chroma_non_intra_quantiser_matrix[i] = get_bits(8);
							}
							break;

						case 9: // Picture Spatial Scalable Extension ID
							ph->picture_spatial_scalable_extension = extension_start_code_identifier;
							ph->lower_layer_temporal_reference = get_bits(10);
							discard_bits(1);
							ph->lower_layer_horizontal_offset = get_bits(15);
							if (ph->lower_layer_horizontal_offset & 0x4000)	// sign extend
								ph->lower_layer_horizontal_offset |= 0xffff8000;
							discard_bits(1);
							ph->lower_layer_vertical_offset = get_bits(15);
							if (ph->lower_layer_vertical_offset & 0x4000)	// sign extend
								ph->lower_layer_vertical_offset |= 0xffff8000;
							ph->spatial_temporal_weight_code_table_index = get_bits(2);
							ph->lower_layer_progressive_frame = get_bits(1);
							ph->lower_layer_deinterlaced_field_select = get_bits(1);
							break;

						case 10: // Picture Temporal Scalable Extension ID
							ph->picture_temporal_scalable_extension = extension_start_code_identifier;
							ph->reference_select_code = get_bits(2);
							ph->forward_temporal_reference = get_bits(10);
							discard_bits(1);
							ph->backward_temporal_reference = get_bits(10);
							break;

						default:
							searchForExtensions = false;
							// back up to start of start code for someone else to process
							bufOffset -= 4;
							set_buf(buf + bufOffset, 0, false);
							break;
						}
					}// end if( get_bits(8) == 0xb5 ) // extension_start_code 
					else
					{
						searchForExtensions = false;
						// back up to start of start code for someone else to process
						bufOffset -= 4;
						set_buf(buf + bufOffset, 0, false);
					}
				}// end while( searchForExtensions && next_start_code( buf, bufOffset, bufSize ))

				if( ph->picture_coding_type == 1 )
					foundIFrame = true;

#ifdef SHOW_INFO_MPEG2_VIDEO
uint64 time_end = rdtsc();

				// info display
				_Error("\t{\ttemporal_reference=%d\n", ph->temporal_reference);
				char *s;
				switch (ph->picture_coding_type)
				{
				case 1: s = "I"; break;
				case 2: s = "P"; break;
				case 3: s = "B"; break;
				default: s = "?"; break;
				}
				_Error("\t\tpicture_coding_type=%d, %s frame\n", (uint)ph->picture_coding_type, s);
				_Error("\t\tvbv_delay=%d\n", ph->vbv_delay);
				if ((ph->picture_coding_type == 2) || (ph->picture_coding_type == 3))
				{
					_Error("\t\tfull_pel_forward_vector=%d\n", ph->full_pel_forward_vector);
					_Error("\t\tforward_f_code=%d\n", (uint)ph->forward_f_code);
				}
				if (ph->picture_coding_type == 3)
				{
					_Error("\t\tfull_pel_backward_vector=%d\n", ph->full_pel_backward_vector);
					_Error("\t\tbackward_f_code=%d\n", (uint)ph->backward_f_code);
				}

				if (ph->picture_coding_extension)
				{
					_Error("\t\tf_code[][]={{%d,%d},{%d,%d}}\n", (uint)ph->f_code[0][0],
							(uint)ph->f_code[0][1], (uint)ph->f_code[1][0], (uint)ph->f_code[1][1]);
					_Error("\t\tintra_dc_precision=%d, %d bits\n", (uint)ph->intra_dc_precision, 8+(uint)ph->intra_dc_precision);
					switch (ph->picture_structure)
					{
					case 1: s = "top field"; break;
					case 2: s = "bottom field"; break;
					case 3: s = "frame picture"; break;
					default: s = "?"; break;
					}
					_Error("\t\tpicture_structure=%d, %s\n", (uint)ph->picture_structure, s);
					_Error("\t\ttop_field_first=%d\n", ph->top_field_first);
					_Error("\t\tframe_pred_frame_dct=%d\n", ph->frame_pred_frame_dct);
					_Error("\t\tconcealment_motion_vectors=%d\n", ph->concealment_motion_vectors);
					_Error("\t\tq_scale_type=%d\n", ph->q_scale_type);
					_Error("\t\tintra_vlc_format=%d\n", ph->intra_vlc_format);
					_Error("\t\talternate_scan=%d\n", ph->alternate_scan);
					_Error("\t\trepeat_first_field=%d\n", ph->repeat_first_field);
					_Error("\t\tchroma_420_type=%d\n", ph->chroma_420_type);
					_Error("\t\tprogressive_frame=%d\n", ph->progressive_frame);
					_Error("\t\tcomposite_display_flag=%d\n", ph->composite_display_flag);
					if (ph->composite_display_flag)
					{
						_Error("\t\tv_axis=%d\n", ph->v_axis);
						_Error("\t\tfield_sequence=%d\n", (uint)ph->field_sequence);
						_Error("\t\tsub_carrier=%d\n", ph->sub_carrier);
						_Error("\t\tburst_amplitude=%d\n", (uint)ph->burst_amplitude);
						_Error("\t\tsub_carrier_phase=%d\n", (uint)ph->sub_carrier_phase);
					}
				}

				if (ph->quant_matrix_extension)
				{
					_Error("\t\tload_intra_quantiser_matrix=%d\n", ph->load_intra_quantiser_matrix);
					if (ph->load_intra_quantiser_matrix)
					{
						for (int i = 0; i < 64; i++)
						{
							if ((i%32) == 0)
								_Error("\t\t\t");
							_Error("%02X ", (uint)ph->intra_quantiser_matrix[i]);
							if ((i%32) == 31)
								_Error("\n");
						}
					}
					_Error("\t\tload_non_intra_quantiser_matrix=%d\n", ph->load_non_intra_quantiser_matrix);
					if (ph->load_non_intra_quantiser_matrix)
					{
						for (int i = 0; i < 64; i++)
						{
							if ((i%32) == 0)
								_Error("\t\t\t");
							_Error("%02X ", (uint)ph->non_intra_quantiser_matrix[i]);
							if ((i%32) == 31)
								_Error("\n");
						}
					}
					_Error("\t\tload_chroma_intra_quantiser_matrix=%d\n", ph->load_chroma_intra_quantiser_matrix);
					if (ph->load_chroma_intra_quantiser_matrix)
					{
						for (int i = 0; i < 64; i++)
						{
							if ((i%32) == 0)
								_Error("\t\t\t");
							_Error("%02X ", (uint)ph->chroma_intra_quantiser_matrix[i]);
							if ((i%32) == 31)
								_Error("\n");
						}
					}
					_Error("\t\tload_chroma_non_intra_quantiser_matrix=%d\n", ph->load_chroma_non_intra_quantiser_matrix);
					if (ph->load_chroma_non_intra_quantiser_matrix)
					{
						for (int i = 0; i < 64; i++)
						{
							if ((i%32) == 0)
								_Error("\t\t\t");
							_Error("%02X ", (uint)ph->chroma_non_intra_quantiser_matrix[i]);
							if ((i%32) == 31)
								_Error("\n");
						}
					}
				}// end if( ph->quant_matrix_extension )

_Error("\t\ttime_taken=%I64d\n",time_end-time_start);

//
#ifdef TESTING_BITS
uint val1, val2, val3, val4, v;

time_start = rdtsc();

for (int i=1000; i>0; i--)
{
	set_buf(buf + bufOffset, 0, false);
	v = 0;

	for (int j=10000; j>0; j--)
	{
#if 1
		val1 = get_bits(3);
		val2 = get_bits(15);
		val3 = get_bits(9);
		val4 = get_bits(2);
		v += val1 + val2 + val3 + val4;
		val1 = get_bits(7);
		val2 = get_bits(23);
		val3 = get_bits(13);
		val4 = get_bits(4);
		v += val1 + val2 + val3 + val4;
		val1 = get_bits(1);
		val2 = get_bits(11);
		val3 = get_bits(5);
		val4 = get_bits(8);
		v += val1 + val2 + val3 + val4;
		val2 = get_bits(6);
		v += val2;
#else
		discard_bits(3);
		discard_bits(15);
		discard_bits(9);
		discard_bits(2);
		discard_bits(7);
		discard_bits(23);
		discard_bits(13);
		discard_bits(4);
		discard_bits(1);
		discard_bits(11);
		discard_bits(5);
		discard_bits(8);
		discard_bits(6);
#endif
	}
}// end for( int i=1000; i>0; i-- )

time_end = rdtsc();
_Error("\t\tTEST time=%I64d\n",time_end-time_start, v);
#endif
//

				_Error("\t}\n");
				Sleep(100);
#endif
			}

			break;

			case 0xb3:	// sequence_header_code
			{
				sh->start_code_identifier = 0x000001b3;

				sh->horizontal_size_value = get_bits(12);
				sh->vertical_size_value = get_bits(12);
				sh->aspect_ratio_information = get_bits(4);
				sh->frame_rate_code = get_bits(4);
				sh->bit_rate_value = get_bits(18);
				discard_bits(1);
				sh->vbv_buffer_size_value = get_bits(10);
				sh->constrained_parameters_flag = get_bits(1);
				sh->load_intra_quantiser_matrix = get_bits(1);
				sh->intra_quantiser_matrix[64];

				if( sh->load_intra_quantiser_matrix ) 
				{
					for( int i = 0; i < 64; i++ )
						sh->intra_quantiser_matrix[i] = get_bits(8);
				}

				sh->load_non_intra_quantiser_matrix = get_bits(1);
				sh->non_intra_quantiser_matrix[64];

				if( sh->load_non_intra_quantiser_matrix )
				{
					for (int i = 0; i < 64; i++)
						sh->non_intra_quantiser_matrix[i] = get_bits(8);
				}

// S - commercial scan info
				if( sh->vertical_size_value == 0x438 )	// 1080
					msi->vertical_size = 0x440;			// 1088, evenly divisible by 16
				else
					msi->vertical_size = sh->vertical_size_value;

				msi->horizontal_size = sh->horizontal_size_value;

				if( msi->vertical_size == 0x440 )
					msi->number_of_frame_rows = ( msi->vertical_size / 16 ) - 1;
				else
					msi->number_of_frame_rows = msi->vertical_size / 16;

				msi->number_of_frame_columns = msi->horizontal_size / 16;
// E - commercial scan info

				bool searchForExtensions = true;
				while( searchForExtensions &&
					   next_start_code( buf, bufOffset, bufSize ))
				{
					if( get_bits(8) == 0xb5 ) // extension_start_code
					{
						// process sequence_extension
						uint extension_start_code_identifier = get_bits(4);
						switch (extension_start_code_identifier)
						{
							case 1:	// sequence extension ID
							{
								sh->sequence_extension = extension_start_code_identifier;

								sh->profile_and_level_indication = get_bits(8);
								sh->progressive_sequence = get_bits(1);
								sh->chroma_format = get_bits(2);
								sh->horizontal_size_extension = get_bits(2);
								sh->vertical_size_extension = get_bits(2);
								sh->bit_rate_extension = get_bits(12);
								discard_bits(1);
								sh->vbv_buffer_size_extension = get_bits(8);
								sh->low_delay = get_bits(1);
								sh->frame_rate_extension_n = get_bits(2);
								sh->frame_rate_extension_d = get_bits(5);

								// apply extension
								sh->horizontal_size_value |= sh->horizontal_size_extension << 12;
								sh->vertical_size_value |= sh->vertical_size_extension << 12;
								sh->bit_rate_value |= sh->bit_rate_extension << 18;
								sh->vbv_buffer_size_value |= sh->vbv_buffer_size_extension << 10;
								break;
							}// end case 1:

							case 5:	// Sequence Scalable Extension ID
							{
								sh->sequence_scalable_extension = extension_start_code_identifier;

								sh->scalable_mode = get_bits(2);
								sh->layer_id  = get_bits(4);

								if (sh->scalable_mode == 1)		// spatial scalability
								{
									sh->lower_layer_prediction_horizontal_size = get_bits(14);
									discard_bits(1);
									sh->lower_layer_prediction_vertical_size = get_bits(14);
									sh->horizontal_subsampling_factor_m = get_bits(5);
									sh->horizontal_subsampling_factor_n = get_bits(5);
									sh->vertical_subsampling_factor_m = get_bits(5);
									sh->vertical_subsampling_factor_n = get_bits(5);
								}
								if (sh->scalable_mode == 3)		// temporal scalability
								{
									sh->picture_mux_enable = get_bits(1);
									if (sh->picture_mux_enable)
										sh->mux_to_progressive_sequence = get_bits(1);
									sh->picture_mux_order = get_bits(3);
									sh->picture_mux_factor = get_bits(3);
								}
								break;
							}// end case 5:

							default:
							{
								searchForExtensions = false;
								// backup to start of start code for some else to process
								bufOffset -= 4;
								set_buf( buf + bufOffset, 0, false );
								break;
							}// end default:
						}// end switch (extension_start_code_identifier)
					}// end if( get_bits(8) == 0xb5 ) // extension_start_code

					else 
					{
						searchForExtensions = false;
						// back up to start of start code for someone else to process
						bufOffset -=4;
						set_buf( buf + bufOffset, 0, false );
					}// end else for if( get_bits(8) == 0xb5 ) // extension_start_code

				}// end while( searchForExtensions && next_start_code( buf, bufOffset, bufSize ))
			}// end case 0xb3:	// sequence_header_code

			default:
			{
				if( stream_id < 0xb0 ) // only want to process I frames
				{
					//	slice_start_code 0x01 through 0xaf ( page 70 )
					if( foundIFrame && (sh->start_code_identifier == 0x000001b3 ) && ( ph->start_code_identifier == 0x00000100 ))
					{
						// process this slice
						uint mb_address = 0; 
						uint rows = 0;

						process_slice( stream_id, false, NULL, sh, ph, &mb_address, &rows, msi );
					}
				}// end if( stream_id < 0x0b )
			}// end default:
			break;
		}// end switch( stream_id )
	}// end while( next_start_code( buf, bufOffset, bufSize )) 
}

// process the video frame and determine the macroblock data
int process_slice( uint slice_vertical_position, bool fullSize, bitmap *bm, sequence_header *sh, picture_header *ph, uint *mb_address, uint *rows, macroblock_state_info *msi )
{
	uchar pattern_code[12];
	uint slice_vertical_position_extension = 0;
	uint priority_breakpoint = 0;
	uchar quantiser_scale_code = 0;
	uint intra_slice = 0;

	int dc_dct_pred[3];
	dc_dct_pred[0] = 128 << ph->intra_dc_precision;
	dc_dct_pred[1] = 128 << ph->intra_dc_precision;
	dc_dct_pred[2] = 128 << ph->intra_dc_precision;

	if (sh->vertical_size_value > 2800)
	{
		slice_vertical_position_extension = get_bits(3);
	}
	if ((sh->sequence_scalable_extension == 5) &&
		(sh->scalable_mode == 0))
	{
		priority_breakpoint = get_bits(7);
	}
	quantiser_scale_code = get_bits(5);
	if (get_bits(1) == 1)
	{
		intra_slice = get_bits(1);
		discard_bits(7);
		while (get_bits(1) == 1)
		{
			discard_bits(8);	// extra_information_slice
		}
	}

	uint mb_row = (slice_vertical_position_extension << 7) + slice_vertical_position - 1;
	uint mb_width = sh->horizontal_size_value >> 4;
	uint macroblock_address = (mb_row * mb_width) - 1;

//_Error("ps %-2d:\n", slice_vertical_position );

#ifdef REPORT_SLICE_DECODE_FAILURES
	uint start_mb_add = macroblock_address + 1;
#endif

	// loop through macroblocks
	do
	{
		// address
		uint inc = 0;
		while (peek_bits(11) == 0x008)		// macroblock_escape
		{
			discard_bits(11);
			inc += 33;
		}
		uint inc2;
		if (!decode_macroblock_address_increment(&inc2))
		{
#ifdef REPORT_SLICE_DECODE_FAILURES
			_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
			return false;
		}
		inc += inc2;
		macroblock_address += inc;

		uint mb_column = macroblock_address % mb_width;

		// modes	(macroblock_modes()) page 72, tables B-2 to B-8
		macroblock_type mbt;
		memset(&mbt, 0, sizeof(mbt));

		if (!decode_macroblock_type(&mbt, ph->picture_coding_type, sh->scalable_mode))
		{
#ifdef REPORT_SLICE_DECODE_FAILURES
			_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
			return false;
		}
		if (mbt.spatial_temporal_weight_code_flag &&
			(ph->spatial_temporal_weight_code_table_index != 0))
		{
			mbt.spatial_temporal_weight_code = get_bits(2);
		}
		if (mbt.macroblock_motion_forward ||
			mbt.macroblock_motion_backward)
		{
			if (ph->picture_structure == 3)	// frame
			{
				if (ph->frame_pred_frame_dct == 0)
					mbt.frame_motion_type = get_bits(2);
			}
			else
			{
				mbt.field_motion_type = get_bits(2);
			}
		}
		if ((ph->picture_structure == 3) &&		// frame
			(ph->frame_pred_frame_dct == 0) &&
			(mbt.macroblock_intra || mbt.macroblock_pattern))
		{
			mbt.dct_type = get_bits(1);
		}

		// reset predictors if a macroblock is skipped, or is non-intra
		if (!mbt.macroblock_intra || (inc > 1))
		{
			dc_dct_pred[0] = 128 << ph->intra_dc_precision;
			dc_dct_pred[1] = 128 << ph->intra_dc_precision;
			dc_dct_pred[2] = 128 << ph->intra_dc_precision;
		}

		// stuff
		if (mbt.macroblock_quant)
			quantiser_scale_code = get_bits(5);

		if (mbt.macroblock_motion_forward ||
			mbt.macroblock_motion_backward ||
			(mbt.macroblock_intra && ph->concealment_motion_vectors))
		{
			// fill in derived motion vector info
			motion_type_table_entry *mtte;
			if (ph->picture_structure == 3)	// frame
			{
				if (mbt.macroblock_intra && ph->concealment_motion_vectors)
					mbt.frame_motion_type = 2;
				// table 6-17 based on frame_motion_type
				// note: not properly looking up values for other than frame_motion_type == 2
				mtte = &table_6_17[mbt.frame_motion_type];
			}
			else
			{
				if (mbt.macroblock_intra && ph->concealment_motion_vectors)
					mbt.field_motion_type = 2;
				// table 6-18 based on field_motion_type
				mtte = &table_6_18[mbt.field_motion_type - 1];
			}

			mbt.prediction_type = mtte->prediction_type;
			mbt.motion_vector_count = mtte->motion_vector_count;
			mbt.mv_format = mtte->mv_format;
			mbt.dmv = mtte->dmv;
		}

		if (mbt.macroblock_motion_forward ||
			(mbt.macroblock_intra && ph->concealment_motion_vectors) )
		{
			// see 7.6.3, page 92, for motion vector use
			// 7.6.3.9 for concealment_motion_vectors specifically
			if (!motion_vectors(0, ph, &mbt))
			{
#ifdef REPORT_SLICE_DECODE_FAILURES
				_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
				return false;
			}
		}
		if (mbt.macroblock_motion_backward)
		{
			if (!motion_vectors(1, ph, &mbt))
			{
#ifdef REPORT_SLICE_DECODE_FAILURES
				_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
				return false;
			}
		}
		if (mbt.macroblock_intra && ph->concealment_motion_vectors)
			discard_bits(1);
		if (mbt.macroblock_pattern)
			__asm int 3		//coded_block_pattern(); // never for I frames

		// blocks
		int block_count = 0;
		switch (sh->chroma_format)
		{
		case 1:	// 4:2:0
			block_count = 6;
			break;
		case 2:	// 4:2:2
			block_count = 8;
			break;
		case 3:	// 4:4:4
			block_count = 12;
			break;
		default:
			__asm int 3		// bleargh!
		}

		for (int i = block_count - 1; i >= 0; i--)
		{
			if (mbt.macroblock_intra)
				pattern_code[i] = 1;
			else
				pattern_code[i] = 0;
		}
		if (mbt.macroblock_pattern)
		{
			__asm int 3		// never for I frames
/*			for (i=0; i<6; i++)
				if ( cbp & (1<<(5-i)) )
					pattern_code[i] = 1;
			if (chromaFormat == 4:2:2)
				for (i=6; i<8; i++)
					if ( coded_block_pattern_1 & (1<<(7-i)) )
						pattern_code[i] = 1;
			if (chromaFormat == 4:4:4)
				for (i=8; i<12; i++)
					if ( coded_block_pattern_2 & (1<<(11-i)) )
						pattern_code[i] = 1;
*/		}

		uint dct_dc_size;
		uint dct_dc_differential;

		// loop through blocks (7.2.1, page 77)
		for( int i = 0; i < block_count; i++)
		{
			if (pattern_code[i])
			{
				uint cc = 0;			// 0 = Y
				if (i >= 4)
					cc = 1 + (i&1);	// 1 = Cb, 2 = Cr

				short *Qd = &QF[i][0][0];
				if (mbt.macroblock_intra)
				{
					if (i < 4)
					{
						dct_dc_size = decode_dct_dc_size_luminance();
					}
					else
					{
						dct_dc_size = decode_dct_dc_size_chrominance();
					}
					dct_dc_differential = 0;
					if (dct_dc_size != 0)
						dct_dc_differential = get_bits(dct_dc_size);

					// calculate QFS[0] (*Qd)
					int dct_diff;
					if (dct_dc_size == 0)
						dct_diff = 0;
					else
					{
						uint half_range = 1 << (dct_dc_size - 1);
						if (dct_dc_differential >= half_range)
							dct_diff = dct_dc_differential;
						else
							dct_diff = (int)(dct_dc_differential + 1) - (int)(2 * half_range);
					}
					*Qd = dc_dct_pred[cc] + dct_diff;
					dc_dct_pred[cc] = *Qd;

					// QFS[0] should have range of 0 to ((256<<ph->intra_dc_precision)-1)
				}
				else
				{
					__asm int 3		// not required for basic I frames (?)
//					First DCT coefficient 2-24
					//??
//					if (!decode_DCT_coefficient(0, ph->intra_vlc_format, &eob_read, &run, Qd))
//						return false;
				}

				if (fullSize)
				{
					// convert QFS[n] to QF[i][v][u] via inverse scan, 7.3, page 80
					int *scan = &scan_QF_QFS[ph->alternate_scan][1];
					uint n = 1;
					bool eob_read = false;
					unsigned char sf = q_scale[ph->q_scale_type][quantiser_scale_code];
					unsigned int quant_sum = *Qd + 1;

					while (!eob_read)
					{
						uint run;
						int signed_level;
						if (!decode_DCT_coefficient(n, ph->intra_vlc_format, &eob_read, &run, &signed_level))
						{
#ifdef REPORT_SLICE_DECODE_FAILURES
							_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
							return false;
						}
						if (eob_read)
						{
							// end of block
							while (n++ < 64)
								Qd[*scan++] = 0;
							Qd[63] ^= quant_sum & 1;

							gpIDCT->transform(Qd, dst);

							// XXX Temp - recalc average to see if it displays correctly.
							int sum = 0;
							for (int idst = 0; idst < 64; idst++)
								sum += dst[idst];
							*Qd = sum / 8;
							if ((sum % 8) & 0x4)
								(*Qd)++;
						}
						else
						{
							for (uint m = 0; (m < run) && (n++ < 64); m++)
								Qd[*scan++] = 0;
							if (n++ < 64)
							{
								int scanCur = *scan++;
								int f = (signed_level * sf * ph->intra_quantiser_matrix[scanCur]) / 16;
								Qd[scanCur] = f;
								quant_sum += f;
							}
						}
					}
				}// end if(fullSize)
				else
				{
					// preview size, we don't actually need to keep the rest of the QFS elements,
					// so just process the bitstream only.
#if 1
					if (!discard_decode_DCT_coefficients( 1, ph->intra_vlc_format, &mbt ))
					{
#ifdef REPORT_SLICE_DECODE_FAILURES
						_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
						return false;
					}
#else
					uint n = 1;
					bool eob_read = false;
					while (!eob_read)
					{
						uint run;
						int signed_level;
						if (!decode_DCT_coefficient(n++, ph->intra_vlc_format, &eob_read, &run, &signed_level))
						{
#ifdef REPORT_SLICE_DECODE_FAILURES
							_Error("decode failure mb %dx%d (%d)\n", macroblock_address % mb_width, mb_row, (macroblock_address - start_mb_add)%mb_width);
#endif
							return false;
						}
					}
#endif
				}
			}
		}


		float invYdiv = 1.0f / (219.0f * (float)(1 << ph->intra_dc_precision));
		float invCdiv = 1.0f / (224.0f * (float)(1 << ph->intra_dc_precision));

		int R_color = 0;
		int G_color = 0;
		int B_color = 0;

		for( int i = 0; i < 4; i++)
		{
			uint x = ((mb_column<<1) + (i&1));
			uint y = ((mb_row<<1) + (i>>1));

			if( 1 ) 
			{
				int Y = QF[i][0][0];
				int Cb = 0;
				int Cr = 0;

				switch (sh->chroma_format)
				{
					case 1:	// 4:2:0
						Cb = QF[4][0][0];
						Cr = QF[5][0][0];
						break;
					case 2:	// 4:2:2
						Cb = QF[4][0][0];
						Cr = QF[5][0][0];
						break;
					case 3:	// 4:4:4
						Cb = QF[4+(i<<1)][0][0];
						Cr = QF[5+(i<<1)][0][0];
						break;
				}

				float EY = (float)(Y - (16 << ph->intra_dc_precision));
				float EPb = (float)(Cb - (128 << ph->intra_dc_precision));
				float EPr = (float)(Cr - (128 << ph->intra_dc_precision));

				EY *= invYdiv;
				EPb *= invCdiv;
				EPr *= invCdiv;

				// see page 56 for matrix coefficients. Unfortunately, they
				// only list the RGB -> YUV matrix, so have to invert it.
				float EG = 0.9998f * EY - 0.1873f * EPb - 0.4580f * EPr;
				float EB = 1.0018f * EY + 1.8556f * EPb + 0.0010f * EPr;
				float ER = 1.0000f * EY + 0.0007f * EPb + 1.5751f * EPr;

				int R = (int)(255.0f * ER);
				if (R > 255) R = 255;
				if (R < 0) R = 0;
				int G = (int)(255.0f * EG);
				if (G > 255) G = 255;
				if (G < 0) G = 0;
				int B = (int)(255.0f * EB);
				if (B > 255) B = 255;
				if (B < 0) B = 0;

//if( macroblock_address < 0xe )
//	_Error("[%x,%d,(%d,%d,%d)]",macroblock_address,i,R,G,B);
//uint m = macroblock_address % msi->number_of_frame_columns;
//if( m >= 0x2c && m <= 0x4c )
//	_Error("[%x,%d,(%d,%d,%d)]",macroblock_address,i,R,G,B);

				R_color |= R;
				G_color |= G;
				B_color |= B;

			}// end if( 1 )
		}// end for (i = 0; i < 4; i++)

//if( macroblock_address < 0xe )
//	_Error("\n");

			int color = 1;
			if(	// black
				(( R_color <= 0x7 ) && ( G_color <= 0x7 ) && ( B_color <= 0x7 )) ||

				// grey
			    (( R_color >= 0x68 && R_color <= 0x6f ) &&
				(  G_color >= 0x68 && G_color <= 0x6f ) &&
				(  B_color >= 0x68 && B_color <= 0x6f ))
			  )   
			{
				color = 0;
			}
//_Error("%d", color );
			uint bitrow = 0;
			uint row = macroblock_address / msi->number_of_frame_columns;
			uint n = macroblock_address % msi->number_of_frame_columns;

			if( color == 0 && row < 0x43 )
			{
				if( msi->horizontal_size == 0x780 )
				{
					// map this macroblock
					// left bar
					if( n <= 0xe )
					{
						// map to lower 16 bits
						bitrow = ~( 0x1 << n );
						// set the bit to zero 
						msi->macroblock_state[ row ] &= bitrow;
					}

					// middle section of frame
					if( n >= 0x2c && n <= 0x4c )
					{
						bitrow = ~( 0x1 << n );
						// set the bit to zero
						msi->macroblock_state_middle[ row ] &= bitrow;
					}

					// right bar
					if( n >= 0x69 )
					{
						// map to upper 16 bits
						bitrow = ~( 0x00000001 << ( n - 0x69 + 16 ));
						// set the bit to zero 
						msi->macroblock_state[ row ] &= bitrow;
					}
				}
				else
				{
					// left bar
					if( n <= 0xa )
					{
						// map to lower 16 bits
						bitrow = ~( 0x1 << n );
						// set the bit to zero 
						msi->macroblock_state[ row ] &= bitrow;
					}

					// middle section of frame
					if( n >= 0x18 && n <= 0x38 )
					{
						bitrow = ~( 0x1 << n );
						// set the bit to zero
						msi->macroblock_state_middle[ row ] &= bitrow;
					}

					// right bar
					if( n >= 0x46 )
					{
						// map to upper 16 bits
						bitrow = ~( 0x00000001 << ( n - 0x46 + 16 )); 
						// set the bit to zero 
						msi->macroblock_state[ row ] &= bitrow;
					}
				}
			}// end if( color == 0 )

	} while (peek_bits(23) != 0);

	return true;
}

// decode bits for the coded_block_pattern_420
// NOTE: Not current being used.
bool decode_pattern_bits( sequence_header *sh, int *cbp )
{
	ASSERT( sh->chroma_format != 0 ); // reserved
	ASSERT( sh->chroma_format != 2 ); // 4:2:2
	ASSERT( sh->chroma_format != 3 ); // 4:4:4

	uint bits = peek_bits( 9 );
	uint size = 0;

	if( bits & 0x100 )
	{
		if( bits & 0x80 )
		{
			if( bits & 0x40 )
			{
				size = 3;
				// 1 11-- ---- 60
				*cbp = 60;
			}
			else if( bits & 0x20 )
			{
				size = 4;
				// 1 101- ---- 4
				*cbp = 4;
			}
			else
			{
				size = 4;
				// 1 100- ---- 8
				*cbp = 8;
			}
		}
		else if( bits & 0x40 )
		{
			if( bits & 0x20 )
			{
				size = 4;
				// 1 011- ---- 16
				*cbp = 16;
			}
			else
			{
				size = 4;
				// 1 010- ---- 32
				*cbp = 32;
			}
		}
		else if( bits & 0x20 )
		{
			if( bits & 0x10 )
			{
				size = 5;
				// 1 0011 ---- 12
				*cbp = 12;
			}
			else
			{
				size = 5;
				// 1 0010 ---- 48
				*cbp = 48;
			}
		}
		else if( bits & 0x10 )
		{
			size = 5;
			// 1 0001 ---- 20
			*cbp = 20;
		}
		else
		{
			size = 5;
			// 1 0000 ---- 40
			*cbp = 40;
		}
	}// end if( bits & 0x100 )
	
	else if( bits & 0x80 )
	{
		if( bits & 0x40 )
		{
			if( bits & 0x20 )
			{
				if( bits & 0x10 )
				{
					size = 5;
					// 0 1111 ---- 28
					*cbp = 28;
				}
				else
				{
					size = 5;
					// 0 1110 ---- 44
					*cbp = 44;
				}
			}
			else if( bits & 0x10 )
			{
				size = 5;
				// 0 1101 ---- 52
				*cbp = 52;
			}
			else
			{
				size = 5;
				// 0 1100 ---- 56
				*cbp = 56;
			}
		}
		else if( bits & 0x20 )
		{
			if( bits & 0x10 )
			{
				size = 5;
				// 0 1011 ---- 1
				*cbp = 1;
			}
			else
			{
				size = 5;
				// 0 1010 ---- 61
				*cbp = 61;
			}
		}
		else if( bits & 0x10 )
		{
			size = 5;
			// 0 1001 ---- 2
			*cbp = 2;
		}
		else
		{
			size = 5;
			// 0 1000 ---- 62
			*cbp = 62;
		}
	}// end if( bits & 0x80 )

	else if( bits & 0x40 )
	{
		if( bits & 0x20 )
		{
			if( bits & 0x10 )
			{
				if( bits & 0x08 )
				{
					size = 6;
					// 0 0111 1--- 24
					*cbp = 24;
				}
				else
				{
					size = 6;
					// 0 0111 0--- 36
					*cbp = 36;
				}
			}
			else if( bits & 0x08 )
			{
				size = 6;
				// 0 0110 1--- 3
				*cbp = 3;
			}
			else
			{
				size = 6;
				// 0 0110 0--- 63
				*cbp = 63;
			}
		}
		else if( bits & 0x10 )
		{
			if( bits & 0x08 )
			{
				if( bits & 0x04 )
				{
					size = 7;
					// 0 0101 11-- 5
					*cbp = 5;
				}
				else
				{
					size = 7;
					// 0 0101 10-- 9
					*cbp = 9;
				}
			}
			else if( bits & 0x04 )
			{
				size = 7;
				// 0 0101 01-- 17
				*cbp = 17;
			}
			else
			{
				size = 7;
				// 0 0101 00-- 33
				*cbp = 33;
			}
		}
		else if( bits & 0x08 )
		{
			if( bits & 0x40 )
			{
				size = 7;
				// 0 0100 11-- 6
				*cbp = 6;
			}
			else
			{
				size = 7;
				// 0 0100 10-- 10
				*cbp = 10;
			}
		}
		else if( bits & 0x04 )
		{
			size = 7;
			// 0 0100 01-- 18
			*cbp = 18;
		}
		else
		{
			size = 7;
			// 0 0100 00-- 34
			*cbp = 34;
		}
	}// end if( bits & 0x40 )
			
	else if( bits & 0x20 )
	{
		if( bits & 0x10 )
		{
			if( bits & 0x08 )
			{
				if( bits & 0x04 )
				{
					if( bits & 0x02 )
					{
						size = 8;
						// 0 0011 111- 7
						*cbp = 7;
					}
					else
					{
						size = 8;
						// 0 0011 110- 11
						*cbp = 11;
					}
				}
				else if( bits & 0x02 )
				{
					size = 8;
					// 0 0011 101- 19
					*cbp = 19;
				}
				else
				{
					size = 8;
					// 0 0011 100- 35
					*cbp = 35;
				}
			}
			else if( bits & 0x04 )
			{
				if( bits & 0x02 )
				{
					size = 8;
					// 0 0011 011- 13
					*cbp = 13;
				}
				else
				{
					size = 8;
					// 0 0011 010- 49
					*cbp = 49;
				}
			}
			else if( bits & 0x02 )
			{
				size = 8;
				// 0 0011 001- 21
				*cbp = 21;
			}
			else
			{
				size = 8;
				// 0 0011 000- 41
				*cbp = 41;
			}
 		}
		else if( bits & 0x08 )
		{
			if( bits & 0x04 )
			{
				if( bits & 0x02 )
				{
					size = 8;
					// 0 0010 111- 14
					*cbp = 14;
				}
				else
				{
					size = 8;
					// 0 0010 110- 50
					*cbp = 50;
				}
			}
			else if( bits & 0x02 )
			{
				size = 8;
				// 0 0010 101- 22
				*cbp = 22;
			}
			else
			{
				size = 8;
				// 0 0010 100- 42
				*cbp = 42;
			}
		}
		else if( bits & 0x04 )
		{
			if( bits & 0x02 )
			{
				size = 8;
				// 0 0010 011- 15
				*cbp = 15;
			}
			else
			{
				size = 8;
				// 0 0010 010- 51
				*cbp = 51;
			}
		}
		else if( bits & 0x02 )
		{
			size = 8;
			// 0 0010 001- 23
			*cbp = 23;
		}
		else
		{
			size = 8;
			// 0 0010 000- 43
			*cbp = 43;
		}
	}// end if( bits & 0x20 )

	else if( bits & 0x10 )
	{
		if( bits & 0x08 )
		{
			if( bits & 0x04 )
			{
				if( bits & 0x02 )
				{
					size = 8;
					// 0 0001 111- 25
					*cbp = 25;
				}
				else
				{
					size = 8;
					// 0 0001 110- 37
					*cbp = 37;
				}
			}
			else if( bits & 0x02 )
			{
				size = 8;
				// 0 0001 101- 26
				*cbp = 26;
			}
			else
			{
				size = 8;
				// 0 0001 100- 38
				*cbp = 38;
			}
		}
		else if( bits & 0x04 )
		{
			if( bits & 0x02 )
			{
				size = 8;
				// 0 0001 011- 29
				*cbp = 29;
			}
			else
			{
				size = 8;
				// 0 0001 010- 45
				*cbp = 45;
			}
		}
		else if( bits & 0x02 )
		{
			size = 8;
			// 0 0001 001- 53
			*cbp = 53;
		}
		else
		{
			size = 8;
			// 0 0001 000- 57
			*cbp = 57;
		}
	}// end if( bits & 0x10 )
			
	else if( bits & 0x08 )
	{
		if( bits & 0x04 )
		{
			if( bits & 0x02 )
			{
				size = 8;
				// 0 0000 111- 30
				*cbp = 30;
			}
			else
			{
				size = 8;
				// 0 0000 110- 46
				*cbp = 46;
			}
		}
		else if( bits & 0x02 )
		{
			size = 8;
			// 0 0000 101- 54
			*cbp = 54;
		}
		else
		{
			size = 8;
			// 0 0000 100- 58
			*cbp = 58;
		}
	}// end if( bits & 0x08 )

	else if( bits & 0x04 )
	{
		if( bits & 0x02 )
		{
			if( bits & 0x01 )
			{
				size = 9;
				// 0 0000 0111 31
				*cbp = 31;
			}
			else
			{
				size = 9;
				// 0 0000 0110 47
				*cbp = 47;
			}
		}
		else if( bits & 0x01 )
		{
			size = 9;
			// 0 0000 0101 55
			*cbp = 55;
		}
		else
		{
			size = 9;
			// 0 0000 0100 59
			*cbp = 59;
		}
	}// end if( bits & 0x04 )

	else if( bits & 0x02 )
	{
		if( bits & 0x01 )
		{
			size = 9;
			// 0 0000 0011 27
			*cbp = 27;
		}
		else
		{
			size = 9;
			// 0 0000 0010 39
			*cbp = 39;
		}
	}// end if( bits & 0x02 )

	else if( bits & 0x01 )
	{
		// 0 0000 0001 Note: shall not be used with 4:2:0 chrominance structure
		// ASSERT( sh->chroma_format != 1 );
		//size = 9;
		*cbp = 0;
#ifdef REPORT_SLICE_DECODE_FAILURES
			_Error("decode pattern code failure. bits are[%x]\n", bits );
#endif
		return false;
	}
	else
		ASSERT( true ); // something went wrong!

	discard_bits( size );

	return true;
}

// old channel scanning routine
int scan_for_excludes(TSFileList *inputList, int vidPid,
		__int64 minExcludeSize, __int64 responseSize, float maxError, float scanRatio,
		FileReadBuffer *readBuffer, convert_callback callback, void *data, HWND parentWnd)
{
	int rVal = 0;
	char errMsgBuf[256];
	int errMsgVal;
	long timeDiscrepancy = 0;

	__int64 filePos = 0;
	__int64 totalRead = 0;
	__int64 lastIndicated = 0;
//	FILE64 *fIn = NULL;

	TSFileInfo *fileInfo = inputList->GetHead();
	if (!fileInfo)
		return 1;

	PIDinfoGroup PIDs = { 1, { { vidPid, 0, STREAM_PES, PID_MPEG2_VIDEO, false, NULL, 0, 0, -1, false, false } } } ;

	PIDs.pids[0].buffer = new uchar[VIDEO_BUFFER_SIZE];
	PIDs.pids[0].bufferSize = VIDEO_BUFFER_SIZE;

	if (!PIDs.pids[0].buffer)
	{
		::MessageBox(parentWnd, "Out of memory", NULL, MB_OK | MB_ICONERROR);
		return 1;
	}

	TSRegionInfo *regionInfo = fileInfo->m_head;
	RegionType regionType = rt_include;
	bool needToSeek = true;

	pdlg_buffer *cTarget = NULL;
	pdlg_buffer *pTarget = NULL;
	__int64 cOffset = 0;
	__int64 pOffset = 0;
	__int64 sOffset = -1;
	__int64 beginOffset = 0;
	bool excluding = false;
	TSRegionInfo *lastInclude = NULL;
	__int64 lastOffset = 0;

	__int64 originalTotalIncludedBytes = inputList->TotalIncludedBytes();

	unsigned char buf[188];
	int bytesReadInPacket = 0;
	for (;;)
	{
		__int64 indicated = fileInfo->m_totalBytesPrior + filePos;
		if (callback && (indicated > lastIndicated + 20000000))
		{
			lastIndicated = indicated;

			bool cont = callback(fileInfo->m_fileName, NULL,
					fileInfo->m_totalBytesPrior + filePos,
					totalRead, originalTotalIncludedBytes,
					-1, timeDiscrepancy, data);
			timeDiscrepancy = 0;

			// Cancelled?
			if (!cont)
				break;
		}

		// check regions, skip data, open new files
		if (regionInfo && (filePos >= regionInfo->m_offset))
		{
			if ((regionType == rt_exclude) && (regionInfo->m_type == rt_include))
			{
				// back into an include from an exclude, reset what we're doing
				cTarget = NULL;
				pTarget = NULL;
				cOffset = 0;
				pOffset = 0;
				sOffset = -1;
				beginOffset = indicated;
				excluding = false;
				lastInclude = NULL;
				lastOffset = 0;
			}
			regionType = regionInfo->m_type;
			regionInfo = regionInfo->m_next;
		}
		if (regionType == rt_exclude)
		{
			if (regionInfo)
			{
				filePos = regionInfo->m_offset;
			}
			else
			{
				if (readBuffer->IsFileOpen())
//				if (fIn)
				{
					readBuffer->Close();
//					fclose64(fIn);
//					fIn = NULL;
				}
				fileInfo = fileInfo->m_next;
				if (!fileInfo)
					break;	// we're complete!
				regionInfo = fileInfo->m_head;
				filePos = 0;
			}
			needToSeek = true;
			continue;
		}
		else
		{
			if (!readBuffer->IsFileOpen())
//			if (!fIn)
			{
				readBuffer->Open(fileInfo->m_fileName);
//				fIn = fopen64(fileInfo->m_fileName, "rb", RDBUFSIZE);
				while (!readBuffer->IsFileOpen())
//				while (!fIn)
				{
					sprintf(errMsgBuf, "Couldn't open input file\n%s", fileInfo->m_fileName);
					long tmp1, tmp2;
					time((time_t*)&tmp1);
					errMsgVal = ::MessageBox(parentWnd, errMsgBuf, NULL, MB_RETRYCANCEL | MB_ICONWARNING);
					time((time_t*)&tmp2);
					timeDiscrepancy += tmp2 - tmp1;
					if (errMsgVal == IDRETRY)
					{
						readBuffer->Open(fileInfo->m_fileName);
//						fIn = fopen64(fileInfo->m_fileName, "rb", RDBUFSIZE);
					}
					else
					{
						rVal = 1;
						break;	// where is this file??
					}
				}
				if (rVal)
					break;
				needToSeek = true;
			}
			if (needToSeek)
			{
				// check packet alignment of new file
				__int64 offset = (188 - ((filePos + 188 - fileInfo->m_firstPacketOffset) % 188)) % 188;
				if (offset != ((188 - bytesReadInPacket) % 188))
				{
					// alignment does not match, discard partial packet
					bytesReadInPacket = 0;
					filePos += offset;
				}
				readBuffer->Seek(filePos);
//				fseek64(fIn, filePos, SEEK_SET);
				needToSeek = false;
			}
		}

		// Try to read packet..
		int bytesRead;
		if ((bytesRead = readBuffer->Read(buf + bytesReadInPacket, 188 - bytesReadInPacket)) != 188 - bytesReadInPacket)
//		if ((bytesRead = fread64(buf + bytesReadInPacket, 1, 188 - bytesReadInPacket, fIn)) != 188 - bytesReadInPacket)
		{
			if (bytesRead < 0)
				bytesRead = 0;
			bytesReadInPacket += bytesRead;
			totalRead += bytesRead;

			if (readBuffer->IsError())
			{
				// say something to the user, and exit!
				sprintf(errMsgBuf, "Error reading input file\n%s", fileInfo->m_fileName);
				::MessageBox(parentWnd, errMsgBuf, NULL, MB_OK | MB_ICONERROR);
				rVal = 1;
				break;
			}

			// Close this file
			readBuffer->Close();
//			fclose64(fIn);
//			fIn = NULL;
			fileInfo = fileInfo->m_next;
			if (!fileInfo)
				break;	// we're complete!
			filePos = 0;

			continue;
		}
		else
		{
			filePos += bytesRead;
			bytesReadInPacket = 0;
			totalRead += bytesRead;
		}

		// Check sync byte
		if ((buf[0] == 0x47) || (buf[0] == 0x72) || (buf[0] == 0x29))
		{
			// Get pid
			int pid = (((buf[1] & 0x1F) << 8) | buf[2]) & 0x1FFF;

			PIDinfo *p = find_PID_no_new(&PIDs, pid);
			if (p)
			{
				p->count++;

				// Get error
				bool errorBit = (buf[1] & 0x80) != 0;
				if (errorBit)
				{
					// invalidate accumulated content of this stream
					p->reading = false;
				}
				else
				{
					// Get adaptation header info
					int adaptLength = 0;
					int adaptation = (buf[3] & 0x30) >> 4;
					if (adaptation != 0)
					{
						if (adaptation == 0x2)
							adaptLength = 184;
						else if (adaptation == 0x3)
							adaptLength = buf[4] + 1;
						if (adaptLength <= 184)
						{
							// Get payload start indicator
							bool start = (buf[1] & 0x40) != 0;

							// process content of packet
							// assuming is a valid PES stream
							if (start)
							{
								if (p->reading && (p->bufferPos > 0))
								{
									if (p->pType == PID_MPEG2_VIDEO)
									{
										// decode I frame into a buffer
										cTarget = &pdlg_buffers[0];
										if (cTarget == pTarget)
											cTarget = &pdlg_buffers[1];
										if (process_PES_video_packet(p->buffer, p->bufferPos, true, NULL, cTarget))
										{
											// compare left and right edges to prior frame
											if (compareFrameEdges(cTarget, pTarget, scanRatio) < maxError)
											{
												// if there is zero difference, possibly a commercial!
												if (sOffset < 0)
												{
													// if not currently in a possible commercial,
													// make a note of prior frame as start frame
													if (beginOffset >= 0)
														sOffset = beginOffset;
													else
														sOffset = pOffset;
												}
												if (!excluding && ((cOffset - sOffset) > responseSize))
												{
													// if duration of current frame - start frame exceeds minimum length,
													// set a begin exclude at start frame
													if (lastInclude && ((sOffset - lastOffset) < responseSize))
													{
														// but wait, the previous exclude zone is too close, thus the
														// intervening include must be a hiccup, so let's remove it,
														// extending the previous exclude zone into this area
_Error("Removing Include %08x%08x\n", (uint)(lastOffset>>32), (uint)(lastOffset&0xffffffff));
														inputList->DeleteRegion(lastInclude);
														lastInclude = NULL;
														excluding = true;
													}
													else if ((cOffset - sOffset) > minExcludeSize)
													{
_Error("Start Exclude %08x%08x\n", (uint)(sOffset>>32), (uint)(sOffset&0xffffffff));
														if (inputList->AddRegion(rt_exclude, sOffset))
														{
															excluding = true;
														}
													}
												}
											}
											else
											{
												// if there is a difference, and we were currently in a begun exclude,
												if (excluding)
												{
													// set an include at this frame
_Error("Start Include %08x%08x\n", (uint)(cOffset>>32), (uint)(cOffset&0xffffffff));
													if (lastInclude = inputList->AddRegion(rt_include, pOffset))
													{
														lastOffset = cOffset;
													}
												}
												sOffset = -1;
												excluding = false;
											}
											if (pOffset)
												beginOffset = -1;
											pTarget = cTarget;
											pOffset = cOffset;
										}
									}
								}
								// start a new PES packet
								if (p->buffer && !p->processed)
								{
									p->reading = true;
									p->bufferPos = 0;
									p->confirmedRead = false;
									cOffset = fileInfo->m_totalBytesPrior + filePos - 188;
								}
							}
							if (p->reading)
							{
								// Add the payload for this packet to the current buffer
								int payloadLength = 184 - adaptLength;
								if (payloadLength > 0)
								{
									if ((p->bufferPos + payloadLength) <= p->bufferSize)
									{
										memcpy(p->buffer + p->bufferPos, buf + 4 + adaptLength, payloadLength);
										p->bufferPos += payloadLength;
									}
									else
									{
										// buffer overflow!
										_Error("Error: buffer overflow! PID=0x%02X, PES packet, buffer size = %d\n", pid, p->bufferSize);
										p->reading = false;
									}
								}

								if (!p->confirmedRead && (p->bufferPos > 512))
								{
									if (p->pType == PID_MPEG2_VIDEO)
									{
										// only want to continue reading this PES packet if it's an I-frame.
										// look for sequence header start code
										uint start_code = 0;
										for (int i = 0; i < p->bufferPos; i++)
										{
											start_code = (start_code << 8) | p->buffer[i];
											if (start_code == 0x000001b3)
											{
												p->confirmedRead = true;
												break;
											}
										}
									}
									if (!p->confirmedRead)
										p->reading = false;
								}
							}
						}
					}
				}
			}
		}
	}

	// Close files
	if (readBuffer->IsFileOpen())
		readBuffer->Close();
//	if (fIn)
//		fclose64(fIn);
	if (PIDs.pids[0].buffer)
		delete []PIDs.pids[0].buffer;

	return rVal;
}




float compareFrameEdges(pdlg_buffer *t1, pdlg_buffer *t2, float scanRatio)
{
	if (!t1 || !t2 ||
		!t1->bm.buffer || !t2->bm.buffer ||
		(t1->rendered == 0) || (t2->rendered == 0) ||
		(t1->renderedWidth != t2->renderedWidth) ||
		(t1->renderedHeight != t2->renderedHeight) ||
		(t1->aspectRatio != t2->aspectRatio))
		return 256.0f;

	uint scan_width = (uint)((float)t1->renderedWidth * scanRatio) - 1;

	float error = 0.0f;
	uchar *b1, *b2;
	int d1, d2, d3;

	for (uint y = t1->bm.height - t1->renderedHeight; y < t1->bm.height; y++)
	{
		b1 = &t1->bm.buffer[y*t1->bm.stride];
		b2 = &t2->bm.buffer[y*t2->bm.stride];
		for (uint x = scan_width; x > 0; x--)
		{
			d1 = (int)(*b1++) - (int)(*b2++);
			d2 = (int)(*b1++) - (int)(*b2++);
			d3 = (int)(*b1++) - (int)(*b2++);
			error += (float)(d1*d1 + d2*d2 + d3*d3);
		}
	}

	for( uint y = t1->bm.height - t1->renderedHeight; y < t1->bm.height; y++)
	{
		b1 = &t1->bm.buffer[(t1->renderedWidth - scan_width)*3 + y*t1->bm.stride];
		b2 = &t2->bm.buffer[(t2->renderedWidth - scan_width)*3 + y*t2->bm.stride];
		for (uint x = scan_width; x > 0; x--)
		{
			d1 = (int)(*b1++) - (int)(*b2++);
			d2 = (int)(*b1++) - (int)(*b2++);
			d3 = (int)(*b1++) - (int)(*b2++);
			error += (float)(d1*d1 + d2*d2 + d3*d3);
		}
	}

	float n_pixels = (float)(2 * t1->renderedHeight * scan_width*3);
	return (float)sqrt(error / n_pixels);
}

